/*
 * (C) Copyright The University of Manchester 2004
 */
package com.ibm.JikesRVM.opt.ir;
// N.B. Not part of the PPC_Emulator package as package access is
// needed to OPT_GenerationContext
import com.ibm.JikesRVM.VM;
import com.ibm.JikesRVM.classloader.VM_PPC_Trace;
import com.ibm.JikesRVM.classloader.VM_Class;
import com.ibm.JikesRVM.classloader.VM_Method;
import com.ibm.JikesRVM.classloader.VM_TypeReference;
import com.ibm.JikesRVM.classloader.VM_MethodReference;
import com.ibm.JikesRVM.classloader.VM_MemberReference;
import com.ibm.JikesRVM.classloader.VM_FieldReference;
import com.ibm.JikesRVM.classloader.VM_SystemClassLoader;
import com.ibm.JikesRVM.classloader.VM_Atom;
import com.ibm.JikesRVM.opt.OPT_Constants;
import org.binarytranslator.PPC_Emulator.PPC_Constants;
import org.binarytranslator.PPC_Emulator.PPC_Options;
import org.binarytranslator.PPC_Emulator.PPC_InstructionTranslator;
import org.binarytranslator.PPC_Emulator.PPC_ProcessSpace;
import org.binarytranslator.PPC_Emulator.PPC_BranchLogic;
import java.util.Enumeration;
import java.util.NoSuchElementException;
import java.util.Vector;
import java.util.Set;
import java.util.Iterator;
import java.util.HashSet;
import java.util.HashMap;

/** 
 * Translation from PPC machine code to HIR.  The public entry point
 * is generateHIR(OPT_GenerationContext gc).
 *
 * <dl><dt>Description of the translation algorithm</dt>
 *     <dd><ol>
 * <li>The translation is set up so that the start and end blocks are
 * created and the start block filled in</li>
 * <li>Translation starts from the PC value given in the VM_PPC_Trace.</li>
 * <li>Translation translates PPC instructions to HIR using
 * PPC_InstructionTranslators. The PPC2IR object contains the utilities
 * required for an instruction translation.</li>
 * <li>The trace ends when the next instructions can't be
 * predicted. The closing basic blocks are generated and things are
 * wrapped up.</li>
 *     </ol></dd>
 * </dl>
 * 
 * How we aim to achieve simplicity and performance:
 * <ol>
 * <li>Keep mapping of PPC instruction to HIR to as few instructions
 * as possible</li>
 * <li>{@see PPC_ProcessSpace} should look like a PPC architecture, but
 * things in here can be grubbier</li>
 * <li>Laziness should be used so that we don't fully calculate all
 * flag bits, reducing number of instructions and helping dead code
 * elimination (ORs aren't killing)</li>
 * </ol>
 *
 * @author Richard Matley, Ian Rogers
 */
public final class PPC2IR implements OPT_IRGenOptions, OPT_Operators, OPT_Constants, PPC_Constants {
  // -oO Things given to us for being in a Java translator Oo-

  /**
   * The generation context.
   */
  private OPT_GenerationContext gc;

  /**
   * The process space object used by the running PPC binary.
   */
  private PPC_ProcessSpace ps;

  /**
   * Current state of local variables.
   */
  private OPT_Operand[] _localState;

  // -oO Implement OPT_Compiler style interface Oo-

  /** 
   * Generate HIR as specified by the OPT_GenerationContext argument. 
   * This is the public entry point to the class.
   */
  public static void generateHIR(OPT_GenerationContext gc)
  {
    new PPC2IR(gc).generateHIR();
  }

  /**
   * Prevent external instantiation.
   */
  private PPC2IR(){}

  /** 
   * Construct the PPC2IR object for the generation context; then
   * we'll be ready to start generating the HIR.
   */
  private PPC2IR(OPT_GenerationContext context)
  {
    // Make copies of popular variables 
    gc = context;
    ps = ((VM_PPC_Trace)(gc.method)).ps;

    // Set up Java local variable state from method information
    _localState = new OPT_Operand[context.method.getLocalWords()];
    for (int argIdx = 0, localIdx = 0; argIdx < context.arguments.length;) {
      VM_TypeReference argType = context.arguments[argIdx].getType();
      _localState[localIdx++] = context.arguments[argIdx++];
      if (argType.isLongType() || argType.isDoubleType()) {
        throw new Error("Unexpected large argument to PPCEmulatorRuntime.invokeCode");
      }
    }

    // Create register maps PPC -> OPT_Register
    intRegMap = new OPT_Register[32];
    intRegInUseMap = new boolean[32];
    fpRegMap = new OPT_Register[32];
    fpRegInUseMap = new boolean[32];
    crFieldMap_LtGtEq_rA = new OPT_Register[8];
    crFieldMap_LtGtEq_rB = new OPT_Register[8];
    crFieldMap_LtGtEq_frA = new OPT_Register[8];
    crFieldMap_LtGtEq_frB = new OPT_Register[8];
    crFieldMap_SO = new OPT_Register[8];

    // Create map of (PC & laziness) -> OPT_BasicBlock
    blockMap = new HashMap();     

	 // Fix up stores
	 unresolvedGoto = new Vector();
	 unresolvedGoto_PC = new Vector();
	 unresolvedGoto_Laziness = new Vector();
    unresolvedIfCmp = new Vector();
    unresolvedIfCmp_PC = new Vector();
    unresolvedIfCmp_Laziness = new Vector();
	 unresolvedLookupSwitchForLR = new Vector();
	 unresolvedLookupSwitchForLR_PC = new Vector();
	 unresolvedLookupSwitchForLR_Laziness = new Vector();
	 unresolvedLookupSwitchForCTR = new Vector();
	 unresolvedLookupSwitchForCTR_PC = new Vector();
	 unresolvedLookupSwitchForCTR_Laziness = new Vector();
	 unresolvedLookupSwitchForCTR_Link = new Vector();

	 // Number of translated instructions
	 numberOfInstructions = 0;
	 // Branch information - either global in the ps or local
	 if (gc.options.getOptLevel() < PPC_Options.globalBranchLevel) {
		branchInfo = new PPC_BranchLogic(ps.branchInfo);
	 }
	 else {
		branchInfo = ps.branchInfo;
	 }

    // Create lazyPreFillBlock, currentBlock and finishBlock
	 gc.prologue.insertOut(gc.epilogue);
    lazyPreFillBlock = createBlockAfter(gc.prologue);

    currentBlock = createBlockAfter(lazyPreFillBlock);

    finishBlock = createBlockAfterCurrent();

    // Debug
    if(PPC_Options.debugCFG) {
      report("CFG at end of constructor:\n" + gc.cfg); 
    }
  }

  // -oO Creations for being a PPC translator Oo-

  // -oO PPC register to HIR register mappings Oo-

  /**
   * The mapping of PPC general purpose registers to HIR
   * registers. All GP registers are loaded into these by the preFill
   * block. This avoids potential inconsistencies caused by using lazy
   * allocation and backward branches.
   */
  private OPT_Register intRegMap[];
  /**
	* Which PPC general purpose registers are in use
	*/
  private boolean intRegInUseMap[];

  /**
   * The mapping of PPC floating point registers to HIR registers.
   */
  private OPT_Register fpRegMap[];
  /**
	* Which PPC floating point registers are in use
	*/
  private boolean fpRegInUseMap[];

  /**
   * The HIR register holding the PPC FPSCR register.
   */
  private OPT_Register fpscrRegMap;
  /**
	* Is the PPC FPSCR register in use
	*/
  private boolean fpscrRegInUse;

  /**
   * The HIR register holding the PPC CTR register.
   */
  private OPT_Register ctrRegMap;
  /**
	* Is the PPC CTR register in use
	*/
  private boolean ctrRegInUse;

  /**
   * The HIR register holding the PPC LR register.
   */
  private OPT_Register lrRegMap;
  /**
	* Is the PPC LR register in use
	*/
  private boolean lrRegInUse;

  /**
   * The HIR register holding the PPC XER register when it is combined
   * in a fillAll or spillAll block
   */
  private OPT_Register xerRegMap;

  /**
   * The HIR register holding the PPC XER register's byte count (bits
   * 25 to 31)
   */
  private OPT_Register xerRegMap_ByteCount;

  /**
   * The HIR register holding thr PPC XER CA (carry) bit. Bit 32 (the
   * 33rd bit) of this long register holds the value.
   */
  private OPT_Register xerRegMap_CA;

  /**
   * The HIR register holding thr PPC XER OV (overflow) bit. If this
   * register is non-zero then the OV bit should be set.
   */
  private OPT_Register xerRegMap_OV;
  /**
   * The HIR register holding thr PPC XER SO (summary overflow)
   * bit. If this register is non-zero then the SO bit should be set.
   */
  private OPT_Register xerRegMap_SO;
  /**
	* Is the PPC XER register in use
	*/
  private boolean xerRegInUse;
  
  /**
   * The HIR register holding PPC condition register however many bits
   * may be lazyily evaluated
   */
  private OPT_Register crRegMap;
  /**
	* Is the PPC CR register in use
	*/
  private boolean crRegInUse;

  /**
   * These 8 registers hold a copy of the rA value used in a
   * comparison
   */
  private OPT_Register crFieldMap_LtGtEq_rA[];

  /**
   * These 8 registers hold a copy of the rB value used in a
   * comparison
   */
  private OPT_Register crFieldMap_LtGtEq_rB[];

  /**
   * These 8 registers hold a copy of the frA value used in a
   * comparison
   */
  private OPT_Register crFieldMap_LtGtEq_frA[];

  /**
   * These 8 registers hold a copy of the frB value used in a
   * comparison
   */
  private OPT_Register crFieldMap_LtGtEq_frB[];

  /**
   * These 8 registers hold a zero or non-zero value depending on the
   * value of the corresponding condition register field's SO
   * bit. They are only valid if the lazy state says they are.
   */
  private OPT_Register crFieldMap_SO[];

  /**
	* Number of translated instructions
	*/
  public int numberOfInstructions;


  // -oO Fill/spill registers between OPT_Registers and the PPC_ProcessSpace Oo-

  /**
   * Copy the value of a general purpose (int) register into its
   * temporary location from the PPC_ProcessSpace.
   * @param r the number of the register to fill.
   */
  private void fillGPRegister(int r) {
    if (VM.VerifyAssertions) VM._assert(r < 32);

    OPT_RegisterOperand result;
	 if (intRegMap[r] == null) {
		result = gc.temps.makeTempInt();
		intRegMap[r]=result.register; // Set mapping
	 }
	 else {
		result = new OPT_RegisterOperand(intRegMap[r], VM_TypeReference.Int);
	 }

    VM_TypeReference psTref = VM_TypeReference.findOrCreate(VM_SystemClassLoader.getVMClassLoader(),
																				VM_Atom.findOrCreateAsciiAtom("Lorg/binarytranslator/PPC_Emulator/PPC_ProcessSpace;")
																				);
    VM_FieldReference ref = VM_MemberReference.findOrCreate(psTref,VM_Atom.findOrCreateAsciiAtom("r"+r),
																				VM_Atom.findOrCreateAsciiAtom("I")
																				).asFieldReference();
    appendInstructionToCurrentBlock(GetField.create(GETFIELD, result,
																	 gc.makeLocal(1,psTref),
																	 new OPT_IntConstantOperand(ref.peekResolvedField().getOffset()),
																	 new OPT_LocationOperand(ref),
																	 new OPT_TrueGuardOperand())
												);
  }


  /**
   * Copy the value of a general purpose (int) register into the
   * PPC_ProcessSpace from its temporary location.
   * @param r the number of the register to spill.
   */
  private void spillGPRegister(int r) {
    if (VM.VerifyAssertions) VM._assert(r < 32);
    if (VM.VerifyAssertions) VM._assert(intRegMap[r] != null);

    OPT_RegisterOperand regOp = new OPT_RegisterOperand(intRegMap[r], VM_TypeReference.Int);

    VM_TypeReference psTref =  VM_TypeReference.findOrCreate(VM_SystemClassLoader.getVMClassLoader(),
																				 VM_Atom.findOrCreateAsciiAtom("Lorg/binarytranslator/PPC_Emulator/PPC_ProcessSpace;")
																				 );
    VM_FieldReference ref = VM_MemberReference.findOrCreate(psTref,VM_Atom.findOrCreateAsciiAtom("r"+r),
																				VM_Atom.findOrCreateAsciiAtom("I")
																				).asFieldReference();
    appendInstructionToCurrentBlock(PutField.create(PUTFIELD, regOp,
																	 gc.makeLocal(1,psTref),
																	 new OPT_IntConstantOperand(ref.peekResolvedField().getOffset()),
																	 new OPT_LocationOperand(ref),
																	 new OPT_TrueGuardOperand())
												);
  }


  /**
   * Copy the value of a floating point register into its temporary
   * location from the PPC_ProcessSpace.
   * @param r the number of the register to fill.
   */
  private void fillFPRegister(int r) {
    if (VM.VerifyAssertions) VM._assert(r < 32);

    OPT_RegisterOperand result;
	 if (fpRegMap[r] == null) {
		result = gc.temps.makeTempDouble();
		fpRegMap[r]=result.register; // Set mapping
	 }
	 else {
		result = new OPT_RegisterOperand(fpRegMap[r], VM_TypeReference.Double);
	 }

    VM_TypeReference psTref = VM_TypeReference.findOrCreate(VM_SystemClassLoader.getVMClassLoader(),
																				VM_Atom.findOrCreateAsciiAtom("Lorg/binarytranslator/PPC_Emulator/PPC_ProcessSpace;")
																				);
    VM_FieldReference ref = VM_MemberReference.findOrCreate(psTref,VM_Atom.findOrCreateAsciiAtom("f"+r),
																				VM_Atom.findOrCreateAsciiAtom("D")
																				).asFieldReference();
    appendInstructionToCurrentBlock(GetField.create(GETFIELD, result,
																	 gc.makeLocal(1,psTref),
																	 new OPT_IntConstantOperand(ref.peekResolvedField().getOffset()),
																	 new OPT_LocationOperand(ref),
																	 new OPT_TrueGuardOperand())
												);
  }


  /**
   * Copy the value of a floating point register into the
   * PPC_ProcessSpace from its temporary location.
   * @param r the number of the register to spill.
   */
  private void spillFPRegister(int r) {
    if (VM.VerifyAssertions) VM._assert(r < 32);
    if (VM.VerifyAssertions) VM._assert(fpRegMap[r] != null);

    OPT_RegisterOperand regOp = new OPT_RegisterOperand(fpRegMap[r], VM_TypeReference.Double);

    VM_TypeReference psTref =  VM_TypeReference.findOrCreate(VM_SystemClassLoader.getVMClassLoader(),
																				 VM_Atom.findOrCreateAsciiAtom("Lorg/binarytranslator/PPC_Emulator/PPC_ProcessSpace;")
																				 );
    VM_FieldReference ref = VM_MemberReference.findOrCreate(psTref,VM_Atom.findOrCreateAsciiAtom("f"+r),
																				VM_Atom.findOrCreateAsciiAtom("D")
																				).asFieldReference();
    appendInstructionToCurrentBlock(PutField.create(PUTFIELD, regOp,
																	 gc.makeLocal(1,psTref),
																	 new OPT_IntConstantOperand(ref.peekResolvedField().getOffset()),
																	 new OPT_LocationOperand(ref),
																	 new OPT_TrueGuardOperand())
												);
  }

  /**
   * Copy the value of the condition register into its temporary
   * locations from the PPC_ProcessSpace.
   */
  private void fillCRRegister()
  {
    OPT_RegisterOperand result;
	 if (crRegMap == null) {
		result = gc.temps.makeTempInt();
		crRegMap=result.register; // Set mapping
		// Creating mappings for lazy registers
		for (int i=0; i < 8; i++) {
		  crFieldMap_LtGtEq_rA[i] = gc.temps.makeTempInt().register;
		  crFieldMap_LtGtEq_rB[i] = gc.temps.makeTempInt().register;
		  crFieldMap_SO[i] = gc.temps.makeTempInt().register;
		  crFieldMap_LtGtEq_frA[i] = gc.temps.makeTempDouble().register;
		  crFieldMap_LtGtEq_frB[i] = gc.temps.makeTempDouble().register;
		}
	 }
	 else {
		result = new OPT_RegisterOperand(crRegMap, VM_TypeReference.Int);
	 }
    VM_TypeReference psTref = VM_TypeReference.findOrCreate(VM_SystemClassLoader.getVMClassLoader(),
																				VM_Atom.findOrCreateAsciiAtom("Lorg/binarytranslator/PPC_Emulator/PPC_ProcessSpace;")
																				);
    VM_FieldReference ref = VM_MemberReference.findOrCreate(psTref,VM_Atom.findOrCreateAsciiAtom("cr"),
																				VM_Atom.findOrCreateAsciiAtom("I")
																				).asFieldReference();
    appendInstructionToCurrentBlock(GetField.create(GETFIELD, result,
																	 gc.makeLocal(1,psTref),
																	 new OPT_IntConstantOperand(ref.peekResolvedField().getOffset()),
																	 new OPT_LocationOperand(ref),
																	 new OPT_TrueGuardOperand())
												);
  }

  /**
   * Copy the value of the condition register into the PPC_ProcessSpace
   * from its temporary location.
   */
  private void spillCRRegister()
  {
    OPT_RegisterOperand regOp = new OPT_RegisterOperand(crRegMap, VM_TypeReference.Int);

    VM_TypeReference psTref =  VM_TypeReference.findOrCreate(VM_SystemClassLoader.getVMClassLoader(),
																				 VM_Atom.findOrCreateAsciiAtom("Lorg/binarytranslator/PPC_Emulator/PPC_ProcessSpace;")
																				 );
    VM_FieldReference ref = VM_MemberReference.findOrCreate(psTref,VM_Atom.findOrCreateAsciiAtom("cr"),
																				VM_Atom.findOrCreateAsciiAtom("I")
																				).asFieldReference();
    appendInstructionToCurrentBlock(PutField.create(PUTFIELD, regOp,
																	 gc.makeLocal(1,psTref),
																	 new OPT_IntConstantOperand(ref.peekResolvedField().getOffset()),
																	 new OPT_LocationOperand(ref),
																	 new OPT_TrueGuardOperand())
												);
  }



  /**
   * Copy the value of the fpscr register into its temporary location
   * from the PPC_ProcessSpace.
   */
  private void fillFPSCRRegister()
  {
    OPT_RegisterOperand result;
	 if (fpscrRegMap == null) {
		result = gc.temps.makeTempInt();
		fpscrRegMap=result.register; // Set mapping
	 }
	 else {
		result = new OPT_RegisterOperand(fpscrRegMap, VM_TypeReference.Int);
	 }

    VM_TypeReference psTref = VM_TypeReference.findOrCreate(VM_SystemClassLoader.getVMClassLoader(),
																				VM_Atom.findOrCreateAsciiAtom("Lorg/binarytranslator/PPC_Emulator/PPC_ProcessSpace;")
																				);
    VM_FieldReference ref = VM_MemberReference.findOrCreate(psTref,VM_Atom.findOrCreateAsciiAtom("fpscr"),
																				VM_Atom.findOrCreateAsciiAtom("I")
																				).asFieldReference();
    appendInstructionToCurrentBlock(GetField.create(GETFIELD, result,
																	 gc.makeLocal(1,psTref),
																	 new OPT_IntConstantOperand(ref.peekResolvedField().getOffset()),
																	 new OPT_LocationOperand(ref),
																	 new OPT_TrueGuardOperand())
												);

  }

  /**
   * Copy the value of the condition register into the PPC_ProcessSpace from its temporary location.
   */
  private void spillFPSCRRegister()
  {
    OPT_RegisterOperand regOp = new OPT_RegisterOperand(fpscrRegMap, VM_TypeReference.Int);

    VM_TypeReference psTref =  VM_TypeReference.findOrCreate(VM_SystemClassLoader.getVMClassLoader(),
																				 VM_Atom.findOrCreateAsciiAtom("Lorg/binarytranslator/PPC_Emulator/PPC_ProcessSpace;")
																				 );
    VM_FieldReference ref = VM_MemberReference.findOrCreate(psTref,VM_Atom.findOrCreateAsciiAtom("fpscr"),
																				VM_Atom.findOrCreateAsciiAtom("I")
																				).asFieldReference();
    appendInstructionToCurrentBlock(PutField.create(PUTFIELD, regOp,
																	 gc.makeLocal(1,psTref),
																	 new OPT_IntConstantOperand(ref.peekResolvedField().getOffset()),
																	 new OPT_LocationOperand(ref),
																	 new OPT_TrueGuardOperand())
												);
  }

  /**
   * Copy the value of the count register into its temporary location
   * from the PPC_ProcessSpace.
   */
  private void fillCTRRegister()
  {
    OPT_RegisterOperand result;
	 if (ctrRegMap == null) {
		result = gc.temps.makeTempInt();
		ctrRegMap=result.register; // Set mapping
	 }
	 else {
		result = new OPT_RegisterOperand(ctrRegMap, VM_TypeReference.Int);
	 }
    VM_TypeReference psTref = VM_TypeReference.findOrCreate(VM_SystemClassLoader.getVMClassLoader(),
																				VM_Atom.findOrCreateAsciiAtom("Lorg/binarytranslator/PPC_Emulator/PPC_ProcessSpace;")
																				);
    VM_FieldReference ref = VM_MemberReference.findOrCreate(psTref,VM_Atom.findOrCreateAsciiAtom("ctr"),
																				VM_Atom.findOrCreateAsciiAtom("I")
																				).asFieldReference();
    appendInstructionToCurrentBlock(GetField.create(GETFIELD, result,
																	 gc.makeLocal(1,psTref),
																	 new OPT_IntConstantOperand(ref.peekResolvedField().getOffset()),
																	 new OPT_LocationOperand(ref),
																	 new OPT_TrueGuardOperand())
												);
  }

  /**
   * Copy the value of the count register into the PPC_ProcessSpace
   * from its temporary location.
   */
  private void spillCTRRegister()
  {
    OPT_RegisterOperand regOp = new OPT_RegisterOperand(ctrRegMap, VM_TypeReference.Int);

    VM_TypeReference psTref =  VM_TypeReference.findOrCreate(VM_SystemClassLoader.getVMClassLoader(),
																				 VM_Atom.findOrCreateAsciiAtom("Lorg/binarytranslator/PPC_Emulator/PPC_ProcessSpace;")
																				 );
    VM_FieldReference ref = VM_MemberReference.findOrCreate(psTref,VM_Atom.findOrCreateAsciiAtom("ctr"),
																				VM_Atom.findOrCreateAsciiAtom("I")
																				).asFieldReference();
    appendInstructionToCurrentBlock(PutField.create(PUTFIELD, regOp,
																	 gc.makeLocal(1,psTref),
																	 new OPT_IntConstantOperand(ref.peekResolvedField().getOffset()),
																	 new OPT_LocationOperand(ref),
																	 new OPT_TrueGuardOperand())
												);
  }


  /**
   * Copy the value of the XER register into its temporary location
   * from the PPC_ProcessSpace.
   */
  private void fillXERRegister()
  {
    OPT_RegisterOperand xerRegMap_Op;
    OPT_RegisterOperand xerRegMap_SO_Op;
    OPT_RegisterOperand xerRegMap_OV_Op;
    OPT_RegisterOperand xerRegMap_ByteCountOp;
    OPT_RegisterOperand xerRegMap_CA_Op;
	 if (xerRegMap_SO == null) {
		xerRegMap_Op = gc.temps.makeTempInt();
		xerRegMap_SO_Op = gc.temps.makeTempInt();
		xerRegMap_OV_Op = gc.temps.makeTempInt();
		xerRegMap_ByteCountOp = gc.temps.makeTempInt();
		xerRegMap_CA_Op = gc.temps.makeTempLong();

		xerRegMap = xerRegMap_Op.register;
		xerRegMap_SO = xerRegMap_SO_Op.register;
		xerRegMap_OV = xerRegMap_OV_Op.register;
		xerRegMap_ByteCount = xerRegMap_ByteCountOp.register;
		xerRegMap_CA = xerRegMap_CA_Op.register;
	 }
	 else {
		xerRegMap_Op = new OPT_RegisterOperand(xerRegMap, VM_TypeReference.Int);
		xerRegMap_SO_Op = new OPT_RegisterOperand(xerRegMap_SO, VM_TypeReference.Int);
		xerRegMap_OV_Op = new OPT_RegisterOperand(xerRegMap_OV, VM_TypeReference.Int);
		xerRegMap_ByteCountOp = new OPT_RegisterOperand(xerRegMap_ByteCount, VM_TypeReference.Int);
		xerRegMap_CA_Op = new OPT_RegisterOperand(xerRegMap_CA, VM_TypeReference.Long);
	 }

    VM_TypeReference psTref = VM_TypeReference.findOrCreate(VM_SystemClassLoader.getVMClassLoader(),
																				VM_Atom.findOrCreateAsciiAtom("Lorg/binarytranslator/PPC_Emulator/PPC_ProcessSpace;")
																				);
    VM_FieldReference ref = VM_MemberReference.findOrCreate(psTref,VM_Atom.findOrCreateAsciiAtom("xer"),
																				VM_Atom.findOrCreateAsciiAtom("I")
																				).asFieldReference();
    appendInstructionToCurrentBlock(GetField.create(GETFIELD, xerRegMap_Op,
																	 gc.makeLocal(1,psTref),
																	 new OPT_IntConstantOperand(ref.peekResolvedField().getOffset()),
																	 new OPT_LocationOperand(ref),
																	 new OPT_TrueGuardOperand())
												);
    // Copy SO bit into XER[SO] register
    appendInstructionToCurrentBlock(Binary.create(INT_AND, xerRegMap_SO_Op,
																  xerRegMap_Op.copyRO(),
																  new OPT_IntConstantOperand(1 << 31)));

    // Copy OV bit into XER[OV] register
    appendInstructionToCurrentBlock(Binary.create(INT_AND, xerRegMap_OV_Op,
																  xerRegMap_Op.copyRO(),
																  new OPT_IntConstantOperand(1 << 30)));

    // Copy byte count bits out
    appendInstructionToCurrentBlock(Binary.create(INT_AND, xerRegMap_ByteCountOp,
																  xerRegMap_Op.copyRO(),
																  new OPT_IntConstantOperand(0x3f)));

    // Copy CA bit into XER[CA] register
    appendInstructionToCurrentBlock(Unary.create(INT_2LONG, xerRegMap_CA_Op, xerRegMap_Op.copyRO()));
    appendInstructionToCurrentBlock(Binary.create(LONG_SHL, xerRegMap_CA_Op.copyRO(),
																  xerRegMap_CA_Op.copyRO(),
																  new OPT_IntConstantOperand(3)));
  }

  /**
   * Copy the value of the XER register into the PPC_ProcessSpace from
   * its temporary location.
   */
  private void spillXERRegister()
  {
    // Set up xer register with byte count
    OPT_RegisterOperand  xer = new OPT_RegisterOperand(xerRegMap, VM_TypeReference.Int);
    appendInstructionToCurrentBlock(Move.create(INT_MOVE, xer,
																new OPT_RegisterOperand(xerRegMap_ByteCount, VM_TypeReference.Int))
												);
    // SO bit
    OPT_RegisterOperand tempInt = getTempInt(0);
    appendInstructionToCurrentBlock(CondMove.create(INT_COND_MOVE, 
																	 tempInt,
																	 new OPT_RegisterOperand(xerRegMap_SO, VM_TypeReference.Int),
																	 new OPT_IntConstantOperand(0),
																	 OPT_ConditionOperand.EQUAL(),
																	 new OPT_IntConstantOperand(0),
																	 new OPT_IntConstantOperand(1 << 31)));
    appendInstructionToCurrentBlock(Binary.create(INT_OR, xer.copyRO(),
																  xer.copy(),
																  tempInt.copyRO())
												);
    // OV bit
    appendInstructionToCurrentBlock(CondMove.create(INT_COND_MOVE, 
																	 tempInt.copyRO(),
																	 new OPT_RegisterOperand(xerRegMap_OV, VM_TypeReference.Int),
																	 new OPT_IntConstantOperand(0),
																	 OPT_ConditionOperand.EQUAL(),
																	 new OPT_IntConstantOperand(0),
																	 new OPT_IntConstantOperand(1 << 30)));
    appendInstructionToCurrentBlock(Binary.create(INT_OR, xer.copyRO(),
																  xer.copy(),
																  tempInt.copyRO())
												);
	 
    // CA bit
    OPT_RegisterOperand xerRegMap_CA_Op = new OPT_RegisterOperand(xerRegMap_CA, VM_TypeReference.Long);
	 appendInstructionToCurrentBlock(Binary.create(LONG_USHR, xerRegMap_CA_Op,
																  xerRegMap_CA_Op.copy(),
																  new OPT_IntConstantOperand(3)));
    appendInstructionToCurrentBlock(Unary.create(LONG_2INT, tempInt.copyRO(), xerRegMap_CA_Op.copy()));
    appendInstructionToCurrentBlock(Binary.create(INT_AND, tempInt.copyRO(),
																  tempInt.copyRO(),
																  new OPT_IntConstantOperand(1<<29)));
    appendInstructionToCurrentBlock(Binary.create(INT_OR, xer.copyRO(),
																  xer.copy(),
																  tempInt.copyRO())
												);
	 
    // Store to process space
    VM_TypeReference psTref =  VM_TypeReference.findOrCreate(VM_SystemClassLoader.getVMClassLoader(),
																				 VM_Atom.findOrCreateAsciiAtom("Lorg/binarytranslator/PPC_Emulator/PPC_ProcessSpace;")
																				 );
    VM_FieldReference ref = VM_MemberReference.findOrCreate(psTref,VM_Atom.findOrCreateAsciiAtom("xer"),
																				VM_Atom.findOrCreateAsciiAtom("I")
																				).asFieldReference();
    appendInstructionToCurrentBlock(PutField.create(PUTFIELD, xer.copy(),
																	 gc.makeLocal(1,psTref),
																	 new OPT_IntConstantOperand(ref.peekResolvedField().getOffset()),
																	 new OPT_LocationOperand(ref),
																	 new OPT_TrueGuardOperand())
												);
  }


  /**
   * Copy the value of the LR register into its temporary location
   * from the PPC_ProcessSpace.
   */
  private void fillLRRegister()
  {
    OPT_RegisterOperand result;
	 if (lrRegMap == null) {
		result = gc.temps.makeTempInt();
		lrRegMap = result.register; // Set mapping
	 }
	 else {
		result = new OPT_RegisterOperand(lrRegMap, VM_TypeReference.Int);
	 }
    VM_TypeReference psTref = VM_TypeReference.findOrCreate(VM_SystemClassLoader.getVMClassLoader(),
																				VM_Atom.findOrCreateAsciiAtom("Lorg/binarytranslator/PPC_Emulator/PPC_ProcessSpace;")
																				);
    VM_FieldReference ref = VM_MemberReference.findOrCreate(psTref,VM_Atom.findOrCreateAsciiAtom("lr"),
																				VM_Atom.findOrCreateAsciiAtom("I")
																				).asFieldReference();
    appendInstructionToCurrentBlock(GetField.create(GETFIELD, result,
																	 gc.makeLocal(1,psTref),
																	 new OPT_IntConstantOperand(ref.peekResolvedField().getOffset()),
																	 new OPT_LocationOperand(ref),
																	 new OPT_TrueGuardOperand())
												);
  }

  /**
   * Copy the value of the LR register into the PPC_ProcessSpace from
   * its temporary location.
   */
  private void spillLRRegister()
  {
    OPT_RegisterOperand regOp = new OPT_RegisterOperand(lrRegMap, VM_TypeReference.Int);

    VM_TypeReference psTref =  VM_TypeReference.findOrCreate(VM_SystemClassLoader.getVMClassLoader(),
																				 VM_Atom.findOrCreateAsciiAtom("Lorg/binarytranslator/PPC_Emulator/PPC_ProcessSpace;")
																				 );
    VM_FieldReference ref = VM_MemberReference.findOrCreate(psTref,VM_Atom.findOrCreateAsciiAtom("lr"),
																				VM_Atom.findOrCreateAsciiAtom("I")
																				).asFieldReference();
    appendInstructionToCurrentBlock(PutField.create(PUTFIELD, regOp,
																	 gc.makeLocal(1,psTref),
																	 new OPT_IntConstantOperand(ref.peekResolvedField().getOffset()),
																	 new OPT_LocationOperand(ref),
																	 new OPT_TrueGuardOperand())
												);
  }

  /**
   * Fill all the registers from the PPC_ProcessSpace.
   */
  private void fillAllRegisters() {
	 for(int r = 0 ; r < 32 ; r++) {
		fillGPRegister(r);
	 }
                  
	 for(int r = 0 ; r < 32 ; r++) {
		fillFPRegister(r);
	 }

	 fillCRRegister();

	 fillCTRRegister();

	 fillXERRegister();

	 fillLRRegister();
                        
	 fillFPSCRRegister();
  }

  /**
   * Spill all the registers
   */
  private void spillAllRegisters() {
	 for(int r = 0 ; r < 32 ; r++) {
		spillGPRegister(r);
	 }
	 for(int r = 0 ; r < 32 ; r++) {
		spillFPRegister(r);
	 }
		
	 spillCRRegister();
		
	 spillCTRRegister();
		
	 spillXERRegister();
		
	 spillLRRegister();
		
	 spillFPSCRRegister();
  }

  // -oO Find a register Oo-

  /**
   * Return the HIR OPT_RegisterOperand (temporary register)
   * corresponding to a general purpose PPC register.
   * @param r the number of the PPC GP register
   */
  public OPT_RegisterOperand getGPRegister(int r) {
    if (VM.VerifyAssertions) VM._assert(r < 32);
	 intRegInUseMap[r] = true;
    return new OPT_RegisterOperand (intRegMap[r], VM_TypeReference.Int);
  }

  /**
   * Return the HIR OPT_RegisterOperand (temporary register)
   * corresponding to a floating point PPC register
   * @param r the number of the PPC FP register
   */
  public OPT_RegisterOperand getFPRegister(int r) {
    if (VM.VerifyAssertions) VM._assert(r < 32);
	 fpRegInUseMap[r] = true;
    return new OPT_RegisterOperand (fpRegMap[r], VM_TypeReference.Double);
  }

  /**
   * Return the HIR OPT_RegisterOperand (temporary register)
   * corresponding to the PPC condition register. NB Laziness should
   * be checked to see if all bits in the register are valid.
   */
  public OPT_RegisterOperand getCRRegister() {
	 crRegInUse = true;
    return new OPT_RegisterOperand (crRegMap, VM_TypeReference.Int);
  }

  /**
   * Return the HIR OPT_RegisterOperand (temporary register)
   * corresponding to a register holding a copy of an operand to a PPC
   * compare instruction
   */
  public OPT_RegisterOperand getCR_LtGtEq_rA_Register(int field) {
    if (VM.VerifyAssertions) VM._assert(field < 8);
    return new OPT_RegisterOperand (crFieldMap_LtGtEq_rA[field], VM_TypeReference.Int);
  }

  /**
   * Return the HIR OPT_RegisterOperand (temporary register)
   * corresponding to a register holding a copy of an operand to a PPC
   * compare instruction
   */
  public OPT_RegisterOperand getCR_LtGtEq_rB_Register(int field) {
    if (VM.VerifyAssertions) VM._assert(field < 8);
    return new OPT_RegisterOperand (crFieldMap_LtGtEq_rB[field], VM_TypeReference.Int);
  }

  /**
   * Return the HIR OPT_RegisterOperand (temporary register)
   * corresponding to a register holding a copy of an operand to a PPC
   * fp compare instruction
   */
  public OPT_RegisterOperand getCR_LtGtEq_frA_Register(int field) {
    if (VM.VerifyAssertions) VM._assert(field < 8);
    return new OPT_RegisterOperand (crFieldMap_LtGtEq_frA[field], VM_TypeReference.Double);
  }

  /**
   * Return the HIR OPT_RegisterOperand (temporary register)
   * corresponding to a register holding a copy of an operand to a PPC
   * fp compare instruction
   */
  public OPT_RegisterOperand getCR_LtGtEq_frB_Register(int field) {
    if (VM.VerifyAssertions) VM._assert(field < 8);
    return new OPT_RegisterOperand (crFieldMap_LtGtEq_frB[field], VM_TypeReference.Double);
  }

  /**
   * Return the HIR OPT_RegisterOperand (temporary register)
   * corresponding to a register holding a copy of the XER[SO] bit at
   * the point of a PPC compare instruction
   */
  public OPT_RegisterOperand getCR_SO_Register(int field) {
    if (VM.VerifyAssertions) VM._assert(field < 8);
    return new OPT_RegisterOperand (crFieldMap_SO[field], VM_TypeReference.Int);
  }

  /**
   * Return the HIR OPT_RegisterOperand (temporary register)
   * corresponding to the PPC fpscr register.
   */
  public OPT_RegisterOperand getFPSCRRegister() {
	 fpscrRegInUse = true;
    return new OPT_RegisterOperand (fpscrRegMap, VM_TypeReference.Int);
  }

  /**
   * Return the HIR OPT_RegisterOperand (temporary register)
   * corresponding to the PPC ctr register
   */
  public OPT_RegisterOperand getCTRRegister() {
	 ctrRegInUse = true;
    return new OPT_RegisterOperand (ctrRegMap, VM_TypeReference.Int);
  }

  /**
   * Return the HIR OPT_RegisterOperand (temporary register)
   * corresponding to the PPC xer register byte count bits
   */
  public OPT_RegisterOperand getXER_ByteCountRegister() {
	 xerRegInUse = true;
    return new OPT_RegisterOperand (xerRegMap_ByteCount, VM_TypeReference.Int);
  }

  /**
   * Return the HIR OPT_RegisterOperand (temporary register)
   * corresponding to the PPC xer register SO bit
   */
  public OPT_RegisterOperand getXER_SO_Register() {
	 xerRegInUse = true;
    return new OPT_RegisterOperand (xerRegMap_SO, VM_TypeReference.Int);
  }

  /**
   * Return the HIR OPT_RegisterOperand (temporary register)
   * corresponding to the PPC xer register OV bit
   */
  public OPT_RegisterOperand getXER_OV_Register() {
	 xerRegInUse = true;
    return new OPT_RegisterOperand (xerRegMap_OV, VM_TypeReference.Int);
  }

  /**
   * Return the HIR OPT_RegisterOperand (temporary register)
   * corresponding to the PPC xer register CA bit
   */
  public OPT_RegisterOperand getXER_CA_Register() {
	 xerRegInUse = true;
    return new OPT_RegisterOperand (xerRegMap_CA, VM_TypeReference.Long);
  }

  /**
   * Return the HIR OPT_RegisterOperand (temporary register)
   * corresponding to the PPC lr register.
   */
  public OPT_RegisterOperand getLRRegister() {
	 lrRegInUse = true;
    return new OPT_RegisterOperand (lrRegMap, VM_TypeReference.Int);
  }

  // -oO Temporaries used during translation Oo-

  /**
   * Temporary int variables
   */
  private OPT_Register intTemps[];

  /**
   * Get/create a temporary int
   * @param num a hint to allow for reuse of temps across instructions
   */
  public OPT_RegisterOperand getTempInt(int num) {
	 if (intTemps == null) {
		intTemps = new OPT_Register[10];
	 }
    OPT_Register result = intTemps[num];
    if(result == null) {
      OPT_RegisterOperand regOp = gc.temps.makeTempInt();
      intTemps[num] = regOp.register;
      return regOp;
    }
    else {
      return new OPT_RegisterOperand(result, VM_TypeReference.Int);
    }
  }

  /**
   * Temporary long variables
   */
  private OPT_Register longTemps[];

  /**
   * Get/create a temporary long
   * @param num a hint to allow for reuse of temps across instructions
   */
  public OPT_RegisterOperand getTempLong(int num) {
	 if (longTemps == null) {
		longTemps = new OPT_Register[10];
	 }
    OPT_Register result = longTemps[num];
    if(result == null) {
      OPT_RegisterOperand regOp = gc.temps.makeTempLong();
      longTemps[num] = regOp.register;
      return regOp;
    }
    else {
      return new OPT_RegisterOperand(result, VM_TypeReference.Long);
    }
  }

  /**
   * Temporary intArray variables
   */

  private OPT_Register intArrayTemp;

  /**
   * Get/create a temporary intArray
   * @param num a hint to allow for reuse of temps across instructions
   */
  public OPT_RegisterOperand getTempIntArray(int num) {
    if (VM.VerifyAssertions) VM._assert(num == 0);

    OPT_Register result = intArrayTemp;
    if(result == null) {
      OPT_RegisterOperand regOp = gc.temps.makeTemp(VM_TypeReference.IntArray);
      intArrayTemp = regOp.register;
      return regOp;
    }
    else {
      return new OPT_RegisterOperand(result, VM_TypeReference.IntArray);
    }
  }

  /**
   * Temporary float variables
   */

  private OPT_Register floatTemps[];

  /**
   * Get/create a temporary float
   * @param num a hint to allow for reuse of temps across instructions
   */
  public OPT_RegisterOperand getTempFloat(int num) {
    if (VM.VerifyAssertions) VM._assert((num == 0)||(num==1));
	 if (floatTemps == null) {
		floatTemps = new OPT_Register[2];
	 }
    OPT_Register result = floatTemps[num];
    if(result == null) {
      OPT_RegisterOperand regOp = gc.temps.makeTempFloat();
      floatTemps[num] = regOp.register;
      return regOp;
    }
    else {
      return new OPT_RegisterOperand(result, VM_TypeReference.Float);
    }
  }

  /**
   * Temporary Double variables
   */

  private OPT_Register doubleTemp;

  /**
   * Get/create a temporary float
   * @param num a hint to allow for reuse of temps across instructions
   */
  public OPT_RegisterOperand getTempDouble(int num) {
    if (VM.VerifyAssertions) VM._assert(num == 0);
    OPT_Register result = doubleTemp;
    if(result == null) {
      OPT_RegisterOperand regOp = gc.temps.makeTempDouble();
      doubleTemp = regOp.register;
      return regOp;
    }
    else {
      return new OPT_RegisterOperand(result, VM_TypeReference.Double);
    }
  }

  /**
   * Temporary validation variables
   */

  private OPT_Register validationTemp;

  /**
   * Get/create a temporary validation variable
   * @param num a hint to allow for reuse of temps across instructions
   */
  public OPT_RegisterOperand getTempValidation(int num) {
    if (VM.VerifyAssertions) VM._assert(num == 0);

    OPT_Register result = validationTemp;
    if(result == null) {
      OPT_RegisterOperand regOp = gc.temps.makeTempValidation();
      validationTemp = regOp.register;
      return regOp;
    }
    else {
      return new OPT_RegisterOperand(result, VM_TypeReference.VALIDATION_TYPE);
    }
  }

  // -oO Lazy state Oo-
  /**
   * A representation of the lazy state. This and a PC address define
   * what a block of code is good for.
   */
  public final class Laziness {

    /**
     * What the state of CR lazy evaluation is:
     * <ul>
     * <li>0 - (default) the field hasn't been defined in this trace
     * so use the bit field in the CR register</li>
     * <li>1 - the crFieldMap_LtGtEq_rA and crFieldMap_LtGtEq_rB
     * values correspond to the result of a signed comparison</li>
     * <li>2 - the crFieldMap_LtGtEq_rA and crFieldMap_LtGtEq_rB
     * values correspond to the result of a unsigned comparison</li>
     * <li>3 - the condition register has been set following a
     * floating-point compare unordered</li>
	  * <li>4 to 15 - room for expansion</li>
     * </ul>
     */
    private int crLaziness;
    public static final int NOT_LAZY = 0;
    public static final int SIGNED_INT_CMP = 1;
    public static final int UNSIGNED_INT_CMP = 2;
    public static final int FP_CMPU = 3;
	 /**
	  * Which CR field is the same as the FPRF. -1 => not in use
	  */
	 private int crf_to_fprf_map;
    /**
     * Default constructor - nothing is lazy yet
     */
    Laziness() {
		crf_to_fprf_map = -1;
    }
    /**
     * Copying constructor
     */
    Laziness(int _crLaziness, int _crf_to_fprf_map) {
      crLaziness = _crLaziness;
		crf_to_fprf_map = _crf_to_fprf_map;
    }
    /**
     * Clone the object
     */
    public Object clone() {
      return new Laziness(crLaziness, crf_to_fprf_map);
	 }
    /**
     * Do these two lazy objects encode a sympathetic lazy state
     * @param other the other lazy object
     * @return true if lazy states are equivalent
     */
    boolean equivalent(Laziness other) {
      return crLaziness == other.crLaziness;
    }
    /**
     * Set the laziness of a condition register field
     * @param field the field we're concerned with
     * @param lazy the laziness of that field
     */
    public void setConditionFieldLaziness(int field, int lazy) {
		if (VM.VerifyAssertions) VM._assert((field >= 0) && (field < 8));
      crLaziness &= ~(0xf0000000 >>> (field * 4));
      switch(lazy){
      case 0:
		  break;
      case 1:
		  crLaziness |= 0x10000000 >>> (field * 4);
		  break;
      case 2:
		  crLaziness |= 0x20000000 >>> (field * 4);
		  break;
      case 3:
		  crLaziness |= 0x30000000 >>> (field * 4);
		  break;
      default:
		  throw new Error("Invalid lazy state encountered");
      }
    }
    /**
     * Get the laziness of a condition register field
     * @param field the field we're concerned with
     * @return its current lazy encoding
     */
    public int getConditionFieldLaziness(int field) {
      return (crLaziness >>> ((7-field) * 4)) & 0x3;
    }
	 /**
	  * Set the mapping from a condition register field to the FPRF
	  * bits of the FPSCR register
	  * @param condition register field holding the FPRF condition codes
	  */
	 public void setCrfToFprfMap(int crf) {
		if (VM.VerifyAssertions) VM._assert((crf >= 0) && (crf < 8) && (getConditionFieldLaziness(crf) == FP_CMPU));
		crf_to_fprf_map = crf;
	 }
	 /**
	  * Get the mapping from a condition register field to the FPRF
	  * bits of the FPSCR register
	  */
	 int getCrfToFprfMap() {
		return crf_to_fprf_map;
	 }
    /**
     * Generate a key value encoding this laziness and a PC value
     * @parm pc the PC value we're trying to make a key for
     */

    public Object makeKey(int pc) {
		class Key {
		  int val1, val2, val3;
		  public int hashCode() {
			 return val1;
		  }
		  public boolean equals(Object o) {
			 return (o instanceof Key) &&
				(((Key)o).val1 == val1) &&
				(((Key)o).val2 == val2) &&
				(((Key)o).val3 == val3);
		  }
		  Key (int v1, int v2, int v3) {
			 val1 = v1;
			 val2 = v2;
			 val3 = v3;
		  }
		  public String toString() {
			 return "0x" + Integer.toHexString(val1) + ".0x" + Integer.toHexString(val2) + "." + val3;
		  }
		}
		return new Key(pc, crLaziness, crf_to_fprf_map);
		/*
		  if((crf_to_fprf_map >= 0)||(crf_to_fprf_map < 8)) {
		  return Long.toHexString(((long)crLaziness << 32) | ((long)pc)) + "." + Integer.toString(crf_to_fprf_map);
		  }
		  else {
		  return Long.toHexString(((long)crLaziness << 32) | ((long)pc)) + ".FF";
		  }
		*/
		//return new Long(((long)crLaziness << 32) | ((long)pc));
	 }
  }

  /**
   * Plant instructions modifying a lazy state into one with no
   * laziness for the specified condition register field
   * @param laziness the laziness to modify
	* @param crf the condition register field
   */
  public void resolveLazinessCrField(Laziness laziness, int crf) {
	 if (PPC_Options.debugLazy) {
		VM.sysWriteln("Resolving laziness in " + currentBlock + " on field " + crf + " of laziness " + laziness.getConditionFieldLaziness(crf));
	 }
	 switch(laziness.getConditionFieldLaziness(crf)) {
	 case Laziness.NOT_LAZY:
		// Nothing to do
		break;
	 case Laziness.SIGNED_INT_CMP:
		// Resolve using signed comparison
		{
		  OPT_RegisterOperand cr = getCRRegister();
		  // mask out bits we're setting
		  appendInstructionToCurrentBlock(Binary.create(INT_AND, cr,
																		cr.copyRO(),
																		new OPT_IntConstantOperand(~(0xf0000000 >>> (crf*4)))));
		  OPT_RegisterOperand cr_lt = getTempInt(0);
		  OPT_RegisterOperand cr_gt = getTempInt(1);
		  OPT_RegisterOperand cr_eq = getTempInt(2);
		  appendInstructionToCurrentBlock(Binary.create(INT_OR, cr_lt,
																		cr.copyRO(),
																		new OPT_IntConstantOperand(0x80000000 >>> (crf*4))));
		  appendInstructionToCurrentBlock(Binary.create(INT_OR, cr_gt,
																		cr.copyRO(),
																		new OPT_IntConstantOperand(0x40000000 >>> (crf*4))));
		  appendInstructionToCurrentBlock(Binary.create(INT_OR, cr_eq,
																		cr.copyRO(),
																		new OPT_IntConstantOperand(0x20000000 >>> (crf*4))));
		  appendInstructionToCurrentBlock(CondMove.create(INT_COND_MOVE, cr.copyRO(),
																		  getCR_LtGtEq_rA_Register(crf),
																		  getCR_LtGtEq_rB_Register(crf),
																		  OPT_ConditionOperand.LESS(),
																		  cr_lt.copyRO(),
																		  cr_eq.copyRO()));
		  appendInstructionToCurrentBlock(CondMove.create(INT_COND_MOVE, cr.copyRO(),
																		  getCR_LtGtEq_rA_Register(crf),
																		  getCR_LtGtEq_rB_Register(crf),
																		  OPT_ConditionOperand.GREATER(),
																		  cr_gt.copyRO(),
																		  cr.copyRO()));
		  OPT_RegisterOperand cr_so = getTempInt(0);
		  appendInstructionToCurrentBlock(Binary.create(INT_OR, cr_so,
																		cr.copyRO(),
																		new OPT_IntConstantOperand(0x10000000 >>> (crf*4))));
		  appendInstructionToCurrentBlock(CondMove.create(INT_COND_MOVE, cr.copyRO(),
																		  getCR_SO_Register(crf),
																		  new OPT_IntConstantOperand(0),
																		  OPT_ConditionOperand.EQUAL(),
																		  cr.copyRO(),
																		  cr_so.copyRO()));			 
		}
		break;
	 case Laziness.UNSIGNED_INT_CMP:
		{
		  OPT_RegisterOperand cr = getCRRegister();
		  // mask out bits we're setting
		  appendInstructionToCurrentBlock(Binary.create(INT_AND, cr,
																		cr.copyRO(),
																		new OPT_IntConstantOperand(~(0xf0000000 >>> (crf*4)))));
		  OPT_RegisterOperand cr_lt = getTempInt(0);
		  OPT_RegisterOperand cr_gt = getTempInt(1);
		  OPT_RegisterOperand cr_eq = getTempInt(2);
		  appendInstructionToCurrentBlock(Binary.create(INT_OR, cr_lt,
																		cr.copyRO(),
																		new OPT_IntConstantOperand(0x80000000 >>> (crf*4))));
		  appendInstructionToCurrentBlock(Binary.create(INT_OR, cr_gt,
																		cr.copyRO(),
																		new OPT_IntConstantOperand(0x40000000 >>> (crf*4))));
		  appendInstructionToCurrentBlock(Binary.create(INT_OR, cr_eq,
																		cr.copyRO(),
																		new OPT_IntConstantOperand(0x20000000 >>> (crf*4))));
		  appendInstructionToCurrentBlock(CondMove.create(INT_COND_MOVE, cr.copyRO(),
																		  getCR_LtGtEq_rA_Register(crf),
																		  getCR_LtGtEq_rB_Register(crf),
																		  OPT_ConditionOperand.LOWER(),
																		  cr_lt.copyRO(),
																		  cr_eq.copyRO()));
		  appendInstructionToCurrentBlock(CondMove.create(INT_COND_MOVE, cr.copyRO(),
																		  getCR_LtGtEq_rA_Register(crf),
																		  getCR_LtGtEq_rB_Register(crf),
																		  OPT_ConditionOperand.HIGHER(),
																		  cr_gt.copyRO(),
																		  cr.copyRO()));
		  OPT_RegisterOperand cr_so = getTempInt(0);
		  appendInstructionToCurrentBlock(Binary.create(INT_OR, cr_so,
																		cr.copyRO(),
																		new OPT_IntConstantOperand(0x10000000 >>> (crf*4))));
		  appendInstructionToCurrentBlock(CondMove.create(INT_COND_MOVE, cr.copyRO(),
																		  getCR_SO_Register(crf),
																		  new OPT_IntConstantOperand(0),
																		  OPT_ConditionOperand.EQUAL(),
																		  cr.copyRO(),
																		  cr_so.copyRO()));
			 
		}
		break;
	 case Laziness.FP_CMPU:
		// Resolve using unsigned comparison
		{
		  OPT_RegisterOperand frA = getCR_LtGtEq_frA_Register(crf);
		  OPT_RegisterOperand frB = getCR_LtGtEq_frB_Register(crf);
		  OPT_RegisterOperand cr = getCRRegister();
		  OPT_RegisterOperand cr_nan = getTempInt(0);
		  OPT_RegisterOperand cr_other = getTempInt(1); // unordered
		  OPT_RegisterOperand cr_lt  = getTempInt(2);
		  OPT_RegisterOperand cr_gt  = getTempInt(3);

		  // mask out bits we're setting for nan and other (==) cases
		  appendInstructionToCurrentBlock(Binary.create(INT_AND, cr,
																		cr.copyRO(),
																		new OPT_IntConstantOperand(~(0xf0000000 >>> (crf*4)))));
		  appendInstructionToCurrentBlock(Binary.create(INT_OR, cr_nan,
																		cr.copyRO(),
																		new OPT_IntConstantOperand(0x10000000 >>> (crf*4))));
		  appendInstructionToCurrentBlock(Binary.create(INT_OR, cr_other,
																		cr.copyRO(),
																		new OPT_IntConstantOperand(0x20000000 >>> (crf*4))));
		  // mask out bits we're setting for < and > cases
		  appendInstructionToCurrentBlock(Binary.create(INT_OR, cr_lt,
																		cr.copyRO(),
																		new OPT_IntConstantOperand(0x80000000 >>> (crf*4))));
		  appendInstructionToCurrentBlock(Binary.create(INT_OR, cr_gt,
																		cr.copyRO(),
																		new OPT_IntConstantOperand(0x40000000 >>> (crf*4))));

		  // Test for NaN (frA != frA)
		  // CR = (frA == NaN) ? CR_other : CR_nan
		  appendInstructionToCurrentBlock(CondMove.create(INT_COND_MOVE, cr.copyRO(),
																		  frA,
																		  frA.copyRO(),
																		  OPT_ConditionOperand.EQUAL(),
																		  cr_other.copyRO(),
																		  cr_nan.copyRO()));
		  // Test for NaN (frB != frB)
		  // CR = (frB == NaN) ? CR: CR_nan
		  appendInstructionToCurrentBlock(CondMove.create(INT_COND_MOVE, cr.copyRO(),
																		  frB,
																		  frB.copyRO(),
																		  OPT_ConditionOperand.EQUAL(),
																		  cr.copyRO(),
																		  cr_nan.copyRO()));
		  // Test for frA < frB
		  // CR = (frA < frB) ? CR_lt: CR
		  appendInstructionToCurrentBlock(CondMove.create(INT_COND_MOVE, cr.copyRO(),
																		  frA.copyRO(),
																		  frB.copyRO(),
																		  OPT_ConditionOperand.GREATER_EQUAL(),
																		  cr.copyRO(),
																		  cr_lt.copyRO()));
		  // Test for frA > frB
		  // CR = (frA > frB) ? CR_gt: CR
		  appendInstructionToCurrentBlock(CondMove.create(INT_COND_MOVE, cr.copyRO(),
																		  frA.copyRO(),
																		  frB.copyRO(),
																		  OPT_ConditionOperand.GREATER(),
																		  cr_gt.copyRO(),
																		  cr.copyRO()));
		
		  // Should we also resolve FPCC bits?
		  if(laziness.getCrfToFprfMap() == crf) {
			 OPT_RegisterOperand fpcc = getTempInt(0);
			 // Get condition code from CR
			 appendInstructionToCurrentBlock(Binary.create(INT_AND, fpcc,
																		  cr.copyRO(),
																		  new OPT_IntConstantOperand(0xf0000000 >>> (crf*4))));
			 // Shift into place (FPCC is at 16-19)
			 if((crf*4) < 16) {
				appendInstructionToCurrentBlock(Binary.create(INT_USHR, fpcc.copyRO(),
																			 fpcc.copyRO(),
																			 new OPT_IntConstantOperand(16-(crf*4))));
			 }
			 else if ((crf * 4) > 16) {
				appendInstructionToCurrentBlock(Binary.create(INT_SHL, fpcc.copyRO(),
																			 fpcc.copyRO(),
																			 new OPT_IntConstantOperand((crf*4)-16)));
			 }
			 // Clear FPCC bits in FPSCR
			 OPT_RegisterOperand fpscr = getFPSCRRegister();
			 appendInstructionToCurrentBlock(Binary.create(INT_AND, fpscr,
																		  fpscr.copyRO(),
																		  new OPT_IntConstantOperand(0xffff0fff)));				
			 // Set to condition bits
			 appendInstructionToCurrentBlock(Binary.create(INT_OR, fpscr.copyRO(),
																		  fpscr.copyRO(),
																		  fpcc.copyRO()));
		  }
		}
		break;
	 default:
		throw new Error("Unhandled lazy state " + laziness.getConditionFieldLaziness(crf));
	 }
	 laziness.setConditionFieldLaziness(crf, Laziness.NOT_LAZY);
  }

  /**
   * Plant instructions modifying a lazy state into one with no
   * laziness
   * @param laziness the laziness to modify
   */
  public void resolveLaziness(Laziness laziness) {
    for(int crf=0; crf<8; crf++) {
		resolveLazinessCrField(laziness, crf);
	 }
  }

  /**
   * Plant instructions modifying a lazy state into one with no
   * laziness for the specified bit
   * @param laziness the laziness to modify
	* @param bit the bit required to be resolved
	*/
  public void resolveLazinessCrBit(Laziness laziness, int bit) {
    int crf=bit >> 2;
	 resolveLazinessCrField(laziness, crf);
  }

  // -oO PPC laziness + PC to HIR basic block map Oo-
  /** 
   * Map to locate HIR basic blocks to re-use translation within a
   * trace
   */
  private HashMap blockMap;

  /**
   * Register a mapping between a pc and lazy and a hir block
   * @param pc
   * @param lazy
   * @param hirBlock
   */
  private void registerMapping(int pc, Laziness lazy, OPT_BasicBlock hirBlock) {
    blockMap.put(lazy.makeKey(pc), hirBlock);
  }
  /**
   * Find if there's already a translation for a given pc and laziness
   * @param pc
   * @param lazy
   * @return basic block or null if no translation exists
   */
  public OPT_BasicBlock findMapping(int pc, Laziness lazy) {
    return (OPT_BasicBlock)blockMap.get(lazy.makeKey(pc));
  }

  // -oO Handling of instructions we can't initially resolve the branch targets for Oo-
  /**
   * List of unresolved IfCmp instructions
   */
  private Vector unresolvedIfCmp;
  /**
   * List of where unresolved IfCmp instructions are trying to go
   */
  private Vector unresolvedIfCmp_PC;
  /**
   * List of what the unresolved IfCmp instruction's target laziness should be
   */
  private Vector unresolvedIfCmp_Laziness;

  /**
   * Register a HIR IfCmp instruction whose target can't be resolved yet
   * @param ifcmp_instr
   * @param targetPC
   * @param targetLaziness
   */
  void registerIfCmpTargetUnresolved(OPT_Instruction ifcmp_instr, int targetPC, Laziness targetLaziness){
    unresolvedIfCmp.add(ifcmp_instr);
    unresolvedIfCmp_PC.add(new Integer(targetPC));
    unresolvedIfCmp_Laziness.add(targetLaziness);
  }

  /**
   * Are all IfCmp instructions resolved
   * @return true if we need to resolve IfCmps
   */
  private boolean areIfCmpsResolved() {
    return unresolvedIfCmp.size() == 0;
  }
  /**
   * Resolve an ifCmp instruction
   */
  private void resolveIfCmp() {
	 if (areIfCmpsResolved() == false) {
		// Try to find if block has now been translated
		int toRemove = unresolvedIfCmp.size() - 1;
		int pc = ((Integer)unresolvedIfCmp_PC.get(toRemove)).intValue();
		Laziness lazy = (Laziness)unresolvedIfCmp_Laziness.get(toRemove);
		OPT_BasicBlock targetBB = findMapping(pc, lazy);
		if (targetBB == null) {
		  // Block not translated so translate
		  translateSubTrace(lazy, pc);
		  targetBB = findMapping(pc, lazy);
		}
		// Fix up instruction
		OPT_Instruction ifCmp = (OPT_Instruction)unresolvedIfCmp.get(toRemove);
		IfCmp.setTarget(ifCmp, targetBB.makeJumpTarget());
		ifCmp.getBasicBlock().insertOut(targetBB);
		// Remove from unresolved list
		unresolvedIfCmp.remove(toRemove);
		unresolvedIfCmp_PC.remove(toRemove);
		unresolvedIfCmp_Laziness.remove(toRemove);
	 }
  }

  /**
   * List of unresolved Goto instructions
   */
  private Vector unresolvedGoto;
  /**
   * List of where unresolved Goto instructions are trying to go
   */
  private Vector unresolvedGoto_PC;
  /**
   * List of what the unresolved Goto instruction's target laziness should be
   */
  private Vector unresolvedGoto_Laziness;

  /**
   * Register a HIR Goto instruction whose target can't be resolved
   * yet. There's a caveat on using this that there are no other out
   * edges for this BB.
   * @param goto_instr
   * @param targetPC
   * @param targetLaziness
   */
  public void registerGotoTargetUnresolved(OPT_Instruction goto_instr, int targetPC, Laziness targetLaziness){
    unresolvedGoto.add(goto_instr);
    unresolvedGoto_PC.add(new Integer(targetPC));
    unresolvedGoto_Laziness.add(targetLaziness);
  }

  /**
   * Are all Goto instructions resolved
   * @return true if we need to resolve Gotos
   */
  private boolean areGotosResolved() {
    return unresolvedGoto.size() == 0;
  }
  /**
   * Resolve a goto instruction
   */
  private void resolveGoto() {
	 if (areGotosResolved() == false) {
		// Try to find if block has now been translated
		int toRemove = unresolvedGoto.size() - 1;
		int pc = ((Integer)unresolvedGoto_PC.get(toRemove)).intValue();
		Laziness lazy = (Laziness)unresolvedGoto_Laziness.get(toRemove);
		OPT_BasicBlock targetBB;
		if (PPC_Options.singleInstrTranslation == false) {
		  targetBB = findMapping(pc, lazy);
		}
		else {
		  // Single instruction mode so resolve gotos to a block that resolves the lazy state and exits
		  if (currentBlock.getNumberOfRealInstructions() != 0) {
			 currentBlock = createBlockAfterCurrentNotInCFG();
		  }
		  targetBB = currentBlock;
		  setReturnValueResolveLazinessAndBranchToFinish(lazy, new OPT_IntConstantOperand(pc));
		}
		if (PPC_Options.debugLazy) {
		  VM.sysWriteln("Resolving goto " + lazy.makeKey(pc) + " " + targetBB);
		}
		if (targetBB == null) {
		  // Block not translated so translate
		  translateSubTrace((Laziness)lazy.clone(), pc);
		  targetBB = findMapping(pc, lazy);
		}
		// Fix up instruction
		OPT_Instruction goto_inst = (OPT_Instruction)unresolvedGoto.get(toRemove);
		Goto.setTarget(goto_inst, targetBB.makeJumpTarget());
		if (PPC_Options.debugLazy) {
		  VM.sysWriteln("Properly resolving goto in block " + goto_inst.getBasicBlock() + " to " + lazy.makeKey(pc) + " " + targetBB);
		}
		goto_inst.getBasicBlock().deleteNormalOut();
		goto_inst.getBasicBlock().insertOut(targetBB);
		if (VM.VerifyAssertions) VM._assert(goto_inst.getBasicBlock().getNumberOfNormalOut() == 1);
		// Remove from unresolved list
		unresolvedGoto.remove(toRemove);
		unresolvedGoto_PC.remove(toRemove);
		unresolvedGoto_Laziness.remove(toRemove);
	 }
  }

  /**
   * List of unresolved LookupSwitch instructions used for branching to the link register
   */
  private Vector unresolvedLookupSwitchForLR;
  /**
   * List of where unresolved Goto instructions are trying to go
   */
  private Vector unresolvedLookupSwitchForLR_PC;
  /**
   * List of what the unresolved Goto instruction's target laziness should be
   */
  private Vector unresolvedLookupSwitchForLR_Laziness;

  /**
   * Register a HIR LookupSwitch instruction whose target can't be
   * resolved yet. There's a caveat on using this that there are no
   * other out edges for this BB.
   * @param lookupswitch_instr
   * @param pc the PC of the branch to link register
   * @param targetLaziness
   */
  public void registerLookupSwitchForLRUnresolved(OPT_Instruction lookupswitch_instr, int pc, Laziness targetLaziness){
	 unresolvedLookupSwitchForLR.add(lookupswitch_instr);
    unresolvedLookupSwitchForLR_PC.add(new Integer(pc));
    unresolvedLookupSwitchForLR_Laziness.add(targetLaziness);
  }
  /**
   * Are all LookupSwitch instructions for branches to LR ready to be
   * resolved
   * @return true if all are ready to be resolved
   */
  private boolean areLookupSwitchLRsReadyToResolve() {
    for(int i=0; i < unresolvedLookupSwitchForLR_PC.size(); i++) {
		int pc = ((Integer)unresolvedLookupSwitchForLR_PC.get(i)).intValue();
		Laziness lazy = (Laziness)unresolvedLookupSwitchForLR_Laziness.get(i);
		Set branchDests = getLikelyCallSites(pc);
		if(branchDests != null) {
		  for(Iterator j=branchDests.iterator(); j.hasNext();) {
			 int dest_pc = ((Integer)j.next()).intValue() + 4;
			 if(findMapping(dest_pc, lazy) == null) {
				return false;
			 }
		  }
		}
	 }
	 return true;
  }
  /**
	* Create sub-traces that will allow for the resolution of all
	* lookup switchs for branch to lr
   */
  private void prepareToResolveLookupSwitchLRs() {
    for(int i=0; i < unresolvedLookupSwitchForLR_PC.size(); i++) {
		int pc = ((Integer)unresolvedLookupSwitchForLR_PC.get(i)).intValue();
		Laziness lazy = (Laziness)unresolvedLookupSwitchForLR_Laziness.get(i);
		Set branchDests = (Set)getLikelyCallSites(pc);
		if (branchDests != null) {
		  branchDests = (Set)((HashSet)getLikelyCallSites(pc)).clone();
		  for(Iterator j=branchDests.iterator(); j.hasNext();) {
			 int dest_pc = ((Integer)j.next()).intValue() + 4;
			 if(findMapping(dest_pc, lazy) == null) {
				// Block not translated so translate
				translateSubTrace((Laziness)lazy.clone(), dest_pc);
			 }
		  }
		}
	 }
  }

  /**
	* Resolve all lookup switch instructions for branch to lr
   */
  private void resolveLookupSwitchLRs() {
    for(int i=0; i < unresolvedLookupSwitchForLR_PC.size(); i++) {
		int pc = ((Integer)unresolvedLookupSwitchForLR_PC.get(i)).intValue();
		Laziness lazy = (Laziness)unresolvedLookupSwitchForLR_Laziness.get(i);
		OPT_Instruction lookupswitch = (OPT_Instruction)unresolvedLookupSwitchForLR.get(i);
		OPT_BranchOperand default_target = LookupSwitch.getDefault(lookupswitch);
		OPT_Operand value = LookupSwitch.getValue(lookupswitch);
		Set branchDests = getLikelyCallSites(pc);
		if (branchDests != null) {
		  if ((branchDests.size() > 1)||(lookupswitch.getBasicBlock().nextBasicBlockInCodeOrder() != default_target.target.getBasicBlock())) {
			 float branchProb = (1.0f - OPT_BranchProfileOperand.UNLIKELY) / (float)branchDests.size();
			 LookupSwitch.mutate(lookupswitch, LOOKUPSWITCH, value, null, null,
										default_target, OPT_BranchProfileOperand.unlikely(),
										branchDests.size() * 3);
			 int match_no = 0;
			 for(Iterator j=branchDests.iterator(); j.hasNext();) {
				int dest_pc = ((Integer)j.next()).intValue() + 4;
				OPT_BasicBlock target = findMapping(dest_pc, lazy);
				if(target == null) {
				  throw new Error("Failed to find trace for " + dest_pc + " with laziness " + lazy);
				}
				LookupSwitch.setMatch(lookupswitch, match_no, new OPT_IntConstantOperand(dest_pc));
				LookupSwitch.setTarget(lookupswitch, match_no, target.makeJumpTarget());
				LookupSwitch.setBranchProfile(lookupswitch, match_no, new OPT_BranchProfileOperand(branchProb)) ;
				lookupswitch.getBasicBlock().insertOut(target);
				match_no++;
			 }
		  }
		  else {
			 int dest_pc = ((Integer)branchDests.iterator().next()).intValue() + 4;
			 OPT_BasicBlock target = findMapping(dest_pc, lazy);
			 if(target == null) {
				throw new Error("Failed to find trace for " + dest_pc + " with laziness " + lazy);
			 }
			 IfCmp.mutate(lookupswitch, INT_IFCMP, null,
							  value, new OPT_IntConstantOperand(dest_pc), OPT_ConditionOperand.EQUAL(),
							  target.makeJumpTarget(),
							  OPT_BranchProfileOperand.likely());
			 lookupswitch.getBasicBlock().insertOut(target);
		  }
		}
		else {
		  // no possibly branch destinations so remove lookupswitch
		  lookupswitch.remove();
		}
	 }
  }

  /**
   * List of unresolved LookupSwitch instructions used for branching to the count register
   */
  private Vector unresolvedLookupSwitchForCTR;
  /**
   * List of where unresolved LookupSwitch instructions are branching from
   */
  private Vector unresolvedLookupSwitchForCTR_PC;
  /**
   * List of what the unresolved LookupSwitch instruction's target laziness should be
   */
  private Vector unresolvedLookupSwitchForCTR_Laziness;
  /**
   * List of whether the unresolved LookupSwitch instruction's was a branch and link
   */
  private Vector unresolvedLookupSwitchForCTR_Link;

  /**
   * Register a HIR LookupSwitch instruction whose target can't be
   * resolved yet. There's a caveat on using this that there are no
   * other out edges for this BB.
   * @param lookupswitch_instr
   * @param pc the PC of the branch to count register
   * @param targetLaziness
   */
  public void registerLookupSwitchForCTRUnresolved(OPT_Instruction lookupswitch_instr, int pc, Laziness targetLaziness, boolean link){
	 unresolvedLookupSwitchForCTR.add(lookupswitch_instr);
    unresolvedLookupSwitchForCTR_PC.add(new Integer(pc));
    unresolvedLookupSwitchForCTR_Laziness.add(targetLaziness);
    unresolvedLookupSwitchForCTR_Link.add(Boolean.valueOf(link));
  }
  /**
   * Are all LookupSwitch instructions for branches to CTR ready to be
   * resolved
   * @return true if all are ready to be resolved
   */
  private boolean areLookupSwitchCTRsReadyToResolve() {
    for(int i=0; i < unresolvedLookupSwitchForCTR_PC.size(); i++) {
		int pc = ((Integer)unresolvedLookupSwitchForCTR_PC.get(i)).intValue();
		Laziness lazy = (Laziness)unresolvedLookupSwitchForCTR_Laziness.get(i);
		Set branchDests = getLikelyCtrDestinations(pc);
		if(branchDests != null) {
		  for(Iterator j=branchDests.iterator(); j.hasNext();) {
			 int dest_pc = ((Integer)j.next()).intValue();
			 if(findMapping(dest_pc, lazy) == null) {
				return false;
			 }
		  }
		}
	 }
	 return true;
  }
  /**
	* Create sub-traces that will allow for the resolution of all
	* lookup switchs for branch to ctr
   */
  private void prepareToResolveLookupSwitchCTRs() {
    for(int i=0; i < unresolvedLookupSwitchForCTR_PC.size(); i++) {
		int pc = ((Integer)unresolvedLookupSwitchForCTR_PC.get(i)).intValue();
		Laziness lazy = (Laziness)unresolvedLookupSwitchForCTR_Laziness.get(i);
		Set branchDests = (Set)getLikelyCtrDestinations(pc);
		if (branchDests != null) {
		  branchDests = (Set)((HashSet)getLikelyCtrDestinations(pc)).clone();
		  for(Iterator j=branchDests.iterator(); j.hasNext();) {
			 int dest_pc = ((Integer)j.next()).intValue();
			 if(((Boolean)unresolvedLookupSwitchForCTR_Link.get(i)).booleanValue()){
				registerBranchAndLink(pc, dest_pc);
			 }
			 if(findMapping(dest_pc, lazy) == null) {
				// Block not translated so translate
				translateSubTrace((Laziness)lazy.clone(), dest_pc);
			 }
		  }
		}
	 }
  }

  /**
	* Resolve all lookup switch instructions for branch to ctr
   */
  private void resolveLookupSwitchCTRs() {
    for(int i=0; i < unresolvedLookupSwitchForCTR_PC.size(); i++) {
		int pc = ((Integer)unresolvedLookupSwitchForCTR_PC.get(i)).intValue();
		Laziness lazy = (Laziness)unresolvedLookupSwitchForCTR_Laziness.get(i);
		OPT_Instruction lookupswitch = (OPT_Instruction)unresolvedLookupSwitchForCTR.get(i);
		OPT_BranchOperand default_target = LookupSwitch.getDefault(lookupswitch);
		OPT_Operand value = LookupSwitch.getValue(lookupswitch);
		Set branchDests = getLikelyCtrDestinations(pc);
		if (branchDests != null) {
		  if ((branchDests.size() > 1)||(lookupswitch.getBasicBlock().nextBasicBlockInCodeOrder() != default_target.target.getBasicBlock())) {
			 float branchProb = (1.0f - OPT_BranchProfileOperand.UNLIKELY) / (float)branchDests.size();
			 LookupSwitch.mutate(lookupswitch, LOOKUPSWITCH, value, null, null,
										default_target, OPT_BranchProfileOperand.unlikely(),
										branchDests.size() * 3);
			 int match_no = 0;
			 for(Iterator j=branchDests.iterator(); j.hasNext();) {
				int dest_pc = ((Integer)j.next()).intValue();
				OPT_BasicBlock target = findMapping(dest_pc, lazy);
				if(target == null) {
				  throw new Error("Failed to find trace for " + dest_pc + " with laziness " + lazy);
				}
				LookupSwitch.setMatch(lookupswitch, match_no, new OPT_IntConstantOperand(dest_pc));
				LookupSwitch.setTarget(lookupswitch, match_no, target.makeJumpTarget());
				LookupSwitch.setBranchProfile(lookupswitch, match_no, new OPT_BranchProfileOperand(branchProb)) ;
				lookupswitch.getBasicBlock().insertOut(target);
				match_no++;
			 }
		  }
		  else {
			 int dest_pc = ((Integer)branchDests.iterator().next()).intValue();
			 OPT_BasicBlock target = findMapping(dest_pc, lazy);
			 if(target == null) {
				throw new Error("Failed to find trace for " + dest_pc + " with laziness " + lazy);
			 }
			 IfCmp.mutate(lookupswitch, INT_IFCMP, null,
							  value, new OPT_IntConstantOperand(dest_pc), OPT_ConditionOperand.EQUAL(),
							  target.makeJumpTarget(),
							  OPT_BranchProfileOperand.likely());
			 lookupswitch.getBasicBlock().insertOut(target);
		  }
		}
		else {
		  // no possibly branch destinations so remove lookupswitch
		  lookupswitch.remove();
		}
	 }
  }

  // -oO Global HIR basic blocks Oo-

  /**
   * The OPT_BasicBlock in which instructions are currently being
   * inserted
   */
  private OPT_BasicBlock currentBlock;

  /**
   * The OPT_BasicBlock which will contain the next translated instruction
   */
  private OPT_BasicBlock nextBlock;

  /**
   * Get the block which is currently having instructions inserted
   * into it
   * @return the current block
   */
  public OPT_BasicBlock getCurrentBlock() {
    return currentBlock;
  }

  /**
   * Get the block which will contain the translation of the next PPC
   * instruction
   * @return the next block
   */
  public OPT_BasicBlock getNextBlock() {
    return nextBlock;
  }

  /**
   * Set the block which is currently having instructions inserted into it
   * @param newCurrentBlock the new current basic block
   */
  public void setCurrentBlock(OPT_BasicBlock newCurrentBlock) {
    currentBlock = newCurrentBlock;
  }

  /**
   * Set the block which will contain the translation of the next PPC
   * instruction
   * @param newCurrentBlock the new next basic block
   */
  public void setNextBlock(OPT_BasicBlock newNextBlock) {
    nextBlock = newNextBlock;
  }
  
  /**
   * Create a basic block immediately after the current block and link
   * its edges into the CFG and code ordering
   * @return the new basic block
   */
  public OPT_BasicBlock createBlockAfterCurrent()
  {
    OPT_BasicBlock nextBlock = currentBlock.nextBasicBlockInCodeOrder();
    OPT_BasicBlock newBlock  = new OPT_BasicBlock(0, gc.inlineSequence, gc.cfg);

    gc.cfg.breakCodeOrder(currentBlock, nextBlock);
    gc.cfg.linkInCodeOrder(currentBlock, newBlock);
    gc.cfg.linkInCodeOrder(newBlock, nextBlock);

	 if (VM.VerifyAssertions) VM._assert(currentBlock.isOut(nextBlock));
    currentBlock.deleteOut(nextBlock);
    currentBlock.insertOut(newBlock);
    newBlock.insertOut(nextBlock);

    if(PPC_Options.debugCFG) {
      report("Created block (" + newBlock + ") after current (" + currentBlock+").");
    }
    return newBlock;
  }

  /**
   * Create a basic block immediately after the current block and link
   * its edges into code ordering but not the CFG
   * @return the new basic block
   */
  public OPT_BasicBlock createBlockAfterCurrentNotInCFG()
  {
    OPT_BasicBlock nextBlock = currentBlock.nextBasicBlockInCodeOrder();
    OPT_BasicBlock newBlock  = new OPT_BasicBlock(0, gc.inlineSequence, gc.cfg);

    gc.cfg.breakCodeOrder(currentBlock, nextBlock);
    gc.cfg.linkInCodeOrder(currentBlock, newBlock);
    gc.cfg.linkInCodeOrder(newBlock, nextBlock);

    if(PPC_Options.debugCFG) {
      report("Created block (" + newBlock + ") after current (" + currentBlock+") but not in CFG.");
    }
    return newBlock;
  }

  /**
   * Create a basic block immediately after the given block and link
   * its edges into the CFG and code ordering
   * @param afterBlock the block to create a block after
   * @return the new basic block
   */
  public OPT_BasicBlock createBlockAfter(OPT_BasicBlock afterBlock)
  {
    OPT_BasicBlock nextBlock = afterBlock.nextBasicBlockInCodeOrder();
    OPT_BasicBlock newBlock  = new OPT_BasicBlock(0, gc.inlineSequence, gc.cfg);

    gc.cfg.breakCodeOrder(afterBlock, nextBlock);
    gc.cfg.linkInCodeOrder(afterBlock, newBlock);
    gc.cfg.linkInCodeOrder(newBlock, nextBlock);

	 if (VM.VerifyAssertions) VM._assert(afterBlock.isOut(nextBlock));
    afterBlock.deleteOut(nextBlock);
    afterBlock.insertOut(newBlock);
    newBlock.insertOut(nextBlock);

    if(PPC_Options.debugCFG) {
      report("Created block (" + newBlock + ") after (" + afterBlock+").");
    }
    return newBlock;
  }

  /**
   * Append a HIR instruction to the current basic block
   * @param i the HIR instruction
   */
  public void appendInstructionToCurrentBlock(OPT_Instruction i) {
    currentBlock.appendInstruction(i);
  }

  /**
   * This block gets instructions to pre-fill registers inserted into
   * it.
   */
  private OPT_BasicBlock lazyPreFillBlock;

  /**
   * Load all the registers from the PPC_ProcessSpace into the pre-fill
   * block
   */
  private void preFillAllRegisters() {
    OPT_BasicBlock temp = currentBlock;
    currentBlock = lazyPreFillBlock;
    fillAllRegisters();
    createPS_MemoryReference();
    currentBlock = temp;
  }

  /** 
   * The basic block is used by finish trace to hold all the code that
   * must be executed before returning to the main run loop
   */
  private OPT_BasicBlock finishBlock;

  /**
   * We're finished translating this trace so ..
   * 1) put all the registers back into the process space
   * 2) return the new PC value
   */
  private void finishTrace() {
    // finishBlock is already linked into the cfg, we just need to
    // make sure we are adding to it.
    currentBlock = finishBlock;
    spillAllRegisters();
	 // Update global branch information to reflect what's been learnt in this trace
	 if (gc.options.getOptLevel() < PPC_Options.globalBranchLevel) {
		ps.branchInfo.merge(branchInfo);
	 }
  }

  /**
   * Set the return value in the currentBlock, resolve its lazy state
   * (so the state is no longer lazy) and then set it to branch to the
   * finish block
   * @param value return value for translated code (the PC value of
   * the next instruction to translate)
   */
  public void setReturnValueResolveLazinessAndBranchToFinish(Laziness laziness, OPT_Operand value) {
    // Copy the value into the register specified by gc.resultReg.
    appendInstructionToCurrentBlock(Move.create(INT_MOVE,
																new OPT_RegisterOperand(gc.resultReg, VM_TypeReference.Int),
																value));
    resolveLaziness(laziness);
    appendInstructionToCurrentBlock(Goto.create(GOTO, finishBlock.makeJumpTarget()));
	 currentBlock.deleteNormalOut();
	 currentBlock.insertOut(finishBlock);
	 if (VM.VerifyAssertions) VM._assert(currentBlock.getNumberOfNormalOut() == 1);
  }

  // -oO Trace control Oo-

  /**
	* Information about branches in this trace
	*/
  PPC_BranchLogic branchInfo;

  /**
	* Is it suitable to stop the trace now?
	* @return true => try to stop the trace
	*/
  boolean suitableToStop() {
	 switch(gc.options.getOptLevel()) {
	 case 0:
		return numberOfInstructions > PPC_Options.instrOpt0;
	 case 1:
		return numberOfInstructions > PPC_Options.instrOpt1;
	 default:
		return numberOfInstructions > PPC_Options.instrOpt2;
	 }
  }
  /**
	* Register a branch and link instruction
	* @param pc the address of the branch instruction (implicity the
	* link value will be assumed to be pc+4)
	* @param dest the destination of the branch instruction
	*/
  public void registerBranchAndLink(int pc, int dest) {
	 branchInfo.registerBranchAndLink(pc, dest);
  }

  /**
	* Returns a vector of likely call sites (typically pc is the
	* address of a return instruction)
	* @param pc location to suggest call sites for
	* @return vector of potential calling instructions
	*/
  public Set getLikelyCallSites(int pc) {
	 return branchInfo.getLikelyCallSites(pc);
  }

  /**
	* Returns a vector of likely bcctr destinations
	* @param pc location to suggest destinations for
	* @return vector of potential destinations
	*/
  public Set getLikelyCtrDestinations(int pc) {
	 return branchInfo.getLikelyCtrDestinations(pc);
  }

  /**
	* Should a trace follow a branch and link instruction or should it
	* terminate the trace?
	* @param pc the address of the branch and link instruction
	* @return whether the trace should continue
	*/
  public boolean traceContinuesAfterBranchAndLink(int pc) {
	 return suitableToStop() == false;
  }

  // -oO Translate sequences of code Oo-

  /**
   * Translate a sequence of instructions upto an instruction that
   * doesn't know its immediate/default successor. The currentBlock
   * should be an empty basic block.
   */
  private void translateSubTrace(Laziness lazy, int pc) {
	 if(suitableToStop()) {
		// Record mapping of this pc value and laziness to this block
		registerMapping(pc, lazy, currentBlock);
		// Create next block
		nextBlock = createBlockAfterCurrent();
		// Finish block to return and exit
		setReturnValueResolveLazinessAndBranchToFinish(lazy, new OPT_IntConstantOperand(pc));
		// Move currentBlock along
		currentBlock = nextBlock;
	 }
	 else {
		do {
		  if (VM.VerifyAssertions) VM._assert(currentBlock.getNumberOfRealInstructions() == 0);
		  // Record mapping of this pc value and laziness to this block
		  registerMapping(pc, lazy, currentBlock);

		  // Create next block
		  nextBlock = createBlockAfterCurrent();

		  // Translare instruction and get address of next instruction
		  pc = PPC_InstructionTranslator.translateInstruction(this, ps, lazy, pc);
		  numberOfInstructions++;

		  // Move currentBlock along
		  currentBlock = nextBlock;
		  if (VM.VerifyAssertions) VM._assert(currentBlock.getNumberOfRealInstructions() == 0);

		  // Are we translating in single instruction mode
		  if (PPC_Options.singleInstrTranslation == true) {
			 if (pc != -1) {
				setReturnValueResolveLazinessAndBranchToFinish(lazy, new OPT_IntConstantOperand(pc));
			 }
			 break;
		  }

		  // Do we already have a translation for this next block?
		  OPT_BasicBlock possibleNextBlock = findMapping(pc, lazy);
		  if(possibleNextBlock != null) {
			 // Yes, branch to that and stop translating
			 appendInstructionToCurrentBlock(Goto.create(GOTO, possibleNextBlock.makeJumpTarget()));
			 OPT_BasicBlock gotoBlock = currentBlock;
			 currentBlock = createBlockAfterCurrent();
			 gotoBlock.deleteNormalOut();
			 gotoBlock.insertOut(possibleNextBlock);
			 if (VM.VerifyAssertions) VM._assert(gotoBlock.getNumberOfNormalOut() == 1);
			 break;
		  }
		} while (pc != -1);
	 }
  }

  /** 
   * This is the main loop to generate the HIR.
   */
  private void generateHIR()
  {
    // Load all the register values to be used in a trace
    preFillAllRegisters();

    // Translate from the current program counter
    translateSubTrace(new Laziness(), ((VM_PPC_Trace)(gc.method)).pc);

	 do {
		if (PPC_Options.resolveProceduresBeforeBranches) {
		  prepareToResolveLookupSwitchLRs();
		  prepareToResolveLookupSwitchCTRs();
		}
		if (PPC_Options.resolveBranchesAtOnce) {
		  do {
			 resolveGoto();
			 resolveIfCmp();
		  } while((areGotosResolved() && areIfCmpsResolved()) == false);
		}
		else {
		  resolveGoto();
		  resolveIfCmp();
		}
		if (!PPC_Options.resolveProceduresBeforeBranches) {
		  prepareToResolveLookupSwitchLRs();
		  prepareToResolveLookupSwitchCTRs();
		}
	 } while((areGotosResolved() && areIfCmpsResolved() && areLookupSwitchLRsReadyToResolve()  && areLookupSwitchCTRsReadyToResolve()) == false);
    // Resolve unresolved branches and branches to ctr
	 resolveLookupSwitchLRs();
	 resolveLookupSwitchCTRs();

    // Finish up the trace
    finishTrace();

	 if (PPC_Options.eliminateRegisterFills) {
		eliminateRegisterFills();
	 }
  }

  // -oO Memory operations Oo-
  /**
   * Register that references the process space memory array
   */
  OPT_Register ps_memory;
  /**
   * Type of process space memory (int[][])
   */
  VM_TypeReference ps_memory_type;
  /**
   * Create a register that references to the process spaces memory
   */
  private void createPS_MemoryReference() {
    VM_TypeReference psTref =  VM_TypeReference.findOrCreate(VM_SystemClassLoader.getVMClassLoader(),
																				 VM_Atom.findOrCreateAsciiAtom
																				 ("Lorg/binarytranslator/PPC_Emulator/PPC_ProcessSpace;")
																				 );
    VM_FieldReference ref = VM_MemberReference.findOrCreate(psTref, VM_Atom.findOrCreateAsciiAtom("memory"),
																				VM_Atom.findOrCreateAsciiAtom("[[I")
																				).asFieldReference();
    ps_memory_type = ref.getFieldContentsType();
    OPT_RegisterOperand ps_memoryOp = gc.temps.makeTemp(ps_memory_type);
    appendInstructionToCurrentBlock(GetField.create(GETFIELD, ps_memoryOp,
																	 gc.makeLocal(1,psTref),
																	 new OPT_IntConstantOperand(ref.peekResolvedField().getOffset()),
																	 new OPT_LocationOperand(ref),
																	 new OPT_TrueGuardOperand()));
    ps_memory = ps_memoryOp.register;
  }

  /** 
   * Store a word value from a register into a memory address in the
   * PPC_ProcessSpace. The generated code should look like:
   * <listing>
   * ps.memory[pageIndex][wordIndex] = reg
   * </listing>
   * Clobbers tempInts 0 - 2
   * @param value the register that holds the value to be stored.
   * @param addr an OPT_RegisterOperand containing the address in
   * PPC_ProcessSpace memory to store the value.
   */
  public void memoryStore(OPT_RegisterOperand value, OPT_RegisterOperand addr) {

	 if(PPC_Options.loadStoreInPS)
		{
		  OPT_Instruction s = Call.create(CALL, null, null, null, null, 3);

		  VM_TypeReference psTref = VM_TypeReference.findOrCreate
			 (VM_SystemClassLoader.getVMClassLoader(),
			  VM_Atom.findOrCreateAsciiAtom
			  ("Lorg/binarytranslator/PPC_Emulator/PPC_ProcessSpace;"));						

		  VM_MemberReference methRef = VM_MemberReference.findOrCreate
			 (psTref,VM_Atom.findOrCreateAsciiAtom("setInt"), 
			  VM_Atom.findOrCreateAsciiAtom("(II)V"));

		  VM_Method method = ((VM_MethodReference)methRef).resolveInvokeSpecial();

		  OPT_MethodOperand methOp = OPT_MethodOperand.VIRTUAL((VM_MethodReference)methRef, method);

		  OPT_Operand psRef = gc.makeLocal(1,psTref);
		  Call.setParam(s, 0, psRef); // Sets 'this' pointer in setInt to ps.
		  Call.setParam(s, 1, addr);
		  Call.setParam(s, 2, value);
		  Call.setGuard(s, new OPT_TrueGuardOperand());
		  Call.setMethod(s, methOp);
		  Call.setAddress(s, new OPT_IntConstantOperand
								(((VM_MethodReference)methRef).peekResolvedMethod().getOffset()));
		  s.position = gc.inlineSequence;
		  s.bcIndex = 0;
								
		  appendInstructionToCurrentBlock(s);

		}
	 else
		{

		  // Extract the memory page number from addr.
		  OPT_RegisterOperand pageIndex = getTempInt(0);
		  appendInstructionToCurrentBlock(Binary.create(INT_USHR, pageIndex,
																		addr,
																		new OPT_IntConstantOperand(PPC_Constants.PAGE_SHIFT)));
		  
		  // Extract the location of the address within the page.
		  OPT_RegisterOperand wordAddr = getTempInt(1);
		  appendInstructionToCurrentBlock(Binary.create(INT_AND, wordAddr,
																		addr.copyRO(),
																		new OPT_IntConstantOperand(PPC_Constants.PAGE_OFFSET)));

		  // Convert to array index (i.e. divide by 4)
		  OPT_RegisterOperand wordIndex = getTempInt(2);
		  appendInstructionToCurrentBlock(Binary.create(INT_USHR, wordIndex, wordAddr.copyRO(),
																		new OPT_IntConstantOperand(2)));
		  
		  
		  // Retrieve the int[] for the correct page into pageReg.
		  OPT_RegisterOperand pageReg = getTempIntArray(0);
		  appendInstructionToCurrentBlock(ALoad.create(REF_ALOAD, pageReg,
																	  new OPT_RegisterOperand(ps_memory, ps_memory_type),
																	  pageIndex.copyRO(),
																	  new OPT_LocationOperand(VM_TypeReference.IntArray),
																	  new OPT_TrueGuardOperand()));
		  
		  // Copy from reg to the correct array element.
		  appendInstructionToCurrentBlock(AStore.create(INT_ASTORE, value,
																		pageReg.copyRO(),
																		wordIndex.copyRO(), 
																		new OPT_LocationOperand(VM_TypeReference.Int),
																		new OPT_TrueGuardOperand()));
		  
		}  

  }
  public static final int TEMP_INT_SAFE_FROM_STORE_CLOBBER=3;

  /** 
   * Store a byte value from a register into a memory address in the
   * PPC_ProcessSpace. The generated code should look like:
   * <listing>
   * tempInt = ps.memory[pageIndex][wordIndex]
   * tempByte = reg & 0xff;
   * switch (addr & 0x3) {
   * default: case 3:
   *    tempInt &= 0xffffff00;
   *    tempInt |= tempByte;
   *    break;
   * case 2:
   *    tempInt &= 0xffff00ff;
   *    tempInt |= (tempByte << 8);
   *    break;
   * case 1:
   *    tempInt &= 0xff00ffff;
   *    tempInt |= (tempByte << 16);
   *    break;
   * case 0:
   *    tempInt &= 0x00ffffff;
   *    tempInt |= (tempByte << 24);
   *    break;
   * }
   * ps.memory[pageIndex][wordIndex] = tempInt
   * </listing>
   * Clobbers tempInt 0-5
   * @param value the register whose lower bits hold the value to be
   * stored.
   * @param addr an OPT_RegisterOperand containing the address in
   * PPC_ProcessSpace memory to store the value.
   */
  public void memoryStoreByte(OPT_RegisterOperand value, OPT_RegisterOperand addr) {
    // tempByte = reg & 0xff
    OPT_RegisterOperand tempByte = getTempInt(0);
    appendInstructionToCurrentBlock(Binary.create(INT_AND, tempByte, 
                                                  value,
                                                  new OPT_IntConstantOperand(0xff)));
    // pageIndex = addr >>> 12
    OPT_RegisterOperand pageIndex = getTempInt(1);
    appendInstructionToCurrentBlock(Binary.create(INT_USHR, pageIndex,
                                                  addr,
                                                  new OPT_IntConstantOperand(PPC_Constants.PAGE_SHIFT)));
    // wordAddr = addr & 0xfff
    OPT_RegisterOperand wordAddr = getTempInt(2);
    appendInstructionToCurrentBlock(Binary.create(INT_AND, wordAddr,
                                                  addr.copyRO(),
                                                  new OPT_IntConstantOperand(PPC_Constants.PAGE_OFFSET)));
    // wordIndex = wordAddr >>> 2
    OPT_RegisterOperand wordIndex = getTempInt(3);
    appendInstructionToCurrentBlock(Binary.create(INT_USHR, wordIndex,
                                                  wordAddr.copyRO(),
                                                  new OPT_IntConstantOperand(2)));
    // byteIndex = addr & 0x3
    OPT_RegisterOperand byteIndex = getTempInt(4);
    appendInstructionToCurrentBlock(Binary.create(INT_AND, byteIndex,
                                                  addr.copyRO(),
                                                  new OPT_IntConstantOperand(0x3)));
    // Retrieve the int[] for the correct page into pageReg.
    OPT_RegisterOperand pageReg = getTempIntArray(0);
    appendInstructionToCurrentBlock(ALoad.create(REF_ALOAD, pageReg,
																 new OPT_RegisterOperand(ps_memory, ps_memory_type),
                                                 pageIndex.copyRO(),
                                                 new OPT_LocationOperand(VM_TypeReference.IntArray),
                                                 new OPT_TrueGuardOperand()));

    // Retrieve the int containing the byte into tempInt
    OPT_RegisterOperand tempInt = getTempInt(5);
    appendInstructionToCurrentBlock(ALoad.create(INT_ALOAD, tempInt,
																 pageReg.copyRO(), 
																 wordIndex.copyRO(),
																 new OPT_LocationOperand(VM_TypeReference.Int),
																 new OPT_TrueGuardOperand()));                                   
    // Generate switch
    OPT_BasicBlock byteIndex3 = createBlockAfterCurrent();
    OPT_BasicBlock byteIndex2 = createBlockAfter(byteIndex3);
    OPT_BasicBlock byteIndex1 = createBlockAfter(byteIndex2);
    OPT_BasicBlock byteIndex0 = createBlockAfter(byteIndex1);
    OPT_BasicBlock storeInt = createBlockAfter(byteIndex0);

    OPT_Instruction s = TableSwitch.create(TABLESWITCH, byteIndex.copyRO(), null, null, 
                                           new OPT_IntConstantOperand(0), 
                                           new OPT_IntConstantOperand(3), 
                                           byteIndex3.makeJumpTarget(), 
                                           OPT_BranchProfileOperand.never(),
                                           4*2);

    //-#if !RVM_FOR_POWERPC
    TableSwitch.setTarget(s, 0, byteIndex0.makeJumpTarget());
    TableSwitch.setTarget(s, 1, byteIndex1.makeJumpTarget());
    TableSwitch.setTarget(s, 2, byteIndex2.makeJumpTarget());
    TableSwitch.setTarget(s, 3, byteIndex3.makeJumpTarget());
	 //-#else
    TableSwitch.setTarget(s, 3, byteIndex0.makeJumpTarget());
    TableSwitch.setTarget(s, 2, byteIndex1.makeJumpTarget());
    TableSwitch.setTarget(s, 1, byteIndex2.makeJumpTarget());
    TableSwitch.setTarget(s, 0, byteIndex3.makeJumpTarget());
	 //-#endif
    TableSwitch.setBranchProfile(s, 0, new OPT_BranchProfileOperand(0.25f));
    TableSwitch.setBranchProfile(s, 1, new OPT_BranchProfileOperand(0.25f));
    TableSwitch.setBranchProfile(s, 2, new OPT_BranchProfileOperand(0.25f));
    TableSwitch.setBranchProfile(s, 3, new OPT_BranchProfileOperand(0.25f));

    appendInstructionToCurrentBlock(s);
    currentBlock.insertOut(byteIndex0);
    currentBlock.insertOut(byteIndex1);
    currentBlock.insertOut(byteIndex2);

    // Generate case 3
    byteIndex3.appendInstruction(Binary.create(INT_AND, tempInt.copyRO(), 
                                               tempInt.copyRO(),
                                               new OPT_IntConstantOperand(0xffffff00)));
    byteIndex3.appendInstruction(Binary.create(INT_OR, tempInt.copyRO(), 
                                               tempInt.copyRO(),
                                               tempByte.copyRO()));
    byteIndex3.appendInstruction(Goto.create(GOTO, storeInt.makeJumpTarget()));
	 byteIndex3.deleteNormalOut();
    byteIndex3.insertOut(storeInt);

    // Generate case 2
    byteIndex2.appendInstruction(Binary.create(INT_AND, tempInt.copyRO(), 
                                               tempInt.copyRO(),
                                               new OPT_IntConstantOperand(0xffff00ff)));
    byteIndex2.appendInstruction(Binary.create(INT_SHL, tempByte.copyRO(), 
                                               tempByte.copyRO(),
                                               new OPT_IntConstantOperand(8)));
    byteIndex2.appendInstruction(Binary.create(INT_OR, tempInt.copyRO(), 
                                               tempInt.copyRO(),
                                               tempByte.copyRO()));
    byteIndex2.appendInstruction(Goto.create(GOTO, storeInt.makeJumpTarget()));
	 byteIndex2.deleteNormalOut();
    byteIndex2.insertOut(storeInt);

    // Generate case 1
    byteIndex1.appendInstruction(Binary.create(INT_AND, tempInt.copyRO(), 
                                               tempInt.copyRO(),
                                               new OPT_IntConstantOperand(0xff00ffff)));
    byteIndex1.appendInstruction(Binary.create(INT_SHL, tempByte.copyRO(), 
                                               tempByte.copyRO(),
                                               new OPT_IntConstantOperand(16)));
    byteIndex1.appendInstruction(Binary.create(INT_OR, tempInt.copyRO(), 
                                               tempInt.copyRO(),
                                               tempByte.copyRO()));
    byteIndex1.appendInstruction(Goto.create(GOTO, storeInt.makeJumpTarget()));
	 byteIndex1.deleteNormalOut();
    byteIndex1.insertOut(storeInt);

    // Generate case 0
    byteIndex0.appendInstruction(Binary.create(INT_AND, tempInt.copyRO(), 
                                               tempInt.copyRO(),
                                               new OPT_IntConstantOperand(0x00ffffff)));
    byteIndex0.appendInstruction(Binary.create(INT_SHL, tempByte.copyRO(), 
                                               tempByte.copyRO(),
                                               new OPT_IntConstantOperand(24)));
    byteIndex0.appendInstruction(Binary.create(INT_OR, tempInt.copyRO(), 
                                               tempInt.copyRO(),
                                               tempByte.copyRO()));
    byteIndex0.appendInstruction(Goto.create(GOTO, storeInt.makeJumpTarget()));
	 byteIndex0.deleteNormalOut();
    byteIndex0.insertOut(storeInt);


    // Copy from reg to the correct array element.
    storeInt.appendInstruction(AStore.create(INT_ASTORE, tempInt.copyRO(),
															pageReg.copyRO(),
                                             wordIndex.copyRO(), 
                                             new OPT_LocationOperand(VM_TypeReference.Int),
                                             new OPT_TrueGuardOperand()));
    setCurrentBlock(storeInt);
  }
  public static final int TEMP_INT_SAFE_FROM_STORE_BYTE_CLOBBER=6;
  
  /** 
   * Store a half-word value from a register into a memory address in the
   * PPC_ProcessSpace. The generated code should look like:
   * <listing>
   * tempInt = ps.memory[pageIndex][wordIndex]
   * tempHalfWord = reg & 0xffff;
   * if(addr & 0x2 != 0) {
   *    tempInt &= 0xffff0000;
   *    tempInt |= tempHalfWord;
   * }
   * else {
   *    tempInt &= 0x0000ffff;
   *    tempInt |= (tempHalfWord << 16);
   * }
   * ps.memory[pageIndex][wordIndex] = tempInt
   * </listing>
   * Clobbers tempInt 0 - 5
   * @param reg the register whose lower bits hold the value to be
   * stored.
   * @param addr an OPT_RegisterOperand containing the address in
   * PPC_ProcessSpace memory to store the value.
   */
  public void memoryStoreHalfWord(OPT_RegisterOperand reg, OPT_RegisterOperand addr) {

	 if(PPC_Options.loadStoreInPS) {
		// Extract lower two bytes.
		OPT_RegisterOperand halfWord = getTempInt(0);
		appendInstructionToCurrentBlock(Binary.create(INT_ADD, halfWord, reg,
																	 new OPT_IntConstantOperand(0xffff)));
		OPT_Instruction s = Call.create(CALL, null, null, null, null, 3);
		VM_TypeReference psTref = VM_TypeReference.findOrCreate(VM_SystemClassLoader.getVMClassLoader(),
																				  VM_Atom.findOrCreateAsciiAtom
																				  ("Lorg/binarytranslator/PPC_Emulator/PPC_ProcessSpace;"));						
		VM_MemberReference methRef = VM_MemberReference.findOrCreate(psTref,VM_Atom.findOrCreateAsciiAtom("setHalfWord"), 
																						 VM_Atom.findOrCreateAsciiAtom("(II)V"));		
		VM_Method method = ((VM_MethodReference)methRef).resolveInvokeSpecial();
		
		OPT_MethodOperand methOp = OPT_MethodOperand.VIRTUAL((VM_MethodReference)methRef, method);
		OPT_Operand psRef = gc.makeLocal(1,psTref);
		Call.setParam(s, 0, psRef); /* Sets 'this' pointer in setInt to ps. */
		Call.setParam(s, 1, addr);
		Call.setParam(s, 2, halfWord.copyRO());
		Call.setGuard(s, new OPT_TrueGuardOperand());
		Call.setMethod(s, methOp);
		Call.setAddress(s, new OPT_IntConstantOperand
							 (((VM_MethodReference)methRef).peekResolvedMethod().getOffset()));
		s.position = gc.inlineSequence;
		s.bcIndex = 0;
		appendInstructionToCurrentBlock(s);
	 }
	 else	{
		// tempHalfWord = reg & 0xffff
		OPT_RegisterOperand tempHalfWord = getTempInt(0);
		appendInstructionToCurrentBlock(Binary.create(INT_AND, tempHalfWord, 
																	 reg,
																	 new OPT_IntConstantOperand(0xffff)));
		// pageIndex = addr >> 12
		OPT_RegisterOperand pageIndex = getTempInt(1);
		appendInstructionToCurrentBlock(Binary.create(INT_USHR, pageIndex,
																	 addr,
																	 new OPT_IntConstantOperand(PPC_Constants.PAGE_SHIFT)));
		// wordAddr = addr & 0xfff
		OPT_RegisterOperand wordAddr = getTempInt(2);
		appendInstructionToCurrentBlock(Binary.create(INT_AND, wordAddr,
																	 addr.copyRO(),
																	 new OPT_IntConstantOperand(PPC_Constants.PAGE_OFFSET)));
		// wordIndex = wordAddr >>> 2
		OPT_RegisterOperand wordIndex = getTempInt(3);
		appendInstructionToCurrentBlock(Binary.create(INT_USHR, wordIndex,
																	 wordAddr.copyRO(),
																	 new OPT_IntConstantOperand(2)));
		// halfWordIndex = addr & 0x2
		OPT_RegisterOperand halfWordIndex = getTempInt(4);
		appendInstructionToCurrentBlock(Binary.create(INT_AND, halfWordIndex,
																	 addr.copyRO(),
																	 new OPT_IntConstantOperand(0x2)));

		// Retrieve the int[] for the correct page into pageReg.
		OPT_RegisterOperand pageReg = getTempIntArray(0);
		appendInstructionToCurrentBlock(ALoad.create(REF_ALOAD, pageReg,
																	new OPT_RegisterOperand(ps_memory, ps_memory_type),
																	pageIndex.copyRO(),
																	new OPT_LocationOperand(VM_TypeReference.IntArray),
																	new OPT_TrueGuardOperand()));

		// Retrieve the int containing the byte into tempInt
		OPT_RegisterOperand tempInt = getTempInt(5);
		appendInstructionToCurrentBlock(ALoad.create(INT_ALOAD, tempInt, pageReg.copyRO(), 
																	wordIndex.copyRO(),
																	new OPT_LocationOperand(VM_TypeReference.Int),
																	new OPT_TrueGuardOperand()));                                   
		// Generate if-then-else
		OPT_BasicBlock halfWordIndex1 = createBlockAfterCurrent();
		OPT_BasicBlock halfWordIndex0 = createBlockAfter(halfWordIndex1);
		OPT_BasicBlock storeInt = createBlockAfter(halfWordIndex0);

		//-#if !RVM_FOR_POWERPC
		OPT_Instruction s = IfCmp.create(INT_IFCMP, null,
													halfWordIndex.copyRO(),
													new OPT_IntConstantOperand(0),
													OPT_ConditionOperand.EQUAL(),
													halfWordIndex0.makeJumpTarget(),
													OPT_BranchProfileOperand.likely());
		//-#else
		OPT_Instruction s = IfCmp.create(INT_IFCMP, null,
													halfWordIndex.copyRO(),
													new OPT_IntConstantOperand(0),
													OPT_ConditionOperand.NOT_EQUAL(),
													halfWordIndex0.makeJumpTarget(),
													OPT_BranchProfileOperand.likely());
		//-#endif
		appendInstructionToCurrentBlock(s);
		currentBlock.insertOut(halfWordIndex0);

		// Generate then case - half word index is 0x2
		halfWordIndex1.appendInstruction(Binary.create(INT_AND, tempInt.copyRO(), 
																	  tempInt.copyRO(),
																	  new OPT_IntConstantOperand(0xffff0000)));
		halfWordIndex1.appendInstruction(Binary.create(INT_OR, tempInt.copyRO(), 
																	  tempInt.copyRO(),
																	  tempHalfWord.copyRO()));
		halfWordIndex1.appendInstruction(Goto.create(GOTO, storeInt.makeJumpTarget()));
		halfWordIndex1.deleteNormalOut();
		halfWordIndex1.insertOut(storeInt);

		// Generate else case - half word index is 0
		halfWordIndex0.appendInstruction(Binary.create(INT_AND, tempInt.copyRO(), 
																	  tempInt.copyRO(),
																	  new OPT_IntConstantOperand(0x0000ffff)));
		halfWordIndex0.appendInstruction(Binary.create(INT_SHL, tempHalfWord.copyRO(), 
																	  tempHalfWord.copyRO(),
																	  new OPT_IntConstantOperand(16)));
		halfWordIndex0.appendInstruction(Binary.create(INT_OR, tempInt.copyRO(), 
																	  tempInt.copyRO(),
																	  tempHalfWord.copyRO()));
		// Copy from reg to the correct array element.
		storeInt.appendInstruction(AStore.create(INT_ASTORE, tempInt.copyRO(), pageReg.copyRO(),
															  wordIndex.copyRO(), 
															  new OPT_LocationOperand(VM_TypeReference.Int),
															  new OPT_TrueGuardOperand()));
		setCurrentBlock(storeInt);
	 } 

  }
  public static final int TEMP_INT_SAFE_FROM_STORE_HALF_WORD_CLOBBER=6;

  /**
   * Load a word into a register from a memory address in the
   * PPC_ProcessSpace.
   * Clobbers tempInt 0 -2
   * @param dest the destination register.
   * @param addr an OPT_RegisterOperand containing the address in PPC_ProcessSpace memory 
   * holding the value (a multiple of four bytes).
   */
  public void memoryLoad(OPT_RegisterOperand dest, OPT_RegisterOperand addr) {
 
	 if(PPC_Options.loadStoreInPS){
		OPT_Instruction s = Call.create(CALL, dest, null, null, null, 
												  2);
		VM_TypeReference psTref = VM_TypeReference.findOrCreate(VM_SystemClassLoader.getVMClassLoader(),
																				  VM_Atom.findOrCreateAsciiAtom("Lorg/binarytranslator/PPC_Emulator/PPC_ProcessSpace;"));
																					  
		VM_MemberReference methRef = VM_MemberReference.findOrCreate(psTref,VM_Atom.findOrCreateAsciiAtom("getInt"), 
																						 VM_Atom.findOrCreateAsciiAtom("(I)I"));
		VM_Method method = ((VM_MethodReference)methRef).resolveInvokeSpecial();
		OPT_MethodOperand methOp = OPT_MethodOperand.VIRTUAL((VM_MethodReference)methRef, method);
		OPT_Operand psRef = gc.makeLocal(1,psTref);
		Call.setParam(s, 0, psRef); /* Sets 'this' pointer in setInt to ps. */
		Call.setParam(s, 1, addr);
		Call.setGuard(s, new OPT_TrueGuardOperand());
		Call.setMethod(s, methOp);
		Call.setAddress(s, new OPT_IntConstantOperand(((VM_MethodReference)methRef).peekResolvedMethod().getOffset()));
		s.position = gc.inlineSequence;
		s.bcIndex = 0;
			
		appendInstructionToCurrentBlock(s);
	 }
	 else {
		// Extract the memory page number from addr.
		OPT_RegisterOperand pageIndex = getTempInt(0);
		appendInstructionToCurrentBlock(Binary.create(INT_USHR, pageIndex,
																	 addr,
																	 new OPT_IntConstantOperand(PPC_Constants.PAGE_SHIFT)));
	 
		// Extract the location of the address within the page.
		OPT_RegisterOperand wordAddr = getTempInt(1);
		appendInstructionToCurrentBlock(Binary.create(INT_AND, wordAddr,
																	 addr.copyRO(),
																	 new OPT_IntConstantOperand(PPC_Constants.PAGE_OFFSET)));

		// Convert to array index (i.e. divide by 4).
		OPT_RegisterOperand wordIndex = getTempInt(2);
		appendInstructionToCurrentBlock(Binary.create(INT_USHR, wordIndex,
																	 wordAddr.copyRO(),
																	 new OPT_IntConstantOperand(2)));
      
		// Retrieve the int[] for the correct page into pageReg.
		OPT_RegisterOperand pageReg = getTempIntArray(0);
		appendInstructionToCurrentBlock(ALoad.create(REF_ALOAD, pageReg,
																	new OPT_RegisterOperand(ps_memory, ps_memory_type),
																	pageIndex.copyRO(),
																	new OPT_LocationOperand(VM_TypeReference.IntArray),
																	new OPT_TrueGuardOperand()));
               
		// Copy to reg from the correct array element.
		appendInstructionToCurrentBlock(ALoad.create(INT_ALOAD, dest,
																	pageReg.copyRO(), 
																	wordIndex.copyRO(),
																	new OPT_LocationOperand(VM_TypeReference.Int),
																	new OPT_TrueGuardOperand()));                               
	 }


  }
  public static final int TEMP_INT_SAFE_FROM_LOAD_CLOBBER=3;

  // -oO System call handling Oo-

  /**
   * Plant a system call
   * @param lazy current translation laziness
   */
  public void plantSystemCall(Laziness lazy) {
    // Need to make sure that the PPC_ProcessSpace registers have
    // the correct values before call to doSysCall().
    resolveLaziness(lazy);
    spillAllRegisters();

    // Plant call
    OPT_Instruction s = Call.create(CALL, null, null, null, null, 1);
    VM_TypeReference psTref = VM_TypeReference.findOrCreate(VM_SystemClassLoader.getVMClassLoader(),
																				VM_Atom.findOrCreateAsciiAtom("Lorg/binarytranslator/PPC_Emulator/PPC_ProcessSpace;"));
    VM_MethodReference methRef = (VM_MethodReference)VM_MemberReference.findOrCreate(psTref,
																												 VM_Atom.findOrCreateAsciiAtom("doSysCall"),
																												 VM_Atom.findOrCreateAsciiAtom("()V"));
    VM_Method method = methRef.resolveInvokeSpecial();

    // VM_CompiledMethod cm = method.getCurrentCompiledMethod();
    // OPT_MethodOperand methOp = OPT_MethodOperand.COMPILED(method, cm.getOsrJTOCoffset());
    OPT_MethodOperand methOp = OPT_MethodOperand.VIRTUAL(methRef, method);

    OPT_Operand psRef = gc.makeLocal(1,psTref); 
    Call.setParam(s, 0, psRef); // Reference to ps, sets 'this' pointer for doSysCall
    Call.setGuard(s, new OPT_TrueGuardOperand());
    Call.setMethod(s, methOp);
    Call.setAddress(s, new OPT_IntConstantOperand(methRef.peekResolvedMethod().getOffset()));
    s.position = gc.inlineSequence;
    s.bcIndex = 1;
    appendInstructionToCurrentBlock(s);

    // Fill all registers again following system call
    fillAllRegisters();
  }

  // -oO Handling bad instructions Oo-

  /**
   * Plant a throw of a bad instruction exception
   * @param lazy current translation laziness
	* @param pc the program counter of the bad instruction
   */
  public void plantThrowBadInstruction(Laziness lazy, int pc) {
	 // There's a bug with this, so for now I'm just commenting it out :-( -- IAR
	 setReturnValueResolveLazinessAndBranchToFinish(lazy, new OPT_IntConstantOperand(0xEBADC0DE));	 
	 if (false) {
    // Need to make sure that the PPC_ProcessSpace registers have
    // the correct values before call to doSysCall().
    resolveLaziness(lazy);
    spillAllRegisters();

	 VM_TypeReference eTref = VM_TypeReference.findOrCreate(VM_SystemClassLoader.getVMClassLoader(),
																			  VM_Atom.findOrCreateAsciiAtom("Lorg/binarytranslator/PPC_Emulator/BadInstructionException;")
																			  );           
	 VM_Class badInstrKlass = (VM_Class)eTref.peekResolvedType();
	 OPT_Operator newOperator;
	 OPT_TypeOperand typeOperand;

	 if ((badInstrKlass != null) && (badInstrKlass.isInitialized() || badInstrKlass.isInBootImage())) {
		newOperator = NEW;
		typeOperand = new OPT_TypeOperand(badInstrKlass);
	 }
	 else {
		newOperator = NEW_UNRESOLVED;
		typeOperand = new OPT_TypeOperand(eTref);
	 }

	 OPT_RegisterOperand eRef = gc.temps.makeTemp(eTref);

	 OPT_Instruction n = New.create(newOperator,eRef, typeOperand);
	 n.position = gc.inlineSequence;
	 n.bcIndex = 0;

	 VM_MemberReference methRef = VM_MemberReference.findOrCreate(eTref, VM_Atom.findOrCreateAsciiAtom("<init>"),
																					  VM_Atom.findOrCreateAsciiAtom("(ILorg/binarytranslator/PPC_Emulator/PPC_ProcessSpace;)V"));
	 VM_Method method = ((VM_MethodReference)methRef).resolveInvokeSpecial();
	 OPT_MethodOperand methOp = OPT_MethodOperand.VIRTUAL((VM_MethodReference)methRef, method);
	 VM_TypeReference psTref = VM_TypeReference.findOrCreate(VM_SystemClassLoader.getVMClassLoader(),
																				VM_Atom.findOrCreateAsciiAtom
																				("Lorg/binarytranslator/PPC_Emulator/PPC_ProcessSpace;"));                       

	 OPT_Operand psRef = gc.makeLocal(1,psTref);

	 OPT_Instruction c = Call.create(CALL, null, null, null, null, 3);

	 Call.setParam(c, 0, eRef.copy()); // 'this' pointer in BadInstructionException.init
	 Call.setParam(c, 1, new OPT_IntConstantOperand(pc));
	 Call.setParam(c, 2, psRef);
	 Call.setGuard(c, new OPT_TrueGuardOperand());
	 Call.setMethod(c, methOp);
	 Call.setAddress(c, new OPT_IntConstantOperand(((VM_MethodReference)methRef).peekResolvedMethod().getOffset()));
	 c.position = gc.inlineSequence;
	 c.bcIndex = 21;

	 OPT_Instruction t = Athrow.create(ATHROW, eRef.copyRO());
	 t.position = gc.inlineSequence;
	 t.bcIndex = 0;

	 appendInstructionToCurrentBlock(n);

	 appendInstructionToCurrentBlock(c);
	 
	 appendInstructionToCurrentBlock(t); 

	 setReturnValueResolveLazinessAndBranchToFinish(lazy, new OPT_IntConstantOperand(0xEBADC0DE));	 
	 }
  }

  // -oO Trace helping methods Oo-

  /**
   * Plant a record bclr call. NB register state won't get resolved for call
   * @param pc the address of the bclr instruction
	* @param lr the link register value
   */
  public void plantRecordUncaughtBclr(int pc, OPT_RegisterOperand lr) {
	 // Is it sensible to record this information?
	 if((gc.options.getOptLevel() > 0) && (PPC_Options.plantUncaughtBclrWatcher)) {
		// Plant call
		OPT_Instruction s = Call.create(CALL, null, null, null, null, 3);
		VM_TypeReference psTref = VM_TypeReference.findOrCreate(VM_SystemClassLoader.getVMClassLoader(),
																				  VM_Atom.findOrCreateAsciiAtom("Lorg/binarytranslator/PPC_Emulator/PPC_ProcessSpace;"));
		VM_MethodReference methRef = (VM_MethodReference)VM_MemberReference.findOrCreate(psTref,
																													VM_Atom.findOrCreateAsciiAtom("recordUncaughtBclr"),
																													VM_Atom.findOrCreateAsciiAtom("(II)V"));
		VM_Method method = methRef.resolve();
		OPT_MethodOperand methOp = OPT_MethodOperand.VIRTUAL(methRef, method);

		OPT_Operand psRef = gc.makeLocal(1,psTref); 
		Call.setParam(s, 0, psRef); // Reference to ps, sets 'this' pointer
		Call.setParam(s, 1, new OPT_IntConstantOperand(pc)); // Address of bclr instruction
		Call.setParam(s, 2, lr);    // Link register value
		Call.setGuard(s, new OPT_TrueGuardOperand());
		Call.setMethod(s, methOp);
		Call.setAddress(s, new OPT_IntConstantOperand(methRef.peekResolvedMethod().getOffset()));
		s.position = gc.inlineSequence;
		s.bcIndex = 7;
		appendInstructionToCurrentBlock(s);
	 }
  }

  /**
   * Plant a record bcctr call. NB register state won't get resolved for call
   * @param pc the address of the bclr instruction
	* @param ctr the count register value
   */
  public void plantRecordUncaughtBcctr(int pc, OPT_RegisterOperand ctr) {
	 if(PPC_Options.plantUncaughtBcctrWatcher) {
		// Plant call
		OPT_Instruction s = Call.create(CALL, null, null, null, null, 3);
		VM_TypeReference psTref = VM_TypeReference.findOrCreate(VM_SystemClassLoader.getVMClassLoader(),
																				  VM_Atom.findOrCreateAsciiAtom("Lorg/binarytranslator/PPC_Emulator/PPC_ProcessSpace;"));
		VM_MethodReference methRef = (VM_MethodReference)VM_MemberReference.findOrCreate(psTref,
																													VM_Atom.findOrCreateAsciiAtom("recordUncaughtBcctr"),
																													VM_Atom.findOrCreateAsciiAtom("(II)V"));
		VM_Method method = methRef.resolve();
		OPT_MethodOperand methOp = OPT_MethodOperand.VIRTUAL(methRef, method);
		OPT_Operand psRef = gc.makeLocal(1,psTref); 
		Call.setParam(s, 0, psRef); // Reference to ps, sets 'this' pointer
		Call.setParam(s, 1, new OPT_IntConstantOperand(pc)); // Address of bcctr instruction
		Call.setParam(s, 2, ctr);   // Count register value
		Call.setGuard(s, new OPT_TrueGuardOperand());
		Call.setMethod(s, methOp);
		Call.setAddress(s, new OPT_IntConstantOperand(methRef.peekResolvedMethod().getOffset()));
		s.position = gc.inlineSequence;
		s.bcIndex = 13;
		appendInstructionToCurrentBlock(s);
	 }
  }

  // -oO Optimisations on the generated HIR Oo-

  /**
	* Eliminate unnecessary register spill and fill code
	*/
  private void eliminateRegisterFills() {
	 Vector unusedRegisterVector = new Vector();
	 for(int i=0; i < 32; i++) {
		if(intRegInUseMap[i] == false) {
		  unusedRegisterVector.add(intRegMap[i]);
		}
		if(fpRegInUseMap[i] == false) {
		  unusedRegisterVector.add(fpRegMap[i]);
		}
	 }
	 if(fpscrRegInUse == false) {
		unusedRegisterVector.add(fpscrRegMap);
	 }
	 if(ctrRegInUse == false) {
		unusedRegisterVector.add(ctrRegMap);
	 }
	 if(crRegInUse == false) {
		unusedRegisterVector.add(crRegMap);
	 }
	 if(lrRegInUse == false) {
		unusedRegisterVector.add(lrRegMap);
	 }
	 if(xerRegInUse == false) {
		unusedRegisterVector.add(xerRegMap);
	 }
	 OPT_Register unusedRegisters[] = (OPT_Register[])unusedRegisterVector.toArray(new OPT_Register[unusedRegisterVector.size()]);
	 OPT_BasicBlock curBB = gc.prologue;
	 while(curBB != null) {
		OPT_Instruction curInstr = curBB.firstInstruction();
		loop_over_instructions:
		while(BBend.conforms(curInstr) == false) {
		  if(PutField.conforms(curInstr)) {
			 OPT_Operand put_value = PutField.getValue(curInstr);
			 if (put_value instanceof OPT_RegisterOperand) {
				OPT_Register put_register = ((OPT_RegisterOperand)put_value).register;
				for (int i=0; i<unusedRegisters.length; i++) {
				  if(unusedRegisters[i] == put_register) {
					 OPT_Instruction toRemove = curInstr;
					 curInstr = curInstr.nextInstructionInCodeOrder();
					 toRemove.remove();
					 continue loop_over_instructions;
				  }
				}
			 }
		  }
		  else if(GetField.conforms(curInstr)) {
			 OPT_Register get_register = GetField.getResult(curInstr).register;
			 if(get_register != xerRegMap) {	// We only remove putfield for the xer register
				for (int i=0; i<unusedRegisters.length; i++) {
				  if(unusedRegisters[i] == get_register) {
					 OPT_Instruction toRemove = curInstr;
					 curInstr = curInstr.nextInstructionInCodeOrder();
					 toRemove.remove();
					 continue loop_over_instructions;
				  }
				}
			 }
		  }
		  curInstr = curInstr.nextInstructionInCodeOrder();
		}
		curBB = curBB.nextBasicBlockInCodeOrder();
	 }
  }

  /**
	* Remove conditional moves
	*/
  static void eliminateConditionalMoves(OPT_IR ir) {
	 OPT_BasicBlock curBB = ir.gc.prologue;
	 while(curBB != null) {
		OPT_Instruction curInstr = curBB.firstInstruction();
		loop_over_instructions:
		while(BBend.conforms(curInstr) == false) {
		  if(CondMove.conforms(curInstr) && (curInstr.operator() == INT_COND_MOVE)) {
			 OPT_BasicBlock compareBlock; // block containing compare instruction
			 OPT_BasicBlock trueBlock;    // block containing true payload
			 OPT_BasicBlock falseBlock;   // block containing false payload
			 OPT_BasicBlock restOfBlock;  // the rest of the block

			 // Set up blocks
			 compareBlock = curBB;
			 trueBlock = new OPT_BasicBlock(0, ir.gc.inlineSequence, ir.gc.cfg);
			 falseBlock = new OPT_BasicBlock(0, ir.gc.inlineSequence, ir.gc.cfg);
			 restOfBlock = compareBlock.splitNodeAt(curInstr, ir);
			 ir.gc.cfg.linkInCodeOrder(compareBlock, falseBlock);
			 ir.gc.cfg.linkInCodeOrder(falseBlock, trueBlock);
			 ir.gc.cfg.linkInCodeOrder(trueBlock, restOfBlock);

			 // Set up values
			 OPT_RegisterOperand result = CondMove.getResult(curInstr);
			 OPT_Operand val1 = CondMove.getVal1(curInstr);
			 OPT_Operand val2 = CondMove.getVal2(curInstr);
			 OPT_ConditionOperand cond = CondMove.getCond(curInstr);
			 OPT_Operand trueValue = CondMove.getTrueValue(curInstr);
			 OPT_Operand falseValue = CondMove.getFalseValue(curInstr);

			 // Create comparison
			 OPT_Operator cmpType;
			 if(val1.isInt()){
				cmpType = INT_IFCMP;
			 }
			 else if (val1.isDouble()){
				cmpType = DOUBLE_IFCMPL;
			 }
			 else {
				throw new Error("Unexpected type of val1: " + val1);
			 }
			 IfCmp.mutate(curInstr, cmpType, null, val1, val2, cond, trueBlock.makeJumpTarget(), OPT_BranchProfileOperand.likely());
			 compareBlock.deleteNormalOut();
			 compareBlock.insertOut(trueBlock);
			 compareBlock.insertOut(falseBlock);

			 // Create false code
			 falseBlock.appendInstruction(Move.create(INT_MOVE, result, falseValue));
			 falseBlock.appendInstruction(Goto.create(GOTO,restOfBlock.makeJumpTarget()));
			 falseBlock.insertOut(restOfBlock);

			 // Create true code
			 trueBlock.appendInstruction(Move.create(INT_MOVE, result, trueValue));
			 trueBlock.insertOut(restOfBlock);
			 
			 // Move along
			 curBB = restOfBlock;
			 curInstr = curBB.firstInstruction();
		  }
		  curInstr = curInstr.nextInstructionInCodeOrder();
		}
		curBB = curBB.nextBasicBlockInCodeOrder();
	 }
  }
  // -oO Debug Oo-

  /**
   * Report some debug output
   */
  private void report(String str) {
    System.out.println(str);
  }

  /**
   * Dump current state of translation
   * @todo implement this
   */
  private void dump() {
  }
} // End of class PPC2IR
