/*
 * (C) Copyright The University of Manchester 2004
 */
package org.binarytranslator.PPC_Emulator;
import java.util.SortedMap;
import java.util.TreeMap;
import java.util.Set;
import java.util.HashSet;
import java.util.Iterator;

public class PPC_BranchLogic {
  /**
	* A set of likely procedure starting points and a vector of all the
	* call sites
	*/
  private SortedMap proceduresAndCallers;

  /**
	* A set of bcctr sites and their destinations
	*/
  private SortedMap bcctrSitesAndDestinations;

  /**
	* Constructor
	*/
  public PPC_BranchLogic() {
	 bcctrSitesAndDestinations = new TreeMap();
	 proceduresAndCallers = new TreeMap();
  }
  /**
	* Constructor
	*/
  public PPC_BranchLogic(PPC_BranchLogic global) {
	 bcctrSitesAndDestinations = global.bcctrSitesAndDestinations;
	 proceduresAndCallers = new TreeMap();
  }
  /**
	* Register a branch and link instruction
	* @param pc the address of the branch instruction (implicity the
	* link value will be assumed to be pc+4)
	* @param dest the destination of the branch instruction
	*/
  public void registerBranchAndLink(int pc, int dest) {
	 Set callers = (Set)proceduresAndCallers.get(new Integer(dest));
	 if (callers != null) {
		if (callers.contains(new Integer(pc))) {
		  // This branch and link is already registered
		  return;
		}
	 }
	 else {
		callers = new HashSet();
		proceduresAndCallers.put(new Integer(dest), callers);
	 }
	 callers.add(new Integer(pc));
  }
  /**
	* Register branch and link instructions
	* @param pcs the set of addresses of the branch instruction
	* (implicity the link value will be assumed to be pc+4)
	* @param dest the destination of the branch instruction
	*/
  public void registerBranchAndLink(Set pcs, Integer dest) {
	 Set callers = (Set)proceduresAndCallers.get(dest);
	 if (callers != null) {
		callers.addAll(pcs);
	 }
	 else {
		proceduresAndCallers.put(dest, pcs);
	 }
  }
  /**
	* Register a branch to the link register
	* @param pc the address of the branch instruction
	* @param lr the value of the link register
	*/
  public void registerBranchToLinkRegister(int pc, int lr) {
	 getLikelyCallSites(pc).add(new Integer(lr - 4));
  }
  /**
	* Register a branch to the count register
	* @param pc the address of the branch instruction
	* @param lr the value of the link register
	*/
  public void registerBranchToCountRegister(int pc, int ctr) {

	 Set dests = (Set)bcctrSitesAndDestinations.get(new Integer(pc));
	 if (dests != null) {
		if (dests.contains(new Integer(ctr))) {
		  // This ctr address is already registered
		  return;
		}
	 }
	 else {
		dests = new HashSet();
		bcctrSitesAndDestinations.put(new Integer(pc), dests);
	 }
	 dests.add(new Integer(ctr));
  }
  /**
	* Returns a vector of likely call sites
	* @param pc location to suggest call sites for
	* @return vector of potential calling instructions
	*/
  public Set getLikelyCallSites(int pc) {
	 if (proceduresAndCallers.size() > 0) {
		SortedMap priorCallDestinations = proceduresAndCallers.headMap(new Integer(pc));
		if (priorCallDestinations.size() > 0) {
		  Integer procedureStart = (Integer)priorCallDestinations.lastKey();
		  return (Set)proceduresAndCallers.get(procedureStart);
		}
	 }
	 return null;
  }
  /**
	* Returns a vector of likely ctr destinations
	* @param pc location of the bcctr
	* @return vector of potential destinations
	*/
  public Set getLikelyCtrDestinations(int pc) {
	 return (Set)bcctrSitesAndDestinations.get(new Integer(pc));
  }
  /**
	* Merge the passed branch information into this object
	* @param to_merge branch information to merge
	*/
  public void merge(PPC_BranchLogic to_merge) {
	 Iterator keys = to_merge.proceduresAndCallers.keySet().iterator();
	 while(keys.hasNext()) {
		Integer dest = (Integer)keys.next();
		Set pcs = (Set)to_merge.proceduresAndCallers.get(dest);
		registerBranchAndLink(pcs, dest);
	 }
  }
}
