/*
 * (C) Copyright The University of Manchester 2004
 */
package org.binarytranslator.PPC_Emulator;

/** 
 * Handle the conbination of a set of bytes into a short or an int, allowing for either byte order. 
 *
 * @author Matthew Horsnell, Richard Matley
 *
 */
public class PPC_ByteHandler
{

  // ELF representation of byte order.
  /**
	* Value of endian argument to a method of this class which indicates little-endian byte ordering.
	*/
  private final static byte LITTLE_ENDIAN = PPC_Constants.ELFDATA2LSB;

  /**
	* Value of endian argument to a method of this class which indicates big-endian byte ordering.
	*/
  private final static byte BIG_ENDIAN = PPC_Constants.ELFDATA2MSB;

  /**
	* 2 to the power of 8.
	*/
  private final static int TWOTO8 = 256;

  /**
	* 2 to the power of 16.
	*/
  private final static int TWOTO16 = 65536;

  /**
	* 2 to the power of 24.
	*/
  private final static int TWOTO24 = 16777216;


  /**
	* Convert an array of four bytes into an integer.
	* @param byte the array of four bytes, starting from the lowest memory/file position.
	* @param endian constant LITTLE_ENDIAN or BIG_ENDIAN as appropriate.
	*/
  public static int toInt(byte[] b, byte endian)
  {

	 int[] theBytes = new int[4];

	 for(int i = 0 ; i < 4 ; i++)
		{

		  theBytes[i] = b[i];
		  // We want a set of unsigned 0-255 values, but Java byte is _signed_, so:
		  if(theBytes[i] < 0 )
			 theBytes[i] += TWOTO8;

		}

	 if(endian == LITTLE_ENDIAN)
		return theBytes[0] + TWOTO8 * theBytes[1] + TWOTO16 * theBytes[2] + TWOTO24 * theBytes[3];
	 else if(endian == BIG_ENDIAN)
		return TWOTO24 * theBytes[0] + TWOTO16 * theBytes[1] + TWOTO8 * theBytes[2] + theBytes[3];
	 else
		return -1; // Should never happen!

  }

  /**
	* Convert an array of two bytes into an integer.
	* @param byte the array of two bytes, starting from the lower memory/file position.
	* @param endian constant LITTLE_ENDIAN or BIG_ENDIAN as appropriate.
	*/
  public static short toShort(byte[] b, byte endian)
  {

	 short[] theBytes = new short[2];

	 for(int i = 0 ; i < 2 ; i++)
		{

		  theBytes[i] = b[i];
		  // We want a set of unsigned 0-255 values, but Java byte is _signed_, so:
		  if(theBytes[i] < 0 )
			 theBytes[i] += TWOTO8;

		}
			
	 if(endian == LITTLE_ENDIAN)
		return (short)(theBytes[0] + TWOTO8 * theBytes[1]);
	 else if(endian == BIG_ENDIAN)
		return (short)(TWOTO8 * theBytes[0] + theBytes[1]);
	 else
		return -1; // Should never happen!

  }

}
