/*
 * (C) Copyright The University of Manchester 2004
 */
package org.binarytranslator.PPC_Emulator;

import java.io.*;

/**
 * Class for loading an ELF file into a PPC_ProcessSpace object. See the documentation for the
 * interface PPC relating to the constants used to describe many of the types/permissions etc. 
 * described below.
 *
 * @author Richard Matley
 */
public class PPC_ELF_Binary
{

  /**
	* The name of the file from which to read the binary.
	*/
  private String filename;

  /**
	* Representation of the ELF file in a form suitable for discovering the canonical path.
	*/
  private File elfFile;

  /**
	* Representation of the ELF file in a form suitable for accessing at specified offsets.
	*/
  private RandomAccessFile rFile;

  /**
	* The ELF header read from the file.
	*/
  private PPC_ELF_Header header;

  /**
	* A table containing the program headers from the ELF file.
	*/
  private PPC_ELF_ProgramHeaderTable programHeaders;

  /** 
	* Construct an ELFBinary object from the contents of the given file.
	* @param filename the name of the file containing the ELF binary.
	* @param ps the process space object into which to copy the binary and in which to execute it.
	* @param argv the command line arguments of the PPC binary (C style, beginning with its name).
	*/
  public PPC_ELF_Binary(String filename, PPC_ProcessSpace ps, String[] argv)
  {

	 this.filename = filename;

	 openFile();
	 readHeader();
	 readProgramHeader();
	 copyToMemory(ps);
	 stackInit(ps, argv);
	 ps.pc = header.entry; // Set program counter.

  }

  /** 
	* Open the ELF file. 
	*/
  private void openFile()
  {

	 try
		{

		  elfFile = new File(filename);
		  rFile = new RandomAccessFile(elfFile, "r");

		}
	 catch(FileNotFoundException e)
		{

		  System.err.println("PPC_ELF_Binary: Error reading file\n");
		  System.exit(-1);

		}



  }

  /** 
	* Read the header from the ELF file. 
	*/
  public void readHeader()
  {

	 header = new PPC_ELF_Header(rFile);

  }

  /** 
	* Read the program header. 
	*/
  public void readProgramHeader()
  {

	 if(header.pHdrPos != 0)
		{

		  if(header.pHdrPos != header.hdrSize)
			 {

				System.err.println("ELFBinary: Program header does not follow ELF header.");
				System.exit(-1);

			 }

		  programHeaders = new PPC_ELF_ProgramHeaderTable(rFile, header);

		}

  }

  /** 
	* Copy the ELF file to memory. 
	*/
  public void copyToMemory(PPC_ProcessSpace ps)
  {

	 for(int s = 0 ; s < header.pHdrCnt ; s++)
		{

		  // Pass the access permissions from the program header 'flags' into mmaps 'prot'.
		  int prot = 0;
		  if( (programHeaders.segment[s].flags & PPC_Constants.PF_X) != 0 )
			 prot = prot | PPC_Constants.PROT_EXEC;
		  if( (programHeaders.segment[s].flags & PPC_Constants.PF_W) != 0 )
			 prot = prot | PPC_Constants.PROT_WRITE;
		  if( (programHeaders.segment[s].flags & PPC_Constants.PF_R) != 0 )
			 prot = prot | PPC_Constants.PROT_READ;

		  if(programHeaders.segment[s].type == PPC_Constants.PT_LOAD)
			 {

				ps.mmap(programHeaders.segment[s].virtAddr & PPC_Constants.PAGE_START, 
						  programHeaders.segment[s].fileSize + 
						  (programHeaders.segment[s].virtAddr & PPC_Constants.PAGE_OFFSET),
						  prot, PPC_Constants.MAP_FIXED | PPC_Constants.MAP_PRIVATE,
						  elfFile, programHeaders.segment[s].offset - 
						  (programHeaders.segment[s].virtAddr & PPC_Constants.PAGE_OFFSET) );

				if(programHeaders.segment[s].memSize > programHeaders.segment[s].fileSize)
				  {

					 /* We need some space for uninitialised data. In general we will already
						 have some of this at the end of the (last) page of initialised data, but
						 are likely to need at least one anonymous page. */
																
					 int uStart = (programHeaders.segment[s].virtAddr + 
										programHeaders.segment[s].fileSize + 0xfff) & 
						PPC_Constants.PAGE_START;
					 // Start address of the area to allocate.

					 int uLength = programHeaders.segment[s].virtAddr + programHeaders.segment[s].memSize - uStart; // Length of area to allocate.

					 if(PPC_Options.debugSyscallMore)
						{

						  System.out.println("\n\nuStart = 0x" + Integer.toHexString(uStart));
						  System.out.println("uLength = 0x" + Integer.toHexString(uLength));

						}

					 if(uLength > 0)
						{

						  ps.mmap(uStart, uLength, prot, PPC_Constants.MAP_FIXED | PPC_Constants.MAP_PRIVATE | PPC_Constants.MAP_ANONYMOUS, null, 0);

						}


				  }

			 }

		} 

	 /***** ??????????????? PROBABLY NEED TO CHANGE ABOVE TO GET THE FLAGS PARAM. RIGHT *****/

	 /* Set the value of ps.brk to the top of the bss section. For now we just assume this is the top of
		 segment 1 (which is probably safe).*/

	 ps.initBrk(programHeaders.segment[1].virtAddr + programHeaders.segment[1].memSize);

  }

  /** 
	* Set up the initial process stack. 
	* @param ps the PPC_ProcessSpace object in which we are operating.
	* @param argv the command line arguments of the binary.
	* Currently (sensible) 'dummy' values are used for the environment variables. The values for the
	* auxiliary vector are read from the ELF file, or set to something which I hope makes sense.
	*/
  public void stackInit(PPC_ProcessSpace ps, String[] argv)
  {
	 /* Set up an auxiliary vector containing suitable values for the binary. */
	 int[] auxVector = {PPC_Constants.AT_IGNOREPPC, PPC_Constants.AT_IGNOREPPC, PPC_Constants.AT_IGNOREPPC, PPC_Constants.AT_IGNOREPPC,
							  PPC_Constants.AT_DCACHEBSIZE, 0x20, PPC_Constants.AT_ICACHEBSIZE, 0x20,
							  PPC_Constants.AT_UCACHEBSIZE, 0x0, PPC_Constants.AT_HWCAP, 0x8c000000, 
							  PPC_Constants.AT_PAGESZ, 0x1000, PPC_Constants.AT_CLKTCK, 0x64, 
							  PPC_Constants.AT_PHDR, header.pHdrPos - programHeaders.segment[0].offset + 
							  programHeaders.segment[0].virtAddr, /* This should give the memory location
																				  of the program header. (Assuming
																				  it is in segment 0). */
							  PPC_Constants.AT_PHENT, header.pHdrEnt, 
							  PPC_Constants.AT_PHNUM, header.pHdrCnt, PPC_Constants.AT_BASE, 0x0, 
							  PPC_Constants.AT_FLAGS, 0x0, PPC_Constants.AT_ENTRY, header.entry, 
							  PPC_Constants.AT_UID, PPC_Constants.UID, PPC_Constants.AT_EUID, PPC_Constants.UID,
							  PPC_Constants.AT_GID, PPC_Constants.GID, PPC_Constants.AT_EGID, PPC_Constants.GID,
							  PPC_Constants.AT_NULL, 0x0};
	 /* The cache sizes and flags are as for a test program running on 
		 the iBook, softwood. AT_BASE will need to be changed for dynamically linked binaries. */

	 /* Environment variables. These are some from my standard setup. */
	 /* String[] env = {"HOSTNAME=softwood",
		 "PVM_RSH=/usr/bin/rsh",
		 "SHELL=/bin/bash",
		 "TERM=xterm",
		 "HISTSIZE=1000",
		 "SSH_TTY=/dev/pts/0",
		 "USER=matleyr",
		 "PATH=/usr/kerberos/bin:/usr/local/bin:/bin:/usr/bin:/usr/X11R6/bin:/opt/lib/j2re1.3.1/bin:/home/M03/cc/matleyr/bin",
		 "MAIL=/var/spool/mail/matleyr",
		 "_=/bin/bash",
		 "PWD=/home/M03/cc/matleyr",
		 "INPUTRC=/etc/inputrc",
		 "LANG=en_GB.iso88591",
		 "SSH_ASKPASS=/usr/libexec/openssh/gnome-ssh-askpass",
		 "HOME=/home/M03/cc/matleyr",
		 "SHLVL=1",
		 "LOGNAME=matleyr",
		 "PRINTER=happy_duplex",
		 "LESSOPEN=|/usr/bin/lesspipe.sh %s",
		 "DISPLAY=localhost:10.0",
		 "G_BROKEN_FILENAMES=1"};
	 */

	 /* Environment variables, exactly as on softwood. Not that the number 8380 in SSH_* varies. */
	 String[] env = {"HOSTNAME=softwood", "PVM_RSH=/usr/bin/rsh", 
						  "HOST_JAVA_HOME=/home/amulocal/linux/appl/j2sdk1.4.2", "SHELL=/bin/bash",
						  "TERM=xterm", "HISTSIZE=1000", "SSH_CLIENT=130.88.194.110 8380 22",
						  "CVSROOT=/home/simgroup/cvsroot", "QTDIR=/usr/lib/qt-3.1", "SSH_TTY=/dev/pts/0",
						  "RVM_HOST_CONFIG=/home/matleyr/cvs/rvm/config/i686-pc-linux-gnu.ManCS",
						  "USER=matleyr", "LS_COLORS=no=00:fi=00:di=00;34:ln=00;36:pi=40;33:so=00;35:bd=40;33;01:cd=40;33;01:or=01;05;37;41:mi=01;05;37;41:ex=00;32:*.cmd=00;32:*.exe=00;32:*.com=00;32:*.btm=00;32:*.bat=00;32:*.sh=00;32:*.csh=00;32:*.tar=00;31:*.tgz=00;31:*.arj=00;31:*.taz=00;31:*.lzh=00;31:*.zip=00;31:*.z=00;31:*.Z=00;31:*.gz=00;31:*.bz2=00;31:*.bz=00;31:*.tz=00;31:*.rpm=00;31:*.cpio=00;31:*.jpg=00;35:*.gif=00;35:*.bmp=00;35:*.xbm=00;35:*.xpm=00;35:*.png=00;35:*.tif=00;35:", "XENVIRONMENT=/home/matleyr/.Xdefaults",
						  "PVM_ROOT=/usr/share/pvm3", "CLASSPATH_ROOT=/home/matleyr/cvs/classpath",
						  "PATH=/home/matleyr/bin:/usr/kerberos/bin:/usr/local/bin:/bin:/usr/bin:/usr/X11R6/bin:/opt/lib/j2re1.3.1/bin:/home/matleyr/cvs/rvm/bin:/home/matleyr/bin", "MAIL=/var/spool/mail/matleyr",
						  "_=/bin/bash", "PWD=/home/matleyr/dhry", "INPUTRC=/etc/inputrc",
						  "LANG=en_GB.iso88591", "LAMHELPFILE=/etc/lam/lam-helpfile",
						  "SSH_ASKPASS=/usr/libexec/openssh/gnome-ssh-askpass",
						  "CSHOME=matleyr@antigua.cs.man.ac.uk:/home/M03/cc/matleyr", "HOME=/home/matleyr",
						  "SHLVL=1", "SIM=/home/simgroup/matleyr", "XPVM_ROOT=/usr/share/pvm3/xpvm",
						  "RVM_ROOT=/home/matleyr/cvs", "LOGNAME=matleyr", "PRINTER=happy_duplex",
						  "SSH_CONNECTION=130.88.194.110 2380 130.88.198.215 22",
						  "LESSOPEN=|/usr/bin/lesspipe.sh %s", "RVM_BUILD=/tmp/RVMbuild",
						  "DISPLAY=localhost:10.0",
						  "RVM_TARGET_CONFIG=/home/matleyr/cvs/rvm/config/i686-pc-linux-gnu.ManCS",
						  "G_BROKEN_FILENAMES=1"};

	 ps.stackInit(argv, env, auxVector);		

  }

}


/*************************************************************************************************/

/** 
 * ELF header. 
 */
class PPC_ELF_Header
{

  /**
	* The magic number read from the file (should be \177 E L F).
	*/
  byte[] magic = new byte[4];

  /**
	* The address size for which the binary was compiled, 1 = 32bit, 2 = 64bit.
	*/
  byte addressSize;

  /**
	* 1 = little-endian, 2 = big-endian.
	*/
  byte byteOrder;

  /**
	* This should always be 1.
	*/
  byte headerVersion; 

  /**
	* The type of ELF file, for an executable this is 2.
	*/
  short fileType; 

  /**
	* The architecture for which the binary was compiled, 20 = PowerPC.
	*/
  short archType;

  /**
	* This should always be 1.
	*/
  int fileVersion; 

  /**
	* The entry point for execution (assuming the ELF is an executable).
	*/
  int  entry;

  /**
	* The position in the file of the program header (must exist for an executable).
	*/
  int pHdrPos; 

  /**
	* The position in the file of the section header, or 0 if none is present.
	*/
  int sHdrPos; 

  /**
	* Some architecture specifc flags.
	*/
  int flags; 

  /**
	* The size in bytes of this header in the file.
	*/
  short hdrSize; 

  /**
	* The size in bytes of a program header entry.
	*/
  short pHdrEnt;

  /**
	* The number of program header entries.
	*/
  short pHdrCnt;

  /**
	* The size in bytes of a section header entry.
	*/
  short sHdrEnt;

  /** 
	* The number of section header entries.
	*/
  short sHdrCnt; 

  /**
	* The number of the section which contains the section name strings.
	*/
  short strSec; 

  /**
	* Used to read in two bytes which can then be interpreted as a short integer
	* quantity according to byte order.
	*/
  byte[] shBytes = new byte[2];

  /**
	* Used to read in four bytes which can then be interpreted as an integer
	* quantity according to byte order.
	*/
  byte[] intBytes = new byte[4];

  /**
	* Construct a PPC_ELF_Header object and fill in with the values read from the file.
	*/
  public PPC_ELF_Header(RandomAccessFile rFile)
  {

	 try
		{

		  // Read in the header data.
		  rFile.read(magic);
		  addressSize = (byte)rFile.read();
		  byteOrder = (byte)rFile.read();
		  headerVersion = (byte)rFile.read();

		  rFile.skipBytes(9); // Padding

		  rFile.read(shBytes);
		  fileType = PPC_ByteHandler.toShort(shBytes, byteOrder);
		  rFile.read(shBytes);
		  archType = PPC_ByteHandler.toShort(shBytes, byteOrder);

		  rFile.read(intBytes);
		  fileVersion = PPC_ByteHandler.toInt(intBytes, byteOrder);
		  rFile.read(intBytes);
		  entry = PPC_ByteHandler.toInt(intBytes, byteOrder);
		  rFile.read(intBytes);
		  pHdrPos = PPC_ByteHandler.toInt(intBytes, byteOrder);
		  rFile.read(intBytes);
		  sHdrPos = PPC_ByteHandler.toInt(intBytes, byteOrder);
		  rFile.read(intBytes);
		  flags = PPC_ByteHandler.toInt(intBytes, byteOrder);

		  rFile.read(shBytes);
		  hdrSize = PPC_ByteHandler.toShort(shBytes, byteOrder);
		  rFile.read(shBytes);
		  pHdrEnt = PPC_ByteHandler.toShort(shBytes, byteOrder);
		  rFile.read(shBytes);
		  pHdrCnt = PPC_ByteHandler.toShort(shBytes, byteOrder);
		  rFile.read(shBytes);
		  sHdrEnt = PPC_ByteHandler.toShort(shBytes, byteOrder);
		  rFile.read(shBytes);
		  sHdrCnt = PPC_ByteHandler.toShort(shBytes, byteOrder);
		  rFile.read(shBytes);
		  strSec = PPC_ByteHandler.toShort(shBytes, byteOrder);

		  if(PPC_Options.debugSyscallMore)
			 {
				System.out.println("Address size = " + addressSize + " Byte order = " + byteOrder + 
										 " Header version = " + headerVersion);
				System.out.println("File type = " + fileType + " Arch = " + archType + 
										 " File version = " + fileVersion);
				System.out.println("Entry = " + entry);
				System.out.println("HdrSize = " + hdrSize);
				System.out.println("PHdrCnt = " + pHdrCnt);
				System.out.println("SHdrCnt = " + sHdrCnt + " StrSec = " + strSec);
			 }

		  // Run some tests to ensure we have a suitable ELF file.
		  if(magic[0] != '\177' || magic[1] != 'E' || magic[2] != 'L' || magic[3] != 'F')
			 {

				System.err.println("ELFBinary: Specified file is not an ELF file.");
				//System.exit(-1);

			 }

		  if(fileType != PPC_Constants.ET_EXEC)
			 {

				System.err.println("ELFBinary: Specified file is not an executable.");
				//System.exit(-1);

			 }

		  if(archType != PPC_Constants.EM_PPC)
			 {

				System.err.println("ELFBinary: Specified file is not for PPC architecture.");
				//System.exit(-1);

			 }

		  if(headerVersion != PPC_Constants.EV_CURRENT || fileVersion != PPC_Constants.EV_CURRENT)
			 {

				System.err.println("ELFBinary: Unknown version.");
				//System.exit(-1);

			 }

		}
	 catch(IOException e)
		{

		  e.printStackTrace();

		}

  }
}



/*************************************************************************************************/

/** 
 * ELF program header table. 
 */
class PPC_ELF_ProgramHeaderTable
{

  /**
	* The program headers for each of the segments in the ELF file.
	*/
  PPC_ELF_ProgramHeader[] segment;

  /**
	* Construct a PPC_ELF_ProgramHeaderTable, and as many PPC_ELF_ProgramHeader objects as
	* required.
	*/
  public PPC_ELF_ProgramHeaderTable(RandomAccessFile rFile, PPC_ELF_Header header)
  {

	 segment = new PPC_ELF_ProgramHeader[header.pHdrCnt];

	 for(int i = 0 ; i < segment.length ; i++)
		{

		  segment[i] = new PPC_ELF_ProgramHeader(rFile, header);

		}

  }

}




/*************************************************************************************************/

/** 
 * ELF program header. 
 */
class PPC_ELF_ProgramHeader
{

  /**
	* The type of the segment to which this header entry relates. Type 1 is loadable code/date.
	*/
  int type; 

  /**
	* The offset in bytes of the segment within the file.
	*/
  int offset; // File offset of the segment.

  /**
	*  The virtual address to which the segment is to be mapped.
	*/
  int virtAddr;

  /**
	* This is not actually used.
	*/
  int physAddr; 

  /**
	* The size of the segment in the file.
	*/
  int fileSize; 

  /**
	* The size of the segment in memory (if uninitialised data is present, this will be larger
	* than fileSize.
	*/
  int memSize;

  /**
	* Read, write and execute permissions.
	*/
  int flags; 

  /**
	* i.e. the hardware page size.
	*/
  int align; 

  /**
	* Used to read in four bytes which can then be interpreted as an integer
	* quantity according to byte order.
	*/
  byte[] intBytes = new byte[4];

  /**
	* Construct a PPC_ELF_ProgramHeader and fill in the values from the ELF file.
	*/
  public PPC_ELF_ProgramHeader(RandomAccessFile rFile, PPC_ELF_Header header)
  {

	 byte byteOrder = header.byteOrder;

	 try
		{

		  rFile.read(intBytes);
		  type = PPC_ByteHandler.toInt(intBytes, byteOrder);
		  rFile.read(intBytes);
		  offset = PPC_ByteHandler.toInt(intBytes, byteOrder);
		  rFile.read(intBytes);
		  virtAddr = PPC_ByteHandler.toInt(intBytes, byteOrder);
		  rFile.read(intBytes);
		  physAddr = PPC_ByteHandler.toInt(intBytes, byteOrder);
		  rFile.read(intBytes);
		  fileSize = PPC_ByteHandler.toInt(intBytes, byteOrder);
		  rFile.read(intBytes);
		  memSize = PPC_ByteHandler.toInt(intBytes, byteOrder);
		  rFile.read(intBytes);
		  flags = PPC_ByteHandler.toInt(intBytes, byteOrder);
		  rFile.read(intBytes);
		  align = PPC_ByteHandler.toInt(intBytes, byteOrder);
								
		  // In case the program header size is > 32 bytes (the eight ints we have just read).
		  rFile.skipBytes(header.pHdrEnt - 32);

		  if(PPC_Options.debugSyscallMore)
			 {
				System.out.println("Read program header. Type = " + type + " Offset = " + offset);
			 }

		}
	 catch(IOException e)
		{

		  System.err.println("ELFBinary: Error reading program header.");
		  System.exit(-1);

		}
				
  }

}
