/*
 * (C) Copyright The University of Manchester 2004
 */
package org.binarytranslator.PPC_Emulator;

import com.ibm.JikesRVM.VM_CompiledMethod;
import com.ibm.JikesRVM.VM_CodeArray;
import com.ibm.JikesRVM.adaptive.VM_Controller;
import com.ibm.JikesRVM.VM_RuntimeCompiler;
import com.ibm.JikesRVM.VM_PragmaNoInline;
import com.ibm.JikesRVM.classloader.*;
import com.ibm.JikesRVM.opt.ir.*;
import com.ibm.JikesRVM.opt.*;
import java.util.*;
import java.io.*;

/** 
 * Simulate the process address space for our PowerPC ELF binary. Also keep track of register contents 
 *		(currently only those used in user models UISA and VEA). 
 *
 * @author Richard Matley, Ian Rogers
 *
 */
public class PPC_ProcessSpace
{

  /**
	* Flag that amongst other things says with doSysCall has been
	* compiled (prevents unresolved errors)
	*/
  private boolean initialized=false;

  /* Here's what would be in the PowerPC's registers if the binary
	  were running on a real PowerPC.  For speed I am using individual
	  variables, not arrays for GRPs and FPRs. */

  /**
	* The contents of the general purpose (int) registers.
	*/
  public int r0, r1, r2, r3, r4, r5, r6, r7, r8, r9, 
	 r10, r11, r12, r13, r14, r15, r16, r17, r18, r19,
	 r20, r21, r22, r23, r24, r25, r26, r27, r28, r29,
	 r30, r31;

  /**
	* The contents of the floating point registers.
	*/
  public double f0, f1, f2, f3, f4, f5, f6, f7, f8, f9, 
	 f10, f11, f12, f13, f14, f15, f16, f17, f18, f19,
	 f20, f21, f22, f23, f24, f25, f26, f27, f28, f29,
	 f30, f31;

  /**
	* The contents of the condition register.
	*/
  private int cr;

  /**
	* The contents of the floating-point status and control register.
	*/
  private int fpscr;

  /**
	* The contents of the XER register.
	*/
  private int xer;

  /**
	* The contents of the link register.
	*/
  private int lr;

  /** 
	* The contents of the counter register.
	*/
  private int ctr;

  /**
	* The contents of the time base registers.
	*/
  private int tbl, tbu;

  /**
	* Pointer to where we are up to in the program (the PPC doesn't
	* actually have a program counter register, but we need to keep
	* track of our place in the binary).
	*/
  public int pc;

  /**
	* The process's address space consisting of1M pages of 4kB size.
	*/
  private int[][] memory = new int[PPC_Constants.TOTAL_PAGES][];

  /**
	* Hashtable mapping PPC address to pre-translated code
	*/
  private Hashtable codeHash;

  /**
	* Information held about branches, call sites, return sites and the
	* like
	*/
  public PPC_BranchLogic branchInfo;

  /**
	* Read permission flag for a page in the memory array.
	*/
  private boolean[] flag_read = new boolean[PPC_Constants.TOTAL_PAGES];

  /**
	* Write permission flag for a page in the memory array.
	*/
  private boolean[] flag_write = new boolean[PPC_Constants.TOTAL_PAGES];

  /**
	* Execute permission flag for a page in the memory array.
	*/
  private boolean[] flag_exec = new boolean[PPC_Constants.TOTAL_PAGES];

  /**
	* Set of all allocated memory regions.
	*/
  private TreeSet regions = new TreeSet(new PPC_MemoryComparator());

  /**
	* The place in the process space memory to begin searching for a free area for mmap.
	*/
  private int freeAreaCache = (PPC_Constants.TASK_SIZE_P/8*3) << PPC_Constants.PAGE_SHIFT;

  /**
	* The top of the bss. 
	*/
  private int brk;

  /**
	* This is true if we had an exit system call.
	*/
  public boolean finished = false;

  private static final int MAX_FILES=1024;

  /**
	* Array of open files, as RandomAccessFile objects, for read/write.
	*/
  private RandomAccessFile[] raFiles;

  /**
	* Array of open files, as File objects for fstat.
	*/
  private File[] files;

  /**
	* The default compiler options
	*/
  private OPT_Options opts;

  /**
	* The default/initial optimization plan
	*/
  private OPT_OptimizationPlanElement[] initialOptimizationPlan;

  /**
	* PPC_ProcessSpace constructor.
	*/
  public PPC_ProcessSpace(String args[]) throws java.io.IOException {
	 // Create the options used for initial compilation of traces
	 opts = new OPT_Options();
	 opts.setOptLevel(PPC_Options.initialOptLevel);
	 processCommandLineOptions(opts,
										PPC_Options.initialOptLevel,
										VM_Controller.options.MAX_OPT_LEVEL,
										VM_Controller.getOptCompilerOptions());

	 // Create the initial optimization plan
	 initialOptimizationPlan = PPC_OptimizationPlanner.createOptimizationPlan(opts);

	 // Initialize the optimizing compiler if necessary (@see OPT_TestHarness)
	 if(OPT_Compiler.isInitialized() == false) {
		if (PPC_Options.debugRuntime) {
		  System.out.println("PPC_EmulatorRuntime: initialising optimizing compiler");
		}
		OPT_Compiler.init(opts);
	 }

	 // Load binary
	 PPC_ELF_Binary elf = new PPC_ELF_Binary(args[0], this, args);
	 // args[0] is the name of the ELF binary, change to full path.
	 File tempFile = new File(args[0]);
	 args[0] = tempFile.getCanonicalPath();

	 // Hashtable mapping PPC address to pre-translated code
	 codeHash = new Hashtable();

	 // Information on branches
	 branchInfo = new PPC_BranchLogic();

	 // Explicitly call doSysCall to force its compilation
	 doSysCall();
	 initialized = true;

	 // Set up files
	 raFiles = new RandomAccessFile[MAX_FILES];
	 files = new File[MAX_FILES];
  }

  /**
   * Process the command line arguments and pass the appropriate ones to the 
   * OPT_Options
   * 
   * @param options The options being constructed
   * @param optLevel The level of the options being constructed
   * @param maxOptLevel The maximum valid opt level
   * @param optCompilerOptions The list of command line options
   */
  static void processCommandLineOptions(OPT_Options options, int optLevel, int maxOptLevel,
                                        String optCompilerOptions[]) {

    String prefix = "opt"+optLevel+":";
    for (int j=0; j<optCompilerOptions.length; j++) {
      if (optCompilerOptions[j].startsWith("opt:")) {
        String option = optCompilerOptions[j].substring(4);
        if (!options.processAsOption("-X:recomp:", option)) {
          System.out.println("vm: Unrecognized optimizing compiler command line argument: \""
									  +option+"\" passed in as "
									  +optCompilerOptions[j]+"\n");
        }
      } else if (optCompilerOptions[j].startsWith(prefix)) {
        String option = optCompilerOptions[j].substring(5);
        if (!options.processAsOption("-X:recomp:"+prefix, option)) {
          System.out.println("vm: Unrecognized optimizing compiler command line argument: \""
									  +option+"\" passed in as "
									  +optCompilerOptions[j]+"\n");
        }
      }
    }
    // TODO: check for optimization levels that are invalid; that is, 
    // greater than optLevelMax.
    //
    for (int j=0; j<optCompilerOptions.length; j++) {
      if (!optCompilerOptions[j].startsWith("opt")) {
        // This should never be the case!
        continue;
      }
      if (!optCompilerOptions[j].startsWith("opt:")) {
        // must specify optimization level!
        int endPoint = optCompilerOptions[j].indexOf(":");
        if (endPoint == -1) {
          System.out.println("vm: Unrecognized optimization level in optimizing compiler command line argument: \""
									  +optCompilerOptions[j]+"\"\n");
        }
        String optLevelS;
        try {
          optLevelS = optCompilerOptions[j].substring(3,endPoint);
        } catch (IndexOutOfBoundsException e) {
          System.out.println("vm internal error: trying to find opt level has thrown indexOutOfBoundsException\n");
          e.printStackTrace();
          continue;
        }
        try {
          Integer optLevelI = new Integer(optLevelS);
          int cmdOptLevel = optLevelI.intValue();
          if (cmdOptLevel > maxOptLevel) {
            System.out.println("vm: Invalid optimization level in optimizing compiler command line argument: \""
										 +optCompilerOptions[j]+"\"\n"+
										 "  Specified optimization level "+cmdOptLevel+
										 " must be less than "+maxOptLevel+"\n");
          }
        } catch (NumberFormatException e) {
          System.out.println("vm: Unrecognized optimization level in optimizing compiler command line argument: \""
									  +optCompilerOptions[j]+"\"\n");
        }
      }
    }
    VM_RuntimeCompiler.setNoCacheFlush(options);
  }

  /**
	* Get a VM_CodeArray object for a specified pc
	*/
  public synchronized VM_CodeArray getCodeForPC(int pc) {
	 VM_CodeArray code = (VM_CodeArray)codeHash.get(new Integer(pc));	 
	 if(code == null) {
		code = translateCode(new VM_PPC_Trace(this, pc));
	 }
	 return code;
  }

  /**
	* Replace the compiled code for a trace (called by the adaptive
	* system)
	*/
  public synchronized void replaceCompiledTrace(VM_CompiledMethod cm, VM_PPC_Trace trace) {
	 VM_CodeArray code = cm.getInstructions();
	 codeHash.put(new Integer(trace.pc), code);
  }

  /**
	* Translate the code for the specified pc value
	* {@see OptTestHarness.compileMethodsInVector}
	* @param pc the location to start translating from
	*/
  private VM_CodeArray translateCode(VM_PPC_Trace trace) {
	 if (PPC_Options.debugRuntime) {
		System.out.println("***** TRANSLATING CODE FOR 0x" + Integer.toHexString(trace.pc) + " *****\n" +
								 "PPC_EmulatorRuntime: trace = " + trace.toString() + "\n" +
								 "Optimisation level: " + opts.getOptLevel());
	 }
	 OPT_CompilationPlan cp = new OPT_CompilationPlan(trace, initialOptimizationPlan, null, opts);
	 VM_CompiledMethod cm = OPT_Compiler.compile(cp);
	 replaceCompiledTrace(cm, trace);
	 return cm.getInstructions();
  }
  
  /**
	* Record a bclr that wasn't caught using the branchInfo object
	*/
  public void recordUncaughtBclr(int pc, int lr) {
	 if (PPC_Options.debugRuntime) {
		System.out.println("Uncaught bclr: 0x" + Integer.toHexString(pc) + " 0x" + Integer.toHexString(lr));
	 }
	 branchInfo.registerBranchToLinkRegister(pc, lr);
  }

  /**
	* Record a bcctr that wasn't caught using the branchInfo object
	*/
  public void recordUncaughtBcctr(int pc, int ctr) {
	 if (PPC_Options.debugRuntime) {
		System.out.println("Uncaught bcctr: 0x" + Integer.toHexString(pc) + " 0x" + Integer.toHexString(ctr));
	 }
	 branchInfo.registerBranchToCountRegister(pc, ctr);
  }

  /**
	* Return the first available file descriptor index. 
	*/
  private int getFileDesc()
  {
	 for(int f = 3 ; f < MAX_FILES ; f++) {
		if(files[f] == null) {
		  return f;
		}
	 }
	 throw new Error("Run out of file descriptors.");
  }

  /** 
	* Find the MemoryRegion containing a particular address if any,
	* otherwise the first one located in a search to higher
	* addresses. Like find_vma in Linux.
	* @param addr an address in the PPC_ProcessSpace memory.
	*/
  public PPC_MemoryRegion findMemoryRegion(int addr)
  {
	 PPC_MemoryRegion m;

	 for(Iterator i = regions.iterator() ; i.hasNext() ; )
		{

		  m = (PPC_MemoryRegion)i.next();

		  if(m.regionEnd > addr)
			 {

				return m;
				
			 }

		}

	 return null; // Means that whole address range above addr is unallocated.

  }

  /**
	* Find MemoryRegion containing a particular address, or with that
	* address as its end point.  Called by setBrk, this allows for the
	* case that the current brk address is the exact end point of a
	* PPC_MemoryRegion.
	*/
  public PPC_MemoryRegion findMemoryRegionInclEnd(int addr)
  {

	 PPC_MemoryRegion m;

	 for(Iterator i = regions.iterator() ; i.hasNext() ; )
		{

		  m = (PPC_MemoryRegion)i.next();
								
		  if(m.regionEnd >= addr)
			 {
												
				return m;
				
			 }

		}

	 return null; // Means that whole address range above addr is unallocated.

  }

  /** 
	* Find a suitably sized free memory area. Like Linux's
	* arch_get_unmapped_area.
	* @param start is the point at which to start searching (if set to
	* 0, the search begins at the value in freeAreaCache).
	* @param length the size of the memory area requested, in bytes.
	*/
  public int getUnmappedArea(int start, int length)
  {

	 PPC_MemoryRegion m;

	 if(start != 0)
		{

		  // Can we start at the requested place?
		  m = findMemoryRegion(start); // First region which ends after 'start'.

		  if(m != null)
			 {

				if( m.regionStart >= (start + length) ) // If we can fit the new region in before it.
				  return start; 
				
			 }
		  else // No region allocated above 'start'.
			 {

				if( ((start >>> PPC_Constants.PAGE_SHIFT) + (length >>> PPC_Constants.PAGE_SHIFT)) < PPC_Constants.TASK_SIZE_P )
				  return start;

			 }
								
		}

	 /* If we get here, either no 'start' was specified, or we cannot put the region there.
		 So we search for a suitable place. */

	 start = freeAreaCache;

	 while(true)
		{

		  m = findMemoryRegion(start);

		  if(m == null) // No region allocated above freeAreaCache.
			 {

				if( ((start >>> PPC_Constants.PAGE_SHIFT) + (length >>> PPC_Constants.PAGE_SHIFT)) < PPC_Constants.TASK_SIZE_P )
				  return start;
				else
				  throw new Error("ProcessSpace: Insufficient free memory.");
														
			 }
		  else // Can we fit in before the region m?
			 {

				if( m.regionStart >= (start + length) )
				  return start;
				else
				  start = m.regionEnd; // The only case which keeps us in the while loop.

			 }
								
		}

  }


		
  /** 
	* Simulate an mmap syscall to allocate a region of (simulated)
	* memory in the process space, either anonymously or to map the
	* contents of a file.
	* @param start requested starting point in memory.
	* @param prot describes the required memory 
	* @param flags various options including shared/private. 
	* @param length number of bytes requested.
	* @param file the file (if any) from which to map into the memory
	* @param offset starting point within the file of the data to map into memory.
	*
	* Both start and offset must be multiples of the size of a page,
	* 4kB. Although only whole pages are allocated, length can be any
	* multiple of 4 bytes, any remaining space in the last page being
	* zeroed.
	*/
  public int mmap(int start, int length, int prot, int flags, File file, int offset)
  {

	 // Do some check of the parameters.
	 if( (flags & PPC_Constants.MAP_GROWSDOWN) != 0 ) // ??????????Don't yet do anything different for this.
		{	
		  if(PPC_Options.debugSyscallMore) {
			 System.out.println("ProcessSpace: Note that the MAP_GROWSDOWN flag is currently ignored.");
		  }
		}
	 if( (flags & PPC_Constants.MAP_SHARED) != 0 ) // For now we only handle MAP_PRIVATE requests
		throw new Error("ProcessSpace: Don't yet support MAP_SHARED");

	 if( (flags & PPC_Constants.MAP_PRIVATE) == 0)
		throw new Error("ProcessSpace: mmap request must specify MAP_PRIVATE or MAP_SHARED.");
				
	 if( (start & 0xfff) != 0 )
		throw new Error("ProcessSpace: Badly aligned mmap request (start).");

	 if( (offset & 0xfff) != 0 )
		throw new Error("ProcessSpace: Badly aligned mmap request (offset).");

	 if(length == 0)
		throw new Error("ProcessSpace: Requesting zero length memory allocation.");

	 if( (length & 0x3) != 0)
		throw new Error("ProcessSpace: Length of area to map must be a multiple of 4 bytes.");

	 if( (length >>> PPC_Constants.PAGE_SHIFT) > PPC_Constants.TASK_SIZE_P)
		throw new Error("ProcessSpace: Requesting memory > PPC_Constants.TASK_SIZE");

	 int numPagesReq; // number of 4kB pages requested.
	 int startPage; // Page number for start of memory region.

	 /* If start is given as 0 this means choose somewhere to put it. */
	 if(start == 0)
		{

		  if( (flags & PPC_Constants.MAP_FIXED) != 0 )
			 {

				throw new Error("ProcessSpace: Cannot specify MAP_FIXED with start=0 in mmap.");

			 }
		  else
			 {

				start = freeAreaCache;

			 }

		}

	 PPC_MemoryRegion m = findMemoryRegion(start); // First region with regionEnd > start.

	 if( (flags & PPC_Constants.MAP_FIXED) != 0 )
		{

		  if( ((start >>> PPC_Constants.PAGE_SHIFT) + (length >>> PPC_Constants.PAGE_SHIFT)) > PPC_Constants.TASK_SIZE_P)
			 {

				throw new Error("ProcessSpace: Requesting memory out of range");

			 }
		  else
			 {

				if(m != null) // If there exists a regions above 'start'.
				  {

					 if( (start + length) > m.regionStart )
						throw new Error("ProcessSpace: Attempting to allocate overlapping regions using MAP_FIXED.");

				  }
				/* else there is no region above 'start' and we have alredy checked PPC_Constants.TASK_SIZE_P.
					So we can allocate at 'start'. */

			 }

		}
	 else // Can relocate if required.
		{

		  if(m != null) // If there exists a regions above 'start'.
			 {

				if( (start + length) > m.regionStart ) // If the space is too small.
				  start = getUnmappedArea(start, length);
				// else we have space to allocate at 'start'.


			 }
		  else
			 {

				if( ((start >>> PPC_Constants.PAGE_SHIFT) + (length >>> PPC_Constants.PAGE_SHIFT)) > PPC_Constants.TASK_SIZE_P )
				  start = getUnmappedArea(0, length); /* Search from default value, since
																	  already know we can't put it at or 
																	  above requested 'start'. */

				// else we have space to allocate at 'start'.

			 }

		}

	 // Create int arrays for new pages at our value of 'start' which may or may not be as requested.
	 numPagesReq = (length + 4095) >>> PPC_Constants.PAGE_SHIFT;
	 startPage = start >>> PPC_Constants.PAGE_SHIFT;

	 for(int page = startPage ; page < startPage + numPagesReq ; page++)
		{
												
		  if(memory[page] == null)
			 {

				memory[page] = new int[1024]; // Allocate array for new page.
				// Set flags for page.
				if((prot & PPC_Constants.PROT_READ) != 0)
				  flag_read[page] = true;
				if((prot & PPC_Constants.PROT_WRITE) != 0)
				  flag_write[page] = true;
				if((prot & PPC_Constants.PROT_EXEC) != 0)
				  flag_exec[page] = true;


			 }
		  else
			 {

				throw new Error("ProcessSpace: Page already allocated but not listed in regions list. SHOULD NOT BE POSSIBLE.");

			 }


		}

	 // If we are mapping from a file, do the copying now.
	 if( (flags & PPC_Constants.MAP_ANONYMOUS) == 0 )
		{

		  try
			 {

				RandomAccessFile rFile = null;
								
				rFile = new RandomAccessFile(file, "r");

				rFile.seek(offset);

				for(int page = startPage ; page < (startPage + numPagesReq - 1); page++) {
				  for(int word = 0 ; word < 1024 ; word++) {
					 memory[page][word] = rFile.readInt(); /* readInt interprets
																			four bytes as big-
																			endian int. */					 
				  }
				}
				for (int word = 0 ; word < ((length & PPC_Constants.PAGE_OFFSET) >>> 2); word++) {
					 memory[startPage + numPagesReq - 1][word] = rFile.readInt(); /* readInt interprets
																										  four bytes as big-
																										  endian int. */					 
				}
			 }
		  catch(IOException e)
			 {

				System.err.println("ProcessSpace: Error reading file into memory.");
				e.printStackTrace();

			 }

		}

	 // Add to the TreeSet of memory regions. length is rounded up to a whole page.
	 regions.add(new PPC_MemoryRegion(start, (length + 0xfff) & PPC_Constants.PAGE_START, prot, flags, file, offset));

	 return start;

  }


  /**
	* Simulate an munmap system call.
	* @param start start of memory area to unmap.
	* @param length length of area.
	*/
  public int munmap(int start, int length)
  {

	 /* Set the relevant pages to null. */
	 int firstPage = start >>> PPC_Constants.PAGE_SHIFT;
	 int lastPage = (start + length - 1) >>> PPC_Constants.PAGE_SHIFT;

	 for(int page = firstPage ; page <= lastPage ; page++)
		memory[page] = null;

	 /* Find the memory region containing start. */
	 PPC_MemoryRegion m = findMemoryRegion(start);

	 if( (m.regionStart == start) & (m.regionEnd == start + length) )
		{

		  regions.remove(m);
	
		}
	 else
		{

		  throw new Error("munmap trying to unmap area not corresponding to a PPC_MemoryRegion. Don't yet handle this case.");

		}

	 return 0;

  }




  /** 
	* Set up a process stack. 
	* @param argv the command line arguments of the PPC binary (starting with its name, C style).
	* @param env the environment variables.
	* @param the auxiliary vector, including the terminating AT_NULL (two zeroes).
	*/
  public void stackInit(String[] argv, String[]  env, int[] auxVector)
  {

	 /* First concatenate the strings into the info block. */
	 String infoBlock = "";

	 for(int a = 0 ; a < argv.length ; a++)
		{

		  infoBlock += argv[a] + "\0"; // Note Java Strings are not null terminated.

		}

	 for(int e = 0 ; e < env.length ; e++)
		{

		  infoBlock += env[e] + "\0";

		}
				
	 infoBlock += argv[0] + "\0\0\0\0\0"; // Need extra copy of this, null terminated, then a zero word.

	 /* Put into a byte array. */
	 byte[] infoBytes = infoBlock.getBytes();

	 int infoBlockStart = PPC_Constants.STACK_TOP - infoBlock.length();

	 int[] argPtr = new int[argv.length]; // Argument pointers.
	 int[] envPtr = new int[env.length]; // Environment pointers.

	 int ptr = infoBlockStart;

	 for(int a = 0 ; a < argv.length ; a++)
		{

		  argPtr[a] = ptr;
		  ptr += argv[a].length() + 1;

		}

	 for(int e = 0 ; e < env.length ; e++)
		{

		  envPtr[e] = ptr;
		  ptr += env[e].length() + 1;

		}

	 /* Now we need to calculate how big it will be. */

	 int argc = argv.length; // Number of command line arguments.
	 int argPtrLength = 4 * argc; // Bytes needed for list of arg pointers.
	 int envPtrLength = 4 * env.length; // Bytes needed for list of env pointers.
	 int auxVectorLength = 4 * auxVector.length; // Bytes needed for auxiliary vector.
	 int infoLength = infoBlock.length();

	 int usedStackLength = 4 /* argc */ + argPtrLength + 4 /* zero word */ + envPtrLength + 
		4 /* zero word */ + auxVectorLength + infoLength;

	 /* We need to round this up to the next multiple of 16, and calculate the number of padding bytes. */
	 int stackLength = (usedStackLength + 0xf) & 0xfffffff0;

	 //????????????? To get the stack to match the iBook
	 stackLength += 16; //TEMP

	 int padLength = stackLength - usedStackLength;

	 /* Call mmap to make some space for the stack. */
	 int stackMapLength = (stackLength + 0xfff) & PPC_Constants.PAGE_START; // Round up to whole page.

	 /* Add a few extra pages. ????????????????? DO SOMETHING MORE ELEGENT! */
	 stackMapLength += 0x10000;

	 mmap( (PPC_Constants.STACK_TOP - stackMapLength), stackMapLength, PPC_Constants.PROT_READ | PPC_Constants.PROT_WRITE | 
			 PPC_Constants.PROT_EXEC, PPC_Constants.MAP_PRIVATE | PPC_Constants.MAP_ANONYMOUS | PPC_Constants.MAP_GROWSDOWN, null, 0);

	 /* Put the appropriate values onto the stack. */
	 int stackPtr = PPC_Constants.STACK_TOP - stackLength;

	 r1 = stackPtr; // Stack pointer register.

	 if(PPC_Options.debugSyscallMore)
		{

		  System.out.println("stackLength = " + Integer.toHexString(stackLength));
		  System.out.println("stackPtr = " + Integer.toHexString(stackPtr));

		}

	 setInt(stackPtr, argc);
	 stackPtr += 4;
	 for(int a = 0 ; a < argv.length ; a++)
		{

		  setInt(stackPtr, argPtr[a]);
		  stackPtr += 4;

		}
	 setInt(stackPtr, 0);
	 stackPtr += 4;
	 for(int e = 0 ; e < env.length ; e++)
		{

		  setInt(stackPtr, envPtr[e]);
		  stackPtr += 4;

		}
	 setInt(stackPtr, 0);
	 stackPtr += 4;
	 for(int v = 0 ; v < auxVector.length ; v++)
		{

		  setInt(stackPtr, auxVector[v]);
		  stackPtr += 4;

		}				
	 for(int p = 0 ; p < padLength ; p++)
		{

		  setByte(stackPtr, (byte)0);
		  stackPtr++;

		}
	 for(int i = 0 ; i < infoLength ; i++)
		{

		  setByte(stackPtr, infoBytes[i]);
		  stackPtr++;
								
		}

  }

  /**
	* Read a null-terminated string from memory.
	*/
  public String getString(int addr)
  {

	 StringBuffer str = new StringBuffer();
	 char c;

	 while( (c = (char)getByte(addr++)) != 0 )
		str.append(c);

	 return str.toString();

  }

  /** 
	* Return an integer from the memory array. 
	* @param wordAddr the address from which to read the value (not
	* necessarily word aligned).
	*/
  public int getInt(int wordAddr)
  {

	 // Is wordAddr word aligned?
	 if((wordAddr & 3) == 0)
		{

		  // Yes
		  int page = wordAddr >>> PPC_Constants.PAGE_SHIFT; // Which page contains word starting at wordAddr?

		  if(!flag_read[page])
			 throw new Error("Attempting to read from a page without read permission."); 
				
		  int word = (wordAddr >>> 2) & 0x3ff; // Address of word within the page.

		  return(memory[page][word]);
				
		}
	 else
		{

		  /* The return value will be derived from two words in memory. */
		  int firstPage = wordAddr >>> PPC_Constants.PAGE_SHIFT;
		  if(!flag_read[firstPage])
			 throw new Error("Attempting to read from a page without read permission.");

		  int firstWord = (wordAddr >>> 2) & 0x3ff;

		  int secondPage, secondWord;

		  /* If the first word in memory to which we are writing is the last of its page, then the
			  second is on the next page. */

		  if(firstWord == 1023)
			 {

				secondPage = firstPage + 1;
				if(!flag_read[secondPage])
				  throw new Error("Attempting to read from a page without read permission.");
												
				secondWord = 0;

			 }
		  else
			 {

				secondPage = firstPage;
				secondWord = firstWord + 1;

			 }
								
		  int firstPartValue, secondPartValue;

		  /* Exactly where is wordAddr aligned? */
		  switch(wordAddr & 3)
			 {

			 case 1:
				/* Last three byte from the first word become first three of the return value and the 
					first one from the second word becomes the last of the return value. */

				firstPartValue = memory[firstPage][firstWord] << 8;
				secondPartValue = memory[secondPage][secondWord] >>> 24;

				break;

			 case 2:
				/* Two bytes from each. */

				firstPartValue = memory[firstPage][firstWord] << 16;
				secondPartValue = memory[secondPage][secondWord] >>> 16;

				break;

			 default: // must be 3
				/* One byte the three bytes. */
												
				firstPartValue = memory[firstPage][firstWord] << 24;
				secondPartValue = memory[secondPage][secondWord] >>> 8;

				break;

			 }

		  /* OR the two part values together and return. */
		  return firstPartValue | secondPartValue;

		}

  }


  /** 
	* Set the value of an integer in the memory. 
	* @param wordAddr the address to which to write (not necessarily
	* word aligned).
	* @param value the value to write to memory.
	*/
  public void setInt(int wordAddr, int value)
  {

	 // Is wordAddr word aligned?
	 if((wordAddr & 3) == 0)
		{

		  // Yes
		  int page = wordAddr >>> PPC_Constants.PAGE_SHIFT;
		  if(!flag_write[page])
			 throw new Error("Attempting to write to a page without write permission.");

		  int word = (wordAddr >>> 2) & 0x3ff;
		  memory[page][word] = value;

		}
	 else
		{
						
		  /* value will be split across two words in memory[][]. */
		  int firstPage = wordAddr >>> PPC_Constants.PAGE_SHIFT;
		  if(!flag_write[firstPage])
			 throw new Error("Attempting to write to a page without write permission.");

		  int firstWord = (wordAddr >>> 2) & 0x3ff;

		  int secondPage, secondWord;

		  /* If the first word in memory to which we are writing is the last of its page, then the
			  second is on the next page. */

		  if(firstWord == 1023)
			 {

				secondPage = firstPage + 1;
				if(!flag_write[secondPage])
				  throw new Error("Attempting to write to a page without write permission.");
												
				secondWord = 0;

			 }
		  else
			 {

				secondPage = firstPage;
				secondWord = firstWord + 1;

			 }
								
		  /* Exactly where is wordAddr aligned? */
								
		  int oldFirstValue = memory[firstPage][firstWord];
		  int oldSecondValue = memory[secondPage][secondWord];
								
		  int newPartOne, newPartTwo; /* The bytes of 'value' separated and shifted to be
													or'ed into the memory[][] words. */
								
		  switch(wordAddr & 3)
			 {

			 case 1:
				/* First three bytes of value go to last three of memory[firstPage][firstWord],
					last to the first byte of memory[secondPage][secondWord]. */

				/* Clear those bits in the old values which we are going to set. */
				oldFirstValue &= 0xff000000;
				oldSecondValue &= 0x00ffffff;

				newPartOne = value >>> 8;
				newPartTwo = value << 24;

				break;

			 case 2:
				/* Two bytes into each memory word. */

				oldFirstValue &= 0xffff0000;
				oldSecondValue &= 0x0000ffff;

				newPartOne = value >>> 16;
				newPartTwo = value << 16;

				break;

			 default: // must be 3
				/* First byte goes to last byte of memory[firstPage][firstWord], rest into
					second word. */
												
				oldFirstValue &= 0xffffff00;
				oldSecondValue &= 0x000000ff;

				newPartOne = value >>> 24;
				newPartTwo = value << 8;

				break;

			 }

		  /* OR the masked old values with the new partial value, and write to memory. */

		  memory[firstPage][firstWord] = oldFirstValue | newPartOne;
		  memory[secondPage][secondWord] = oldSecondValue | newPartTwo;

		}

  }


  /**
	* Return a half word from the memory.
	* @param hwAddr the address from which to read the value (not
	* necessarily half word aligned).
	*/
  public int getHalfWord(int hwAddr)
  {
				
	 /* Page and word indices of the integer in the memory array from which we are reading our half
		 word. If we span two ints this is more complicated. */

	 int page = hwAddr >>> PPC_Constants.PAGE_SHIFT;
	 if(!flag_read[page])
		throw new Error("Attempting to read from a page without read permission.");
				
	 int word = (hwAddr >>> 2) & 0x3ff;

	 /* Where is hwAddr aligned wrt a word boundary? */
	 switch(hwAddr & 3)
		{

		case 0:
		  /* We want the first two bytes of a word in memory. */
		  return (memory[page][word] >>> 16);

		  // unreachable, so no break;

		case 1:
		  /* We want the second and third bytes. */
		  return ((memory[page][word] & 0x00ffff00) >>> 8);

		  // unreachable, so no break;

		case 2:
		  /* We want the last two bytes. */
		  return (memory[page][word] & 0xffff);

		  // unreachable, so no break;
								
		default: // must be 3
		  /* More complicated, we want the last byte from memory[page][word] as our first
			  byte and the first from the next memory integer as our second. */

		  int secondPage, secondWord;
		  if(word == 1023)
			 {

				secondPage = page + 1;
				if(!flag_read[secondPage])
				  throw new Error("Attempting to read from a page without read permission.");

				secondWord = 0;

			 }
		  else
			 {

				secondPage = page;
				secondWord = word + 1;

			 }

		  int firstPart = ((memory[page][word] & 0xff) << 8);
		  int secondPart = (memory[secondPage][secondWord] >>> 24);

		  return (firstPart | secondPart);

		}

  }

  /**
	* Set the value of a half word in memory.
	* @param hwAddr the address to which to write (not necessarily half word aligned).
	* @param value the value to write to memory.
	*/
  public void setHalfWord(int hwAddr, int iValue)
  {

	 short value = (short)iValue;

	 /* Page and word indices of the integer in the memory array from which we are reading our half
		 word. If we span two ints this is more complicated. */

	 int page = hwAddr >>> PPC_Constants.PAGE_SHIFT;
	 if(!flag_write[page])
		throw new Error("Attempting to write to a page without write permission.");
				
	 int word = (hwAddr >>> 2) & 0x3ff;

	 int oldValue;
	 int newValue;

	 /* Where is hwAddr aligned wrt a word boundary? */
	 switch(hwAddr & 3)
		{

		case 0:
		  /* We want to store value as the first two bytes of memory[page][word]. */
		  oldValue = memory[page][word] & 0xffff;
		  newValue = (((int)value) << 16) | oldValue;
		  memory[page][word] = newValue;

		  break;

		case 1:
		  /* We want to store into the second and third bytes. */
		  oldValue = memory[page][word] & 0xff0000ff;
		  newValue = (((int)value) << 8) | oldValue;
		  memory[page][word] = newValue;

		  break;

		case  2:
		  /* We want to store into the last two bytes. */
		  oldValue = memory[page][word] & 0xffff0000;
		  newValue = ((int)value) | oldValue;
		  memory[page][word] = newValue;

		  break;

		default: // must be 3
		  /* More complicated, we want to store to the last byte of memory[page][word], and the
			  first byte of the next memory integer. */
								
		  int secondPage, secondWord;
		  if(word == 1023)
			 {

				secondPage = page + 1;
				if(!flag_write[secondPage])
				  throw new Error("Attempting to write to a page without write permission.");

				secondWord = 0;

			 }
		  else
			 {

				secondPage = page;
				secondWord = word + 1;

			 }

		  int oldFirstValue = memory[page][word] & 0xffffff00;
		  int oldSecondValue = memory[secondPage][secondWord] & 0x00ffffff;

		  int newFirstValue = (value >>> 8) | oldFirstValue;
		  int newSecondValue = (value << 24) | oldSecondValue;

		  memory[page][word] = newFirstValue;
		  memory[secondPage][secondWord] = newSecondValue;

		}

  }

  /** 
	* Return a bytes from the memory. 
	* @param byteAddr the address from which to read.
	*/
  public byte getByte(int byteAddr)
  {

	 int page = byteAddr >>> PPC_Constants.PAGE_SHIFT; // Which page contains the required byte?
	 if(!flag_read[page])
		throw new Error("Attempting to read from a page without read permission.");

	 int word = (byteAddr >>> 2) & 0x3ff; // Which word contains the required byte?
	 int i = memory[page][word]; // Get that word.

	 int b = byteAddr & 3; // Which byte within the word? 
	 // Note: RandomAccessFile.readInt takes four bytes as a big-endian int, hence the following:

	 switch (b) {
	 case 1:
		return (byte)(i >>> 16);
	 case 2:
		return (byte)(i >>> 8);
	 case 3:
		return (byte)(i);
	 default:
		return (byte)(i >>> 24);
	 }

  }

  /** 
	* Set the value of a byte in the memory. 
	* @param byteAddr the address to which to write
	* @param value the value to write.
	*/
  public void setByte(int byteAddr, byte byteValue)
  {

	 int value = (int)byteValue & 0xff; // Avoid spurious sign extending

	 int page = byteAddr >>> PPC_Constants.PAGE_SHIFT;
	 if(!flag_write[page])
		throw new Error("Attempting to write to a page without write permission.");

	 int word = (byteAddr >>> 2) & 0x3ff;
	 int b = byteAddr & 3;

	 /* Generate an int, i with the byte shifted to the right position. 
		 Zero the corresponding bits in the word in memory.	*/
	 int i;
	 switch(b)
		{

		case 1:
		  i = value << 16;
		  memory[page][word] &= 0xff00ffff;
		  break;
		case 2:
		  i = value << 8;
		  memory[page][word] &= 0xffff00ff;
		  break;
		case 3:
		  i = value;
		  memory[page][word] &= 0xffffff00;
		  break;
		default:
		  i = value << 24;
		  memory[page][word] &= 0x00ffffff;

		}
				
	 /* Now fill in those bits with the value we wish to store. */
	 memory[page][word] |= i;

  }



  /** 
	* Return list of used memory regions, in a format rather like linux's /proc/.../maps. 
	*/
  public String listRegions()
  {

	 String str = "";

	 for(Iterator i = regions.iterator() ; i.hasNext(); )
		{

		  PPC_MemoryRegion m = (PPC_MemoryRegion)i.next();
		  str += "0x" + Integer.toHexString(m.regionStart) + "-0x" + Integer.toHexString(m.regionEnd)
			 + " ";
		  if(m.flag_read)
			 str += "r";
		  else
			 str += "-";
		  if(m.flag_write)
			 str += "w";
		  else
			 str += "-";
		  if(m.flag_exec)
			 str += "x";
		  else
			 str += "-";
												
		  if(m.file != null)
			 {

				try
				  {

					 str += " 0x" + Integer.toHexString(m.offset) + " " + m.file.getCanonicalPath();

				  }
				catch(IOException e)
				  {

					 throw new Error("ProcessSpace: File access error.");

				  }

			 }

		  str += "\n";

		}

	 return str;

  }



  /** 
	* Return a list of the memory pages in use, with read, write and execute permissions indicated.
	*/
  public String listUsedPages()
  {

	 String str = "";

	 for(int page = 0 ; page < PPC_Constants.TOTAL_PAGES ; page++)
		{

		  if(memory[page] != null)
			 {

				str += "0x" + Integer.toHexString(page) + " ";
				if(flag_read[page])
				  str += "r";
				else
				  str += "-";
				if(flag_write[page])
				  str += "w";
				else
				  str += "-";
				if(flag_exec[page])
				  str += "x  ";
				else
				  str += "-  ";
												

			 }

		}

	 str += "\n";

	 return str;

  }

  static int startTime = (int)System.currentTimeMillis();
  /**
	* Handle a system call
	*/
  public void doSysCall() throws VM_PragmaNoInline
  {
	 if (initialized) {				
		if(PPC_Options.debugSyscallMore)
		  System.out.println("SYSTEM CALL NUMBER: " + r0 + " AT 0x" + Integer.toHexString(pc));
						
		switch(r0)
		  {

		  case 1: // sys_exit

			 System.exit(r3);

			 break;

		  case 3: // sys_read

			 /* Check that r3 is a valid file descriptor. */

			 if(files[r3] == null)
				{

				  r3 = PPC_Constants.EBADF;
				  cr |= PPC_Constants.CR0_SO;

				}
			 else
				{

				  /* Where to put the data. */
				  int addr = r4;

				  /* Copy from file into buffer. */
				  int b = 0; // Number of bytes read.

				  int i;

				  try
					 {

						while( ((i = raFiles[r3].read()) != -1) & (b < r5) )
						  {
			
							 byte by = (byte)i;

							 b++;
							 setByte(addr++, by);

						  }

						r3 = b; // Return number of bytes read.

					 }
				  catch(IOException e)
					 {

						r3 = PPC_Constants.EIO;
						cr |= PPC_Constants.CR0_SO;

					 }

				}

			 break;

		  case 4: // sys_write

			 if(r3 == 1) // stdout
				{

				  for(int c = 0 ; c < r5 ; c++)
					 {

						System.out.print((char)getByte(r4 + c));

					 }

				  r3 = r5;

				}
			 else if(r3 == 2) // sterr
				{

				  for(int c = 0 ; c < r5 ; c++)
					 {

						System.err.print((char)getByte(r4 + c));

					 }

				  r3 = r5;

				}
			 else
				{

				  /* Check that r3 is a valid file descriptor. */

				  if(files[r3] == null)
					 {

						r3 = PPC_Constants.EBADF;
						cr |= PPC_Constants.CR0_SO;

					 }
				  else
					 {

						/* Where to get the data. */
						int addr = r4;

						/* Copy from buffer into file. */
						int b; //  Number of bytes written.

						byte by;

						try
						  {

							 for(b = 1 ; b <= r5 ; b++)
								{

								  by = getByte(addr++);
								  raFiles[r3].write(by);

								}

							 r3 = --b; /* Return number of bytes written, having accounted for
											  b being incremented once too many at the end of the
											  writing. */

						  }
						catch(IOException e)
						  {

							 r3 = PPC_Constants.EIO;
							 cr |= PPC_Constants.CR0_SO;

						  }
																		
					 }

				}
										

			 break;

		  case 5: // sys_open

			 int flags = r4;

			 /* Examine the flags argument and open read or read-write accordingly. 
				 r3 points to the file name. */

			 String fileName = getString(r3);

			 /* Set r3 to 0, change if an error occurs. */
			 r3 = 0;

			 /* Create a File object so we can test for the existance and properties of the file. */
			 File testFile = new File(fileName);

			 if((flags & PPC_Constants.O_CREAT) != 0)
				{

				  /* O_CREAT specified. If O_EXCL also specified, we require that the file does not
					  already exist. */

				  if( ((flags & PPC_Constants.O_EXCL) != 0) & testFile.exists() )
					 {
																		
						r3 = PPC_Constants.EEXIST;
						cr |= PPC_Constants.CR0_SO;
																
					 }
																		
				}
			 else
				{

				  /* O_CREAT not specified. We require that the file exists. */

				  if(!testFile.exists())
					 {

						r3 = PPC_Constants.ENOENT;
						cr |= PPC_Constants.CR0_SO;

					 }

				}

			 /* If r3 still 0, we have not found an error, so we go ahead and try to open the file. */

			 if(r3 == 0)
				{

				  int fileDesc = getFileDesc(); // Get a vacant file descriptor.

				  try
					 {

						if((flags & 0x3) == PPC_Constants.O_RDONLY)
						  {

							 raFiles[fileDesc] = new RandomAccessFile(fileName, "r");
							 files[fileDesc] = new File(fileName);

							 if(PPC_Options.debugSyscallMore)
								System.out.println("Opened file " + fileName + 
														 " with file descriptor" + fileDesc);
															
						  }
						else // (Java RandomAccessFile has no write only).
						  {

							 raFiles[fileDesc] = new RandomAccessFile(fileName, "rw"); 
							 files[fileDesc] = new File(fileName);
																						
						  }

					 }
				  catch(FileNotFoundException e)
					 {

						throw new Error("File not found: " + fileName + " " + e);

					 }

				  /* Return the file descriptor. */
				  r3 = fileDesc;

				}


			 //????????? NOT YET HANDLING ALL THE flags OPTIONS
			 if((flags & ~(PPC_Constants.O_WRONLY | PPC_Constants.O_RDWR | PPC_Constants.O_CREAT | PPC_Constants.O_EXCL)) != 0)
				throw new Error("Not yet implemented option to sys_open.");

			 break;

		  case 6: // sys_close

			 /* Check that r3 is a valid file descriptor. */
			 if(files[r3] == null)
				{

				  r3 = PPC_Constants.EBADF;
				  cr |= PPC_Constants.CR0_SO;

				}
			 else
				{

				  try
					 {

						raFiles[r3].close();

						files[r3] = null;
						raFiles[r3] = null;

						r3 = 0; // return success code

					 }
				  catch(IOException e)
					 {

						r3 = PPC_Constants.EIO;
						cr |= PPC_Constants.CR0_SO;

					 }

				}

			 break;

		  case 10: // sys_unlink

			 fileName = getString(r3);

			 File ulFile = new File(fileName);

			 if(!ulFile.exists())
				{

				  r3 = PPC_Constants.ENOENT;
				  cr |= PPC_Constants.CR0_SO;

				}
			 else if(!ulFile.getAbsoluteFile().getParentFile().canWrite())
				{

				  r3 = PPC_Constants.EACCES;
				  cr |= PPC_Constants.CR0_SO;

				}
			 // ????????????? MORE POSSIBLE ERRORS
			 else
				{

				  if(ulFile.delete())
					 {

						r3 = 0; // success
																		
					 }
				  else
					 {

						r3 = PPC_Constants.EIO;
						cr |= PPC_Constants.CR0_SO;

					 }

				}

			 break;

		  case 13: // sys_time

			 int time = (int)(System.currentTimeMillis() / 1000);

			 /* If a non-null pointer has been passed in r3, store the time at that address. */
			 if(r3 != 0)
				setInt(r3, time);

			 /* Return the time. */
			 r3 = time;

			 break;
								
		  case 15: // sys_chmod

			 // ???????????? 'FAKE'

			 r3 = 0;

			 System.out.println("sys_chmod NOT IMPLEMENTED");

			 break;

		  case 24: // sys_getuid

			 r3 = PPC_Constants.UID;

			 break;

		  case 30: // sys_utime

			 /* Get filename. */
			 fileName = getString(r3);

			 File uFile = new File(fileName);

			 if(!uFile.exists())
				{

				  r3 = PPC_Constants.EBADF;
				  cr |= PPC_Constants.CR0_SO;

				}
			 else if(!uFile.canWrite())
				{

				  r3 = PPC_Constants.EACCES;
				  cr |= PPC_Constants.CR0_SO;

				}
			 else
				{

				  long mTime = (long)getInt(r4 + 4) * 1000L; // Seconds to ms.
										
				  if(uFile.setLastModified(mTime))
					 {

						r3 = 0; // success

					 }
				  else
					 {

						throw new Error("Error in sys_utime");

					 }
										
				}

			 break;

		  case 43:
			 {
				// r3 points to tms structure
				int currTime = (int)System.currentTimeMillis();
				setInt(r3, currTime - startTime);
				setInt(r3+4, 0);
				setInt(r3+8, 0);
				setInt(r3+12, 0);
				r3 = currTime;
				//System.out.println(toString());
				//cr |= PPC_Constants.CR0_SO;
				//r3 = PPC_Constants.ENOSYS;
			 }
			 break;

		  case 45: // sys_brk

			 if(r3 == 0)
				{

				  /* Requesting current top of bss. */
				  r3 = brk;

				}
			 else
				{

				  /* Changing the value. */
				  setBrk(r3);

				}
			 break; // sys_brk

		  case 47: // sys_getgid

			 r3 = PPC_Constants.GID;

			 break;

		  case 49: // sys_geteuid

			 r3 = PPC_Constants.UID;

			 break;
		
		  case 50: // sys_getegid
										
			 r3 = PPC_Constants.GID;

			 break;

		  case 90: // sys_mmap

			 if( (r6 & PPC_Constants.MAP_ANONYMOUS) != 0 )
				{

				  r3 = mmap(r3, r4, r5, r6, null, 0);

				}
			 else
				{

				  throw new Error("Non-anonymous sys_mmap not yet implemented.");

				}

			 break;

		  case 91: // sys_munmap

			 r3 = munmap(r3, r4);

			 break;

			 //?????? IN FOLLOWING TREAT lstat as stat.

		  case 106: // sys_newstat

			 /* Fall through. */

		  case 107: // sys_newlstat

			 /* Fall through. */

		  case 108: // sys_newfstat

			 File stFile;

			 if(r0 == 108)
				{

				  /* sys_newfstat */	

				  /* Get the file from the array with file descriptor given by r3. */

				  stFile = files[r3];
														

				}
			 else
				{

				  /* sys_newstat or sys_newlstat

				  /* Get the filename from the location pointed to by r3. */

				  stFile = new File(getString(r3));
													
				}
	
			 /* This is the equivalent of Linux's struct stat. Some of these values are 'fudged' by
				 copying from the iBook. ???????????*/

			 int stDev = 0x305;
			 int stIno = 0x14ad1;
			 int stMode = 0x81a4;
			 int stNlink = 0x1;
			 int stUid = PPC_Constants.UID;
			 int stGid = PPC_Constants.GID;
			 int stRdev = 0;
			 int stSize = (int)stFile.length();
			 int stBlkSize = 0x1000;
			 int stBlocks = 0x8;
			 int stMtime = (int)(stFile.lastModified() / 1000);
			 int stCtime = stMtime;
			 int stAtime = stMtime;

			 int addr = r4; // Where to start writing the 'struct'.

			 setInt(addr, stDev);
			 addr += 4;

			 setInt(addr, stIno);
			 addr += 4;

			 setInt(addr, stMode);
			 addr += 4;

			 setInt(addr, stNlink << 16);
			 addr += 4;

			 setInt(addr, stUid);
			 addr += 4;

			 setInt(addr, stGid);
			 addr += 4;

			 setInt(addr, stRdev);
			 addr += 4;

			 setInt(addr, stSize);
			 addr += 4;

			 setInt(addr, stBlkSize);
			 addr += 4;

			 setInt(addr, stBlocks);
			 addr += 4;

			 setInt(addr, stAtime);
			 addr += 4;

			 setInt(addr, 0); // unused
			 addr += 4;

			 setInt(addr, stMtime);
			 addr += 4;

			 setInt(addr, 0); // unused
			 addr += 4;

			 setInt(addr, stCtime);
			 addr += 4;

			 setInt(addr, 0); // unused
			 addr += 4;

			 setInt(addr, 0); // unused
			 addr += 4;

			 setInt(addr, 0); // unused
			 addr += 4;

			 //??????????????? FINISH ERRORS, at least EACCES
			 /* If any error, set r3 to the code and set CR0.SO. */

			 if(!stFile.exists())
				{

				  r3 = PPC_Constants.ENOENT;
				  cr |= PPC_Constants.CR0_SO;

				}
			 else
				{

				  r3 = 0;

				}
			 break;

		  case 173: // sys_rt_sigaction

			 // ???????????? 'FAKE'

			 r3 = 0; //TEMP

			 //TEMP r3 = PPC_Constants.ENOSYS;
			 //TEMP cr |= PPC_Constants.CR0_SO;

			 System.out.println("sys_rt_sigaction NOT IMPLEMENTED");

			 break;

		  case 181: // sys_chown
										
			 // ???????????? 'FAKE'

			 r3 = 0;

			 System.out.println("sys_chown NOT IMPLEMENTED");

			 break; 

		  case 197: // sys_fstat64

			 if(r3 == 1) // stdout
				{

				  /* These are the values for the equivalent of Linux's struct stat64,
					  based on values from softwood. */

				  long st64Dev = 6;
				  long st64Ino = 2;
				  int st64Mode = 0x2190;
				  int st64Nlink = 1;
				  int st64Uid = PPC_Constants.UID;
				  int st64Gid = 5; // group tty
				  long st64Rdev = 0x8800;
				  long st64Size = 0;
				  int st64BlkSize = 0;
				  long st64Blocks = 0x40000000000L;
				  int st64Atime = 0;
				  int st64Mtime = (int)(System.currentTimeMillis() / 1000);
				  int st64Ctime = st64Mtime;

				  addr = r4; // Where to start writing the 'struct'.

				  setInt(addr, (int)(st64Dev >>> 32));
				  addr += 4;
				  setInt(addr, (int)(st64Dev & PPC_Constants.LW_MASK));
				  addr += 4;

				  setInt(addr, (int)(st64Ino >>> 32));
				  addr += 4;
				  setInt(addr, (int)(st64Ino & PPC_Constants.LW_MASK));
				  addr += 4;

				  setInt(addr, st64Mode);
				  addr += 4;

				  setInt(addr, st64Nlink);
				  addr += 4;
														
				  setInt(addr, st64Uid);
				  addr += 4;

				  setInt(addr, st64Gid);
				  addr += 4;

				  setInt(addr, (int)(st64Rdev >>> 32));
				  addr += 4;
				  setInt(addr, (int)(st64Rdev & PPC_Constants.LW_MASK));
				  addr += 4;
	
				  setInt(addr, 0); // padding
				  addr += 4;

				  setInt(addr, (int)(st64Size >>> 32));
				  addr += 4;
				  setInt(addr, (int)(st64Size & PPC_Constants.LW_MASK));
				  addr += 4;

				  setInt(addr, st64BlkSize);
				  addr += 4;
														
				  setInt(addr, (int)(st64Blocks >>> 32));
				  addr += 4;
				  setInt(addr, (int)(st64Blocks & PPC_Constants.LW_MASK));
				  addr += 4;

				  setInt(addr, st64Atime);
				  addr += 4;
													
				  setInt(addr, 0); // unused
				  addr += 4;

				  setInt(addr, st64Mtime);
				  addr += 4;
													
				  setInt(addr, 0); // unused
				  addr += 4;
														
				  setInt(addr, st64Ctime);
				  addr += 4;
														
				  setInt(addr, 0); // unused
				  addr += 4;

				  setInt(addr, 0); // unused
				  addr += 4;
														
				  setInt(addr, 0); // unused
				  addr += 4;

				  r3 = 0; /* If we failed writing to the stack we would not be here,
								 so return success. ??????????????? MIGHT ALSO NEED CHANGING*/

				}

			 break; // case 197

		  case 204: // sys_fcntl64
										
			 /* This is complicated so fudge it for now. */
										
			 int fd = r3;
			 int cmd = r4;
										
			 if( ((fd == 0) | (fd == 1) | (fd == 2)) & (cmd == 1) )
				r3 = 0;
			 else
				throw new Error("Unrecognised system call.");
										
			 break; // case 204
										
		  default:
			 throw new Error("Unrecognised system call: " + r0);
										
		  }
	 }
  }

  /**
	* Set the initial value of brk. 
	*/
  public void initBrk(int brk)
  {

	 this.brk = brk;

  }


  /**
	* Change the value of brk and allocate any necesary memory.
	*/
  public void setBrk(int newBrk)
  {

	 /* Is this within the currently allocated space?
		 Find the memory region which contains the current brk address, or has brk as its end point. 
		 If the new brk address is not within the region, mmap some more memory. */

	 PPC_MemoryRegion m = findMemoryRegionInclEnd(brk);
				
	 if(m.regionEnd < newBrk)
		{

		  /* Need to allocate more space. */
		  mmap(m.regionEnd, newBrk - m.regionEnd, m.getProt(),
				 PPC_Constants.MAP_FIXED | PPC_Constants.MAP_PRIVATE | PPC_Constants.MAP_ANONYMOUS, null, 0);
		  PPC_MemoryRegion new_region = findMemoryRegionInclEnd(newBrk);
		  m.merge(new_region);
		  regions.remove(new_region);
		}

	 this.brk = newBrk;

  }

  /** 
	* Return a string representation of the process space. The contents of the string are register values 
	* and regions of memory space allocated. 
	*/
  public String toString()
  {

	 String str = "";
				
	 str += " r0 = 0x" + Integer.toHexString(r0) + "\n";
	 str += " r1 = 0x" + Integer.toHexString(r1) + "\n";
	 str += " r2 = 0x" + Integer.toHexString(r2) + "\n";
	 str += " r3 = 0x" + Integer.toHexString(r3) + "\n";
	 str += " r4 = 0x" + Integer.toHexString(r4) + "\n";
	 str += " r5 = 0x" + Integer.toHexString(r5) + "\n";
	 str += " r6 = 0x" + Integer.toHexString(r6) + "\n";
	 str += " r7 = 0x" + Integer.toHexString(r7) + "\n";
	 str += " r8 = 0x" + Integer.toHexString(r8) + "\n";
	 str += " r9 = 0x" + Integer.toHexString(r9) + "\n";
	 str += "r10 = 0x" + Integer.toHexString(r10) + "\n";
	 str += "r11 = 0x" + Integer.toHexString(r11) + "\n";
	 str += "r12 = 0x" + Integer.toHexString(r12) + "\n";
	 str += "r13 = 0x" + Integer.toHexString(r13) + "\n";
	 str += "r14 = 0x" + Integer.toHexString(r14) + "\n";
	 str += "r15 = 0x" + Integer.toHexString(r15) + "\n";
	 str += "r16 = 0x" + Integer.toHexString(r16) + "\n";
	 str += "r17 = 0x" + Integer.toHexString(r17) + "\n";
	 str += "r18 = 0x" + Integer.toHexString(r18) + "\n";
	 str += "r19 = 0x" + Integer.toHexString(r19) + "\n";
	 str += "r20 = 0x" + Integer.toHexString(r20) + "\n";
	 str += "r21 = 0x" + Integer.toHexString(r21) + "\n";
	 str += "r22 = 0x" + Integer.toHexString(r22) + "\n";
	 str += "r23 = 0x" + Integer.toHexString(r23) + "\n";
	 str += "r24 = 0x" + Integer.toHexString(r24) + "\n";
	 str += "r25 = 0x" + Integer.toHexString(r25) + "\n";
	 str += "r26 = 0x" + Integer.toHexString(r26) + "\n";
	 str += "r27 = 0x" + Integer.toHexString(r27) + "\n";
	 str += "r28 = 0x" + Integer.toHexString(r28) + "\n";
	 str += "r29 = 0x" + Integer.toHexString(r29) + "\n";
	 str += "r30 = 0x" + Integer.toHexString(r30) + "\n";
	 str += "r31 = 0x" + Integer.toHexString(r31) + "\n\n";

	 str += "pc = 0x" + Integer.toHexString(pc) + "\n\n";

	 str += "cr = 0x" + Integer.toHexString(cr) + "\n";
	 str += "lr = 0x" + Integer.toHexString(lr) + "\n";
	 str += "ctr = 0x" + Integer.toHexString(ctr) + "\n";
	 str += "xer = 0x" + Integer.toHexString(xer) + "\n\n";
	 str += "fpscr = 0x" + Integer.toHexString(fpscr) + "\n\n";

	 if(!PPC_Options.debugPS_OmitFP)	{
		  str += " f0 = " + f0 + " bit pattern 0x" + Long.toHexString(Double.doubleToRawLongBits(f0)) + "\n";
		  str += " f1 = " + f1 + " bit pattern 0x" + Long.toHexString(Double.doubleToRawLongBits(f1)) + "\n";
		  str += " f2 = " + f2 + " bit pattern 0x" + Long.toHexString(Double.doubleToRawLongBits(f2)) + "\n";
		  str += " f3 = " + f3 + " bit pattern 0x" + Long.toHexString(Double.doubleToRawLongBits(f3)) + "\n";
		  str += " f4 = " + f4 + " bit pattern 0x" + Long.toHexString(Double.doubleToRawLongBits(f4)) + "\n";
		  str += " f5 = " + f5 + " bit pattern 0x" + Long.toHexString(Double.doubleToRawLongBits(f5)) + "\n";
		  str += " f6 = " + f6 + " bit pattern 0x" + Long.toHexString(Double.doubleToRawLongBits(f6)) + "\n";
		  str += " f7 = " + f7 + " bit pattern 0x" + Long.toHexString(Double.doubleToRawLongBits(f7)) + "\n";
		  str += " f8 = " + f8 + " bit pattern 0x" + Long.toHexString(Double.doubleToRawLongBits(f8)) + "\n";
		  str += " f9 = " + f9 + " bit pattern 0x" + Long.toHexString(Double.doubleToRawLongBits(f9)) + "\n";
		  str += "f10 = " + f10 + " bit pattern 0x" + Long.toHexString(Double.doubleToRawLongBits(f10)) + "\n";
		  str += "f11 = " + f11 + " bit pattern 0x" + Long.toHexString(Double.doubleToRawLongBits(f11)) + "\n";
		  str += "f12 = " + f12 + " bit pattern 0x" + Long.toHexString(Double.doubleToRawLongBits(f12)) + "\n";
		  str += "f13 = " + f13 + " bit pattern 0x" + Long.toHexString(Double.doubleToRawLongBits(f13)) + "\n";
		  str += "f14 = " + f14 + " bit pattern 0x" + Long.toHexString(Double.doubleToRawLongBits(f14)) + "\n";
		  str += "f15 = " + f15 + " bit pattern 0x" + Long.toHexString(Double.doubleToRawLongBits(f15)) + "\n";
		  str += "f16 = " + f16 + " bit pattern 0x" + Long.toHexString(Double.doubleToRawLongBits(f16)) + "\n";
		  str += "f17 = " + f17 + " bit pattern 0x" + Long.toHexString(Double.doubleToRawLongBits(f17)) + "\n";
		  str += "f18 = " + f18 + " bit pattern 0x" + Long.toHexString(Double.doubleToRawLongBits(f18)) + "\n";
		  str += "f19 = " + f19 + " bit pattern 0x" + Long.toHexString(Double.doubleToRawLongBits(f19)) + "\n";
		  str += "f20 = " + f20 + " bit pattern 0x" + Long.toHexString(Double.doubleToRawLongBits(f20)) + "\n";
		  str += "f21 = " + f21 + " bit pattern 0x" + Long.toHexString(Double.doubleToRawLongBits(f21)) + "\n";
		  str += "f22 = " + f22 + " bit pattern 0x" + Long.toHexString(Double.doubleToRawLongBits(f22)) + "\n";
		  str += "f23 = " + f23 + " bit pattern 0x" + Long.toHexString(Double.doubleToRawLongBits(f23)) + "\n";
		  str += "f24 = " + f24 + " bit pattern 0x" + Long.toHexString(Double.doubleToRawLongBits(f24)) + "\n";
		  str += "f25 = " + f25 + " bit pattern 0x" + Long.toHexString(Double.doubleToRawLongBits(f25)) + "\n";
		  str += "f26 = " + f26 + " bit pattern 0x" + Long.toHexString(Double.doubleToRawLongBits(f26)) + "\n";
		  str += "f27 = " + f27 + " bit pattern 0x" + Long.toHexString(Double.doubleToRawLongBits(f27)) + "\n";
		  str += "f28 = " + f28 + " bit pattern 0x" + Long.toHexString(Double.doubleToRawLongBits(f28)) + "\n";
		  str += "f29 = " + f29 + " bit pattern 0x" + Long.toHexString(Double.doubleToRawLongBits(f29)) + "\n";
		  str += "f30 = " + f30 + " bit pattern 0x" + Long.toHexString(Double.doubleToRawLongBits(f30)) + "\n";
		  str += "f31 = " + f31 + " bit pattern 0x" + Long.toHexString(Double.doubleToRawLongBits(f31)) + 
			 "\n\n";


		}

	 //				str += "tbl = 0x" + Integer.toHexString(tbl) + "\n";
	 //				str += "tbu = 0x" + Integer.toHexString(tbu) + "\n\n";



	 return str + listRegions();
  }

  /**
	* Print out the stack contents.
	*/
  public void printStack() {
	 // w > 0 is a kludge due to 0x800000000 being treated as negative. 
	 for(int w = r1 - 64 ; (w <= 0x7fffffff) & (w > 0) ; w += 4) {
		System.out.print("0x" + Integer.toHexString(w) + "  ");
		for(int b = w ; b < w + 4 ; b++) {
		  char c = (char)getByte(b);
		  if( (c < 0x20) | (c > 0x7e) )
			 c = ' ';
		  System.out.print(c);
		}
		System.out.println("  0x" + Integer.toHexString(getInt(w)));
	 }
  }
}

/*************************************************************************************************/

/** Representation of a region of memory in the process space (i.e. a
 * set of consecutive pages with identical flags). It serves the same
 * purpose as the struct vm_area_struct in Linux.
 * @author Richard Matley
 */

public class PPC_MemoryRegion
{

  int regionStart; // First address inside the region.
  int regionEnd; // First address after the region, hence end-start=length.

  boolean flag_read, flag_write, flag_exec; // Access permissions.

  File file; // If not anonymous, the file which is mapped here.
  int offset; // Offset within file.

  /**
	*  Construct a PPC_MemoryRegion object for a newly allocated region. 
	*/
  public PPC_MemoryRegion(int start, int length, int prot, int flags, File file, int offset)
  {
				
	 this.file = file;

	 regionStart = start;
	 regionEnd = start + length;
	 this.offset = offset;

	 if( (prot & PPC_Constants.PROT_READ) != 0 )
		this.flag_read = true;
	 if( (prot & PPC_Constants.PROT_WRITE) != 0 )
		this.flag_write = true;
	 if( (prot & PPC_Constants.PROT_EXEC) != 0 )
		this.flag_exec = true;

  }
  /**
	* Merge two memory regions
	*/
  public void merge(PPC_MemoryRegion other) {
	 if ((file != other.file) || (flag_read != other.flag_read) || 
		  (flag_write != other.flag_write) || (flag_exec != other.flag_exec)) {
		throw new Error("Invalid file or permissions in memory region merge: " + this + " " + other);
	 }
	 if (regionStart < other.regionStart) {
		if (regionEnd != other.regionStart) {
		  throw new Error("Non-contiguous memory regions in merge: " + this + " " + other);
		}
		regionEnd = other.regionEnd;
	 }
	 else {
		if (other.regionEnd != regionStart) {
		  throw new Error("Non-contiguous memory regions in merge: " + this + " " + other);
		}
		regionStart = other.regionStart;		
		offset = other.offset;
	 }
  }
  /**
	* Return the protection flags.
	*/
  public int getProt()
  {

	 int prot = 0;

	 if( this.flag_read == true )
		prot |= PPC_Constants.PROT_READ;

	 if( this.flag_write == true )
		prot |= PPC_Constants.PROT_WRITE;

	 if( this.flag_exec == true )
		prot |= PPC_Constants.PROT_EXEC;
				
	 return prot;
  }
  /**
	* String representation of object
	*/
  public String toString() {
	 return "0x" + Integer.toHexString(regionStart) + "-" +
		"0x" + Integer.toHexString(regionEnd) + " " +
		(flag_read ? "r" : "-") +
		(flag_write ? "w" : "-") +
		(flag_exec ? "x" : "-") + " " +
		file + ":" + 
		offset;
  }
}


/** Used to order PPCMemoryRegion objects. */
class PPC_MemoryComparator implements Comparator
{

  public int compare(Object mr1, Object mr2)
  {

	 return ((PPC_MemoryRegion)mr1).regionStart - ((PPC_MemoryRegion)mr2).regionStart;

  }

}
