/*
 * (C) Copyright The University of Manchester 2004
 */
package org.binarytranslator.PPC_Emulator;
import com.ibm.JikesRVM.opt.ir.*;
import com.ibm.JikesRVM.*;
import com.ibm.JikesRVM.classloader.*;
import com.ibm.JikesRVM.opt.*;
import java.util.Enumeration;
import java.util.NoSuchElementException;

import java.util.*;

/**
 * This class is the super class of all instruction translators, it
 * implements default methods that just throw errors
 */
public class PPC_InstructionTranslator implements OPT_Operators {
  public PPC_InstructionTranslator() {
  }

  /* Different instruction formats */
  
  protected static final int  INVALID_OP = 0;
  protected static final int  D_FORM   = 1;
  protected static final int  B_FORM   = 2;
  protected static final int  I_FORM   = 3;
  protected static final int  SC_FORM  = 4;
  protected static final int  X_FORM   = 5;
  protected static final int  XL_FORM  = 6;
  protected static final int  XFX_FORM = 7;
  protected static final int  XFL_FORM = 8;
  protected static final int  XO_FORM  = 9;
  protected static final int  A_FORM   = 10;
  protected static final int  M_FORM   = 11;
  protected static final int  EXTENDED = 12;

  /**
   * Look up table to find instruction translator, performed using the
   * instruction opcode.
   */
  static OpCodeLookUp[] opCodeLookUp = {
    /*   OPCD                        form    mnemonic    translator */
    /*   ----                        ----    --------    ---------- */
    /*    0,   */  null,
    /*    1,   */  null,
    /*    2,   */  new OpCodeLookUp (D_FORM, "tdi",      new tdi_translator()),
    /*    3,   */  new OpCodeLookUp (D_FORM, "twi",      new twi_translator()), 
    /*    4,   */  null,
       
    /*    5,   */  null,
    /*    6,   */  null,
    /*    7,   */  new OpCodeLookUp (D_FORM, "mulli",    new mulli_translator()),
    /*    8,   */  new OpCodeLookUp (D_FORM, "subfic",   new subfic_translator()),
    /*    9,   */  null,
    
    /*    10,  */  new OpCodeLookUp (D_FORM, "cmpli",    new cmpli_translator()),
    /*    11,  */  new OpCodeLookUp (D_FORM, "cmpi",     new cmpi_translator()),
    /*    12,  */  new OpCodeLookUp (D_FORM, "addic",    new addic_translator()),
    /*    13,  */  new OpCodeLookUp (D_FORM, "addic.",   new addic__translator()),
    /*    14,  */  new OpCodeLookUp (D_FORM, "addi",     new addi_translator()),
       
    /*    15,  */  new OpCodeLookUp (D_FORM, "addis",    new addis_translator()),
    /*    16,  */  new OpCodeLookUp (B_FORM, "bc",       new bc_translator()),
    /*    17,  */  new OpCodeLookUp (SC_FORM,"sc",       new sc_translator()),
    /*    18,  */  new OpCodeLookUp (I_FORM, "b",        new b_translator()),
    /*    19,  */  new OpCodeLookUp (XL_FORM,"XL_FORM",  new xlForm_translator()),
    
    /*    20,  */  new OpCodeLookUp (M_FORM, "rlwimi",   new rlwimi_translator()),
    /*    21,  */  new OpCodeLookUp (M_FORM, "rlwinm",   new rlwinm_translator()),

    /*    22,  */  null,
    /*    23,  */  new OpCodeLookUp (M_FORM, "rlwnm",    new rlwnm_translator()),
    /*    24,  */  new OpCodeLookUp (D_FORM, "ori",      new ori_translator()),
    
    /*    25,  */  new OpCodeLookUp (D_FORM, "oris",     new oris_translator()),
    /*    26,  */  new OpCodeLookUp (D_FORM, "xori",     new xori_translator()),
    /*    27,  */  new OpCodeLookUp (D_FORM, "xoris",    new xoris_translator()),
    /*    28,  */  new OpCodeLookUp (D_FORM, "andi.",    new andi__translator()),
    /*    29,  */  new OpCodeLookUp (D_FORM, "andis.",   new andis__translator()),
    
    /*    30,  */  null,
    /*    31,  */  new OpCodeLookUp (EXTENDED, "EXT 31", new extended31_translator()),
    /*    32,  */  new OpCodeLookUp (D_FORM, "lwz",      new lwz_translator()),
    /*    33,  */  new OpCodeLookUp (D_FORM, "lwzu",     new lwzu_translator()),
    /*    34,  */  new OpCodeLookUp (D_FORM, "lbz",      new lbz_translator()),
    
    /*    35,  */  new OpCodeLookUp (D_FORM, "lbzu",     new lbzu_translator()),
    /*    36,  */  new OpCodeLookUp (D_FORM, "stw",      new stw_translator()),
    /*    37,  */  new OpCodeLookUp (D_FORM, "stwu",     new stwu_translator()),
    /*    38,  */  new OpCodeLookUp (D_FORM, "stb",      new stb_translator()),
    /*    39,  */  new OpCodeLookUp (D_FORM, "stbu",     new stbu_translator()),
    
    /*    40,  */  new OpCodeLookUp (D_FORM, "lhz",      new lhz_translator()),
    /*    41,  */  new OpCodeLookUp (D_FORM, "lhzu",     new lhzu_translator()),
    /*    42,  */  new OpCodeLookUp (D_FORM, "lha",      new lha_translator()),
    /*    43,  */  new OpCodeLookUp (D_FORM, "lhau",     new lhau_translator()),
    /*    44,  */  new OpCodeLookUp (D_FORM, "sth",      new sth_translator()),
    
    /*    45,  */  new OpCodeLookUp (D_FORM, "sthu",     new sthu_translator()),
    /*    46,  */  new OpCodeLookUp (D_FORM, "lmw",      new lmw_translator()),
    /*    47,  */  new OpCodeLookUp (D_FORM, "stmw",     new stmw_translator()),
    /*    48,  */  new OpCodeLookUp (D_FORM, "lfs",      new lfs_translator()),
    /*    49,  */  new OpCodeLookUp (D_FORM, "lfsu",     new lfsu_translator()),
    
    /*    50,  */  new OpCodeLookUp (D_FORM, "lfd",      new lfd_translator()),
    /*    51,  */  new OpCodeLookUp (D_FORM, "lfdu",     new lfdu_translator()),
    /*    52,  */  new OpCodeLookUp (D_FORM, "stfs",     new stfs_translator()),
    /*    53,  */  new OpCodeLookUp (D_FORM, "stfsu",    new stfsu_translator()),
    /*    54,  */  new OpCodeLookUp (D_FORM, "stfd",     new stfd_translator()),
    
    /*    55,  */  new OpCodeLookUp (D_FORM, "stfdu",    new stfdu_translator()),
    /*    56,  */  null,
    /*    57,  */  null,
    /*    58,  */  null,
    /*    59,  */  new OpCodeLookUp (EXTENDED, "EXT 59", new extended59_translator()),

    /*    60,  */  null,
    /*    61,  */  null,
    /*    62,  */  null,
    /*    63,  */  new OpCodeLookUp (EXTENDED, "EXT 63", new extended63_translator())
  };

  /**
   * Return a range of bits from an int
   * @param x value to return range of bits from
   * @param n the start of the range of bits
   * @param m the end of the range of bits
   * @return the specified bits
   */
  protected static final int bits (int x, int n, int m) {
    return ((x >> (31-m)) & ((1 << (m-n+1)) - 1));
  }


  /**
   * Translate a single instruction
   * @param ppc2ir the object containing the translation sequence
   * @param ps the process space of the translation
   * @param pc the address of the instruction to translate
   * @return the address of the next instruction or -1 if this
   * instruction has branched to the end of the trace
   */
  public static int translateInstruction(PPC2IR ppc2ir, PPC_ProcessSpace ps, PPC2IR.Laziness lazy, int pc) {
    int instr = ps.getInt(pc);

	 if(PPC_Options.debugInstr) {
		System.out.println(lazy.makeKey(pc) + PPC_Disassembler.disasm(instr, pc) + " " + ppc2ir.getCurrentBlock() + " " + ppc2ir.numberOfInstructions);
	 }
	 
    int primaryOpcode =  bits(instr, 0, 5);
	 if (opCodeLookUp[primaryOpcode] == null) {
		if(PPC_Options.debugInstr) {
		  System.out.println("Unknown extended instruction. Primary " + primaryOpcode);
		}
		ppc2ir.plantThrowBadInstruction((PPC2IR.Laziness)lazy.clone(), pc);
		return -1;
		// throw new Error("Unknown extended instruction. Primary " + primaryOpcode);
	 }
    int form = opCodeLookUp[primaryOpcode].getForm();
	 
    switch(form)  /* decode known instruction format */
      {
      case D_FORM:
		  return opCodeLookUp[primaryOpcode].translator.translateD_FORM(ppc2ir, lazy, pc, instr, primaryOpcode, bits(instr,6,10), bits(instr,11,15), bits(instr,16,31));
      case B_FORM:
		  return opCodeLookUp[primaryOpcode].translator.translateB_FORM(ppc2ir, lazy, pc, instr, primaryOpcode, bits(instr,6,10), bits(instr,11,15), bits(instr,16,29), bits(instr,30,30), instr & 0x1);
      case I_FORM:
		  return opCodeLookUp[primaryOpcode].translator.translateI_FORM(ppc2ir, lazy, pc, instr, primaryOpcode, bits(instr,6,29), bits(instr,30,30), instr & 0x1);
      case SC_FORM:
		  return opCodeLookUp[primaryOpcode].translator.translateSC_FORM(ppc2ir, lazy, pc, instr, primaryOpcode, bits(instr,6,10), bits(instr,11,15), bits(instr,16,29), bits(instr,30,30), bits(instr,31,31));
		case X_FORM:
		  return opCodeLookUp[primaryOpcode].translator.translateX_FORM(ppc2ir, lazy, pc, instr, primaryOpcode, bits(instr,6,10), bits(instr,11,15), bits(instr,16,20), bits(instr,21,30), instr & 0x1);
		case XFX_FORM:
		  return opCodeLookUp[primaryOpcode].translator.translateXFX_FORM(ppc2ir, lazy, pc, instr, primaryOpcode, bits(instr,6,10), bits(instr,11,20), bits(instr,21,30), instr & 0x1);
		case XFL_FORM:
		  return opCodeLookUp[primaryOpcode].translator.translateXFL_FORM(ppc2ir, lazy, pc, instr, primaryOpcode, bits(instr,6,6), bits(instr,7,14), bits(instr,15,15), bits(instr,16,20), bits(instr,21,30), instr & 0x1);
      case XL_FORM:
		  return opCodeLookUp[primaryOpcode].translator.translateXL_FORM(ppc2ir, lazy, pc, instr, primaryOpcode, bits(instr,6,10), bits(instr,11,15), bits(instr,16,20), bits(instr,21,30), instr & 0x1);
      case XO_FORM:
		  return opCodeLookUp[primaryOpcode].translator.translateXO_FORM(ppc2ir, lazy, pc, instr, primaryOpcode, bits(instr,6,10), bits(instr,11,15), bits(instr,16,20), bits(instr,21,21), bits(instr,22,30), instr & 0x1);
      case M_FORM:
		  return opCodeLookUp[primaryOpcode].translator.translateM_FORM(ppc2ir, lazy, pc, instr, primaryOpcode, bits(instr,6,10), bits(instr,11,15), bits(instr,16,20), bits(instr,21,25), bits(instr,26,30), instr & 0x1);
      case A_FORM:
		  return opCodeLookUp[primaryOpcode].translator.translateA_FORM(ppc2ir, lazy, pc, instr, primaryOpcode, bits(instr,6,10), bits(instr,11,15), bits(instr,16,20), bits(instr,21,25), bits(instr,26,30), instr & 0x1);
      case EXTENDED:
		  return opCodeLookUp[primaryOpcode].translator.translateEXTENDED(ppc2ir, lazy, pc, instr, primaryOpcode);
      case INVALID_OP:      /* More work to do... */
      default:
		  throw new Error("Invalid opcode");
      }
  }

  protected int translateD_FORM(PPC2IR ppc2ir, PPC2IR.Laziness lazy, int pc, int inst, int opcode, int bits6_to_10, int bits11_to_15, int bits16_to_31){
    throw new Error("Attempting to translateD_Form on a non D_Form instruction\n" +
						  "Address=0x" + Integer.toHexString(pc) +
						  " Instruction 0x" + Integer.toHexString(inst) + "\n" +
						  "Opcode 0x" + Integer.toHexString(opcode) +
						  " bits6_to_10=" + bits6_to_10 +
						  " bits11_to_15=" + bits11_to_15 +
						  " bits16_to_31=" + bits16_to_31 +
						  "\nDisassembly: " + PPC_Disassembler.disasm(inst, pc));
  }
  protected int translateDS_FORM(int inst, int opcode){
    throw new Error("Attempting to translateDS_Form on a non DS_Form instruction\n" +
						  "Instruction 0x" + Integer.toHexString(inst) + "\n" +
						  "Opcode 0x" + Integer.toHexString(inst));
  }
  protected int translateB_FORM(PPC2IR ppc2ir, PPC2IR.Laziness lazy, int pc, int inst, int primaryOpcode, int bits_6_to_10, int bits_11_to_15, int bits16_to_29, int bit30, int bit31) {
    throw new Error("Attempting to translateB_Form on a non B_Form instruction\n" +
						  "Instruction 0x" + Integer.toHexString(inst) + "\n" +
						  "Opcode 0x" + Integer.toHexString(primaryOpcode) +
						  "\nDisassembly: " + PPC_Disassembler.disasm(inst, pc));
  }
  protected int translateI_FORM(PPC2IR ppc2ir, PPC2IR.Laziness lazy, int pc, int inst, int opcode, int bits_6_to_29, int bit30, int bit31) {
    throw new Error("Attempting to translateI_Form on a non I_Form instruction\n" +
						  "Instruction 0x" + Integer.toHexString(inst) + "\n" +
						  "Opcode 0x" + Integer.toHexString(opcode) +
						  "\nDisassembly: " + PPC_Disassembler.disasm(inst, pc));
  }
  protected int translateSC_FORM(PPC2IR ppc2ir, PPC2IR.Laziness lazy, int pc, int inst, int primaryOpcode, int bits_6_to_10, int bits_11_to_15, int bits_16_to_29, int bit30, int bit31){
    throw new Error("Attempting to translateSC_Form on a non SC_Form instruction\n" +
						  "Instruction 0x" + Integer.toHexString(inst) + "\n" +
						  "Opcode 0x" + Integer.toHexString(primaryOpcode) +
						  " bits_6_to_10=" + bits_6_to_10 +
						  " bits_11_to_15=" + bits_11_to_15 +
						  " bits_16_to_29=" + bits_16_to_29 +
						  " bit30=" + bit30 +
						  " bit31=" + bit31 +
						  "\nDisassembly: " + PPC_Disassembler.disasm(inst, pc));
  }
  protected int translateX_FORM(PPC2IR ppc2ir, PPC2IR.Laziness lazy, int pc, int inst, int opcode, int bits6_to_10, int bits11_to_15, int bits16_to_20, int secondaryOpcode, int bit31) {
    throw new Error("Attempting to translateX_Form on a non X_Form instruction\n" +
						  "Address=0x" + Integer.toHexString(pc) +
						  " Instruction 0x" + Integer.toHexString(inst) + "\n" +
						  "Opcode 0x" + Integer.toHexString(opcode) +
						  " bits6_to_10=" + bits6_to_10 +
						  " bits11_to_15=" + bits11_to_15 +
						  " bits16_to_20=" + bits16_to_20 +
						  " secondaryOpcode=" + secondaryOpcode +
						  " bit31=" + bit31+
						  "\nDisassembly: " + PPC_Disassembler.disasm(inst, pc));
  }
  protected int translateXFX_FORM(PPC2IR ppc2ir, PPC2IR.Laziness lazy, int pc, int inst, int opcode, int bits6_to_10, int bits11_to_20, int secondaryOpcode, int bit31) {
    throw new Error("Attempting to translateXFX_Form on a non XFX_Form instruction\n" +
						  "Address=0x" + Integer.toHexString(pc) +
						  " Instruction 0x" + Integer.toHexString(inst) + "\n" +
						  "Opcode 0x" + Integer.toHexString(opcode) +
						  " bits6_to_10=" + bits6_to_10 +
						  " bits11_to_20=" + bits11_to_20 +
						  " secondaryOpcode=" + secondaryOpcode +
						  " bit31=" + bit31 +
						  "\nDisassembly: " + PPC_Disassembler.disasm(inst, pc));
  }
  protected int translateXFL_FORM(PPC2IR ppc2ir, PPC2IR.Laziness lazy,int  pc, int inst, int opcode, int bit6, int bits7_to_14, int bit15, int bits16_to_20, int secondaryOpcode, int bit31) {
    throw new Error("Attempting to translateXFL_Form on a non XFL_Form instruction\n" +
						  "Address=0x" + Integer.toHexString(pc) +
						  " Instruction 0x" + Integer.toHexString(inst) + "\n" +
						  "Opcode 0x" + Integer.toHexString(opcode) +
						  " bit6=" + bit6 +
						  " bits7_to_14=" + bits7_to_14 +
						  " bit15=" + bit15 +
						  " secondaryOpcode=" + secondaryOpcode +
						  " bit31=" + bit31 +
						  "\nDisassembly: " + PPC_Disassembler.disasm(inst, pc));
  }
  protected int translateXL_FORM(PPC2IR ppc2ir, PPC2IR.Laziness lazy, int pc, int inst, int opcode, int bits6_to_10, int bits11_to_15, int bits16_to_20, int secondaryOpcode, int bit31){
    throw new Error("Attempting to translateXL_Form on a non XL_Form instruction\n" +
						  "Address=0x" + Integer.toHexString(pc) +
						  " Instruction 0x" + Integer.toHexString(inst) + "\n" +
						  "Opcode 0x" + Integer.toHexString(inst) +
						  " bits6_to_10=" + bits6_to_10 +
						  " bits11_to_15=" + bits11_to_15 +
						  " bits16_to_20=" + bits16_to_20 +
						  " secondaryOpcode=" + secondaryOpcode +
						  " bit31=" + bit31 +
						  "\nDisassembly: " + PPC_Disassembler.disasm(inst, pc));
  }
  protected int translateXO_FORM(PPC2IR ppc2ir, PPC2IR.Laziness lazy, int pc, int inst, int opcode, int bits6_to_10, int bits11_to_15, int bits16_to_20, int bit21, int secondaryOpcode, int bit31) {
    throw new Error("Attempting to translateXO_Form on a non XO_Form instruction\n" +
						  "Address=0x" + Integer.toHexString(pc) +
						  " Instruction 0x" + Integer.toHexString(inst) + "\n" +
						  "Opcode 0x" + Integer.toHexString(inst) +
						  " bits6_to_10=" + bits6_to_10 +
						  " bits11_to_15=" + bits11_to_15 +
						  " bits16_to_20=" + bits16_to_20 +
						  " bit21=" + bit21 +
						  " secondaryOpcode=" + secondaryOpcode +
						  " bit31=" + bit31 +
						  "\nDisassembly: " + PPC_Disassembler.disasm(inst, pc));
  }
  protected int translateM_FORM(PPC2IR ppc2ir, PPC2IR.Laziness lazy, int pc, int inst, int opcode, int bits6_to_10, int bits11_to_15, int bits16_to_20, int bits21_to_25, int bits26_to_30, int bit31){
    throw new Error("Attempting to translateM_Form on a non M_Form instruction\n" +
						  "Address=0x" + Integer.toHexString(pc) +
						  " Instruction 0x" + Integer.toHexString(inst) + "\n" +
						  "Opcode 0x" + Integer.toHexString(inst) +
						  " bits6_to_10=" + bits6_to_10 +
						  " bits11_to_15=" + bits11_to_15 +
						  " bits16_to_20=" + bits16_to_20 +
						  " bits21_to_25=" + bits21_to_25 +
						  " bits26_to_30=" + bits26_to_30 +
						  " bit31=" + bit31 +
						  "\nDisassembly: " + PPC_Disassembler.disasm(inst, pc));
  }
  protected int translateA_FORM(PPC2IR ppc2ir, PPC2IR.Laziness lazy, int pc, int inst, int opcode, int bits6_to_10, int bits11_to_15, int bits16_to_20, int bits21_to_25, int bits26_to_30, int bit31){
    throw new Error("Attempting to translateM_Form on a non M_Form instruction\n" +
						  "Address=0x" + Integer.toHexString(pc) +
						  " Instruction 0x" + Integer.toHexString(inst) + "\n" +
						  "Opcode 0x" + Integer.toHexString(inst) +
						  " bits6_to_10=" + bits6_to_10 +
						  " bits11_to_15=" + bits11_to_15 +
						  " bits16_to_20=" + bits16_to_20 +
						  " bits21_to_25=" + bits21_to_25 +
						  " bits26_to_30=" + bits26_to_30 +
						  " bit31=" + bit31 +
						  "\nDisassembly: " + PPC_Disassembler.disasm(inst, pc));
  }
  protected OpCodeLookUp getOpCodeLookUp(int primaryOpcode, int secondaryOpcode) {
    throw new Error("Unknown opcode in opcode look up. Primary opcode " + primaryOpcode + " secondary opcode " + secondaryOpcode);
  }

  protected int translateEXTENDED(PPC2IR ppc2ir, PPC2IR.Laziness lazy, int pc, int instr, int primaryOpcode){
    int secondaryOpcode =  bits(instr, 21, 30);
	 OpCodeLookUp opCodeLookUp = getOpCodeLookUp(primaryOpcode, secondaryOpcode);
	 if (opCodeLookUp == null) {
		if(PPC_Options.debugInstr) {
		  System.out.println("Unknown extended instruction. Primary " + primaryOpcode + " secondary " + secondaryOpcode);
		}
		ppc2ir.plantThrowBadInstruction((PPC2IR.Laziness)lazy.clone(), pc);
		return -1;
		// throw new Error("Unknown extended instruction. Primary " + primaryOpcode + " secondary " + secondaryOpcode);
	 }
    switch(opCodeLookUp.getForm())  /* decode known instruction format */
      {
      case D_FORM:
		  return opCodeLookUp.translator.translateD_FORM(ppc2ir, lazy, pc, instr, primaryOpcode, bits(instr,6,10), bits(instr,11,15), bits(instr,16,31));
      case B_FORM:
		  return opCodeLookUp.translator.translateB_FORM(ppc2ir, lazy, pc, instr, primaryOpcode, bits(instr,6,10), bits(instr,11,15), bits(instr,16,29), bits(instr,30,30), instr & 0x1);
      case I_FORM:
		  return opCodeLookUp.translator.translateI_FORM(ppc2ir, lazy, pc, instr, primaryOpcode, bits(instr,6,29), bits(instr,30,30), instr & 0x1);
      case SC_FORM:
		  return opCodeLookUp.translator.translateSC_FORM(ppc2ir, lazy, pc, instr, primaryOpcode, bits(instr,6,10), bits(instr,11,15), bits(instr,16,29), bits(instr,30,30), instr & 0x1);
		case X_FORM:
		  return opCodeLookUp.translator.translateX_FORM(ppc2ir, lazy, pc, instr, primaryOpcode, bits(instr,6,10), bits(instr,11,15), bits(instr,16,20), secondaryOpcode, instr & 0x1);
		case XFX_FORM:
		  return opCodeLookUp.translator.translateXFX_FORM(ppc2ir, lazy, pc, instr, primaryOpcode, bits(instr,6,10), bits(instr,11,20), secondaryOpcode, instr & 0x1);
		case XFL_FORM:
		  return opCodeLookUp.translator.translateXFL_FORM(ppc2ir, lazy, pc, instr, primaryOpcode, bits(instr,6,6), bits(instr,7,14), bits(instr,15,15), bits(instr,16,20), secondaryOpcode, instr & 0x1);
      case XL_FORM:
		  return opCodeLookUp.translator.translateXL_FORM(ppc2ir, lazy, pc, instr, primaryOpcode, bits(instr,6,10), bits(instr,11,15), bits(instr,16,20), secondaryOpcode, instr & 0x1);
      case XO_FORM:
		  return opCodeLookUp.translator.translateXO_FORM(ppc2ir, lazy, pc, instr, primaryOpcode, bits(instr,6,10), bits(instr,11,15), bits(instr,16,20), bits(instr,21,21), bits(instr,22,30), instr & 0x1);
      case M_FORM:
		  return opCodeLookUp.translator.translateM_FORM(ppc2ir, lazy, pc, instr, primaryOpcode, bits(instr,6,10), bits(instr,11,15), bits(instr,16,20), bits(instr,21,25), bits(instr,26,30), instr & 0x1);
      case A_FORM:
		  return opCodeLookUp.translator.translateA_FORM(ppc2ir, lazy, pc, instr, primaryOpcode, bits(instr,6,10), bits(instr,11,15), bits(instr,16,20), bits(instr,21,25), bits(instr,26,30), instr & 0x1);
      case EXTENDED:
		  return opCodeLookUp.translator.translateEXTENDED(ppc2ir, lazy, pc, instr, primaryOpcode);
      case INVALID_OP:      /* More work to do... */
      default:
		  throw new Error("Invalid opcode. primary " + primaryOpcode + " secondary " + secondaryOpcode);
      }
  }
  /** 
   * Sign extend for &lt;32 bit quantities copied into an integer.
   * @param value a signed number occupying &lt;32 bits. 
   * @param n the number of digits used in the original value.
   */
  protected static int EXTS(int value, int n) {
    return (value << (32 - n)) >> (32 - n);
  }

  /**
   * Decrement CTR register
   */
  protected static void plantDecrementCTR(PPC2IR ppc2ir) {
    OPT_RegisterOperand ctr = ppc2ir.getCTRRegister();
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_SUB, ctr,
							 ctr.copyRO(),
							 new OPT_IntConstantOperand(1)));
  }

  /**
   * Set a condition register field
   * @param ppc2ir the current translation
   * @param lazy the current lazy state
   * @param crf the field to change
   * @param rA the lhs operand for the compare
   * @param rB the rhs operand for the compare
   * @param laziness whether to do a signed or unsigned comparison on rA and rB
   */
  protected static void setCRfield(PPC2IR ppc2ir, PPC2IR.Laziness lazy, int crf, OPT_Operand rA, OPT_Operand rB, int laziness) {
	 switch (laziness) {
	 case PPC2IR.Laziness.UNSIGNED_INT_CMP:
	 case PPC2IR.Laziness.SIGNED_INT_CMP:
		// Copy operands for lazy state
		ppc2ir.appendInstructionToCurrentBlock(Move.create(INT_MOVE,
																			ppc2ir.getCR_LtGtEq_rA_Register(crf),
																			rA));
		ppc2ir.appendInstructionToCurrentBlock(Move.create(INT_MOVE,
																			ppc2ir.getCR_LtGtEq_rB_Register(crf),
																			rB));
		// Copy SO bit
		ppc2ir.appendInstructionToCurrentBlock(Move.create(INT_MOVE,
																			ppc2ir.getCR_SO_Register(crf),
																			ppc2ir.getXER_SO_Register()));
		break;
	 case PPC2IR.Laziness.FP_CMPU:
		// Copy operands for lazy state
		ppc2ir.appendInstructionToCurrentBlock(Move.create(DOUBLE_MOVE,
																			ppc2ir.getCR_LtGtEq_frA_Register(crf),
																			rA));
		ppc2ir.appendInstructionToCurrentBlock(Move.create(DOUBLE_MOVE,
																			ppc2ir.getCR_LtGtEq_frB_Register(crf),
																			rB));
		break;
	 default:
		break;
	 }
    // Set lazy state for condition register field
    lazy.setConditionFieldLaziness(crf, laziness);
  }

  /**
   * Plant (likely) branch to block if ctr is zero (ctr_zero == true)
   * or not zero (ctr_zero == false). NB as this code is likely to be
   * used to branch to a fall through case the likely and ctr_zero
   * values are usually the opposite of what the instruction encodes
   */
  protected static void plantBranchToBlockDependentOnCTR(PPC2IR ppc2ir, OPT_BasicBlock targetBlock, boolean ctr_zero, boolean likely) {
    OPT_RegisterOperand ctr = ppc2ir.getCTRRegister();
    OPT_RegisterOperand guardResult = ppc2ir.getTempValidation(0);
    OPT_ConditionOperand condOp;
    if (ctr_zero) {
      condOp = OPT_ConditionOperand.EQUAL();
    }
    else {
      condOp = OPT_ConditionOperand.NOT_EQUAL();
    }
    OPT_BranchProfileOperand likelyOp;
    if (likely) {
      likelyOp = OPT_BranchProfileOperand.likely();
    }
    else {
      likelyOp = OPT_BranchProfileOperand.unlikely();
    }
    OPT_BasicBlock fallThrough = ppc2ir.createBlockAfterCurrent();
    ppc2ir.appendInstructionToCurrentBlock(IfCmp.create(INT_IFCMP, guardResult,
							ctr,
							new OPT_IntConstantOperand(0),
							condOp,
							targetBlock.makeJumpTarget(),
							likelyOp));
	 ppc2ir.getCurrentBlock().insertOut(targetBlock);
    ppc2ir.setCurrentBlock(fallThrough);
  }

  /**
   * Plant (likely) branch to block if condition is cond_true
   */
  protected static void plantBranchToBlockDependentOnCondition(PPC2IR ppc2ir, OPT_BasicBlock targetBlock, PPC2IR.Laziness lazy, int BI, boolean cond_true, boolean likely) {
    if (VM.VerifyAssertions) VM._assert ((BI >= 0) && (BI < 32));

    // Calculate the condition register field being tested
    int crf = BI >> 2;

    // Block entered after the test
    OPT_BasicBlock fallThrough = ppc2ir.createBlockAfterCurrent();

    // Create a branch likelihood operand
    OPT_BranchProfileOperand likelyOp;
    if (likely) {
      likelyOp = OPT_BranchProfileOperand.likely();
    }
    else {
      likelyOp = OPT_BranchProfileOperand.unlikely();
    }

    switch(lazy.getConditionFieldLaziness(crf)) {
    case PPC2IR.Laziness.NOT_LAZY:
      {
		  OPT_ConditionOperand condOp;
		  if (cond_true) { // NB invert sense of cond_true as we're comparing with 0
			 condOp = OPT_ConditionOperand.NOT_EQUAL();
		  }
		  else {
			 condOp = OPT_ConditionOperand.EQUAL();
		  }
		  OPT_RegisterOperand testResult = ppc2ir.getTempInt(0);
		  ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_AND, testResult,
																				 ppc2ir.getCRRegister(),
																				 new OPT_IntConstantOperand(0x80000000 >>> BI)));
		  OPT_RegisterOperand guardResult = ppc2ir.getTempValidation(0);
		  ppc2ir.appendInstructionToCurrentBlock(IfCmp.create(INT_IFCMP, guardResult,
																				testResult.copyRO(),
																				new OPT_IntConstantOperand(0),
																				condOp,
																				targetBlock.makeJumpTarget(),
																				likelyOp));
		  ppc2ir.getCurrentBlock().insertOut(targetBlock);
      }
      break;
    case PPC2IR.Laziness.SIGNED_INT_CMP:
      {
		  if ((BI & 3) < 3) {
			 // Lt, Gt or Eq cases
			 OPT_ConditionOperand condOp;
			 switch (BI & 3) {
			 default:
			 case 0: // LT
				if (cond_true) {
				  condOp = OPT_ConditionOperand.LESS();
				}
				else {
				  condOp = OPT_ConditionOperand.GREATER_EQUAL();
				}
				break;
			 case 1: // GT
				if (cond_true) {
				  condOp = OPT_ConditionOperand.GREATER();
				}
				else {
				  condOp = OPT_ConditionOperand.LESS_EQUAL();
				}
				break;
			 case 2: // EQ
				if (cond_true) {
				  condOp = OPT_ConditionOperand.EQUAL();
				}
				else {
				  condOp = OPT_ConditionOperand.NOT_EQUAL();
				}
				break;
			 }
			 OPT_RegisterOperand guardResult = ppc2ir.getTempValidation(0);
			 ppc2ir.appendInstructionToCurrentBlock(IfCmp.create(INT_IFCMP, guardResult,
																				  ppc2ir.getCR_LtGtEq_rA_Register(crf),
																				  ppc2ir.getCR_LtGtEq_rB_Register(crf),
																				  condOp,
																				  targetBlock.makeJumpTarget(),
																				  likelyOp));
			 ppc2ir.getCurrentBlock().insertOut(targetBlock);
		  }
		  else {
			 // SO case
			 OPT_ConditionOperand condOp;
			 if (cond_true) { // NB invert sense of cond_true as we're comparing with 0
				condOp = OPT_ConditionOperand.NOT_EQUAL();
			 }
			 else {
				condOp = OPT_ConditionOperand.EQUAL();
			 }
			 OPT_RegisterOperand guardResult = ppc2ir.getTempValidation(0);
			 ppc2ir.appendInstructionToCurrentBlock(IfCmp.create(INT_IFCMP, guardResult,
																				  ppc2ir.getCR_SO_Register(crf),
																				  new OPT_IntConstantOperand(0),
																				  condOp,
																				  targetBlock.makeJumpTarget(),
																				  likelyOp));
			 ppc2ir.getCurrentBlock().insertOut(targetBlock);
		  }
      }
		break;
    case PPC2IR.Laziness.UNSIGNED_INT_CMP:
      {
		  if ((BI & 3) < 3) {
			 // Lt, Gt or Eq cases
			 OPT_ConditionOperand condOp;
			 switch (BI & 3) {
			 default:
			 case 0: // LT
				if (cond_true) {
				  condOp = OPT_ConditionOperand.LOWER();
				}
				else {
				  condOp = OPT_ConditionOperand.HIGHER_EQUAL();
				}
				break;
			 case 1: // GT
				if (cond_true) {
				  condOp = OPT_ConditionOperand.HIGHER();
				}
				else {
				  condOp = OPT_ConditionOperand.LOWER_EQUAL();
				}
				break;
			 case 2: // EQ
				if (cond_true) {
				  condOp = OPT_ConditionOperand.EQUAL();
				}
				else {
				  condOp = OPT_ConditionOperand.NOT_EQUAL();
				}
				break;
			 }
			 OPT_RegisterOperand guardResult = ppc2ir.getTempValidation(0);
			 ppc2ir.appendInstructionToCurrentBlock(IfCmp.create(INT_IFCMP, guardResult,
																				  ppc2ir.getCR_LtGtEq_rA_Register(crf),
																				  ppc2ir.getCR_LtGtEq_rB_Register(crf),
																				  condOp,
																				  targetBlock.makeJumpTarget(),
																				  likelyOp));
			 ppc2ir.getCurrentBlock().insertOut(targetBlock);
		  }
		  else {
			 // SO case
			 OPT_ConditionOperand condOp;
			 if (cond_true) { // NB invert sense of cond_true as we're comparing with 0
				condOp = OPT_ConditionOperand.NOT_EQUAL();
			 }
			 else {
				condOp = OPT_ConditionOperand.EQUAL();
			 }
			 OPT_RegisterOperand guardResult = ppc2ir.getTempValidation(0);
			 ppc2ir.appendInstructionToCurrentBlock(IfCmp.create(INT_IFCMP, guardResult,
																				  ppc2ir.getCR_SO_Register(crf),
																				  new OPT_IntConstantOperand(0),
																				  condOp,
																				  targetBlock.makeJumpTarget(),
																				  likelyOp));
			 ppc2ir.getCurrentBlock().insertOut(targetBlock);
		  }
      }
		break;
	 case PPC2IR.Laziness.FP_CMPU:
		/*
		ppc2ir.resolveLazinessCrField(lazy, crf);
		plantBranchToBlockDependentOnCondition(ppc2ir, targetBlock, lazy, BI, cond_true, likely);
		if (false)
		*/
		{
		  if ((BI & 3) < 3) {
			 // Lt, Gt or Eq cases
			 OPT_ConditionOperand condOp;
			 OPT_Operator operator;
			 switch (BI & 3) {
			 default:
			 case 0: // LT
				condOp = OPT_ConditionOperand.LESS();
				operator = DOUBLE_IFCMPG;
				break;
			 case 1: // GT
				condOp = OPT_ConditionOperand.GREATER();
				operator = DOUBLE_IFCMPL;
				break;
			 case 2: // EQ
				condOp = OPT_ConditionOperand.EQUAL();
				operator = DOUBLE_IFCMPG;
				break;
			 }
			 OPT_RegisterOperand guardResult = ppc2ir.getTempValidation(0);
			 if (cond_true) {
				// Branch if <, > or ==
				ppc2ir.appendInstructionToCurrentBlock(IfCmp.create(operator, guardResult,
																					 ppc2ir.getCR_LtGtEq_frA_Register(crf),
																					 ppc2ir.getCR_LtGtEq_frB_Register(crf),
																					 condOp,
																					 targetBlock.makeJumpTarget(),
																					 likelyOp));
				ppc2ir.getCurrentBlock().insertOut(targetBlock);
			 }
			 else {
				// Branch if not <, > or == - n.b. care taken about unordered
				OPT_BasicBlock gotoTarget = ppc2ir.createBlockAfterCurrent();
				OPT_BasicBlock test = ppc2ir.getCurrentBlock();
				test.appendInstruction(IfCmp.create(operator, guardResult,
																ppc2ir.getCR_LtGtEq_frA_Register(crf),
																ppc2ir.getCR_LtGtEq_frB_Register(crf),
																condOp,
																fallThrough.makeJumpTarget(),
																likelyOp));
				test.insertOut(fallThrough);
				gotoTarget.appendInstruction(Goto.create(GOTO, targetBlock.makeJumpTarget()));
				gotoTarget.deleteNormalOut();
				gotoTarget.insertOut(targetBlock);
			 }
		  }
		  else {
			 // FU case
			 OPT_RegisterOperand guardResult = ppc2ir.getTempValidation(0);
			 if (cond_true) {
				ppc2ir.appendInstructionToCurrentBlock(IfCmp.create(DOUBLE_IFCMPL, guardResult,
																					 ppc2ir.getCR_LtGtEq_frA_Register(crf),
																					 ppc2ir.getCR_LtGtEq_frA_Register(crf),
																					 OPT_ConditionOperand.EQUAL(),
																					 targetBlock.makeJumpTarget(),
																					 likelyOp));
				ppc2ir.getCurrentBlock().insertOut(targetBlock);
				ppc2ir.setCurrentBlock(ppc2ir.createBlockAfterCurrent());
				ppc2ir.appendInstructionToCurrentBlock(IfCmp.create(DOUBLE_IFCMPL, guardResult.copyRO(),
																					 ppc2ir.getCR_LtGtEq_frB_Register(crf),
																					 ppc2ir.getCR_LtGtEq_frB_Register(crf),
																					 OPT_ConditionOperand.EQUAL(),
																					 targetBlock.makeJumpTarget(),
																					 likelyOp));
				ppc2ir.getCurrentBlock().insertOut(targetBlock);
			 }
			 else{
				ppc2ir.appendInstructionToCurrentBlock(IfCmp.create(DOUBLE_IFCMPL, guardResult,
																					 ppc2ir.getCR_LtGtEq_frA_Register(crf),
																					 ppc2ir.getCR_LtGtEq_frA_Register(crf),
																					 OPT_ConditionOperand.NOT_EQUAL(),
																					 targetBlock.makeJumpTarget(),
																					 likelyOp));
				ppc2ir.getCurrentBlock().insertOut(targetBlock);
				ppc2ir.setCurrentBlock(ppc2ir.createBlockAfterCurrent());
				ppc2ir.appendInstructionToCurrentBlock(IfCmp.create(DOUBLE_IFCMPL, guardResult.copyRO(),
																					 ppc2ir.getCR_LtGtEq_frB_Register(crf),
																					 ppc2ir.getCR_LtGtEq_frB_Register(crf),
																					 OPT_ConditionOperand.NOT_EQUAL(),
																					 targetBlock.makeJumpTarget(),
																					 likelyOp));
				ppc2ir.getCurrentBlock().insertOut(targetBlock);
			 }
		  }
		}
		break;
    default:
      throw new Error("Unhandled laziness " + lazy.getConditionFieldLaziness(crf));
    }
    // Move currentBlock onto fall through
    ppc2ir.setCurrentBlock(fallThrough);
  }
}

// -oO OpCodeLookUp Oo-

/**
 * A class holding what's looked up from an instruction
 */
class OpCodeLookUp {
  private int form;
  String mnemonic;
  PPC_InstructionTranslator translator;
  OpCodeLookUp(int _form, String _mnemonic, PPC_InstructionTranslator _instrTranslator) {
	 form = _form;
	 mnemonic = _mnemonic;
	 translator=_instrTranslator;
  }
  int getForm() {
	 return form;
  }
}

// -oO Extended Instructions Oo-
/**
 * The translator for the XL_FORM instructions
 */
class xlForm_translator extends PPC_InstructionTranslator {
  /**
	* Table for the XL instruction format
	*/
  private static final OpCodeLookUp[] XLform = {
	 /*   OPCD      EO                           form     mnemonic         translator */
	 /*   ----      --                           ----     --------         ---------- */ 
	 /*    19,      0,      */  new OpCodeLookUp(XL_FORM, "mcrf",          new mcrf_translator()),
	 /*    19,      16,     */  new OpCodeLookUp(XL_FORM, "bclr or bclrl", new bclr_translator()),
	 /*    19,      33,     */  new OpCodeLookUp(XL_FORM, "crnor",         new crnor_translator()),
	 /*    19,      50,     */  new OpCodeLookUp(XL_FORM, "rfi",           new rfi_translator()),
	 /*    19,      XXX,    */  null,
    
	 /*    19,      XXX,    */  null,
	 /*    19,      XXX,    */  null,
	 /*    19,      XXX,    */  null,
	 /*    19,      129,    */  new OpCodeLookUp(XL_FORM, "crandc",        new crandc_translator()),
	 /*    19,      150,    */  new OpCodeLookUp(XL_FORM, "isync",         new isync_transltor()),
    
	 /*    19,      XXX,    */  null,
	 /*    19,      XXX,    */  null,
	 /*    19,      193,    */  new OpCodeLookUp(XL_FORM, "crxor",         new crxor_translator()),
	 /*    19,      XXX,    */  null,
	 /*    19,      225,    */  new OpCodeLookUp(XL_FORM, "crnand",        new crnand_translator()),
    
	 /*    19,      XXX,    */  null,
	 /*    19,      257,    */  new OpCodeLookUp(XL_FORM, "crand",         new crand_translator()),
	 /*    19,      XXX,    */  null,
	 /*    19,      289,    */  new OpCodeLookUp(XL_FORM, "creqv",         new creqv_translator()),
	 /*    19,      XXX,    */  null,
    
	 /*    19,      XXX,    */  null,
	 /*    19,      XXX,    */  null,
	 /*    19,      XXX,    */  null,
	 /*    19,      XXX,    */  null,
	 /*    19,      XXX,    */  null,
    
	 /*    19,      XXX,    */  null,
	 /*    19,      417,    */  new OpCodeLookUp(XL_FORM, "crorc",         new croc_translator()),
	 /*    19,      XXX,    */  null,
	 /*    19,      449,    */  new OpCodeLookUp(XL_FORM, "cror",          new cror_traslator()),
	 /*    19,      XXX,    */  null,
    
	 /*    19,      XXX,    */  null,
	 /*    19,      XXX,    */  null,
	 /*    19,      XXX,    */  null,
	 /*    19,      528,    */  new OpCodeLookUp(XL_FORM, "bcctr or bcctrl", new bcctr_translator())
  };
  protected int translateXL_FORM(PPC2IR ppc2ir, PPC2IR.Laziness lazy, int pc, int inst, int opcode, int bits6_to_10, int bits11_to_15, int bits16_to_20, int secondaryOpcode, int bit31){
	 return XLform[secondaryOpcode >> 4].translator.translateXL_FORM(ppc2ir, lazy, pc, inst, opcode, bits6_to_10, bits11_to_15, bits16_to_20, secondaryOpcode, bit31);
  }
}

/**
 * The translator for the EXTENDED 31 instructions
 */
class extended31_translator extends PPC_InstructionTranslator {

  // The translators for the XO_Form instructions
  private static OpCodeLookUp subfc_lookup = new OpCodeLookUp(XO_FORM, "subfc", new subfc_translator());
  private static OpCodeLookUp addc_lookup = new OpCodeLookUp(XO_FORM, "addc", new addc_translator());
  private static OpCodeLookUp mulhwu_lookup = new OpCodeLookUp(XO_FORM, "mulhwu", new mulhwu_translator());
  private static OpCodeLookUp subf_lookup = new OpCodeLookUp(XO_FORM, "subf", new subf_translator());
  private static OpCodeLookUp mulhw_lookup = new OpCodeLookUp(XO_FORM, "mulhw", new mulhw_translator());
  private static OpCodeLookUp neg_lookup = new OpCodeLookUp(XO_FORM, "neg", new neg_translator());
  private static OpCodeLookUp subfe_lookup = new OpCodeLookUp(XO_FORM, "subfe", new subfe_translator());
  private static OpCodeLookUp adde_lookup = new OpCodeLookUp(XO_FORM, "adde", new adde_translator());
  private static OpCodeLookUp subfze_lookup = new OpCodeLookUp(XO_FORM, "subfze", new subfze_translator());
  private static OpCodeLookUp addze_lookup = new OpCodeLookUp(XO_FORM, "addze", new addze_translator());
  private static OpCodeLookUp subfme_lookup = new OpCodeLookUp(XO_FORM, "subfme", new subfme_translator());
  private static OpCodeLookUp addme_lookup = new OpCodeLookUp(XO_FORM, "addme", new addme_translator());
  private static OpCodeLookUp mullw_lookup = new OpCodeLookUp(XO_FORM, "mullw", new mullw_translator());
  private static OpCodeLookUp add_lookup = new OpCodeLookUp(XO_FORM, "add", new add_translator());
  private static OpCodeLookUp divwu_lookup = new OpCodeLookUp(XO_FORM, "divwu", new divwu_translator());
  private static OpCodeLookUp divw_lookup = new OpCodeLookUp(XO_FORM, "divw", new divw_translator());
  private static OpCodeLookUp subfco_lookup = new OpCodeLookUp(XO_FORM, "subfco", new subfco_translator());
  private static OpCodeLookUp addco_lookup = new OpCodeLookUp(XO_FORM, "addco", new addco_translator());
  private static OpCodeLookUp subfo_lookup = new OpCodeLookUp(XO_FORM, "subfo", new subfo_translator());
  private static OpCodeLookUp subfeo_lookup = new OpCodeLookUp(XO_FORM, "subfeo", new subfeo_translator());
  private static OpCodeLookUp addeo_lookup = new OpCodeLookUp(XO_FORM, "addeo", new addeo_translator());
  private static OpCodeLookUp subfzeo_lookup = new OpCodeLookUp(XO_FORM, "subfzeo", new subfzeo_translator());
  private static OpCodeLookUp addzeo_lookup = new OpCodeLookUp(XO_FORM, "addzeo", new addzeo_translator());
  private static OpCodeLookUp subfmeo_lookup = new OpCodeLookUp(XO_FORM, "subfmeo", new subfmeo_translator());
  private static OpCodeLookUp addmeo_lookup = new OpCodeLookUp(XO_FORM, "addmeo", new addmeo_translator());
  private static OpCodeLookUp mullwo_lookup = new OpCodeLookUp(XO_FORM, "mullwo", new mullwo_translator());
  private static OpCodeLookUp nego_lookup = new OpCodeLookUp(XO_FORM, "nego", new nego_translator());
  private static OpCodeLookUp addo_lookup = new OpCodeLookUp(XO_FORM, "addo", new addo_translator());
  private static OpCodeLookUp divwuo_lookup = new OpCodeLookUp(XO_FORM, "divwuo", new divwuo_translator());
  private static OpCodeLookUp divwo_lookup = new OpCodeLookUp(XO_FORM, "divwo", new divwo_translator());

  // The translators for the XFX_Form instructions
  private static OpCodeLookUp mtcrf_lookup = new OpCodeLookUp(XFX_FORM, "mtcrf", new mtcrf_translator());
  private static OpCodeLookUp mfspr_lookup = new OpCodeLookUp(XFX_FORM, "mfspr", new mfspr_translator());
  private static OpCodeLookUp mtspr_lookup = new OpCodeLookUp(XFX_FORM, "mtspr", new mtspr_translator());

  // The translators for the X_Form instructions
  private static OpCodeLookUp cmp_lookup = new OpCodeLookUp(X_FORM, "cmp", new cmp_translator());
  private static OpCodeLookUp tw_lookup = new OpCodeLookUp(X_FORM, "tw", new tw_translator());
  private static OpCodeLookUp lvsl_lookup = new OpCodeLookUp(X_FORM, "lvsl", new lvsl_translator());
  private static OpCodeLookUp lvebx_lookup = new OpCodeLookUp(X_FORM, "lvebx", new lvebx_translator());
  private static OpCodeLookUp mulhdu_lookup = new OpCodeLookUp(X_FORM, "mulhdu", new mulhdu_translator());
  private static OpCodeLookUp mfcr_lookup = new OpCodeLookUp(X_FORM, "mfcr", new mfcr_translator());
  private static OpCodeLookUp lwarx_lookup = new OpCodeLookUp(X_FORM, "lwarx", new lwarx_translator());
  private static OpCodeLookUp ldx_lookup = new OpCodeLookUp(X_FORM, "ldx", new ldx_translator());
  private static OpCodeLookUp lwzx_lookup = new OpCodeLookUp(X_FORM, "lwzx", new lwzx_translator());
  private static OpCodeLookUp slw_lookup = new OpCodeLookUp(X_FORM, "slw", new slw_translator());
  private static OpCodeLookUp cntlzw_lookup = new OpCodeLookUp(X_FORM, "cntlzw", new cntlzw_translator());
  private static OpCodeLookUp sld_lookup = new OpCodeLookUp(X_FORM, "sld", new sld_translator());
  private static OpCodeLookUp and_lookup = new OpCodeLookUp(X_FORM, "and", new and_translator());
  private static OpCodeLookUp cmpl_lookup = new OpCodeLookUp(X_FORM, "cmpl", new cmpl_translator());
  private static OpCodeLookUp lvsr_lookup = new OpCodeLookUp(X_FORM, "lvsr", new lvsr_translator());
  private static OpCodeLookUp lvehx_lookup = new OpCodeLookUp(X_FORM, "lvehx", new lvehx_translator());
  private static OpCodeLookUp ldux_lookup = new OpCodeLookUp(X_FORM, "ldux", new ldux_translator());
  private static OpCodeLookUp dcbst_lookup = new OpCodeLookUp(X_FORM, "dcbst", new dcbst_translator());
  private static OpCodeLookUp lwzux_lookup = new OpCodeLookUp(X_FORM, "lwzux", new lwzux_translator());
  private static OpCodeLookUp andc_lookup = new OpCodeLookUp(X_FORM, "andc", new andc_translator());
  private static OpCodeLookUp td_lookup = new OpCodeLookUp(X_FORM, "td", new td_translator());
  private static OpCodeLookUp mulhd_lookup = new OpCodeLookUp(X_FORM, "mulhd", new mulhd_translator());
  private static OpCodeLookUp mfmsr_lookup = new OpCodeLookUp(X_FORM, "mfmsr", new mfmsr_translator());
  private static OpCodeLookUp ldarx_lookup = new OpCodeLookUp(X_FORM, "ldarx", new ldarx_translator());
  private static OpCodeLookUp dcbf_lookup = new OpCodeLookUp(X_FORM, "dcbf", new dcbf_translator());
  private static OpCodeLookUp lbzx_lookup = new OpCodeLookUp(X_FORM, "lbzx", new lbzx_translator());
  private static OpCodeLookUp lbzux_lookup = new OpCodeLookUp(X_FORM, "lbzux", new lbzux_translator());
  private static OpCodeLookUp nor_lookup = new OpCodeLookUp(X_FORM, "nor", new nor_translator());
  private static OpCodeLookUp mtmsr_lookup = new OpCodeLookUp(X_FORM, "mtmsr", new mtmsr_translator());
  private static OpCodeLookUp stdx_lookup = new OpCodeLookUp(X_FORM, "stdx", new stdx_translator());
  private static OpCodeLookUp stwcx_lookup = new OpCodeLookUp(X_FORM, "stwcx", new stwcx_translator());
  private static OpCodeLookUp stwx_lookup = new OpCodeLookUp(X_FORM, "stwx", new stwx_translator());
  private static OpCodeLookUp stdux_lookup = new OpCodeLookUp(X_FORM, "stdux", new stdux_translator());
  private static OpCodeLookUp stwux_lookup = new OpCodeLookUp(X_FORM, "stwux", new stwux_translator());
  private static OpCodeLookUp mtsr_lookup = new OpCodeLookUp(X_FORM, "mtsr", new mtsr_translator());
  private static OpCodeLookUp stdcx_lookup = new OpCodeLookUp(X_FORM, "stdcx", new stdcx_translator());
  private static OpCodeLookUp stbx_lookup = new OpCodeLookUp(X_FORM, "stbx", new stbx_translator());
  private static OpCodeLookUp mulld_lookup = new OpCodeLookUp(X_FORM, "mulld", new mulld_translator());
  private static OpCodeLookUp mtsrin_lookup = new OpCodeLookUp(X_FORM, "mtsrin", new mtsrin_translator());
  private static OpCodeLookUp dcbtst_lookup = new OpCodeLookUp(X_FORM, "dcbtst", new dcbtst_translator());
  private static OpCodeLookUp stbux_lookup = new OpCodeLookUp(X_FORM, "stbux", new stbux_translator());
  private static OpCodeLookUp dcbt_lookup = new OpCodeLookUp(X_FORM, "dcbt", new dcbt_translator());
  private static OpCodeLookUp lhzx_lookup = new OpCodeLookUp(X_FORM, "lhzx", new lhzx_translator());
  private static OpCodeLookUp eqv_lookup = new OpCodeLookUp(X_FORM, "eqv", new eqv_translator());
  private static OpCodeLookUp tlbie_lookup = new OpCodeLookUp(X_FORM, "tlbie", new tlbie_translator());
  private static OpCodeLookUp eciwx_lookup = new OpCodeLookUp(X_FORM, "eciwx", new eciwx_translator());
  private static OpCodeLookUp lhzux_lookup = new OpCodeLookUp(X_FORM, "lhzux", new lhzux_translator());
  private static OpCodeLookUp xor_lookup = new OpCodeLookUp(X_FORM, "xor", new xor_translator());
  private static OpCodeLookUp lwax_lookup = new OpCodeLookUp(X_FORM, "lwax", new lwax_translator());
  private static OpCodeLookUp dst_lookup = new OpCodeLookUp(X_FORM, "dst", new dst_translator());
  private static OpCodeLookUp lhax_lookup = new OpCodeLookUp(X_FORM, "lhax", new lhax_translator());
  private static OpCodeLookUp lvxl_lookup = new OpCodeLookUp(X_FORM, "lvxl", new lvxl_translator());
  private static OpCodeLookUp tlbia_lookup = new OpCodeLookUp(X_FORM, "tlbia", new tlbia_translator());
  private static OpCodeLookUp mftb_lookup = new OpCodeLookUp(X_FORM, "mftb", new mftb_translator());
  private static OpCodeLookUp lwaux_lookup = new OpCodeLookUp(X_FORM, "lwaux", new lwaux_translator());
  private static OpCodeLookUp dstst_lookup = new OpCodeLookUp(X_FORM, "dstst", new dstst_translator());
  private static OpCodeLookUp lhaux_lookup = new OpCodeLookUp(X_FORM, "lhaux", new lhaux_translator());
  private static OpCodeLookUp sthx_lookup = new OpCodeLookUp(X_FORM, "sthx", new sthx_translator());
  private static OpCodeLookUp orc_lookup = new OpCodeLookUp(X_FORM, "orc", new orc_translator());
  private static OpCodeLookUp slbie_lookup = new OpCodeLookUp(X_FORM, "slbie", new slbie_translator());
  private static OpCodeLookUp ecowx_lookup = new OpCodeLookUp(X_FORM, "ecowx", new ecowx_translator());
  private static OpCodeLookUp sthux_lookup = new OpCodeLookUp(X_FORM, "sthux", new sthux_translator());
  private static OpCodeLookUp or_lookup = new OpCodeLookUp(X_FORM, "or", new or_translator());
  private static OpCodeLookUp divdu_lookup = new OpCodeLookUp(X_FORM, "divdu", new divdu_translator());
  private static OpCodeLookUp dcbi_lookup = new OpCodeLookUp(X_FORM, "dcbi", new dcbi_translator());
  private static OpCodeLookUp nand_lookup = new OpCodeLookUp(X_FORM, "nand", new nand_translator());
  private static OpCodeLookUp divd_lookup = new OpCodeLookUp(X_FORM, "divd", new divd_translator());
  private static OpCodeLookUp slbia_lookup = new OpCodeLookUp(X_FORM, "slbia", new slbia_translator());
  private static OpCodeLookUp mcrxr_lookup = new OpCodeLookUp(X_FORM, "mcrxr", new mcrxr_translator());
  private static OpCodeLookUp lswx_lookup = new OpCodeLookUp(X_FORM, "lswx", new lswx_translator());
  private static OpCodeLookUp lwbrx_lookup = new OpCodeLookUp(X_FORM, "lwbrx", new lwbrx_translator());
  private static OpCodeLookUp lfsx_lookup = new OpCodeLookUp(X_FORM, "lfsx", new lfsx_translator());
  private static OpCodeLookUp srw_lookup = new OpCodeLookUp(X_FORM, "srw", new srw_translator());
  private static OpCodeLookUp srd_lookup = new OpCodeLookUp(X_FORM, "srd", new srd_translator());
  private static OpCodeLookUp cntlzd_lookup = new OpCodeLookUp(X_FORM, "cntlzd", new cntlzd_translator());
  private static OpCodeLookUp lfsux_lookup = new OpCodeLookUp(X_FORM, "lfsux", new lfsux_translator());
  private static OpCodeLookUp mfsr_lookup = new OpCodeLookUp(X_FORM, "mfsr", new mfsr_translator());
  private static OpCodeLookUp lswi_lookup = new OpCodeLookUp(X_FORM, "lswi", new lswi_translator());
  private static OpCodeLookUp sync_lookup = new OpCodeLookUp(X_FORM, "sync", new sync_translator());
  private static OpCodeLookUp lfdx_lookup = new OpCodeLookUp(X_FORM, "lfdx", new lfdx_translator());
  private static OpCodeLookUp lfdux_lookup = new OpCodeLookUp(X_FORM, "lfdux", new lfdux_translator());
  private static OpCodeLookUp mfsrin_lookup = new OpCodeLookUp(X_FORM, "mfsrin", new mfsrin_translator());
  private static OpCodeLookUp stswx_lookup = new OpCodeLookUp(X_FORM, "stswx", new stswx_translator());
  private static OpCodeLookUp stwbrx_lookup = new OpCodeLookUp(X_FORM, "stwbrx", new stwbrx_translator());
  private static OpCodeLookUp stfsx_lookup = new OpCodeLookUp(X_FORM, "stfsx", new stfsx_translator());
  private static OpCodeLookUp stfsux_lookup = new OpCodeLookUp(X_FORM, "stfsux", new stfsux_translator());
  private static OpCodeLookUp stswi_lookup = new OpCodeLookUp(X_FORM, "stswi", new stswi_translator());
  private static OpCodeLookUp stfdx_lookup = new OpCodeLookUp(X_FORM, "stfdx", new stfdx_translator());
  private static OpCodeLookUp dcba_lookup = new OpCodeLookUp(X_FORM, "dcba", new dcba_translator());
  private static OpCodeLookUp stfdux_lookup = new OpCodeLookUp(X_FORM, "stfdux", new stfdux_translator());
  private static OpCodeLookUp lhbrx_lookup = new OpCodeLookUp(X_FORM, "lhbrx", new lhbrx_translator());
  private static OpCodeLookUp sraw_lookup = new OpCodeLookUp(X_FORM, "sraw", new sraw_translator());
  private static OpCodeLookUp srad_lookup = new OpCodeLookUp(X_FORM, "srad", new srad_translator());
  private static OpCodeLookUp dss_lookup = new OpCodeLookUp(X_FORM, "dss", new dss_translator());
  private static OpCodeLookUp srawi_lookup = new OpCodeLookUp(X_FORM, "srawi", new srawi_translator());
  private static OpCodeLookUp sradi_lookup = new OpCodeLookUp(X_FORM, "sradi", new sradi_translator());
  // private static OpCodeLookUp sradi_lookup = new OpCodeLookUp(X_FORM, "sradi", new sradi_translator());
  private static OpCodeLookUp eieio_lookup = new OpCodeLookUp(X_FORM, "eieio", new eieio_translator());
  private static OpCodeLookUp sthbrx_lookup = new OpCodeLookUp(X_FORM, "sthbrx", new sthbrx_translator());
  private static OpCodeLookUp extsh_lookup = new OpCodeLookUp(X_FORM, "extsh", new extsh_translator());
  private static OpCodeLookUp extsb_lookup = new OpCodeLookUp(X_FORM, "extsb", new extsb_translator());
  private static OpCodeLookUp tlbld_lookup = new OpCodeLookUp(X_FORM, "tlbld", new tlbld_translator());
  private static OpCodeLookUp icbi_lookup = new OpCodeLookUp(X_FORM, "icbi", new icbi_translator());
  private static OpCodeLookUp stfiwx_lookup = new OpCodeLookUp(X_FORM, "stfiwx", new stfiwx_translator());
  private static OpCodeLookUp extsw_lookup = new OpCodeLookUp(X_FORM, "extsw", new extsw_translator());
  private static OpCodeLookUp tlbli_lookup = new OpCodeLookUp(X_FORM, "tlbli", new tlbli_translator());
  private static OpCodeLookUp dcbz_lookup = new OpCodeLookUp(X_FORM, "dcbz", new dcbz_translator());
  protected OpCodeLookUp getOpCodeLookUp(int primaryOpcode, int secondaryOpcode) {
	 switch(secondaryOpcode){
	 case 0: return cmp_lookup;
	 case 4: return tw_lookup;
	 case 6: return lvsl_lookup;
	 case 7: return lvebx_lookup;
	 case 8: return subfc_lookup;
	 case 9: return mulhdu_lookup;
	 case 10: return addc_lookup;
	 case 11: return mulhwu_lookup;
	 case 19: return mfcr_lookup;
	 case 20: return lwarx_lookup;
	 case 21: return ldx_lookup;
	 case 23: return lwzx_lookup;
	 case 24: return slw_lookup;
	 case 26: return cntlzw_lookup;
	 case 27: return sld_lookup;
	 case 28: return and_lookup;
	 case 32: return cmpl_lookup;
	 case 38: return lvsr_lookup;
	 case 39: return lvehx_lookup;
	 case 40: return subf_lookup;
	 case 53: return ldux_lookup;
	 case 54: return dcbst_lookup;
	 case 55: return lwzux_lookup;
	 case 60: return andc_lookup;
	 case 68: return td_lookup;
	 case 73: return mulhd_lookup;
	 case 75: return mulhw_lookup;
	 case 83: return mfmsr_lookup;
	 case 84: return ldarx_lookup;
	 case 86: return dcbf_lookup;
	 case 87: return lbzx_lookup;
	 case 104: return neg_lookup;
	 case 119: return lbzux_lookup;
	 case 124: return nor_lookup;
	 case 136: return subfe_lookup;
	 case 138: return adde_lookup;
	 case 144: return mtcrf_lookup;
	 case 146: return mtmsr_lookup;
	 case 149: return stdx_lookup;
	 case 150: return stwcx_lookup;
	 case 151: return stwx_lookup;
	 case 181: return stdux_lookup;
	 case 183: return stwux_lookup;
	 case 200: return subfze_lookup;
	 case 202: return addze_lookup;
	 case 210: return mtsr_lookup;
	 case 214: return stdcx_lookup;
	 case 215: return stbx_lookup;
	 case 232: return subfme_lookup;
	 case 233: return mulld_lookup;
	 case 234: return addme_lookup;
	 case 235: return mullw_lookup;
	 case 242: return mtsrin_lookup;
	 case 246: return dcbtst_lookup;
	 case 247: return stbux_lookup;
	 case 266: return add_lookup;
	 case 278: return dcbt_lookup;
	 case 279: return lhzx_lookup;
	 case 284: return eqv_lookup;
	 case 306: return tlbie_lookup;
	 case 310: return eciwx_lookup;
	 case 311: return lhzux_lookup;
	 case 316: return xor_lookup;
	 case 339: return mfspr_lookup;
	 case 341: return lwax_lookup;
	 case 342: return dst_lookup;
	 case 343: return lhax_lookup;
	 case 359: return lvxl_lookup;
	 case 370: return tlbia_lookup;
	 case 371: return mftb_lookup;
	 case 373: return lwaux_lookup;
	 case 374: return dstst_lookup;
	 case 375: return lhaux_lookup;
	 case 407: return sthx_lookup;
	 case 412: return orc_lookup;
	 case 434: return slbie_lookup;
	 case 438: return ecowx_lookup;
	 case 439: return sthux_lookup;
	 case 444: return or_lookup;
	 case 457: return divdu_lookup;
	 case 459: return divwu_lookup;
	 case 467: return mtspr_lookup;
	 case 470: return dcbi_lookup;
	 case 476: return nand_lookup;
	 case 489: return divd_lookup;
	 case 491: return divw_lookup;
	 case 498: return slbia_lookup;
	 case 512: return mcrxr_lookup;
	 case 520: return subfco_lookup;
	 case 522: return addco_lookup;
	 case 533: return lswx_lookup;
	 case 534: return lwbrx_lookup;
	 case 535: return lfsx_lookup;
	 case 536: return srw_lookup;
	 case 539: return srd_lookup;
	 case 552: return subfo_lookup;
	 case 558: return cntlzd_lookup;
	 case 566: return lfsux_lookup;
	 case 595: return mfsr_lookup;
	 case 597: return lswi_lookup;
	 case 598: return sync_lookup;
	 case 599: return lfdx_lookup;
	 case 631: return lfdux_lookup;
	 case 648: return subfeo_lookup;
	 case 650: return addeo_lookup;
	 case 659: return mfsrin_lookup;
	 case 661: return stswx_lookup;
	 case 662: return stwbrx_lookup;
	 case 663: return stfsx_lookup;
	 case 695: return stfsux_lookup;
	 case 712: return subfzeo_lookup;
	 case 714: return addzeo_lookup;
	 case 725: return stswi_lookup;
	 case 727: return stfdx_lookup;
	 case 744: return subfmeo_lookup;
	 case 746: return addmeo_lookup;
	 case 747: return mullwo_lookup;
	 case 758: return dcba_lookup;
	 case 759: return stfdux_lookup;
	 case 772: return nego_lookup;
	 case 778: return addo_lookup;
	 case 790: return lhbrx_lookup;
	 case 792: return sraw_lookup;
	 case 794: return srad_lookup;
	 case 822: return dss_lookup;
	 case 824: return srawi_lookup;
	 case 826: // both 826 & 827 are sradi
	 case 827: return sradi_lookup;
	 case 854: return eieio_lookup;
	 case 918: return sthbrx_lookup;
	 case 922: return extsh_lookup;
	 case 954: return extsb_lookup;
	 case 971: return divwuo_lookup;
	 case 978: return tlbld_lookup;
	 case 982: return icbi_lookup;
	 case 983: return stfiwx_lookup;
	 case 986: return extsw_lookup;
	 case 1003: return divwo_lookup;
	 case 1010: return tlbli_lookup;
	 case 1014: return dcbz_lookup;
	 default: return null;
	 }
  }
}

/**
 * The translator for the EXTENDED 59 instructions
 */
class extended59_translator extends PPC_InstructionTranslator {
  private static OpCodeLookUp fdivs_lookup = new OpCodeLookUp(A_FORM, "fdivs", new fdivs_translator());
  private static OpCodeLookUp fsubs_lookup = new OpCodeLookUp(A_FORM, "fsubs", new fsubs_translator());  
  private static OpCodeLookUp fadds_lookup = new OpCodeLookUp(A_FORM, "fadds", new fadds_translator());
  private static OpCodeLookUp fsqrts_lookup = new OpCodeLookUp(A_FORM, "fsqrts", new fsqrts_translator());
  private static OpCodeLookUp fres_lookup = new OpCodeLookUp(A_FORM, "fres", new fres_translator());
  private static OpCodeLookUp fmuls_lookup = new OpCodeLookUp(A_FORM, "fmuls", new fmuls_translator());  
  private static OpCodeLookUp fmsubs_lookup = new OpCodeLookUp(A_FORM, "fmsubs", new fmsubs_translator());  
  private static OpCodeLookUp fmadds_lookup = new OpCodeLookUp(A_FORM, "fmadds", new fmadds_translator());  
  private static OpCodeLookUp fnmsubs_lookup = new OpCodeLookUp(A_FORM, "fnmsubs", new fnmsubs_translator());
  private static OpCodeLookUp fnmadds_lookup = new OpCodeLookUp(A_FORM, "fnmadds", new fnmadds_translator()); 

  protected OpCodeLookUp getOpCodeLookUp(int primaryOpcode, int secondaryOpcode) {
	 switch(secondaryOpcode){
	 case 18: return fdivs_lookup;   
	 case 20: return fsubs_lookup;  
	 case 21: return fadds_lookup;
	 case 22: return fsqrts_lookup;
	 case 24: return fres_lookup;
	 default: break;
	 }
	 switch(secondaryOpcode & 0x1f) {
	 case 25: return fmuls_lookup;  
	 case 28: return fmsubs_lookup;  
	 case 29: return fmadds_lookup;
	 case 30: return fnmsubs_lookup;
	 case 31: return fnmadds_lookup; 
	 default: break;
	 }
	 return null;
  }
}

/**
 * The translator for the EXTENDED 63 instructions
 */
class extended63_translator extends PPC_InstructionTranslator {
  // The translators for the X_Form instructions
  private static OpCodeLookUp fcmpu_lookup = new OpCodeLookUp(X_FORM, "fcmpu", new fcmpu_translator());
  private static OpCodeLookUp frsp_lookup = new OpCodeLookUp(X_FORM, "frsp", new frsp_translator());
  private static OpCodeLookUp fctiw_lookup = new OpCodeLookUp(X_FORM, "fctiw", new fctiw_translator());  
  private static OpCodeLookUp fctiwz_lookup = new OpCodeLookUp(X_FORM, "fctiwz", new fctiwz_translator());  
  private static OpCodeLookUp fcmpo_lookup = new OpCodeLookUp(X_FORM, "fcmpo", new fcmpo_translator());
  private static OpCodeLookUp mtfsb1_lookup = new OpCodeLookUp(X_FORM, "mtfsb1", new mtfsb1_translator());
  private static OpCodeLookUp fneg_lookup = new OpCodeLookUp(X_FORM, "fneg", new fneg_translator());
  private static OpCodeLookUp mcrfs_lookup = new OpCodeLookUp(X_FORM, "mcrfs", new mcrfs_translator());
  private static OpCodeLookUp mtfsb0_lookup = new OpCodeLookUp(X_FORM, "mtfsb0", new mtfsb0_translator());
  private static OpCodeLookUp fmr_lookup = new OpCodeLookUp(X_FORM, "fmr", new fmr_translator());
  private static OpCodeLookUp mtfsfi_lookup = new OpCodeLookUp(X_FORM, "mtfsfi", new mtfsfi_translator());
  private static OpCodeLookUp fnabs_lookup = new OpCodeLookUp(X_FORM, "fnabs", new fnabs_translator());
  private static OpCodeLookUp fabs_lookup = new OpCodeLookUp(X_FORM, "fabs", new fabs_translator());
  private static OpCodeLookUp mffs_lookup = new OpCodeLookUp(X_FORM, "mffs", new mffs_translator());
  private static OpCodeLookUp fctid_lookup = new OpCodeLookUp(X_FORM, "fctid", new fctid_translator());
  private static OpCodeLookUp fctidz_lookup = new OpCodeLookUp(X_FORM, "fctidz", new fctidz_translator());
  private static OpCodeLookUp fcfid_lookup = new OpCodeLookUp(X_FORM, "fcfid", new fcfid_translator());

  // The translators for the XFL_Form instructions
  private static OpCodeLookUp mtfsf_lookup = new OpCodeLookUp(XFL_FORM, "mtfsf", new mtfsf_translator());

  // The translators for the A_Form instructions
  private static OpCodeLookUp fdiv_lookup = new OpCodeLookUp(A_FORM, "fdiv", new fdiv_translator());  
  private static OpCodeLookUp fsub_lookup = new OpCodeLookUp(A_FORM, "fsub", new fsub_translator());  
  private static OpCodeLookUp fadd_lookup = new OpCodeLookUp(A_FORM, "fadd", new fadd_translator());  
  private static OpCodeLookUp fmul_lookup = new OpCodeLookUp(A_FORM, "fmul", new fmul_translator()); 
  private static OpCodeLookUp fmsub_lookup = new OpCodeLookUp(A_FORM, "fmsub", new fmsub_translator());  
  private static OpCodeLookUp fmadd_lookup = new OpCodeLookUp(A_FORM, "fmadd", new fmadd_translator());  
  private static OpCodeLookUp fnmsub_lookup = new OpCodeLookUp(A_FORM, "fnmsub", new fnmsub_translator());  
  private static OpCodeLookUp fnmadd_lookup = new OpCodeLookUp(A_FORM, "fnmadd", new fnmadd_translator());  

  protected OpCodeLookUp getOpCodeLookUp(int primaryOpcode, int secondaryOpcode) {
	 switch(secondaryOpcode){
	 case 0: return fcmpu_lookup;
	 case 12: return frsp_lookup;
	 case 14: return fctiw_lookup;  
	 case 15: return fctiwz_lookup;
	 case 18: return fdiv_lookup;
	 case 20: return fsub_lookup;
	 case 21: return fadd_lookup;
	 case 22: throw new Error("fsqrt!");
	 case 32: return fcmpo_lookup;
	 case 38: return mtfsb1_lookup;
	 case 40: return fneg_lookup;
	 case 64: return mcrfs_lookup;
	 case 70: return mtfsb0_lookup;
	 case 72: return fmr_lookup;
	 case 134: return mtfsfi_lookup;
	 case 136: return fnabs_lookup;
	 case 264: return fabs_lookup;
	 case 583: return mffs_lookup;
	 case 711: return mtfsf_lookup;
	 case 814: return fctid_lookup;
	 case 815: return fctidz_lookup;
	 case 846: return fcfid_lookup;
	 default: break;
	 }
	 switch(secondaryOpcode & 0x1f) {
	 case 23: throw new Error("fsel!");
	 case 25: return fmul_lookup;
	 case 28: return fmsub_lookup;
	 case 29: return fmadd_lookup;
	 case 30: return fnmsub_lookup;
	 case 31: return fnmadd_lookup;
	 default: break;
	 }
	 return null;
  }
}

// -oO Start of D form instructions Oo-

/**
 * The translator for the tdi instruction
 */
class tdi_translator extends PPC_InstructionTranslator {}
/**
 * The translator for the twi instruction
 */
class twi_translator extends PPC_InstructionTranslator {} 
/**
 * The translator for the mulli instruction
 */
class mulli_translator extends PPC_InstructionTranslator {
  /**
   * Translate mulli (multiply low immediate):
   * prod[0-48] <- (rA) * SIMM
   * (rD) <- prod[16-48]
   */
  protected int translateD_FORM(PPC2IR ppc2ir, PPC2IR.Laziness lazy, int pc, int inst, int opcode, int rD, int rA, int simm) {
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_MUL, ppc2ir.getGPRegister(rD),
							 ppc2ir.getGPRegister(rA),
							 new OPT_IntConstantOperand(EXTS(simm,16))));
    return pc+4;
  }
}
/**
 * The translator for the subfic instruction
 */
class subfic_translator extends PPC_InstructionTranslator {
  /**
   * Translate subfic (subtract from immediate carrying):
   * <listing>
   * rD <-  (rA) + EXTS(SIMM) + 1
   * </listing>
   * Also modifies XER[CA]. Equivalent to:
   * <listing>
   * rD <- EXTS(SIMM) - (rA)
   * </listing>
   */
  protected int translateD_FORM(PPC2IR ppc2ir, PPC2IR.Laziness lazy, int pc, int inst, int opcode, int rD, int rA, int imm) {
    // Get rA & simm
    OPT_RegisterOperand reg_rA = ppc2ir.getGPRegister(rA);
    int simm = EXTS(imm,16);

    // Copy to a long to set XER CA bit
    OPT_RegisterOperand xer_CA = ppc2ir.getXER_CA_Register();
    ppc2ir.appendInstructionToCurrentBlock(Unary.create(INT_2LONG, xer_CA, reg_rA));
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(LONG_AND, xer_CA.copyRO(),
																			xer_CA.copyRO(),
																			new OPT_LongConstantOperand(0xffffffffl)));
    // Calculate XER CA bit
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(LONG_SUB, xer_CA.copyRO(),
																			new OPT_LongConstantOperand(((long)simm) & 0xffffffffl),
																			xer_CA.copyRO()));
    ppc2ir.appendInstructionToCurrentBlock(Unary.create(INT_NOT, xer_CA.copyRO(), xer_CA.copyRO()));
    // Perform subtract for rD
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_SUB, ppc2ir.getGPRegister(rD),
																			new OPT_IntConstantOperand(simm),
																			reg_rA.copyRO()));
    return pc+4;
  }
}
/**
 * The translator for the cmpli instruction
 */
class cmpli_translator extends PPC_InstructionTranslator {
  /**
   * Translate cmpli (compare logical immediate):
   * <listing>
   * a <- (rA)
   * if a <U ((16)0 || UIMM)
   *    then c <- 0b100
   *    else if a >U ((16)0 || UIMM)
   *            then c <- 0b010
   *            else c <- 0b001
   * CR[(4 * crfD)-(4 * crfD + 3)] <- c || XER[SO]
   * </listing>
   */
  protected int translateD_FORM(PPC2IR ppc2ir, PPC2IR.Laziness lazy, int pc, int inst, int opcode, int crfD0L, int rA, int uimm) {
    int crfD = crfD0L >>> 2;

    setCRfield(ppc2ir, lazy, crfD, ppc2ir.getGPRegister(rA), new OPT_IntConstantOperand(uimm), PPC2IR.Laziness.UNSIGNED_INT_CMP);

    return pc+4;
  }
}
/**
 * The translator for the cmpi instruction
 */
class cmpi_translator extends PPC_InstructionTranslator {
  /**
   * Translate cmpi (compare immediate):
   * <listing>
   * a <- (rA)
   * if a < EXTS(SIMM)
   *    then c <- 0b100
   *    else if a > EXTS(SIMM)
   *            then c <- 0b010
   *            else c <- 0b001
   * CR[(4 * crfD)-(4 * crfD + 3)] <- c || XER[SO]
   * </listing>
   */
  protected int translateD_FORM(PPC2IR ppc2ir, PPC2IR.Laziness lazy, int pc, int inst, int opcode, int crfD0L, int rA, int imm) {
    int crfD = crfD0L >>> 2;

    setCRfield(ppc2ir, lazy, crfD, ppc2ir.getGPRegister(rA), new OPT_IntConstantOperand(EXTS(imm, 16)), PPC2IR.Laziness.SIGNED_INT_CMP);

    return pc+4;
  }
}
/**
 * The translator for the addic instruction
 */
class addic_translator extends PPC_InstructionTranslator {
  /**
   * Translate addic (add immediate carrying)
   * <listing>
   * rD <- (rA) + EXTS(SIMM)
   * </listing>
   * Other simplified mnemonics: subic
   */
  protected int translateD_FORM(PPC2IR ppc2ir, PPC2IR.Laziness lazy, int pc, int inst, int opcode, int rD, int rA, int imm) {
    // Get rA & simm
    OPT_RegisterOperand reg_rA = ppc2ir.getGPRegister(rA);
    int simm = EXTS(imm,16);

    // Copy to a long to set XER CA bit
    OPT_RegisterOperand xer_CA = ppc2ir.getXER_CA_Register();
    ppc2ir.appendInstructionToCurrentBlock(Unary.create(INT_2LONG, xer_CA, reg_rA));
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(LONG_AND, xer_CA.copyRO(),
																			xer_CA.copyRO(),
																			new OPT_LongConstantOperand(0xffffffffl)));
    // Calculate XER CA bit
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(LONG_ADD, xer_CA.copyRO(),
																			new OPT_LongConstantOperand(((long)simm) & 0xffffffffl),
																			xer_CA.copyRO()));
    // Perform add for rD
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_ADD, ppc2ir.getGPRegister(rD),
							 new OPT_IntConstantOperand(simm),
							 reg_rA.copyRO()));
    return pc+4;
  }
}
/**
 * The translator for the addic. instruction
 */
class addic__translator extends PPC_InstructionTranslator {
  /**
   * Translate addic. (add immediate carrying and record)
   * <listing>
   * rD <- (rA) + EXTS(SIMM)
   * </listing>
   * Other simplified mnemonics: subfic.
   */
  protected int translateD_FORM(PPC2IR ppc2ir, PPC2IR.Laziness lazy, int pc, int inst, int opcode, int rD, int rA, int imm) {
    // Get rA & simm
    OPT_RegisterOperand reg_rA = ppc2ir.getGPRegister(rA);
    int simm = EXTS(imm,16);

    // Copy to a long to set XER CA bit
    OPT_RegisterOperand xer_CA = ppc2ir.getXER_CA_Register();
    ppc2ir.appendInstructionToCurrentBlock(Unary.create(INT_2LONG, xer_CA, reg_rA));
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(LONG_AND, xer_CA.copyRO(),
																			xer_CA.copyRO(),
																			new OPT_LongConstantOperand(0xffffffffl)));
    // Calculate XER CA bit
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(LONG_ADD, xer_CA.copyRO(),
																			new OPT_LongConstantOperand(((long)simm) & 0xffffffffl),
																			xer_CA.copyRO()));
    // Perform add for rD
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_ADD, ppc2ir.getGPRegister(rD),
							 new OPT_IntConstantOperand(simm),
							 reg_rA.copyRO()));
    // Make copies of operands for lazy state
    setCRfield(ppc2ir, lazy, 0, ppc2ir.getGPRegister(rD), new OPT_IntConstantOperand(0), PPC2IR.Laziness.SIGNED_INT_CMP);

    return pc+4;
  }
}
/**
 * The translator for the addi instruction
 */
class addi_translator extends PPC_InstructionTranslator {
  /**
   * Translate addi (add immediate)
   * <listing>
   * if rA = 0 then rD <- EXTS(SIMM)
   * else rD <- (rA) + EXTS(SIMM)
   * </listing>
   * Other simplified mnemonics: li, la, subi
   */
  protected int translateD_FORM(PPC2IR ppc2ir, PPC2IR.Laziness lazy, int pc, int inst, int opcode, int rD, int rA, int imm) {
    int simm = EXTS(imm, 16);

    if (rA == 0) {
      ppc2ir.appendInstructionToCurrentBlock(Move.create(INT_MOVE,
							 ppc2ir.getGPRegister(rD),
							 new OPT_IntConstantOperand(simm)));        
    }
    else {
      ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_ADD, ppc2ir.getGPRegister(rD),
																			  ppc2ir.getGPRegister(rA),
																			  new OPT_IntConstantOperand(simm)));
    }
    return pc+4;
  }
}
/**
 * The translator for the addis instruction
 */
class addis_translator extends PPC_InstructionTranslator {
  /**
   * Translate addis (add immediate shifted)
   * <listing>
   * if rA = 0 then rD <- EXTS(SIMM || (16)0)
   * else rD <- (rA) + EXTS(SIMM || (16)0)
   * </listing>
   * Other simplified mnemonics: lis, subis
   */
  protected int translateD_FORM(PPC2IR ppc2ir, PPC2IR.Laziness lazy, int pc, int inst, int opcode, int rD, int rA, int imm) {
    int simm = EXTS(imm, 16);

    if (rA == 0) {
      ppc2ir.appendInstructionToCurrentBlock(Move.create(INT_MOVE,
							 ppc2ir.getGPRegister(rD),
							 new OPT_IntConstantOperand(simm << 16)));
    }
    else {
      ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_ADD, ppc2ir.getGPRegister(rD),
							   ppc2ir.getGPRegister(rA),
							   new OPT_IntConstantOperand(simm << 16)));
    }
    return pc+4;
  }
}

/**
 * The translator for the ori instruction
 */
class ori_translator extends PPC_InstructionTranslator {
  /**
   * Translate ori (or immediate)
   * <listing>
   * rA <- (rS) | ((16)0 || UIMM)
   * </listing>
   * Other simplified mnemonics: nop
   */
  protected int translateD_FORM(PPC2IR ppc2ir, PPC2IR.Laziness lazy, int pc, int inst, int opcode, int rS, int rA, int uimm) {
    if((uimm == 0) && (rA == rS)) {
      // Don't translate nop cases
    }
    else {
      ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_OR, ppc2ir.getGPRegister(rA),
							   ppc2ir.getGPRegister(rS),
							   new OPT_IntConstantOperand(uimm)));
    }
    return pc+4;
  }
}

/**
 * The translator for the oris instruction
 */
class oris_translator extends PPC_InstructionTranslator {
  /**
   * Translate oris (or immediate shifted)
   * <listing>
   * rA <- (rS) | (UIMM || (16)0)
   * </listing>
   */
  protected int translateD_FORM(PPC2IR ppc2ir, PPC2IR.Laziness lazy, int pc, int inst, int opcode, int rS, int rA, int uimm) {
    if((uimm == 0) && (rA == rS)) {
      // Don't translate nop case
    }
    else {
      ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_OR, ppc2ir.getGPRegister(rA),
							   ppc2ir.getGPRegister(rS),
							   new OPT_IntConstantOperand(uimm << 16)));
    }
    return pc+4;
  }
}
/**
 * The translator for the xori instruction
 */
class xori_translator extends PPC_InstructionTranslator {
  /**
   * Translate xori (xor immediate)
   * <listing>
   * rA <- (rS) (+) ((16)0 || UIMM)
   * </listing>
   */
  protected int translateD_FORM(PPC2IR ppc2ir, PPC2IR.Laziness lazy, int pc, int inst, int opcode, int rS, int rA, int uimm) {
    if((uimm == 0) && (rA == rS)) {
      // Don't translate nop case
    }
    else {
      ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_XOR, ppc2ir.getGPRegister(rA),
							   ppc2ir.getGPRegister(rS),
							   new OPT_IntConstantOperand(uimm)));
    }
    return pc+4;
  }
}
/**
 * The translator for the xoris instruction
 */
class xoris_translator extends PPC_InstructionTranslator {
  /**
   * Translate xoris (xor immediate shifted)
   * <listing>
   * rA <- (rS) (+) (UIMM || (16)0)
   * </listing>
   */
  protected int translateD_FORM(PPC2IR ppc2ir, PPC2IR.Laziness lazy, int pc, int inst, int opcode, int rS, int rA, int uimm) {
    if((uimm == 0) && (rA == rS)) {
      // Don't translate nop case
    }
    else {
      ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_XOR, ppc2ir.getGPRegister(rA),
							   ppc2ir.getGPRegister(rS),
							   new OPT_IntConstantOperand(uimm << 16)));
    }
    return pc+4;
  }
}
/**
 * The translator for the andi. instruction
 */
class andi__translator extends PPC_InstructionTranslator {
  /**
   * Translate andi. (and immediate (and record))
   * <listing>
   * rA <- (rS) & ((16)0 || UIMM)
   * </listing>
   */
  protected int translateD_FORM(PPC2IR ppc2ir, PPC2IR.Laziness lazy, int pc, int inst, int opcode, int rS, int rA, int uimm) {
    OPT_RegisterOperand result =  ppc2ir.getGPRegister(rA);
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_AND, result,
							 ppc2ir.getGPRegister(rS),
							 new OPT_IntConstantOperand(uimm)));
    // Make copies of operands for lazy state
    setCRfield(ppc2ir, lazy, 0, ppc2ir.getGPRegister(rA), new OPT_IntConstantOperand(0), PPC2IR.Laziness.SIGNED_INT_CMP);
    return pc+4;
  }
}

/**
 * The translator for the andis. instruction
 */
class andis__translator extends PPC_InstructionTranslator {
  /**
   * Translate andis. (and immediate shifted (and record))
   * <listing>
   * rA <- (rS) & (UIMM || (16)0)
   * </listing>
   */
  protected int translateD_FORM(PPC2IR ppc2ir, PPC2IR.Laziness lazy, int pc, int inst, int opcode, int rS, int rA, int uimm) {
    OPT_RegisterOperand result =  ppc2ir.getGPRegister(rA);
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_AND, result,
							 ppc2ir.getGPRegister(rS),
							 new OPT_IntConstantOperand(uimm << 16)));
    // Make copies of operands for lazy state
    setCRfield(ppc2ir, lazy, 0, ppc2ir.getGPRegister(rA), new OPT_IntConstantOperand(0), PPC2IR.Laziness.SIGNED_INT_CMP);
    return pc+4;
  }
}

/**
 * The translator for the lwz instruction
 */
class lwz_translator extends PPC_InstructionTranslator {
  /**
   * Translate lwz (load word and zero)
   * <listing>
   * if rA = 0 then b <- 0
   * else b <- (rA)
   * EA <- b+ EXTS(d)
   * rD <- MEM(EA, 4)
   * </listing>
   */
  protected int translateD_FORM(PPC2IR ppc2ir, PPC2IR.Laziness lazy, int pc, int inst, int opcode, int rD, int rA, int d) {
    OPT_RegisterOperand EA = ppc2ir.getTempInt(PPC2IR.TEMP_INT_SAFE_FROM_LOAD_CLOBBER);
    d = EXTS(d,16);
    if (rA == 0) {
      // EA = EXTS(d)
      ppc2ir.appendInstructionToCurrentBlock(Move.create(INT_MOVE,
							 EA,
							 new OPT_IntConstantOperand(d)));
    }
    else {
      // EA = rA + EXTS(d)
      ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_ADD, EA,
							   ppc2ir.getGPRegister(rA),
							   new OPT_IntConstantOperand(d)));
    }
    // Load value from memory at EA into rD.
    ppc2ir.memoryLoad(ppc2ir.getGPRegister(rD), EA.copyRO());
    return pc+4;
  }
}
/**
 * The translator for the lwzu instruction
 */
class lwzu_translator extends PPC_InstructionTranslator {
  /**
   * Translate lwzu (load word and zero with update)
   * <listing>
   * EA <- rA + EXTS(d)
   * rD <- MEM(EA, 4)
   * rA <- EA
   * </listing>
   */
  protected int translateD_FORM(PPC2IR ppc2ir, PPC2IR.Laziness lazy, int pc, int inst, int opcode, int rD, int rA, int d) {
    OPT_RegisterOperand EA = ppc2ir.getTempInt(PPC2IR.TEMP_INT_SAFE_FROM_LOAD_CLOBBER);
    d = EXTS(d,16);
    // EA = rA + EXTS(d)
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_ADD, EA,
							 ppc2ir.getGPRegister(rA),
							 new OPT_IntConstantOperand(d)));
    // Load value from memory at EA into rD.
    ppc2ir.memoryLoad(ppc2ir.getGPRegister(rD), EA.copyRO());
    // Place EA into rA
    ppc2ir.appendInstructionToCurrentBlock(Move.create(INT_MOVE, ppc2ir.getGPRegister(rA), 
																		 EA.copyRO()));
    return pc+4;
  }
}
/**
 * The translator for the lbz instruction
 */
class lbz_translator extends PPC_InstructionTranslator {
  /**
   * Translate lbz (load byte and zero)
   * <listing>
   * if rA = 0 then b <- 0
   * else b <- (rA)
   * EA <- b+ EXTS(d)
   * rD <- (24)0 || MEM(EA, 1)
   * </listing>
   */
  protected int translateD_FORM(PPC2IR ppc2ir, PPC2IR.Laziness lazy, int pc, int inst, int opcode, int rD, int rA, int d) {
    OPT_RegisterOperand EA = ppc2ir.getTempInt(PPC2IR.TEMP_INT_SAFE_FROM_LOAD_CLOBBER);
    d = EXTS(d,16);
    if (rA == 0) {
      // EA = EXTS(d)
      ppc2ir.appendInstructionToCurrentBlock(Move.create(INT_MOVE,
																			EA,
																			new OPT_IntConstantOperand(d)));
    }
    else {
      // EA = rA + EXTS(d)
      ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_ADD, EA,
							   ppc2ir.getGPRegister(rA),
							   new OPT_IntConstantOperand(d)));
    }
    // Load value from memory at EA into rD.
    OPT_RegisterOperand result = ppc2ir.getGPRegister(rD);
    ppc2ir.memoryLoad(result, EA.copyRO());
    // Turn 32bit address into appropriate shift using last 2 bits
    // of the address
    //-#if RVM_FOR_POWERPC
    // EA = (EA & 0x3) * 8
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_AND, EA.copyRO(),
																			EA.copyRO(),
																			new OPT_IntConstantOperand(3)));
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_SHL, EA.copyRO(),
																			EA.copyRO(),
																			new OPT_IntConstantOperand(3)));
	 //-#else
    // EA = (3 - (EA & 0x3)) * 8
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_AND, EA.copyRO(),
																			EA.copyRO(),
																			new OPT_IntConstantOperand(3)));
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_SUB, EA.copyRO(),
																			new OPT_IntConstantOperand(3),
																			EA.copyRO()));
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_SHL, EA.copyRO(),
																			EA.copyRO(),
																			new OPT_IntConstantOperand(3)));
	 //-#endif
    // rD >>>= EA
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_USHR, result.copyRO(),
																			result.copyRO(),
																			EA.copyRO()));
    // rD &= 0xff
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_AND, result.copyRO(),
																			result.copyRO(),
																			new OPT_IntConstantOperand(0xff)));
    return pc+4;
  }
}
/**
 * The translator for the lbzu instruction
 */
class lbzu_translator extends PPC_InstructionTranslator {
  /**
   * Translate lbzu (load byte and zero with update)
   * <listing>
   * EA <- rA + EXTS(d)
   * rD <- (24)0 || MEM(EA, 1)
   * rA <- EA
   * </listing>
   */
  protected int translateD_FORM(PPC2IR ppc2ir, PPC2IR.Laziness lazy, int pc, int inst, int opcode, int rD, int rA, int d) {
    OPT_RegisterOperand EA = ppc2ir.getTempInt(PPC2IR.TEMP_INT_SAFE_FROM_LOAD_CLOBBER);
    d = EXTS(d,16);
    // EA = rA + EXTS(d)
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_ADD, EA,
																			ppc2ir.getGPRegister(rA),
																			new OPT_IntConstantOperand(d)));
    // Load value from memory at EA into rD.
    OPT_RegisterOperand result = ppc2ir.getGPRegister(rD);
    ppc2ir.memoryLoad(result, EA.copyRO());
    // Place EA into rA
    ppc2ir.appendInstructionToCurrentBlock(Move.create(INT_MOVE, ppc2ir.getGPRegister(rA), 
																		 EA.copyRO()));
    // Turn 32bit address into appropriate shift using last 2 bits
    // of the address
    //-#if RVM_FOR_POWERPC
    // EA = (EA & 0x3) * 8
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_AND, EA.copyRO(),
																			EA.copyRO(),
																			new OPT_IntConstantOperand(3)));
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_SHL, EA.copyRO(),
																			EA.copyRO(),
																			new OPT_IntConstantOperand(3)));
	 //-#else
    // EA = (3 - (EA & 0x3)) * 8
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_AND, EA.copyRO(),
																			EA.copyRO(),
																			new OPT_IntConstantOperand(0x3)));
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_SUB, EA.copyRO(),
																			new OPT_IntConstantOperand(3),
																			EA.copyRO()));
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_SHL, EA.copyRO(),
																			EA.copyRO(),
																			new OPT_IntConstantOperand(3)));
	 //-#endif
    // rD >>>= EA
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_USHR, result.copyRO(),
																			result.copyRO(),
																			EA.copyRO()));
    // rD &= 0xff
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_AND, result.copyRO(),
																			result.copyRO(),
																			new OPT_IntConstantOperand(0xff)));
    return pc+4;
  }
}
/**
 * The translator for the stw instruction
 */
class stw_translator extends PPC_InstructionTranslator {
  /**
   * Translate stw (store word)
   * <listing>
   * if rA = 0 then b <- 0
   * else b <- (rA)
   * EA <- b+ EXTS(d)
   * MEM(EA, 4) <- rS
   * </listing>
   */
  protected int translateD_FORM(PPC2IR ppc2ir, PPC2IR.Laziness lazy, int pc, int inst, int opcode, int rS, int rA, int d) {
    OPT_RegisterOperand EA = ppc2ir.getTempInt(PPC2IR.TEMP_INT_SAFE_FROM_STORE_CLOBBER);
    d = EXTS(d,16);
    if (rA == 0) {
      // EA = EXTS(d)
      ppc2ir.appendInstructionToCurrentBlock(Move.create(INT_MOVE,
																			EA,
																			new OPT_IntConstantOperand(d)));
    }
    else {
      // EA = rA + EXTS(d)
      ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_ADD, EA,
																			  ppc2ir.getGPRegister(rA),
																			  new OPT_IntConstantOperand(d)));
    }
    // Store value from rS into memory at address EA
    ppc2ir.memoryStore(ppc2ir.getGPRegister(rS), EA.copyRO());
    return pc+4;
  }
}
/**
 * The translator for the stwu instruction
 */
class stwu_translator extends PPC_InstructionTranslator {
  /**
   * Translate stwu (store word with update)
   * <listing>
   * EA <- rA + EXTS(d)
   * MEM(EA, 4) <- rS
   * rA <- EA
   * </listing>
   */
  protected int translateD_FORM(PPC2IR ppc2ir, PPC2IR.Laziness lazy, int pc, int inst, int opcode, int rS, int rA, int d) {
    OPT_RegisterOperand EA = ppc2ir.getTempInt(PPC2IR.TEMP_INT_SAFE_FROM_STORE_CLOBBER);
    d = EXTS(d,16);
    // EA = rA + EXTS(d)
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_ADD, EA,
							 ppc2ir.getGPRegister(rA),
							 new OPT_IntConstantOperand(d)));
    // Load value from memory at EA into rD.
    ppc2ir.memoryStore(ppc2ir.getGPRegister(rS), EA.copyRO());
    // Place EA into rA
    ppc2ir.appendInstructionToCurrentBlock(Move.create(INT_MOVE, ppc2ir.getGPRegister(rA), 
						       EA.copyRO()));
    return pc+4;
  }
}
/**
 * The translator for the stb instruction
 */
class stb_translator extends PPC_InstructionTranslator {
  /**
   * Translate stb (store byte)
   * <listing>
   * if rA = 0 then b <- 0
   * else b <- (rA)
   * EA <- b+ EXTS(d)
   * MEM(EA, 1) <- rS[24-31]
   * </listing>
   */
  protected int translateD_FORM(PPC2IR ppc2ir, PPC2IR.Laziness lazy, int pc, int inst, int opcode, int rS, int rA, int d) {
    OPT_RegisterOperand EA = ppc2ir.getTempInt(PPC2IR.TEMP_INT_SAFE_FROM_STORE_BYTE_CLOBBER);
    d = EXTS(d,16);
    if (rA == 0) {
      // EA = EXTS(d)
      ppc2ir.appendInstructionToCurrentBlock(Move.create(INT_MOVE,
																			EA,
																			new OPT_IntConstantOperand(d)));
    }
    else {
      // EA = rA + EXTS(d)
      ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_ADD, EA,
																			  ppc2ir.getGPRegister(rA),
																			  new OPT_IntConstantOperand(d)));
    }
    // Store value from rS into memory at address EA
    ppc2ir.memoryStoreByte(ppc2ir.getGPRegister(rS), EA.copyRO());
    return pc+4;
  }
}

/**
 * The translator for the stbu instruction
 */
class stbu_translator extends PPC_InstructionTranslator {
  /**
   * Translate stbu (store byte with update)
   * <listing>
   * EA <- rA + EXTS(d)
   * MEM(EA, 1) <- rS[24-31]
   * rA <- EA
   * </listing>
   */
  protected int translateD_FORM(PPC2IR ppc2ir, PPC2IR.Laziness lazy, int pc, int inst, int opcode, int rS, int rA, int d) {
    OPT_RegisterOperand EA = ppc2ir.getTempInt(PPC2IR.TEMP_INT_SAFE_FROM_STORE_BYTE_CLOBBER);
    d = EXTS(d,16);

    if(rA == 0) {
      throw new Error("Invalid form of stbu at 0x" + Integer.toHexString(pc));
    }

    // EA = rA + EXTS(d)
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_ADD, EA,
																			ppc2ir.getGPRegister(rA),
																			new OPT_IntConstantOperand(d)));
    // Load value from memory at EA into rS.
    ppc2ir.memoryStoreByte(ppc2ir.getGPRegister(rS), EA.copyRO());
    // Place EA into rA
    ppc2ir.appendInstructionToCurrentBlock(Move.create(INT_MOVE, ppc2ir.getGPRegister(rA), 
																		 EA.copyRO()));
    return pc+4;
  }
}

/**
 * The translator for the lhz instruction
 */
class lhz_translator extends PPC_InstructionTranslator {
  /**
   * Translate lhz (load half word and zero)
   * <listing>
   * if rA = 0 then b <- 0
   * else b <- (rA)
   * EA <- b+ EXTS(d)
   * rD <- (16)0 || MEM(EA, 2)
   * </listing>
   */
  protected int translateD_FORM(PPC2IR ppc2ir, PPC2IR.Laziness lazy, int pc, int inst, int opcode, int rD, int rA, int d) {
    OPT_RegisterOperand EA = ppc2ir.getTempInt(PPC2IR.TEMP_INT_SAFE_FROM_LOAD_CLOBBER);
    d = EXTS(d,16);
    if (rA == 0) {
      // EA = EXTS(d)
      ppc2ir.appendInstructionToCurrentBlock(Move.create(INT_MOVE,
																			EA,
																			new OPT_IntConstantOperand(d)));
    }
    else {
      // EA = rA + EXTS(d)
      ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_ADD, EA,
																			  ppc2ir.getGPRegister(rA),
																			  new OPT_IntConstantOperand(d)));
    }
    // Load value from memory at EA into rD.
    OPT_RegisterOperand result = ppc2ir.getGPRegister(rD);
    ppc2ir.memoryLoad(result, EA.copyRO());
    // Turn 32bit address into appropriate shift using last 2 bits
    // of the address
    //-#if RVM_FOR_POWERPC
    // EA = (EA & 0x2) * 8
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_AND, EA.copyRO(),
							 EA.copyRO(),
							 new OPT_IntConstantOperand(0x2)));
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_SHL, EA.copyRO(),
							 EA.copyRO(),
							 new OPT_IntConstantOperand(3)));
	 //-#else
    // EA = (~EA & 0x2) * 8
    ppc2ir.appendInstructionToCurrentBlock(Unary.create(INT_NOT, EA.copyRO(),
																		  EA.copyRO()));
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_AND, EA.copyRO(),
																			EA.copyRO(),
																			new OPT_IntConstantOperand(0x2)));
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_SHL, EA.copyRO(),
																			EA.copyRO(),
																			new OPT_IntConstantOperand(3)));
	 //-#endif
    // rD >>>= EA
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_USHR, result.copyRO(),
																			result.copyRO(),
																			EA.copyRO()));
    // rD &= 0xffff
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_AND, result.copyRO(),
																			result.copyRO(),
																			new OPT_IntConstantOperand(0xffff)));
    return pc+4;
  }
}
/**
 * The translator for the lhzu instruction
 */
class lhzu_translator extends PPC_InstructionTranslator {
  /**
   * Translate lhz (load half word and zero with update)
   * <listing>
   * if rA = 0 then b <- 0
   * else b <- (rA)
   * EA <- b+ EXTS(d)
   * rD <- (16)0 || MEM(EA, 2)
   * </listing>
   */
  protected int translateD_FORM(PPC2IR ppc2ir, PPC2IR.Laziness lazy, int pc, int inst, int opcode, int rD, int rA, int d) {
    OPT_RegisterOperand EA = ppc2ir.getTempInt(PPC2IR.TEMP_INT_SAFE_FROM_LOAD_CLOBBER);
    d = EXTS(d,16);
    if (rA == 0) {
      // EA = EXTS(d)
      ppc2ir.appendInstructionToCurrentBlock(Move.create(INT_MOVE,
																			EA,
																			new OPT_IntConstantOperand(d)));
    }
    else {
      // EA = rA + EXTS(d)
      ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_ADD, EA,
																			  ppc2ir.getGPRegister(rA),
																			  new OPT_IntConstantOperand(d)));
    }
    // Load value from memory at EA into rD.
    OPT_RegisterOperand result = ppc2ir.getGPRegister(rD);
    ppc2ir.memoryLoad(result, EA.copyRO());
	 ppc2ir.appendInstructionToCurrentBlock(Move.create(INT_MOVE,
																		 ppc2ir.getGPRegister(rA),
																		 EA.copyRO()));
    // Turn 32bit address into appropriate shift using last 2 bits
    // of the address
    //-#if RVM_FOR_POWERPC
    // EA = (EA & 0x2) * 8
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_AND, EA.copyRO(),
							 EA.copyRO(),
							 new OPT_IntConstantOperand(0x2)));
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_SHL, EA.copyRO(),
							 EA.copyRO(),
							 new OPT_IntConstantOperand(3)));
	 //-#else
    // EA = (~EA & 0x2) * 8
    ppc2ir.appendInstructionToCurrentBlock(Unary.create(INT_NOT, EA.copyRO(),
																		  EA.copyRO()));
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_AND, EA.copyRO(),
																			EA.copyRO(),
																			new OPT_IntConstantOperand(0x2)));
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_SHL, EA.copyRO(),
																			EA.copyRO(),
																			new OPT_IntConstantOperand(3)));
	 //-#endif
    // rD >>>= EA
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_USHR, result.copyRO(),
																			result.copyRO(),
																			EA.copyRO()));
    // rD &= 0xffff
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_AND, result.copyRO(),
																			result.copyRO(),
																			new OPT_IntConstantOperand(0xffff)));
    return pc+4;
  }
}
/**
 * The translator for the lha instruction
 */
class lha_translator extends PPC_InstructionTranslator {
  /**
   * Translate lha (load half word algabraeic)
   * <listing>
   * if rA = 0 then b <- 0
   * else b <- (rA)
   * EA <- b+ EXTS(d)
   * rD <- EXTS(MEM(EA, 2))
   * </listing>
   */
  protected int translateD_FORM(PPC2IR ppc2ir, PPC2IR.Laziness lazy, int pc, int inst, int opcode, int rD, int rA, int d) {
    OPT_RegisterOperand EA = ppc2ir.getTempInt(PPC2IR.TEMP_INT_SAFE_FROM_LOAD_CLOBBER);
    d = EXTS(d,16);
    if (rA == 0) {
      // EA = EXTS(d)
      ppc2ir.appendInstructionToCurrentBlock(Move.create(INT_MOVE,
																			EA,
																			new OPT_IntConstantOperand(d)));
    }
    else {
      // EA = rA + EXTS(d)
      ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_ADD, EA,
																			  ppc2ir.getGPRegister(rA),
																			  new OPT_IntConstantOperand(d)));
    }
    // Load value from memory at EA into rD.
    OPT_RegisterOperand result = ppc2ir.getGPRegister(rD);
    ppc2ir.memoryLoad(result, EA.copyRO());
    // Turn 32bit address into appropriate shift using last 2 bits
    // of the address
    //-#if RVM_FOR_POWERPC
    // EA = (EA & 0x2) * 8
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_AND, EA.copyRO(),
							 EA.copyRO(),
							 new OPT_IntConstantOperand(0x2)));
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_SHL, EA.copyRO(),
							 EA.copyRO(),
							 new OPT_IntConstantOperand(3)));
	 //-#else
    // EA = (~EA & 0x2) * 8
    ppc2ir.appendInstructionToCurrentBlock(Unary.create(INT_NOT, EA.copyRO(),
																		  EA.copyRO()));
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_AND, EA.copyRO(),
																			EA.copyRO(),
																			new OPT_IntConstantOperand(0x2)));
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_SHL, EA.copyRO(),
																			EA.copyRO(),
																			new OPT_IntConstantOperand(3)));
	 //-#endif
    // rD >>>= EA
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_SHR, result.copyRO(),
																			result.copyRO(),
																			EA.copyRO()));
    // rD = (rD << 16) >> 16
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_SHL, result.copyRO(),
																			result.copyRO(),
																			new OPT_IntConstantOperand(16)));
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_SHR, result.copyRO(),
																			result.copyRO(),
																			new OPT_IntConstantOperand(16)));
    return pc+4;
  }
}
/**
 * The translator for the lhau instruction
 */
class lhau_translator extends PPC_InstructionTranslator {
  /**
   * Translate lhau (load half word algabraeic with update)
   * <listing>
   * EA <- (rA)+ EXTS(d)
   * rD <- EXTS(MEM(EA, 2))
	* rA <- EA
   * </listing>
   */
  protected int translateD_FORM(PPC2IR ppc2ir, PPC2IR.Laziness lazy, int pc, int inst, int opcode, int rD, int rA, int d) {
    OPT_RegisterOperand EA = ppc2ir.getTempInt(PPC2IR.TEMP_INT_SAFE_FROM_LOAD_CLOBBER);
    d = EXTS(d,16);
	 // EA = rA + EXTS(d)
	 ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_ADD, EA,
																			ppc2ir.getGPRegister(rA),
																			new OPT_IntConstantOperand(d)));
    // Load value from memory at EA into rD.
    OPT_RegisterOperand result = ppc2ir.getGPRegister(rD);
    ppc2ir.memoryLoad(result, EA.copyRO());
    // Turn 32bit address into appropriate shift using last 2 bits
    // of the address
    //-#if RVM_FOR_POWERPC
    // EA = (EA & 0x2) * 8
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_AND, EA.copyRO(),
							 EA.copyRO(),
							 new OPT_IntConstantOperand(0x2)));
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_SHL, EA.copyRO(),
							 EA.copyRO(),
							 new OPT_IntConstantOperand(3)));
	 //-#else
    // EA = (~EA & 0x2) * 8
    ppc2ir.appendInstructionToCurrentBlock(Unary.create(INT_NOT, EA.copyRO(),
																		  EA.copyRO()));
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_AND, EA.copyRO(),
																			EA.copyRO(),
																			new OPT_IntConstantOperand(0x2)));
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_SHL, EA.copyRO(),
																			EA.copyRO(),
																			new OPT_IntConstantOperand(3)));
	 //-#endif
    // rD >>>= EA
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_SHR, result.copyRO(),
																			result.copyRO(),
																			EA.copyRO()));
    // rD = (rD << 16) >> 16
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_SHL, result.copyRO(),
																			result.copyRO(),
																			new OPT_IntConstantOperand(16)));
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_SHR, result.copyRO(),
																			result.copyRO(),
																			new OPT_IntConstantOperand(16)));
    // Place EA into rA
    ppc2ir.appendInstructionToCurrentBlock(Move.create(INT_MOVE, ppc2ir.getGPRegister(rA), 
																		 EA.copyRO()));
    return pc+4;
  }
}
/**
 * The translator for the sth instruction
 */
class sth_translator extends PPC_InstructionTranslator {
  /**
   * Translate sth (store half word)
   * <listing>
   * if rA = 0 then b <- 0
   * else b <- (rA)
   * EA <- b+ EXTS(d)
   * MEM(EA, 2) <- rS[16-31]
   * </listing>
   */
  protected int translateD_FORM(PPC2IR ppc2ir, PPC2IR.Laziness lazy, int pc, int inst, int opcode, int rS, int rA, int d) {
    OPT_RegisterOperand EA = ppc2ir.getTempInt(PPC2IR.TEMP_INT_SAFE_FROM_STORE_HALF_WORD_CLOBBER);
    d = EXTS(d,16);
    if (rA == 0) {
      // EA = EXTS(d)
      ppc2ir.appendInstructionToCurrentBlock(Move.create(INT_MOVE,
																			EA,
																			new OPT_IntConstantOperand(d)));
    }
    else {
      // EA = rA + EXTS(d)
      ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_ADD, EA,
																			  ppc2ir.getGPRegister(rA),
																			  new OPT_IntConstantOperand(d)));
    }
    // Store value from rS into memory at address EA
    ppc2ir.memoryStoreHalfWord(ppc2ir.getGPRegister(rS), EA.copyRO());
    return pc+4;
  }
}

/**
 * The translator for the sthu instruction
 */
class sthu_translator extends PPC_InstructionTranslator {
  /**
   * Translate sthu (store half word with update)
   * <listing>
   * EA <- (rA) + EXTS(d)
   * MEM(EA, 2) <- rS[16-31]
	* rA <- EA
   * </listing>
   */
  protected int translateD_FORM(PPC2IR ppc2ir, PPC2IR.Laziness lazy, int pc, int inst, int opcode, int rS, int rA, int d) {
    OPT_RegisterOperand EA = ppc2ir.getTempInt(PPC2IR.TEMP_INT_SAFE_FROM_STORE_HALF_WORD_CLOBBER);
    d = EXTS(d,16);
    if (rA == 0) {
      throw new Error("Invalid form of sthu at 0x" + Integer.toHexString(pc));
    }
    else {
      // EA = rA + EXTS(d)
      ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_ADD, EA,
																			  ppc2ir.getGPRegister(rA),
																			  new OPT_IntConstantOperand(d)));
    }
    // Store value from rS into memory at address EA
    ppc2ir.memoryStoreHalfWord(ppc2ir.getGPRegister(rS), EA.copyRO());
    // Place EA into rA
    ppc2ir.appendInstructionToCurrentBlock(Move.create(INT_MOVE, ppc2ir.getGPRegister(rA), 
																		 EA.copyRO()));
    return pc+4;
  }
}
/**
 * The translator for the lmw instruction
 */
class lmw_translator extends PPC_InstructionTranslator {}
/**
 * The translator for the stmw instruction
 */
class stmw_translator extends PPC_InstructionTranslator {}
/**
 * The translator for the lfs instruction
 */
class lfs_translator extends PPC_InstructionTranslator {
  /**
   * Translate lfs (load floating-point single)
   * <listing>
   * if rA = 0 then b <- 0
   * else b <- (rA)
   * EA <- b+ EXTS(d)
   * frD <- DOUBLE(MEM(EA, 4))
   * </listing>
   */
  protected int translateD_FORM(PPC2IR ppc2ir, PPC2IR.Laziness lazy, int pc, int inst, int opcode, int frD, int rA, int d) {
    OPT_RegisterOperand EA = ppc2ir.getTempInt(PPC2IR.TEMP_INT_SAFE_FROM_LOAD_CLOBBER);
    d = EXTS(d,16);
    if (rA == 0) {
      // EA = EXTS(d)
      ppc2ir.appendInstructionToCurrentBlock(Move.create(INT_MOVE,
																			EA,
																			new OPT_IntConstantOperand(d)));
    }
    else {
      // EA = rA + EXTS(d)
      ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_ADD, EA,
																			  ppc2ir.getGPRegister(rA),
																			  new OPT_IntConstantOperand(d)));
    }
    // Load value from memory at EA into tempInt
    OPT_RegisterOperand tempInt = ppc2ir.getTempInt(PPC2IR.TEMP_INT_SAFE_FROM_LOAD_CLOBBER+1);
    ppc2ir.memoryLoad(tempInt, EA.copyRO());
    // Convert bit pattern to a float.
    OPT_RegisterOperand tempFloat = ppc2ir.getTempFloat(0);
    ppc2ir.appendInstructionToCurrentBlock(Unary.create(INT_BITS_AS_FLOAT, tempFloat,
																		  tempInt.copyRO()));

    // Now to a double, and put into register.
    ppc2ir.appendInstructionToCurrentBlock(Unary.create(ppc2ir.FLOAT_2DOUBLE, ppc2ir.getFPRegister(frD),
																		  tempFloat.copyRO()));
    return pc+4;
  }
}

/**
 * The translator for the lfsu instruction
 */
class lfsu_translator extends PPC_InstructionTranslator {
  /**
   * Translate lfsu (load floating-point single with update)
   * <listing>
   * EA <- (rA)+ EXTS(d)
   * frD <- DOUBLE(MEM(EA, 4))
	* rA <- EA
   * </listing>
   */
  protected int translateD_FORM(PPC2IR ppc2ir, PPC2IR.Laziness lazy, int pc, int inst, int opcode, int frD, int rA, int d) {
    OPT_RegisterOperand EA = ppc2ir.getTempInt(PPC2IR.TEMP_INT_SAFE_FROM_LOAD_CLOBBER);
    d = EXTS(d,16);
	 // EA = rA + EXTS(d)
	 ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_ADD, EA,
																			ppc2ir.getGPRegister(rA),
																			new OPT_IntConstantOperand(d)));
    // Load value from memory at EA into tempInt
    OPT_RegisterOperand tempInt = ppc2ir.getTempInt(PPC2IR.TEMP_INT_SAFE_FROM_LOAD_CLOBBER+1);
    ppc2ir.memoryLoad(tempInt, EA.copyRO());
    // Convert bit pattern to a float.
    OPT_RegisterOperand tempFloat = ppc2ir.getTempFloat(0);
    ppc2ir.appendInstructionToCurrentBlock(Unary.create(INT_BITS_AS_FLOAT, tempFloat,
																		  tempInt.copyRO()));

    // Now to a double, and put into register.
    ppc2ir.appendInstructionToCurrentBlock(Unary.create(ppc2ir.FLOAT_2DOUBLE, ppc2ir.getFPRegister(frD),
																		  tempFloat.copyRO()));
    // Place EA into rA
    ppc2ir.appendInstructionToCurrentBlock(Move.create(INT_MOVE, ppc2ir.getGPRegister(rA), 
																		 EA.copyRO()));
    return pc+4;
  }
}
/**
 * The translator for the lfd instruction
 */
class lfd_translator extends PPC_InstructionTranslator {
  /**
   * Translate lfd (load floating-point double)
   * <listing>
   * if rA = 0 then b <- 0
   * else b <- (rA)
   * EA <- b+ EXTS(d)
   * frD <- MEM(EA, 8)
   * </listing>
   */
  protected int translateD_FORM(PPC2IR ppc2ir, PPC2IR.Laziness lazy, int pc, int inst, int opcode, int frD, int rA, int d) {
    OPT_RegisterOperand EA = ppc2ir.getTempInt(PPC2IR.TEMP_INT_SAFE_FROM_LOAD_CLOBBER);
    OPT_RegisterOperand EA2 = ppc2ir.getTempInt(PPC2IR.TEMP_INT_SAFE_FROM_LOAD_CLOBBER+1);
    d = EXTS(d,16);
    if (rA == 0) {
      // EA = EXTS(d)
      ppc2ir.appendInstructionToCurrentBlock(Move.create(INT_MOVE,
																			EA,
																			new OPT_IntConstantOperand(d)));
      ppc2ir.appendInstructionToCurrentBlock(Move.create(INT_MOVE,
																			EA2,
																			new OPT_IntConstantOperand(d+4)));
    }
    else {
      // EA = rA + EXTS(d)
      ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_ADD, EA,
																			  ppc2ir.getGPRegister(rA),
																			  new OPT_IntConstantOperand(d)));
      ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_ADD, EA2,
																			  ppc2ir.getGPRegister(rA),
																			  new OPT_IntConstantOperand(d+4)));
    }
    // Load value from memory at EA & EA2 into tempInt & tempInt2 (msb at low address)
    OPT_RegisterOperand tempInt  = ppc2ir.getTempInt(PPC2IR.TEMP_INT_SAFE_FROM_LOAD_CLOBBER+2);
    OPT_RegisterOperand tempInt2 = ppc2ir.getTempInt(PPC2IR.TEMP_INT_SAFE_FROM_LOAD_CLOBBER+3);
    ppc2ir.memoryLoad(tempInt, EA.copyRO());
    ppc2ir.memoryLoad(tempInt2,EA2.copyRO());

    // Merge ints into a long
    // tempLong = ((long)tempInt << 32) | (long)tempInt2
    OPT_RegisterOperand tempLong = ppc2ir.getTempLong(0);
    OPT_RegisterOperand tempLong2 = ppc2ir.getTempLong(1);
    ppc2ir.appendInstructionToCurrentBlock(Unary.create(INT_2LONG, tempLong, tempInt.copyRO()));
    ppc2ir.appendInstructionToCurrentBlock(Unary.create(INT_2LONG, tempLong2, tempInt2.copyRO()));
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(LONG_SHL, tempLong.copyRO(), 
																			tempLong.copyRO(),
																			new OPT_IntConstantOperand(32)));
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(LONG_AND, tempLong2.copyRO(), 
																			tempLong2.copyRO(),
																			new OPT_LongConstantOperand(0xffffffffl)));
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(LONG_OR, tempLong.copyRO(), 
																			tempLong.copyRO(),
																			tempLong2.copyRO()));
    // Convert long to a double, and put into register.
    ppc2ir.appendInstructionToCurrentBlock(Unary.create(LONG_BITS_AS_DOUBLE, ppc2ir.getFPRegister(frD),
																		  tempLong.copyRO()));
    return pc+4;
  }
}

/**
 * The translator for the lfdu instruction
 */
class lfdu_translator extends PPC_InstructionTranslator {
  /**
   * Translate lfdu (load floating-point double with update)
   * <listing>
   * EA <- (rA)+ EXTS(d)
   * frD <- MEM(EA, 8)
	* rA <- EA
   * </listing>
   */
  protected int translateD_FORM(PPC2IR ppc2ir, PPC2IR.Laziness lazy, int pc, int inst, int opcode, int frD, int rA, int d) {
    OPT_RegisterOperand EA = ppc2ir.getTempInt(PPC2IR.TEMP_INT_SAFE_FROM_LOAD_CLOBBER);
    OPT_RegisterOperand EA2 = ppc2ir.getTempInt(PPC2IR.TEMP_INT_SAFE_FROM_LOAD_CLOBBER+1);
    d = EXTS(d,16);
	 // EA = rA + EXTS(d)
	 ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_ADD, EA,
																			ppc2ir.getGPRegister(rA),
																			new OPT_IntConstantOperand(d)));
	 ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_ADD, EA2,
																			ppc2ir.getGPRegister(rA),
																			new OPT_IntConstantOperand(d+4)));
    // Load value from memory at EA & EA2 into tempInt & tempInt2 (msb at low address)
    OPT_RegisterOperand tempInt  = ppc2ir.getTempInt(PPC2IR.TEMP_INT_SAFE_FROM_LOAD_CLOBBER+2);
    OPT_RegisterOperand tempInt2 = ppc2ir.getTempInt(PPC2IR.TEMP_INT_SAFE_FROM_LOAD_CLOBBER+3);
    ppc2ir.memoryLoad(tempInt, EA.copyRO());
    ppc2ir.memoryLoad(tempInt2,EA2.copyRO());

    // Merge ints into a long
    // tempLong = ((long)tempInt << 32) | (long)tempInt2
    OPT_RegisterOperand tempLong = ppc2ir.getTempLong(0);
    OPT_RegisterOperand tempLong2 = ppc2ir.getTempLong(1);
    ppc2ir.appendInstructionToCurrentBlock(Unary.create(INT_2LONG, tempLong, tempInt.copyRO()));
    ppc2ir.appendInstructionToCurrentBlock(Unary.create(INT_2LONG, tempLong2, tempInt2.copyRO()));
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(LONG_SHL, tempLong.copyRO(), 
																			tempLong.copyRO(),
																			new OPT_IntConstantOperand(32)));
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(LONG_AND, tempLong2.copyRO(), 
																			tempLong2.copyRO(),
																			new OPT_LongConstantOperand(0xffffffffl)));
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(LONG_OR, tempLong.copyRO(), 
																			tempLong.copyRO(),
																			tempLong2.copyRO()));
    // Convert long to a double, and put into register.
    ppc2ir.appendInstructionToCurrentBlock(Unary.create(LONG_BITS_AS_DOUBLE, ppc2ir.getFPRegister(frD),
																		  tempLong.copyRO()));
    // Place EA into rA
    ppc2ir.appendInstructionToCurrentBlock(Move.create(INT_MOVE, ppc2ir.getGPRegister(rA), 
																		 EA.copyRO()));
    return pc+4;
  }
}
/**
 * The translator for the stfs instruction
 */
class stfs_translator extends PPC_InstructionTranslator {
  /**
   * Translate stfs (store floating point single)
   * <listing>
   * if rA = 0 then b <- 0
   * else b <- (rA)
   * EA <- b+ EXTS(d)
   * MEM(EA, 4) <- SINGLE(frS)
   * </listing>
   */
  protected int translateD_FORM(PPC2IR ppc2ir, PPC2IR.Laziness lazy, int pc, int inst, int opcode, int frS, int rA, int d) {
    OPT_RegisterOperand EA = ppc2ir.getTempInt(PPC2IR.TEMP_INT_SAFE_FROM_STORE_CLOBBER);
    d = EXTS(d,16);
    if (rA == 0) {
      // EA = EXTS(d)
      ppc2ir.appendInstructionToCurrentBlock(Move.create(INT_MOVE,
																			EA,
																			new OPT_IntConstantOperand(d)));
    }
    else {
      // EA = rA + EXTS(d)
      ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_ADD, EA,
																			  ppc2ir.getGPRegister(rA),
																			  new OPT_IntConstantOperand(d)));
    }
    // tempInt = SINGLE(frS)
    OPT_RegisterOperand tempFloat = ppc2ir.getTempFloat(0);
    ppc2ir.appendInstructionToCurrentBlock(Unary.create(DOUBLE_2FLOAT, tempFloat,
																		  ppc2ir.getFPRegister(frS)));
    OPT_RegisterOperand tempInt = ppc2ir.getTempInt(PPC2IR.TEMP_INT_SAFE_FROM_STORE_CLOBBER+1);
    ppc2ir.appendInstructionToCurrentBlock(Unary.create(FLOAT_AS_INT_BITS, tempInt, 
																		  tempFloat.copyRO()));

    // Store value from rS into memory at address EA
    ppc2ir.memoryStore(tempInt.copyRO(), EA.copyRO());
    return pc+4;
  }
}
  
/**
 * The translator for the stfsu instruction
 */
class stfsu_translator extends PPC_InstructionTranslator {
  /**
   * Translate stfsu (store floating point single with update)
   * <listing>
   * EA <- (rA)+ EXTS(d)
   * MEM(EA, 4) <- SINGLE(frS)
	* rA <- EA
   * </listing>
   */
  protected int translateD_FORM(PPC2IR ppc2ir, PPC2IR.Laziness lazy, int pc, int inst, int opcode, int frS, int rA, int d) {
    OPT_RegisterOperand EA = ppc2ir.getTempInt(PPC2IR.TEMP_INT_SAFE_FROM_STORE_CLOBBER);
    d = EXTS(d,16);
    if (rA == 0) {
      // EA = EXTS(d)
      ppc2ir.appendInstructionToCurrentBlock(Move.create(INT_MOVE,
																			EA,
																			new OPT_IntConstantOperand(d)));
    }
    else {
      // EA = rA + EXTS(d)
      ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_ADD, EA,
																			  ppc2ir.getGPRegister(rA),
																			  new OPT_IntConstantOperand(d)));
    }
    // tempInt = SINGLE(frS)
    OPT_RegisterOperand tempFloat = ppc2ir.getTempFloat(0);
    ppc2ir.appendInstructionToCurrentBlock(Unary.create(DOUBLE_2FLOAT, tempFloat,
																		  ppc2ir.getFPRegister(frS)));
    OPT_RegisterOperand tempInt = ppc2ir.getTempInt(PPC2IR.TEMP_INT_SAFE_FROM_STORE_CLOBBER+1);
    ppc2ir.appendInstructionToCurrentBlock(Unary.create(FLOAT_AS_INT_BITS, tempInt, 
																		  tempFloat.copyRO()));

    // Store value from rS into memory at address EA
    ppc2ir.memoryStore(tempInt.copyRO(), EA.copyRO());
    // Place EA into rA
    ppc2ir.appendInstructionToCurrentBlock(Move.create(INT_MOVE, ppc2ir.getGPRegister(rA), 
																		 EA.copyRO()));
    return pc+4;
  }
}
/**
 * The translator for the stfd instruction
 */
class stfd_translator extends PPC_InstructionTranslator {
  /**
   * Translate stfd (store floating-point double)
   * <listing>
   * if rA = 0 then b <- 0
   * else b <- (rA)
   * EA <- b+ EXTS(d)
   * MEM(EA, 8) <- (frS)
   * </listing>
   */
  protected int translateD_FORM(PPC2IR ppc2ir, PPC2IR.Laziness lazy, int pc, int inst, int opcode, int frS, int rA, int d) {
    OPT_RegisterOperand EA = ppc2ir.getTempInt(PPC2IR.TEMP_INT_SAFE_FROM_STORE_CLOBBER);
    OPT_RegisterOperand EA2 = ppc2ir.getTempInt(PPC2IR.TEMP_INT_SAFE_FROM_STORE_CLOBBER+1);
    d = EXTS(d,16);
    if (rA == 0) {
      // EA = EXTS(d)
      ppc2ir.appendInstructionToCurrentBlock(Move.create(INT_MOVE,
																			EA,
																			new OPT_IntConstantOperand(d)));
      ppc2ir.appendInstructionToCurrentBlock(Move.create(INT_MOVE,
																			EA2,
																			new OPT_IntConstantOperand(d+4)));
    }
    else {
      // EA = rA + EXTS(d)
      ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_ADD, EA,
																			  ppc2ir.getGPRegister(rA),
																			  new OPT_IntConstantOperand(d)));
      ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_ADD, EA2,
																			  ppc2ir.getGPRegister(rA),
																			  new OPT_IntConstantOperand(d+4)));
    }
    // Split double into ints
	 // tempLong = DOUBLE_AS_LONG_BITS(frS)
	 // tempLong2 = tempLong >>> 32
	 // tempLong = tempLong & 0xffffffff
	 // tempInt = (int)tempLong
	 // tempInt2 = (int)tempLong2 <- msb
	 // mem[EA] = tempInt2
	 // mem[EA2] = tempInt
    OPT_RegisterOperand tempLong = ppc2ir.getTempLong(0);
    OPT_RegisterOperand tempLong2 = ppc2ir.getTempLong(1);
    OPT_RegisterOperand tempInt = ppc2ir.getTempInt(PPC2IR.TEMP_INT_SAFE_FROM_STORE_CLOBBER+2);
    OPT_RegisterOperand tempInt2 = ppc2ir.getTempInt(PPC2IR.TEMP_INT_SAFE_FROM_STORE_CLOBBER+3);
    ppc2ir.appendInstructionToCurrentBlock(Unary.create(DOUBLE_AS_LONG_BITS, tempLong,
																		  ppc2ir.getFPRegister(frS)));
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(LONG_USHR, tempLong2, tempLong.copyRO(),
																			new OPT_IntConstantOperand(32)));
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(LONG_AND, tempLong.copyRO(), tempLong.copyRO(),
																			new OPT_LongConstantOperand(0xffffffffl)));
    ppc2ir.appendInstructionToCurrentBlock(Unary.create(LONG_2INT, tempInt, tempLong.copyRO()));
    ppc2ir.appendInstructionToCurrentBlock(Unary.create(LONG_2INT, tempInt2, tempLong2.copyRO()));

    // Store value into memory at address EA (msb at low address)
    ppc2ir.memoryStore(tempInt2.copyRO(), EA.copyRO());
    ppc2ir.memoryStore(tempInt.copyRO(), EA2.copyRO());
    return pc+4;
  }
}

/**
 * The translator for the stfdu instruction
 */
class stfdu_translator extends PPC_InstructionTranslator {
  /**
   * Translate stfdu (store floating-point double with update)
   * <listing>
   * EA <- (rA)+ EXTS(d)
   * MEM(EA, 8) <- (frS)
	* rA <- EA
   * </listing>
   */
  protected int translateD_FORM(PPC2IR ppc2ir, PPC2IR.Laziness lazy, int pc, int inst, int opcode, int frS, int rA, int d) {
    OPT_RegisterOperand EA = ppc2ir.getTempInt(PPC2IR.TEMP_INT_SAFE_FROM_STORE_CLOBBER);
    OPT_RegisterOperand EA2 = ppc2ir.getTempInt(PPC2IR.TEMP_INT_SAFE_FROM_STORE_CLOBBER+1);
    d = EXTS(d,16);
	 // EA = rA + EXTS(d)
	 ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_ADD, EA,
																			ppc2ir.getGPRegister(rA),
																			new OPT_IntConstantOperand(d)));
	 ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_ADD, EA2,
																			ppc2ir.getGPRegister(rA),
																			new OPT_IntConstantOperand(d+4)));
    // Split double into ints
	 // tempLong = DOUBLE_AS_LONG_BITS(frS)
	 // tempLong2 = tempLong >>> 32
	 // tempLong = tempLong & 0xffffffff
	 // tempInt = (int)tempLong
	 // tempInt2 = (int)tempLong2 <- msb
	 // mem[EA] = tempInt2
	 // mem[EA2] = tempInt
    OPT_RegisterOperand tempLong = ppc2ir.getTempLong(0);
    OPT_RegisterOperand tempLong2 = ppc2ir.getTempLong(1);
    OPT_RegisterOperand tempInt = ppc2ir.getTempInt(PPC2IR.TEMP_INT_SAFE_FROM_STORE_CLOBBER+2);
    OPT_RegisterOperand tempInt2 = ppc2ir.getTempInt(PPC2IR.TEMP_INT_SAFE_FROM_STORE_CLOBBER+3);
    ppc2ir.appendInstructionToCurrentBlock(Unary.create(DOUBLE_AS_LONG_BITS, tempLong,
																		  ppc2ir.getFPRegister(frS)));
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(LONG_USHR, tempLong2, tempLong.copyRO(),
																			new OPT_IntConstantOperand(32)));
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(LONG_AND, tempLong.copyRO(), tempLong.copyRO(),
																			new OPT_LongConstantOperand(0xffffffffl)));
    ppc2ir.appendInstructionToCurrentBlock(Unary.create(LONG_2INT, tempInt, tempLong.copyRO()));
    ppc2ir.appendInstructionToCurrentBlock(Unary.create(LONG_2INT, tempInt2, tempLong2.copyRO()));

    // Store value into memory at address EA (msb at low address)
    ppc2ir.memoryStore(tempInt2.copyRO(), EA.copyRO());
    ppc2ir.memoryStore(tempInt.copyRO(), EA2.copyRO());
    // Place EA into rA
    ppc2ir.appendInstructionToCurrentBlock(Move.create(INT_MOVE, ppc2ir.getGPRegister(rA), 
																		 EA.copyRO()));
    return pc+4;
  }
}
  
// -oO Start of XL form instructions Oo-
/**
 * The translator for the mcrf instruction
 */
class mcrf_translator extends PPC_InstructionTranslator {
  /**
   * Translate mcrf (move condition register field)
   * <listing>
   * CR[(4 * crfD) through (4 * crfD + 3)] <- CR[(4 * crfS) through (4 * crfS + 3)]
   * </listing>
   */
  protected int translateXL_FORM(PPC2IR ppc2ir, PPC2IR.Laziness lazy, int pc, int inst, int opcode, int crfD00, int crfS00, int zero, int secondaryOpcode, int zero2){
    if (VM.VerifyAssertions) VM._assert((secondaryOpcode == 0) && (zero == 0) && (zero2 == 0));
    int crfD = crfD00 >> 2;
    int crfS = crfS00 >> 2;

    if (crfD != crfS) {
      switch(lazy.getConditionFieldLaziness(crfS)) {
      case PPC2IR.Laziness.SIGNED_INT_CMP:
      case PPC2IR.Laziness.UNSIGNED_INT_CMP:
		  // Copy operands for lazy state
		  ppc2ir.appendInstructionToCurrentBlock(Move.create(INT_MOVE,
																			  ppc2ir.getCR_LtGtEq_rA_Register(crfD),
																			  ppc2ir.getCR_LtGtEq_rA_Register(crfS)));
		  ppc2ir.appendInstructionToCurrentBlock(Move.create(INT_MOVE,
																			  ppc2ir.getCR_LtGtEq_rB_Register(crfD),
																			  ppc2ir.getCR_LtGtEq_rB_Register(crfS)));
		  // Copy SO bit
		  ppc2ir.appendInstructionToCurrentBlock(Move.create(INT_MOVE,
																			  ppc2ir.getCR_SO_Register(crfD),
																			  ppc2ir.getCR_SO_Register(crfS)));
		  break;
		case PPC2IR.Laziness.NOT_LAZY:
		  {
			 OPT_RegisterOperand tempInt = ppc2ir.getTempInt(0);
			 // Get bits to be copied
			 ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_AND, tempInt,
																					ppc2ir.getCRRegister(),
																					new OPT_IntConstantOperand((0xf0000000 >>> (crfS * 4))))
																 );
			 // Align bits to be copied with destination field
			 if (crfS > crfD) {
				ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_SHL, tempInt.copyRO(),
																					  tempInt.copyRO(),
																					  new OPT_IntConstantOperand((crfS - crfD)*4))
																	);
			 }
			 else { // crfS < crfD
				ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_USHR, tempInt.copyRO(),
																					  tempInt.copyRO(),
																					  new OPT_IntConstantOperand((crfD - crfS)*4))
																	);
			 }
			 // Mask out bits that will be copied into
			 ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_AND, ppc2ir.getCRRegister(),
																					ppc2ir.getCRRegister(),
																					new OPT_IntConstantOperand( ~(0xf0000000 >>> (crfD * 4))))
																 );
			 // Combine into CR register
			 ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_OR, ppc2ir.getCRRegister(),
																					ppc2ir.getCRRegister(),
																					tempInt.copyRO())
																 );
		  }
		  break;
      default:
		  throw new Error("More work to do for mcrf\n");
      }
		// Set lazy state for condition register field
		lazy.setConditionFieldLaziness(crfD, lazy.getConditionFieldLaziness(crfS));
    }
    return pc+ 4;
  }
}

/**
 * The translator for the bclr instruction
 */
class bclr_translator extends PPC_InstructionTranslator {
  /**
   * Translate bclr (branch conditional to link register)
   * <listing>
   * if  BO[2] then CTR <- CTR - 1
   * ctr_ok <- BO[2] | ((CTR != 0) (+) BO[3])
   * cond_ok <- BO[0] | (CR[BI] == BO[1])
   * if ctr_ok & cond_ok then
   *   NIA <-iea LR || 0b00
   *   if LK then LR <-iea CIA + 4
   * </listing>
   * Other simplified mnemonics: bltlr, bnelr, bdnzlr
   */
  protected int translateXL_FORM(PPC2IR ppc2ir, PPC2IR.Laziness lazy, int pc, int inst, int opcode, int BO, int BI, int zero, int secondaryOpcode, int lk){
    if (VM.VerifyAssertions) VM._assert((secondaryOpcode == 16) && (zero == 0));

    // Translation process:
    // decode BO and optionally plant any of:
    //  1: decrement CTR
    //  2: branch to instructionEndBlock if CTR == 0 (or CTR != 0)
    //  3: condition field test and branch to instructionEndBlock if true (or false)
    // plant branch block (and if lk alter lr value)

    // The block gone to if the branch isn't taken
    OPT_BasicBlock instructionEndBlock = ppc2ir.getNextBlock();

    // Decode BO
    boolean likely_to_fallthrough = ((BO & 1) == 0);
    boolean branch_if_ctr_zero = ((BO & 2) == 0);
    if ((BO & 4) == 0) {
      plantDecrementCTR(ppc2ir);
      plantBranchToBlockDependentOnCTR(ppc2ir, instructionEndBlock, branch_if_ctr_zero, likely_to_fallthrough);
    }
    if ((BO & 16) == 0) {
      boolean branch_if_cond_true = ((BO & 8) != 0);
      plantBranchToBlockDependentOnCondition(ppc2ir, instructionEndBlock, lazy, BI, !branch_if_cond_true, likely_to_fallthrough);
    }

    // Plant branch block
    OPT_RegisterOperand branchAddress;
    if (lk != 0){
      // @todo: record the pc as an address that set lr
      OPT_RegisterOperand lr = ppc2ir.getLRRegister();
      branchAddress = ppc2ir.getTempInt(0);
      ppc2ir.appendInstructionToCurrentBlock(Move.create(INT_MOVE,
																			branchAddress.copyRO(),
																			lr));
      ppc2ir.appendInstructionToCurrentBlock(Move.create(INT_MOVE, lr.copyRO(),
																			new OPT_IntConstantOperand(pc+4)));
    }
    else {
      branchAddress = ppc2ir.getLRRegister();
    }
    OPT_BasicBlock fallThrough = ppc2ir.createBlockAfterCurrent();
	 OPT_Instruction lookupswitch_instr;
	 lookupswitch_instr = LookupSwitch.create(LOOKUPSWITCH, branchAddress, null, null, 
															fallThrough.makeJumpTarget(), 
															null, 0);
	 ppc2ir.appendInstructionToCurrentBlock(lookupswitch_instr);
	 ppc2ir.registerLookupSwitchForLRUnresolved(lookupswitch_instr, pc, (PPC2IR.Laziness)lazy.clone());
	 ppc2ir.setCurrentBlock(fallThrough);
	 ppc2ir.plantRecordUncaughtBclr(pc, branchAddress.copyRO());
    ppc2ir.setReturnValueResolveLazinessAndBranchToFinish((PPC2IR.Laziness)lazy.clone(), branchAddress.copyRO());

	 // stop translation on branch always
	 if (BO == 0x14) {
		return -1;
	 }
	 else {
		// continue translation of next instruction 
		return pc+4;
	 }
  }
}

/**
 * The translator for the crnor instruction
 */
class crnor_translator extends PPC_InstructionTranslator {
  /**
   * Translate crnor (condition register nor)
   * <listing>
   * CR[crbD] <-  (CR[crbA] | CR[crbB])
   * </listing>
   * Other simplified mnemonics: crnot
   */
  protected int translateXL_FORM(PPC2IR ppc2ir, PPC2IR.Laziness lazy, int pc, int inst, int opcode, int crbD, int crbA, int crbB, int secondaryOpcode, int zero){
    if (VM.VerifyAssertions) VM._assert((secondaryOpcode == 33) && (zero == 0));      

    // Resolve laziness as this instruction encodes lots of
    // nonsensical states for the condition register :-(
    ppc2ir.resolveLazinessCrBit(lazy, crbA);

    OPT_RegisterOperand cr = ppc2ir.getCRRegister();
    if((crbA == crbB)&&(crbA == crbD)) { // crnot crbA, crbA -> just twiddle bit in position
      int bitMask = 0x80000000 >>> crbA;
      ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_XOR, cr.copyRO(),
																			  cr,
																			  new OPT_IntConstantOperand(bitMask)));
    }
    else {
      ppc2ir.resolveLazinessCrBit(lazy, crbB);
      ppc2ir.resolveLazinessCrBit(lazy, crbD);

      // Move crbA into position
      OPT_RegisterOperand tempInt1;
      if (crbA < crbD) {
		  tempInt1 = ppc2ir.getTempInt(0);
		  ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_USHR, tempInt1,
																				 cr,
																				 new OPT_IntConstantOperand(crbD - crbA)));
      }
      else if (crbA > crbD) {
		  tempInt1 = ppc2ir.getTempInt(0);
		  ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_SHL, tempInt1,
																				 cr,
																				 new OPT_IntConstantOperand(crbA - crbD)));
      }
      else {
		  tempInt1 = cr;
      }
      // Move crbB into position
      OPT_RegisterOperand tempInt2;
      if (crbB < crbD) {
		  tempInt2 = ppc2ir.getTempInt(1);
		  ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_USHR, tempInt2,
																				 cr.copyRO(),
																				 new OPT_IntConstantOperand(crbD - crbB)));
      }
      else if (crbB > crbD) {
		  tempInt2 = ppc2ir.getTempInt(1);
		  ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_SHL, tempInt2,
																				 cr.copyRO(),
																				 new OPT_IntConstantOperand(crbB - crbD)));
      }
      else {
		  tempInt2 = cr;
      }
      // Perform NOR
      OPT_RegisterOperand tempInt3 = ppc2ir.getTempInt(2);
      ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_OR, tempInt3,
																			  tempInt1.copyRO(),
																			  tempInt2.copyRO()));
      ppc2ir.appendInstructionToCurrentBlock(Unary.create(INT_NOT, tempInt3.copyRO(),
																			 tempInt3.copyRO()));
      // Mask out bits that don't concern this instruction
      ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_AND, tempInt3.copyRO(),
																			  tempInt3.copyRO(),
																			  new OPT_IntConstantOperand(0x80000000 >>> crbD)));
      // Clear desitination bit
      ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_AND, cr.copyRO(),
																			  cr.copyRO(),
																			  new OPT_IntConstantOperand(~(0x80000000 >>> crbD))));
      // Combine result
      ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_OR, cr.copyRO(),
																			  cr.copyRO(),
																			  tempInt3.copyRO()));
    }
    return pc+4;
  }
}
/**
 * The translator for the rfi instruction
 */
class rfi_translator extends PPC_InstructionTranslator {}
/**
 * The translator for the crandc instruction
 */
class crandc_translator extends PPC_InstructionTranslator {}
/**
 * The translator for the isync instruction
 */
class isync_transltor extends PPC_InstructionTranslator {}
/**
 * The translator for the crxor instruction
 */
class crxor_translator extends PPC_InstructionTranslator {
  /**
   * Translate crxor (condition register xor)
   * <listing>
   * CR[crbD] <-  CR[crbA] (+) CR[crbB]
   * </listing>
   * Other simplified mnemonics: crclr
   */
  protected int translateXL_FORM(PPC2IR ppc2ir, PPC2IR.Laziness lazy, int pc, int inst, int opcode, int crbD, int crbA, int crbB, int secondaryOpcode, int zero){
    if (VM.VerifyAssertions) VM._assert((secondaryOpcode == 193) && (zero == 0));      

    // Resolve laziness as this instruction encodes lots of
    // nonsensical states for the condition register :-(
    ppc2ir.resolveLazinessCrBit(lazy, crbA);

    OPT_RegisterOperand cr = ppc2ir.getCRRegister();
    if((crbA == crbB) && (crbA == crbD)) { // crclr crbD -> just zero bit in position
      int bitMask = ~(0x80000000 >>> crbA);
      ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_AND, cr.copyRO(),
																			  cr,
																			  new OPT_IntConstantOperand(bitMask)));
    }
    else {
      // Move crbA into position
      OPT_RegisterOperand tempInt1;
      if (crbA < crbD) {
		  ppc2ir.resolveLazinessCrBit(lazy, crbB);
		  ppc2ir.resolveLazinessCrBit(lazy, crbD);

		  tempInt1 = ppc2ir.getTempInt(0);
		  ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_USHR, tempInt1,
																				 cr,
																				 new OPT_IntConstantOperand(crbD - crbA)));
      }
      else if (crbA > crbD) {
		  tempInt1 = ppc2ir.getTempInt(0);
		  ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_SHL, tempInt1,
																				 cr,
																				 new OPT_IntConstantOperand(crbA - crbD)));
      }
      else {
		  tempInt1 = cr;
      }
      // Move crbB into position
      OPT_RegisterOperand tempInt2;
      if (crbB < crbD) {
		  tempInt2 = ppc2ir.getTempInt(1);
		  ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_USHR, tempInt2,
																				 cr.copyRO(),
																				 new OPT_IntConstantOperand(crbD - crbB)));
      }
      else if (crbB > crbD) {
		  tempInt2 = ppc2ir.getTempInt(1);
		  ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_SHL, tempInt2,
																				 cr.copyRO(),
																				 new OPT_IntConstantOperand(crbB - crbD)));
      }
      else {
		  tempInt2 = cr;
		}
      // Perform XOR
      OPT_RegisterOperand tempInt3 = ppc2ir.getTempInt(2);
      ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_XOR, tempInt3,
																			  tempInt1.copyRO(),
																			  tempInt2.copyRO()));
      // Mask out bits that don't concern this instruction
      ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_AND, tempInt3.copyRO(),
																			  tempInt3.copyRO(),
																			  new OPT_IntConstantOperand(0x80000000 >>> crbD)));
      // Clear desitination bit
      ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_AND, cr.copyRO(),
																			  cr.copyRO(),
																			  new OPT_IntConstantOperand(~(0x80000000 >>> crbD))));
      // Combine result
      ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_OR, cr.copyRO(),
																			  cr.copyRO(),
																			  tempInt3.copyRO()));
    }
    return pc+4;
  }
}

/**
 * The translator for the crnand instruction
 */
class crnand_translator extends PPC_InstructionTranslator {}
/**
 * The translator for the crand instruction
 */
class crand_translator extends PPC_InstructionTranslator {}
/**
 * The translator for the creqv instruction
 */
class creqv_translator extends PPC_InstructionTranslator {
  /**
   * Translate creqv (condition register equivalent)
   * <listing>
   * CR[crbD] <-  CR[crbA] == CR[crbB]
   * </listing>
   * Other simplified mnemonics: crset
   */
  protected int translateXL_FORM(PPC2IR ppc2ir, PPC2IR.Laziness lazy, int pc, int inst, int opcode, int crbD, int crbA, int crbB, int secondaryOpcode, int zero){
    if (VM.VerifyAssertions) VM._assert((secondaryOpcode == 289) && (zero == 0));      

    // Resolve laziness as this instruction encodes lots of
    // nonsensical states for the condition register :-(
    ppc2ir.resolveLazinessCrBit(lazy, crbA);

    OPT_RegisterOperand cr = ppc2ir.getCRRegister();
    if((crbA == crbB)&&(crbA == crbD)) { // crset crbD -> just set bit in position
      int bitMask = 0x80000000 >>> crbA;
      ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_OR, cr.copyRO(),
																			  cr,
																			  new OPT_IntConstantOperand(bitMask)));
    }
    else {
      ppc2ir.resolveLazinessCrBit(lazy, crbB);
      ppc2ir.resolveLazinessCrBit(lazy, crbD);

      // Move crbA into position
      OPT_RegisterOperand tempInt1;
      if (crbA < crbD) {
		  tempInt1 = ppc2ir.getTempInt(0);
		  ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_USHR, tempInt1,
																				 cr,
																				 new OPT_IntConstantOperand(crbD - crbA)));
      }
      else if (crbA > crbD) {
		  tempInt1 = ppc2ir.getTempInt(0);
		  ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_SHL, tempInt1,
																				 cr,
																				 new OPT_IntConstantOperand(crbA - crbD)));
      }
      else {
		  tempInt1 = cr;
      }
      // Move crbB into position
      OPT_RegisterOperand tempInt2;
      if (crbB < crbD) {
		  tempInt2 = ppc2ir.getTempInt(1);
		  ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_USHR, tempInt2,
																				 cr.copyRO(),
																				 new OPT_IntConstantOperand(crbD - crbB)));
      }
      else if (crbB > crbD) {
		  tempInt2 = ppc2ir.getTempInt(1);
		  ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_SHL, tempInt2,
																				 cr.copyRO(),
																				 new OPT_IntConstantOperand(crbB - crbD)));
      }
      else {
		  tempInt2 = cr;
		}
      // Perform equivalence
      OPT_RegisterOperand tempInt3 = ppc2ir.getTempInt(2);
      ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_XOR, tempInt3,
																			  tempInt1.copyRO(),
																			  tempInt2.copyRO()));
      ppc2ir.appendInstructionToCurrentBlock(Unary.create(INT_NOT, tempInt3.copyRO(),
																			 tempInt3.copyRO()));
      // Mask out bits that don't concern this instruction
      ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_AND, tempInt3.copyRO(),
																			  tempInt3.copyRO(),
																			  new OPT_IntConstantOperand(0x80000000 >>> crbD)));
      // Clear desitination bit
      ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_AND, cr.copyRO(),
																			  cr.copyRO(),
																			  new OPT_IntConstantOperand(~(0x80000000 >>> crbD))));
      // Combine result
      ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_OR, cr.copyRO(),
																			  cr.copyRO(),
																			  tempInt3.copyRO()));
    }
    return pc+4;
  }
}

/**
 * The translator for the croc instruction
 */
class croc_translator extends PPC_InstructionTranslator {}
/**
 * The translator for the cror instruction
 */
class cror_traslator extends PPC_InstructionTranslator {
  /**
   * Translate cror (condition register or)
   * <listing>
   * CR[crbD] <-  CR[crbA] | CR[crbB]
   * </listing>
   * Other simplified mnemonics: crclr
   */
  protected int translateXL_FORM(PPC2IR ppc2ir, PPC2IR.Laziness lazy, int pc, int inst, int opcode, int crbD, int crbA, int crbB, int secondaryOpcode, int zero){
    if (VM.VerifyAssertions) VM._assert((secondaryOpcode == 449) && (zero == 0));      

    // Resolve laziness as this instruction encodes lots of
    // nonsensical states for the condition register :-(
    ppc2ir.resolveLazinessCrBit(lazy, crbA);


    if((crbA == crbB) && (crbA == crbD)) {
		// nop
    }
	 else if (crbA == crbB) {
		// crmove
		OPT_RegisterOperand cr = ppc2ir.getCRRegister();
      // Move crbA into position
      OPT_RegisterOperand tempInt1;
      if (crbA < crbD) {
		  ppc2ir.resolveLazinessCrBit(lazy, crbB);
		  ppc2ir.resolveLazinessCrBit(lazy, crbD);

		  tempInt1 = ppc2ir.getTempInt(0);
		  ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_USHR, tempInt1,
																				 cr,
																				 new OPT_IntConstantOperand(crbD - crbA)));
      }
      else if (crbA > crbD) {
		  tempInt1 = ppc2ir.getTempInt(0);
		  ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_SHL, tempInt1,
																				 cr,
																				 new OPT_IntConstantOperand(crbA - crbD)));
      }
      else {
		  tempInt1 = cr;
      }
      // Mask out bits that don't concern this instruction
      ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_AND, tempInt1.copyRO(),
																			  tempInt1.copyRO(),
																			  new OPT_IntConstantOperand(0x80000000 >>> crbD)));
      // Clear desitination bit
      ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_AND, cr.copyRO(),
																			  cr.copyRO(),
																			  new OPT_IntConstantOperand(~(0x80000000 >>> crbD))));
      // Combine result
      ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_OR, cr.copyRO(),
																			  cr.copyRO(),
																			  tempInt1.copyRO()));		
	 }
    else {
		OPT_RegisterOperand cr = ppc2ir.getCRRegister();
      // Move crbA into position
      OPT_RegisterOperand tempInt1;
      if (crbA < crbD) {
		  ppc2ir.resolveLazinessCrBit(lazy, crbB);
		  ppc2ir.resolveLazinessCrBit(lazy, crbD);

		  tempInt1 = ppc2ir.getTempInt(0);
		  ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_USHR, tempInt1,
																				 cr,
																				 new OPT_IntConstantOperand(crbD - crbA)));
      }
      else if (crbA > crbD) {
		  tempInt1 = ppc2ir.getTempInt(0);
		  ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_SHL, tempInt1,
																				 cr,
																				 new OPT_IntConstantOperand(crbA - crbD)));
      }
      else {
		  tempInt1 = cr;
      }
      // Move crbB into position
      OPT_RegisterOperand tempInt2;
      if (crbB < crbD) {
		  tempInt2 = ppc2ir.getTempInt(1);
		  ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_USHR, tempInt2,
																				 cr.copyRO(),
																				 new OPT_IntConstantOperand(crbD - crbB)));
      }
      else if (crbB > crbD) {
		  tempInt2 = ppc2ir.getTempInt(1);
		  ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_SHL, tempInt2,
																				 cr.copyRO(),
																				 new OPT_IntConstantOperand(crbB - crbD)));
      }
      else {
		  tempInt2 = cr;
		}
      // Perform OR
      OPT_RegisterOperand tempInt3 = ppc2ir.getTempInt(2);
      ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_OR, tempInt3,
																			  tempInt1.copyRO(),
																			  tempInt2.copyRO()));
      // Mask out bits that don't concern this instruction
      ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_AND, tempInt3.copyRO(),
																			  tempInt3.copyRO(),
																			  new OPT_IntConstantOperand(0x80000000 >>> crbD)));
      // Clear desitination bit
      ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_AND, cr.copyRO(),
																			  cr.copyRO(),
																			  new OPT_IntConstantOperand(~(0x80000000 >>> crbD))));
      // Combine result
      ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_OR, cr.copyRO(),
																			  cr.copyRO(),
																			  tempInt3.copyRO()));
    }
    return pc+4;
  }
}
/**
 * The translator for the bcctr instruction
 */
class bcctr_translator extends PPC_InstructionTranslator {
  /**
   * Translate bcctr (branch conditional to count register)
   * <listing>
   * if  BO[2] then CTR <- CTR - 1
   * ctr_ok <- BO[2] | ((CTR != 0) (+) BO[3])
   * cond_ok <- BO[0] | (CR[BI] == BO[1])
   * if ctr_ok & cond_ok then
   *   NIA <-iea CTR || 0b00
   *   if LK then LR <-iea CIA + 4
   * </listing>
   * Other simplified mnemonics: bltctr, bnectr
   */
  protected int translateXL_FORM(PPC2IR ppc2ir, PPC2IR.Laziness lazy, int pc, int inst, int opcode, int BO, int BI, int zero, int secondaryOpcode, int lk){
    if (VM.VerifyAssertions) VM._assert((secondaryOpcode == 528) && (zero == 0));
	 
    // Translation process:
    // decode BO and optionally plant any of:
    //  1: decrement CTR
    //  2: branch to instructionEndBlock if CTR == 0 (or CTR != 0)
    //  3: condition field test and branch to instructionEndBlock if true (or false)
    // plant branch block (and if lk alter lr value)

    // The block gone to if the branch isn't taken
    OPT_BasicBlock instructionEndBlock = ppc2ir.getNextBlock();

    // Decode BO
    boolean likely_to_fallthrough = ((BO & 1) == 0);
    boolean branch_if_ctr_zero = ((BO & 2) == 0);
    if ((BO & 4) == 0) {
      plantDecrementCTR(ppc2ir);
      plantBranchToBlockDependentOnCTR(ppc2ir, instructionEndBlock, branch_if_ctr_zero, likely_to_fallthrough);
    }
    if ((BO & 16) == 0) {
      boolean branch_if_cond_true = ((BO & 8) != 0);
      plantBranchToBlockDependentOnCondition(ppc2ir, instructionEndBlock, lazy, BI, !branch_if_cond_true, likely_to_fallthrough);
    }

    // Plant branch block
    if (lk != 0){
      // @todo: record the pc as an address that set lr
      OPT_RegisterOperand lr = ppc2ir.getLRRegister();
      ppc2ir.appendInstructionToCurrentBlock(Move.create(INT_MOVE,
																			lr,
																			new OPT_IntConstantOperand(pc+4)));
    }
	 OPT_RegisterOperand branchAddress = ppc2ir.getCTRRegister();
    OPT_BasicBlock fallThrough = ppc2ir.createBlockAfterCurrent();
	 OPT_Instruction lookupswitch_instr;
	 lookupswitch_instr = LookupSwitch.create(LOOKUPSWITCH, branchAddress, null, null, 
															fallThrough.makeJumpTarget(), 
															null, 0);
	 ppc2ir.appendInstructionToCurrentBlock(lookupswitch_instr);
	 ppc2ir.registerLookupSwitchForCTRUnresolved(lookupswitch_instr, pc, (PPC2IR.Laziness)lazy.clone(), lk != 0);
	 ppc2ir.setCurrentBlock(fallThrough);
	 ppc2ir.plantRecordUncaughtBcctr(pc, branchAddress.copyRO());
    ppc2ir.setReturnValueResolveLazinessAndBranchToFinish((PPC2IR.Laziness)lazy.clone(), branchAddress.copyRO());

	 // stop translation on branch always
	 if (BO == 0x14) {
		return -1;
	 }
	 else {
		// continue translation of next instruction 
		return pc+4;
	 }
  }
}

// -oO Start of XO form instructions Oo-

/**
 * The translator for the subfc instruction
 */
class subfc_translator extends PPC_InstructionTranslator {
  /**
   * Translate subfc (subtract from carrying):
   * <listing>
   * rD <-  (rA) + (rB) + 1
   * </listing>
   * Also modifies XER[CA]. Equivalent to:
   * <listing>
   * rD <- (rB) - (rA)
   * </listing>
   */
  protected int translateXO_FORM(PPC2IR ppc2ir, PPC2IR.Laziness lazy, int pc, int inst, int opcode, int rD, int rA, int rB, int OE, int secondaryOpcode, int Rc) {
    // Get rA & rB
    OPT_RegisterOperand reg_rA = ppc2ir.getGPRegister(rA);
    OPT_RegisterOperand reg_rB = ppc2ir.getGPRegister(rB);

    // Copy to a long to set XER CA bit
    OPT_RegisterOperand xer_CA = ppc2ir.getXER_CA_Register();
    ppc2ir.appendInstructionToCurrentBlock(Unary.create(INT_2LONG, xer_CA, reg_rB));
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(LONG_AND, xer_CA.copyRO(),
																			xer_CA.copyRO(),
																			new OPT_LongConstantOperand(0xffffffffl)));
    OPT_RegisterOperand tempLong = ppc2ir.getTempLong(0);
    ppc2ir.appendInstructionToCurrentBlock(Unary.create(INT_2LONG, tempLong, reg_rA));
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(LONG_AND, tempLong.copyRO(),
																			tempLong.copyRO(),
																			new OPT_LongConstantOperand(0xffffffffl)));
    // Calculate XER CA bit
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(LONG_SUB, xer_CA.copyRO(),
							 xer_CA.copyRO(),
							 tempLong.copyRO()));
    ppc2ir.appendInstructionToCurrentBlock(Unary.create(INT_NOT, xer_CA.copyRO(), xer_CA.copyRO()));
    // Perform subtract for rD
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_SUB, ppc2ir.getGPRegister(rD),
							 reg_rB.copyRO(),
							 reg_rA.copyRO()));
    if (Rc != 0) {
      // Set condition codes
      setCRfield(ppc2ir, lazy, 0, ppc2ir.getGPRegister(rD), new OPT_IntConstantOperand(0), PPC2IR.Laziness.SIGNED_INT_CMP);
    }

    return pc+4;
  }
}
/**
 * The translator for the addc instruction
 */
class addc_translator extends PPC_InstructionTranslator {
  /**
   * Translate addc (add carrying):
   * <listing>
   * rD <- (rA) + (rB)
   * </listing>
   * Also modifies XER[CA].
   */
  protected int translateXO_FORM(PPC2IR ppc2ir, PPC2IR.Laziness lazy, int pc, int inst, int opcode, int rD, int rA, int rB, int OE, int secondaryOpcode, int Rc) {
    // Get rA & rB
    OPT_RegisterOperand reg_rA = ppc2ir.getGPRegister(rA);
    OPT_RegisterOperand reg_rB = ppc2ir.getGPRegister(rB);

    // Copy to a long to set XER CA bit
    OPT_RegisterOperand xer_CA = ppc2ir.getXER_CA_Register();
    OPT_RegisterOperand long_rA = ppc2ir.getTempLong(0);
    OPT_RegisterOperand long_rB = ppc2ir.getTempLong(1);
    ppc2ir.appendInstructionToCurrentBlock(Unary.create(INT_2LONG, long_rA, reg_rA));
    ppc2ir.appendInstructionToCurrentBlock(Unary.create(INT_2LONG, long_rB, reg_rB));
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(LONG_AND, long_rA.copyRO(),
																			long_rA.copyRO(),
																			new OPT_LongConstantOperand(0xffffffffl)));
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(LONG_AND, long_rB.copyRO(),
																			long_rB.copyRO(),
																			new OPT_LongConstantOperand(0xffffffffl)));
    // Calculate new XER CA bit
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(LONG_ADD, xer_CA.copyRO(),
																			long_rB.copyRO(),
																			long_rA.copyRO()));																			
    // Perform operation for rD
    OPT_RegisterOperand reg_rD = ppc2ir.getGPRegister(rD);
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_ADD, reg_rD,
																			reg_rB.copyRO(),
																			reg_rA.copyRO()));
    if (Rc != 0) {
      // Set condition codes
      setCRfield(ppc2ir, lazy, 0, reg_rD.copyRO(), new OPT_IntConstantOperand(0), PPC2IR.Laziness.SIGNED_INT_CMP);
    }

    return pc+4;
  }
}

/**
 * The translator for the mulhwu instruction
 */
class mulhwu_translator extends PPC_InstructionTranslator {
  /**
   * Translate mulhwu (multiply high word unsigned):
   * <listing>
   * prod[0-63] <- rA * rB
   * rD <- prod[0-31]
   * </listing>
   */
  protected int translateXO_FORM(PPC2IR ppc2ir, PPC2IR.Laziness lazy, int pc, int inst, int opcode, int rD, int rA, int rB, int OE, int secondaryOpcode, int Rc) {
    // Get rA & rB as longs
    OPT_RegisterOperand long_rA = ppc2ir.getTempLong(0);
    OPT_RegisterOperand long_rB = ppc2ir.getTempLong(1);
    OPT_RegisterOperand long_rD = ppc2ir.getTempLong(2);
    ppc2ir.appendInstructionToCurrentBlock(Unary.create(INT_2LONG, long_rA, ppc2ir.getGPRegister(rA)));
    ppc2ir.appendInstructionToCurrentBlock(Unary.create(INT_2LONG, long_rB, ppc2ir.getGPRegister(rB)));
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(LONG_AND, long_rA.copyRO(),
																			long_rA.copyRO(),
																			new OPT_LongConstantOperand(0xffffffffL)));
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(LONG_AND, long_rB.copyRO(),
																			long_rB.copyRO(),
																			new OPT_LongConstantOperand(0xffffffffL)));
    // Perform multiply
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(LONG_MUL, long_rD,
																			long_rB.copyRO(),
																			long_rA.copyRO()));
    // Get high 32bits
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(LONG_USHR, long_rD.copyRO(),
																			long_rD.copyRO(),
																			new OPT_IntConstantOperand(32)));
    // Set rD
    ppc2ir.appendInstructionToCurrentBlock(Unary.create(LONG_2INT, ppc2ir.getGPRegister(rD),
																		  long_rD.copyRO()));
	 
    if (Rc != 0) {
      // Set condition codes
      setCRfield(ppc2ir, lazy, 0, ppc2ir.getGPRegister(rD), new OPT_IntConstantOperand(0), PPC2IR.Laziness.SIGNED_INT_CMP);
    }

    return pc+4;
  }
}

/**
 * The translator for the subf instruction
 */
class subf_translator extends PPC_InstructionTranslator {
  /**
   * Translate subf (subtract from):
   * <listing>
   * rD <-  (rA) + (rB) + 1
   * </listing>
   * Equivalent to:
   * <listing>
   * rD <- (rB) - (rA)
   * </listing>
   * Other simplified mnemonics: subf
   */
  protected int translateXO_FORM(PPC2IR ppc2ir, PPC2IR.Laziness lazy, int pc, int inst, int opcode, int rD, int rA, int rB, int OE, int secondaryOpcode, int Rc) {
    // Get rA & rB
    OPT_RegisterOperand reg_rA = ppc2ir.getGPRegister(rA);
    OPT_RegisterOperand reg_rB = ppc2ir.getGPRegister(rB);

    // Perform subtract for rD
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_SUB, ppc2ir.getGPRegister(rD),
							 reg_rB.copyRO(),
							 reg_rA.copyRO()));
    if (Rc != 0) {
      // Set condition codes
      setCRfield(ppc2ir, lazy, 0, ppc2ir.getGPRegister(rD), new OPT_IntConstantOperand(0), PPC2IR.Laziness.SIGNED_INT_CMP);
    }

    return pc+4;
  }
}

/**
 * The translator for the mulhw instruction
 */
class mulhw_translator extends PPC_InstructionTranslator {
  /**
   * Translate mulhw (multiply high word):
   * <listing>
   * prod[0-63] <- rA * rB
   * rD <- prod
   * </listing>
   */
  protected int translateXO_FORM(PPC2IR ppc2ir, PPC2IR.Laziness lazy, int pc, int inst, int opcode, int rD, int rA, int rB, int OE, int secondaryOpcode, int Rc) {
    // Get rA & rB as longs
    OPT_RegisterOperand long_rA = ppc2ir.getTempLong(0);
    OPT_RegisterOperand long_rB = ppc2ir.getTempLong(1);
    OPT_RegisterOperand long_rD = ppc2ir.getTempLong(2);
    ppc2ir.appendInstructionToCurrentBlock(Unary.create(INT_2LONG, long_rA, ppc2ir.getGPRegister(rA)));
    ppc2ir.appendInstructionToCurrentBlock(Unary.create(INT_2LONG, long_rB, ppc2ir.getGPRegister(rB)));
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(LONG_AND, long_rA.copyRO(),
																			long_rA.copyRO(),
																			new OPT_LongConstantOperand(0xffffffffl)));
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(LONG_AND, long_rB.copyRO(),
																			long_rB.copyRO(),
																			new OPT_LongConstantOperand(0xffffffffl)));


    // Perform multiply
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(LONG_MUL, long_rD,
							 long_rB.copyRO(),
							 long_rA.copyRO()));
    // Get high 32bits
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(LONG_USHR, long_rD.copyRO(),
							 long_rD.copyRO(),
							 new OPT_IntConstantOperand(32)));
    // Set rD
    ppc2ir.appendInstructionToCurrentBlock(Unary.create(LONG_2INT, ppc2ir.getGPRegister(rD),
							long_rD.copyRO()));

    if (Rc != 0) {
      // Set condition codes
      setCRfield(ppc2ir, lazy, 0, ppc2ir.getGPRegister(rD), new OPT_IntConstantOperand(0), PPC2IR.Laziness.SIGNED_INT_CMP);
    }

    return pc+4;
  }
}

/**
 * The translator for the neg instruction
 */
class neg_translator extends PPC_InstructionTranslator {
  /**
   * Translate neg (negate):
   * <listing>
   * rD <-  (rA) + 1
   * </listing>
   * Equivalent to:
   * <listing>
   * rD <- -(rA)
   * </listing>
   */
  protected int translateXO_FORM(PPC2IR ppc2ir, PPC2IR.Laziness lazy, int pc, int inst, int opcode, int rD, int rA, int rB, int OE, int secondaryOpcode, int Rc) {
    if (VM.VerifyAssertions) VM._assert(rB == 0);

    // Perform negation
    ppc2ir.appendInstructionToCurrentBlock(Unary.create(INT_NEG, ppc2ir.getGPRegister(rD), ppc2ir.getGPRegister(rA)));

    if (Rc != 0) {
      // Set condition codes
      setCRfield(ppc2ir, lazy, 0, ppc2ir.getGPRegister(rD), new OPT_IntConstantOperand(0), PPC2IR.Laziness.SIGNED_INT_CMP);
    }

    return pc+4;
  }
}

/**
 * The translator for the subfe instruction
 */
class subfe_translator extends PPC_InstructionTranslator {
  /**
   * Translate subfe (subtract from extended):
   * <listing>
   * rD <-  (rA) + (rB) + XER[CA]
   * </listing>
   * Also modifies XER[CA].
   */
  protected int translateXO_FORM(PPC2IR ppc2ir, PPC2IR.Laziness lazy, int pc, int inst, int opcode, int rD, int rA, int rB, int OE, int secondaryOpcode, int Rc) {
    // Get rA & rB
    OPT_RegisterOperand reg_rA = ppc2ir.getGPRegister(rA);
    OPT_RegisterOperand reg_rB = ppc2ir.getGPRegister(rB);

    // Copy to a long to set XER CA bit
    OPT_RegisterOperand xer_CA = ppc2ir.getXER_CA_Register();
    OPT_RegisterOperand long_rA = ppc2ir.getTempLong(0);
    OPT_RegisterOperand long_rB = ppc2ir.getTempLong(1);
    ppc2ir.appendInstructionToCurrentBlock(Unary.create(INT_2LONG, long_rA, reg_rA));
    ppc2ir.appendInstructionToCurrentBlock(Unary.create(INT_2LONG, long_rB, reg_rB));
    ppc2ir.appendInstructionToCurrentBlock(Unary.create(LONG_NOT, long_rA.copyRO(),
																		  long_rA.copyRO()));
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(LONG_AND, long_rA.copyRO(),
																			long_rA.copyRO(),
																			new OPT_LongConstantOperand(0xffffffffl)));
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(LONG_AND, long_rB.copyRO(),
																			long_rB.copyRO(),
																			new OPT_LongConstantOperand(0xffffffffl)));

    // Get XER CA bit
    OPT_RegisterOperand xer_CA_bit = ppc2ir.getTempLong(2);
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(LONG_USHR, xer_CA,
																			xer_CA.copyRO(),
																			new OPT_IntConstantOperand(32)));
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(LONG_AND, xer_CA_bit,
																			xer_CA.copyRO(),
																			new OPT_LongConstantOperand(1)));
    // Calculate new XER CA bit
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(LONG_ADD, xer_CA.copyRO(),
																			long_rB.copyRO(),
																			long_rA.copyRO()));
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(LONG_ADD, xer_CA.copyRO(),
																			xer_CA.copyRO(),
																			xer_CA_bit.copyRO()));
																			
    // Perform operation for rD
    OPT_RegisterOperand tempInt = ppc2ir.getTempInt(0);
    OPT_RegisterOperand tempInt2 = ppc2ir.getTempInt(1);
    ppc2ir.appendInstructionToCurrentBlock(Unary.create(LONG_2INT, tempInt, xer_CA_bit.copyRO()));	 
    ppc2ir.appendInstructionToCurrentBlock(Unary.create(INT_NOT, tempInt2, reg_rA.copyRO()));	 
    OPT_RegisterOperand reg_rD = ppc2ir.getGPRegister(rD);
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_ADD, reg_rD,
																			reg_rB.copyRO(),
																			tempInt.copyRO()));
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_ADD, reg_rD.copyRO(),
																			reg_rD.copyRO(),
																			tempInt2.copyRO()));
    if (Rc != 0) {
      // Set condition codes
      setCRfield(ppc2ir, lazy, 0, reg_rD.copyRO(), new OPT_IntConstantOperand(0), PPC2IR.Laziness.SIGNED_INT_CMP);
    }

    return pc+4;
  }
}

/**
 * The translator for the adde instruction
 */
class adde_translator extends PPC_InstructionTranslator {
  /**
   * Translate adde (add extended):
   * <listing>
   * rD <- (rA) + (rB) + XER[CA]
   * </listing>
   * Also modifies XER[CA].
   */
  protected int translateXO_FORM(PPC2IR ppc2ir, PPC2IR.Laziness lazy, int pc, int inst, int opcode, int rD, int rA, int rB, int OE, int secondaryOpcode, int Rc) {
    // Get rA & rB
    OPT_RegisterOperand reg_rA = ppc2ir.getGPRegister(rA);
    OPT_RegisterOperand reg_rB = ppc2ir.getGPRegister(rB);

    // Copy to a long to set XER CA bit
    OPT_RegisterOperand xer_CA = ppc2ir.getXER_CA_Register();
    OPT_RegisterOperand long_rA = ppc2ir.getTempLong(0);
    OPT_RegisterOperand long_rB = ppc2ir.getTempLong(1);
    ppc2ir.appendInstructionToCurrentBlock(Unary.create(INT_2LONG, long_rA, reg_rA));
    ppc2ir.appendInstructionToCurrentBlock(Unary.create(INT_2LONG, long_rB, reg_rB));
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(LONG_AND, long_rA.copyRO(),
																			long_rA.copyRO(),
																			new OPT_LongConstantOperand(0xffffffffl)));
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(LONG_AND, long_rB.copyRO(),
																			long_rB.copyRO(),
																			new OPT_LongConstantOperand(0xffffffffl)));

    // Get XER CA bit
    OPT_RegisterOperand xer_CA_bit = ppc2ir.getTempLong(2);
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(LONG_USHR, xer_CA,
																			xer_CA.copyRO(),
																			new OPT_IntConstantOperand(32)));
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(LONG_AND, xer_CA_bit,
																			xer_CA.copyRO(),
																			new OPT_LongConstantOperand(1)));
    // Calculate new XER CA bit
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(LONG_ADD, xer_CA.copyRO(),
																			long_rB.copyRO(),
																			long_rA.copyRO()));
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(LONG_ADD, xer_CA.copyRO(),
																			xer_CA.copyRO(),
																			xer_CA_bit.copyRO()));
																			
    // Perform operation for rD
    OPT_RegisterOperand tempInt = ppc2ir.getTempInt(0);
    ppc2ir.appendInstructionToCurrentBlock(Unary.create(LONG_2INT, tempInt, xer_CA_bit.copyRO()));
    OPT_RegisterOperand reg_rD = ppc2ir.getGPRegister(rD);
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_ADD, reg_rD,
																			reg_rB.copyRO(),
																			reg_rA.copyRO()));
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_ADD, reg_rD.copyRO(),
																			reg_rD.copyRO(),
																			tempInt.copyRO()));
    if (Rc != 0) {
      // Set condition codes
      setCRfield(ppc2ir, lazy, 0, reg_rD.copyRO(), new OPT_IntConstantOperand(0), PPC2IR.Laziness.SIGNED_INT_CMP);
    }

    return pc+4;
  }
}
/**
 * The translator for the subfze instruction
 */
class subfze_translator extends PPC_InstructionTranslator {
  /**
   * Translate subfze (subtract from zero extended):
   * <listing>
   * rD <- (rA) + XER[CA]
   * </listing>
   * Also modifies XER[CA].
   */
  protected int translateXO_FORM(PPC2IR ppc2ir, PPC2IR.Laziness lazy, int pc, int inst, int opcode, int rD, int rA, int rB, int OE, int secondaryOpcode, int Rc) {
    if (VM.VerifyAssertions) VM._assert(rB == 0);
    // Get rA
    OPT_RegisterOperand reg_rA = ppc2ir.getGPRegister(rA);

    // Copy to a long to set XER CA bit
    OPT_RegisterOperand xer_CA = ppc2ir.getXER_CA_Register();
    OPT_RegisterOperand long_rA = ppc2ir.getTempLong(0);
    ppc2ir.appendInstructionToCurrentBlock(Unary.create(INT_2LONG, long_rA, reg_rA));
    ppc2ir.appendInstructionToCurrentBlock(Unary.create(LONG_NOT, long_rA.copyRO(), long_rA.copyRO()));
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(LONG_AND, long_rA.copyRO(),
																			long_rA.copyRO(),
																			new OPT_LongConstantOperand(0xffffffffl)));
    // Get XER CA bit
    OPT_RegisterOperand xer_CA_bit = ppc2ir.getTempLong(2);
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(LONG_USHR, xer_CA,
							 xer_CA.copyRO(),
							 new OPT_IntConstantOperand(32)));
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(LONG_AND, xer_CA_bit,
							 xer_CA.copyRO(),
							 new OPT_LongConstantOperand(1)));
    // Calculate new XER CA bit
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(LONG_ADD, xer_CA.copyRO(),
							 xer_CA_bit.copyRO(),
							 long_rA.copyRO()));
																			
    // Perform operation for rD
    OPT_RegisterOperand tempInt = ppc2ir.getTempInt(0);
    ppc2ir.appendInstructionToCurrentBlock(Unary.create(LONG_2INT, tempInt, xer_CA_bit.copyRO()));
    OPT_RegisterOperand reg_rD = ppc2ir.getGPRegister(rD);
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_ADD, reg_rD,
																			reg_rA.copyRO(),
																			tempInt.copyRO()));
    if (Rc != 0) {
      // Set condition codes
      setCRfield(ppc2ir, lazy, 0, reg_rD.copyRO(), new OPT_IntConstantOperand(0), PPC2IR.Laziness.SIGNED_INT_CMP);
    }

    return pc+4;
  }
}
/**
 * The translator for the addze instruction
 */
class addze_translator extends PPC_InstructionTranslator {
  /**
   * Translate adde (add to zero extended):
   * <listing>
   * rD <- (rA) + XER[CA]
   * </listing>
   * Also modifies XER[CA].
   */
  protected int translateXO_FORM(PPC2IR ppc2ir, PPC2IR.Laziness lazy, int pc, int inst, int opcode, int rD, int rA, int rB, int OE, int secondaryOpcode, int Rc) {
    if (VM.VerifyAssertions) VM._assert(rB == 0);
    // Get rA
    OPT_RegisterOperand reg_rA = ppc2ir.getGPRegister(rA);

    // Copy to a long to set XER CA bit
    OPT_RegisterOperand xer_CA = ppc2ir.getXER_CA_Register();
    OPT_RegisterOperand long_rA = ppc2ir.getTempLong(0);
    ppc2ir.appendInstructionToCurrentBlock(Unary.create(INT_2LONG, long_rA, reg_rA));
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(LONG_AND, long_rA.copyRO(),
																			long_rA.copyRO(),
																			new OPT_LongConstantOperand(0xffffffffl)));

    // Get XER CA bit
    OPT_RegisterOperand xer_CA_bit = ppc2ir.getTempLong(2);
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(LONG_USHR, xer_CA,
							 xer_CA.copyRO(),
							 new OPT_IntConstantOperand(32)));
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(LONG_AND, xer_CA_bit,
							 xer_CA.copyRO(),
							 new OPT_LongConstantOperand(1)));
    // Calculate new XER CA bit
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(LONG_ADD, xer_CA.copyRO(),
							 xer_CA_bit.copyRO(),
							 long_rA.copyRO()));
																			
    // Perform operation for rD
    OPT_RegisterOperand tempInt = ppc2ir.getTempInt(0);
    ppc2ir.appendInstructionToCurrentBlock(Unary.create(LONG_2INT, tempInt, xer_CA_bit.copyRO()));
    OPT_RegisterOperand reg_rD = ppc2ir.getGPRegister(rD);
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_ADD, reg_rD,
																			reg_rA.copyRO(),
																			tempInt.copyRO()));
    if (Rc != 0) {
      // Set condition codes
      setCRfield(ppc2ir, lazy, 0, reg_rD.copyRO(), new OPT_IntConstantOperand(0), PPC2IR.Laziness.SIGNED_INT_CMP);
    }

    return pc+4;
  }
}
/**
 * The translator for the subfme instruction
 */
class subfme_translator extends PPC_InstructionTranslator {}
/**
 * The translator for the addme instruction
 */
class addme_translator extends PPC_InstructionTranslator {
  /**
   * Translate addme (add to minus one extended):
   * <listing>
   * rD <- (rA) + XER[CA] - 1
   * </listing>
   * Also modifies XER[CA].
   */
  protected int translateXO_FORM(PPC2IR ppc2ir, PPC2IR.Laziness lazy, int pc, int inst, int opcode, int rD, int rA, int rB, int OE, int secondaryOpcode, int Rc) {
    if (VM.VerifyAssertions) VM._assert(rB == 0);
    // Get rA
    OPT_RegisterOperand reg_rA = ppc2ir.getGPRegister(rA);

    // Copy to a long to set XER CA bit
    OPT_RegisterOperand xer_CA = ppc2ir.getXER_CA_Register();
    OPT_RegisterOperand long_rA = ppc2ir.getTempLong(0);
    ppc2ir.appendInstructionToCurrentBlock(Unary.create(INT_2LONG, long_rA, reg_rA));
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(LONG_AND, long_rA.copyRO(),
																			long_rA.copyRO(),
																			new OPT_LongConstantOperand(0xffffffffl)));

    // Get XER CA bit ((xer_CA >> 32) & 1)
    OPT_RegisterOperand xer_CA_bit = ppc2ir.getTempLong(1);
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(LONG_USHR, xer_CA,
																			xer_CA.copyRO(),
																			new OPT_IntConstantOperand(32)));
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(LONG_AND, xer_CA_bit,
																			xer_CA.copyRO(),
																			new OPT_LongConstantOperand(1)));
    // Calculate new XER CA bit
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(LONG_ADD, xer_CA.copyRO(),
																			xer_CA_bit.copyRO(),
																			long_rA.copyRO()));
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(LONG_ADD, xer_CA.copyRO(),
																			xer_CA.copyRO(),
																			new OPT_LongConstantOperand(0xffffffffl)));	 
    // Perform operation for rD
    OPT_RegisterOperand tempInt = ppc2ir.getTempInt(0);
    ppc2ir.appendInstructionToCurrentBlock(Unary.create(LONG_2INT, tempInt, xer_CA_bit.copyRO()));
    OPT_RegisterOperand reg_rD = ppc2ir.getGPRegister(rD);
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_ADD, reg_rD,
							 reg_rA.copyRO(),
							 tempInt.copyRO()));
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_ADD, reg_rD.copyRO(),
																			reg_rD.copyRO(),
																			new OPT_IntConstantOperand(0xffffffff)));	 
    if (Rc != 0) {
      // Set condition codes
      setCRfield(ppc2ir, lazy, 0, reg_rD.copyRO(), new OPT_IntConstantOperand(0), PPC2IR.Laziness.SIGNED_INT_CMP);
    }

    return pc+4;
  }
}

/**
 * The translator for the mullw instruction
 */
class mullw_translator extends PPC_InstructionTranslator {
  /**
   * Translate mullw (multiply low word):
   * <listing>
   * rD <- rA * rB
   * </listing>
   */
  protected int translateXO_FORM(PPC2IR ppc2ir, PPC2IR.Laziness lazy, int pc, int inst, int opcode, int rD, int rA, int rB, int OE, int secondaryOpcode, int Rc) {
    // Perform multiply
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_MUL, ppc2ir.getGPRegister(rD),
							 ppc2ir.getGPRegister(rA),
							 ppc2ir.getGPRegister(rB)));
    if (Rc != 0) {
      // Set condition codes
      setCRfield(ppc2ir, lazy, 0, ppc2ir.getGPRegister(rD), new OPT_IntConstantOperand(0), PPC2IR.Laziness.SIGNED_INT_CMP);
    }

    return pc+4;
  }
}

/**
 * The translator for the add instruction
 */
class add_translator extends PPC_InstructionTranslator {
  /**
   * Translate add
   * <listing>
   * rD <- rA + rB
   * </listing>
   */
  protected int translateXO_FORM(PPC2IR ppc2ir, PPC2IR.Laziness lazy, int pc, int inst, int opcode, int rD, int rA, int rB, int OE, int secondaryOpcode, int Rc) {
    // Perform add
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_ADD, ppc2ir.getGPRegister(rD),
							 ppc2ir.getGPRegister(rA),
							 ppc2ir.getGPRegister(rB)));
    if (Rc != 0) {
      // Set condition codes
      setCRfield(ppc2ir, lazy, 0, ppc2ir.getGPRegister(rD), new OPT_IntConstantOperand(0), PPC2IR.Laziness.SIGNED_INT_CMP);
    }

    return pc+4;
  }
}

/**
 * The translator for the divwu instruction
 */
class divwu_translator extends PPC_InstructionTranslator {
  /**
   * Translate divwu (divide word unsigned):
   * <listing>
   * dividend <- (rA)
   * divisor <- (rB)
   * rD <- dividend / divisor
   * </listing>
   */
  protected int translateXO_FORM(PPC2IR ppc2ir, PPC2IR.Laziness lazy, int pc, int inst, int opcode, int rD, int rA, int rB, int OE, int secondaryOpcode, int Rc) {
    // Get rA & rB as longs
    OPT_RegisterOperand long_rA = ppc2ir.getTempLong(0);
    OPT_RegisterOperand long_rB = ppc2ir.getTempLong(1);
    OPT_RegisterOperand long_rD = ppc2ir.getTempLong(2);
    ppc2ir.appendInstructionToCurrentBlock(Unary.create(INT_2LONG, long_rA, ppc2ir.getGPRegister(rA)));
    ppc2ir.appendInstructionToCurrentBlock(Unary.create(INT_2LONG, long_rB, ppc2ir.getGPRegister(rB)));
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(LONG_AND, long_rA.copyRO(),
							 long_rA.copyRO(),
							 new OPT_LongConstantOperand(0xffffffffL)));
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(LONG_AND, long_rB.copyRO(),
							 long_rB.copyRO(),
							 new OPT_LongConstantOperand(0xffffffffL)));
    // Perform divide
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(LONG_DIV, long_rD,
							 long_rA.copyRO(),
							 long_rB.copyRO()));
    // Set rD
    ppc2ir.appendInstructionToCurrentBlock(Unary.create(LONG_2INT, ppc2ir.getGPRegister(rD),
																		  long_rD.copyRO()));

    if (Rc != 0) {
      // Set condition codes
      setCRfield(ppc2ir, lazy, 0, ppc2ir.getGPRegister(rD), new OPT_IntConstantOperand(0), PPC2IR.Laziness.SIGNED_INT_CMP);
    }

    return pc+4;
  }
}

/**
 * The translator for the divw instruction
 */
class divw_translator extends PPC_InstructionTranslator {
  /**
   * Translate divw (divide word)
   * <listing>
   * dividend <- (rA)
   * divisor <- (rB)
   * rD <- dividend / divisor
   * </listing>
   */
  protected int translateXO_FORM(PPC2IR ppc2ir, PPC2IR.Laziness lazy, int pc, int inst, int opcode, int rD, int rA, int rB, int OE, int secondaryOpcode, int Rc) {
    // Perform divide
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_DIV, ppc2ir.getGPRegister(rD),
							 ppc2ir.getGPRegister(rA),
							 ppc2ir.getGPRegister(rB)));
    if (Rc != 0) {
      // Set condition codes
      setCRfield(ppc2ir, lazy, 0, ppc2ir.getGPRegister(rD), new OPT_IntConstantOperand(0), PPC2IR.Laziness.SIGNED_INT_CMP);
    }

    return pc+4;
  }
}

/**
 * The translator for the subfco instruction
 */
class subfco_translator extends PPC_InstructionTranslator {}
/**
 * The translator for the addco instruction
 */
class addco_translator extends PPC_InstructionTranslator {}
/**
 * The translator for the subfo instruction
 */
class subfo_translator extends PPC_InstructionTranslator {}
/**
 * The translator for the subfeo instruction
 */
class subfeo_translator extends PPC_InstructionTranslator {}
/**
 * The translator for the addeo instruction
 */
class addeo_translator extends PPC_InstructionTranslator {}
/**
 * The translator for the subfzeo instruction
 */
class subfzeo_translator extends PPC_InstructionTranslator {}
/**
 * The translator for the addzeo instruction
 */
class addzeo_translator extends PPC_InstructionTranslator {}
/**
 * The translator for the subfmeo instruction
 */
class subfmeo_translator extends PPC_InstructionTranslator {}
/**
 * The translator for the addmeo instruction
 */
class addmeo_translator extends PPC_InstructionTranslator {}
/**
 * The translator for the mullwo instruction
 */
class mullwo_translator extends PPC_InstructionTranslator {}
/**
 * The translator for the nego instruction
 */
class nego_translator extends PPC_InstructionTranslator {}
/**
 * The translator for the addo instruction
 */
class addo_translator extends PPC_InstructionTranslator {}
/**
 * The translator for the divwuo instruction
 */
class divwuo_translator extends PPC_InstructionTranslator {}
/**
 * The translator for the divwo instruction
 */
class divwo_translator extends PPC_InstructionTranslator {}

// -oO Start of XFL form instructions Oo-

/**
 * The translator for the mtfsf instruction
 */
class mtfsf_translator extends PPC_InstructionTranslator {
  /**
   * Update the VX bit of the FPSCR (bit 2).
   */
  private void setVX(PPC2IR ppc2ir) {            
    // Clear the VX bit then test if we need to set it
    OPT_RegisterOperand FPSCR = ppc2ir.getFPSCRRegister();
    // FPSCR = FPSCR & 0b 1101 1111 1111 1111
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_AND, FPSCR.copyRO(),
							 FPSCR,
							 new OPT_IntConstantOperand(0xdfffffff)));
    OPT_BasicBlock testVX = ppc2ir.getCurrentBlock();
    OPT_BasicBlock setVX = ppc2ir.createBlockAfterCurrent();
    OPT_BasicBlock end = ppc2ir.createBlockAfter(setVX);

    // Test whether to set VX bit
    {
      // If any one (or more) of a set of bits (VXxxx) is non-zero,
      // the VX bit is set.
      OPT_RegisterOperand vxBits = ppc2ir.getTempInt(1);
      testVX.appendInstruction(Binary.create(INT_AND, vxBits, FPSCR.copyRO(),
					     new OPT_IntConstantOperand(0x1f80700)));
        
      OPT_RegisterOperand guardResult = null;
      testVX.appendInstruction(IfCmp.create(INT_IFCMP, guardResult, vxBits.copyRO(),
					    new OPT_IntConstantOperand(0), OPT_ConditionOperand.EQUAL(),
					    end.makeJumpTarget(),
					    OPT_BranchProfileOperand.likely()));
      testVX.insertOut(end);
    }
    // Set the VX bit
    {
      setVX.appendInstruction(Binary.create(INT_OR, FPSCR.copyRO(),
														  FPSCR.copyRO(),
														  new OPT_IntConstantOperand(0x20000000)));
    }
    // Block entered from all paths through setting the VX bit      
    ppc2ir.setCurrentBlock(end);
  }
  /**
   * Update the FEX bit of the FPSCR (bit 1).
   */
  protected void setFEX(PPC2IR ppc2ir) {
	 OPT_RegisterOperand guardResult;
	 OPT_BasicBlock currentBlock = ppc2ir.getCurrentBlock();
    // Clear the FEX bit then test if we need to set it
    OPT_RegisterOperand FPSCR = ppc2ir.getFPSCRRegister();
    // FPSCR = FPSCR & 0b 1011 1111 1111 1111
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_AND, FPSCR.copyRO(),
							 FPSCR,
							 new OPT_IntConstantOperand(0xbfffffff)));     

    // We need to make several tests of the bits of the FPSCR, so
    // make some blocks to do this.
    OPT_BasicBlock testVE_OE_UE_ZE_XE = ppc2ir.getCurrentBlock();
    OPT_BasicBlock testV = ppc2ir.createBlockAfter(testVE_OE_UE_ZE_XE);
    OPT_BasicBlock testO = ppc2ir.createBlockAfter(testV);
    OPT_BasicBlock testU = ppc2ir.createBlockAfter(testO);
    OPT_BasicBlock testZ = ppc2ir.createBlockAfter(testU);
    OPT_BasicBlock testX = ppc2ir.createBlockAfter(testZ);
    // The block to set the FEX bit
    OPT_BasicBlock setFEX = ppc2ir.createBlockAfter(testX);
    // The block after everything is done
    OPT_BasicBlock end = ppc2ir.createBlockAfter(setFEX);

    // Make a block to set the FEX bit
    {
      // FPSCR = FPSCR | 0b 0100 0000 0000 0000
      setFEX.appendInstruction(Binary.create(INT_OR, FPSCR.copyRO(),
					     FPSCR.copyRO(),
					     new OPT_IntConstantOperand(0x40000000)));
    }

    // Make a test to see if any exception bits are enabled, if not goto end
    {
      OPT_RegisterOperand VE_OE_UE_ZE_XE = ppc2ir.getTempInt(1);
      testVE_OE_UE_ZE_XE.appendInstruction(Binary.create(INT_AND, VE_OE_UE_ZE_XE,
							 FPSCR.copyRO(),
							 new OPT_IntConstantOperand(0xf8)));
      testVE_OE_UE_ZE_XE.appendInstruction(IfCmp.create(INT_IFCMP, null,
							VE_OE_UE_ZE_XE.copyRO(),
							new OPT_IntConstantOperand(0),
							OPT_ConditionOperand.EQUAL(),
							end.makeJumpTarget(),
							OPT_BranchProfileOperand.likely()));
      testVE_OE_UE_ZE_XE.insertOut(end);
        
	}
      
    /* First we test (VX & VE), i.e. (bit 2 & bit bit 24). */
    {
      OPT_RegisterOperand tempVX = ppc2ir.getTempInt(1);
      OPT_RegisterOperand tempVE = ppc2ir.getTempInt(2);
      OPT_RegisterOperand VX = ppc2ir.getTempInt(3);
      OPT_RegisterOperand VE = ppc2ir.getTempInt(4);
      OPT_RegisterOperand andV = ppc2ir.getTempInt(5);
      // Extract VX bit.
      testV.appendInstruction(Binary.create(INT_AND,  tempVX, ppc2ir.getFPSCRRegister(),
					    new OPT_IntConstantOperand(0x20000000)));
      // Shift to the last bit position.
      testV.appendInstruction(Binary.create(INT_USHR, VX, tempVX.copyRO(),
					    new OPT_IntConstantOperand(29)));
      // Extract VE bit.
      testV.appendInstruction(Binary.create(INT_AND, tempVE, ppc2ir.getFPSCRRegister(),
					    new OPT_IntConstantOperand(0x80)));
      // Shift to the last bit position.
      testV.appendInstruction(Binary.create(INT_USHR, VE, tempVE.copyRO(),
					    new OPT_IntConstantOperand(7)));
      // AND them together.
      testV.appendInstruction(Binary.create(INT_AND, andV, VX.copyRO(),
					    VE.copyRO()));
      // If the result is 1, jump to set the FEX bit.
      guardResult = null;
      testV.appendInstruction(IfCmp.create(INT_IFCMP, guardResult, andV,
					   new OPT_IntConstantOperand(1), OPT_ConditionOperand.EQUAL(),
					   setFEX.makeJumpTarget(),
					   OPT_BranchProfileOperand.unlikely()));
      // Add out edge to the jump block. 
      testV.insertOut(setFEX);
    }

    /* Now test (OX & OE), i.e. (bit 3 & bit 25). */
    {
      OPT_RegisterOperand tempOX = ppc2ir.getTempInt(1);
      OPT_RegisterOperand tempOE = ppc2ir.getTempInt(2);
      OPT_RegisterOperand OX = ppc2ir.getTempInt(3);
      OPT_RegisterOperand OE = ppc2ir.getTempInt(4);
      OPT_RegisterOperand andO = ppc2ir.getTempInt(5);
      // Extract OX bit.
      testO.appendInstruction(Binary.create(INT_AND,  tempOX, ppc2ir.getFPSCRRegister(),
					    new OPT_IntConstantOperand(0x10000000)));
      // Shift to the last bit position.
      testO.appendInstruction(Binary.create(INT_USHR, OX, tempOX.copyRO(),
					    new OPT_IntConstantOperand(28)));
      // Extract OE bit.
      testO.appendInstruction(Binary.create(INT_AND, tempOE, ppc2ir.getFPSCRRegister(),
					    new OPT_IntConstantOperand(0x40)));
      // Shift to the last bit position.
      testO.appendInstruction(Binary.create(INT_USHR, OE, tempOE.copyRO(),
					    new OPT_IntConstantOperand(6)));
      // AND them together.
      testO.appendInstruction(Binary.create(INT_AND, andO, OX.copyRO(),
					    OE.copyRO()));
      // If the result is 1, jump to set the FEX bit.
      guardResult = null;
      testO.appendInstruction(IfCmp.create(INT_IFCMP, guardResult, andO,
					   new OPT_IntConstantOperand(1), OPT_ConditionOperand.EQUAL(),
					   setFEX.makeJumpTarget(),
					   OPT_BranchProfileOperand.unlikely()));
      // Add out edge to the jump block. 
      testO.insertOut(setFEX);
    }
      
    /* Now test (UX & UE), i.e. (bit 4 & bit 26). */
    {
      OPT_RegisterOperand tempUX = ppc2ir.getTempInt(1);
      OPT_RegisterOperand tempUE = ppc2ir.getTempInt(2);
      OPT_RegisterOperand UX = ppc2ir.getTempInt(3);
      OPT_RegisterOperand UE = ppc2ir.getTempInt(4);
      OPT_RegisterOperand andU = ppc2ir.getTempInt(5);
      // Extract UX bit.
      testU.appendInstruction(Binary.create(INT_AND,  tempUX, ppc2ir.getFPSCRRegister(),
					    new OPT_IntConstantOperand(0x8000000)));
      // Shift to the last bit position.
      testU.appendInstruction(Binary.create(INT_USHR, UX, tempUX.copyRO(),
					    new OPT_IntConstantOperand(27)));
      // Extract UE bit.
      testU.appendInstruction(Binary.create(INT_AND, tempUE, ppc2ir.getFPSCRRegister(),
					    new OPT_IntConstantOperand(0x20)));
      // Shift to the last bit position.
      testU.appendInstruction(Binary.create(INT_USHR, UE, tempUE.copyRO(),
					    new OPT_IntConstantOperand(5)));
      // AND them together.
      testU.appendInstruction(Binary.create(INT_AND, andU, UX.copyRO(),
					    UE.copyRO()));
      // If the result is 1, jump to set the FEX bit.
      guardResult = null;
      testU.appendInstruction(IfCmp.create(INT_IFCMP, guardResult, andU,
					   new OPT_IntConstantOperand(1), OPT_ConditionOperand.EQUAL(),
					   setFEX.makeJumpTarget(),
					   OPT_BranchProfileOperand.unlikely()));
      // Add out edge to the jump block. 
      testU.insertOut(setFEX);
    }
      
    /* Now test (ZX & ZE), i.e. (bit 5 & bit 27). */
    {
      OPT_RegisterOperand tempZX = ppc2ir.getTempInt(1);
      OPT_RegisterOperand tempZE = ppc2ir.getTempInt(2);
      OPT_RegisterOperand ZX = ppc2ir.getTempInt(3);
      OPT_RegisterOperand ZE = ppc2ir.getTempInt(4);
      OPT_RegisterOperand andZ = ppc2ir.getTempInt(5);
      // Extract ZX bit.
      testZ.appendInstruction(Binary.create(INT_AND,  tempZX, ppc2ir.getFPSCRRegister(),
					    new OPT_IntConstantOperand(0x4000000)));
      // Shift to the last bit position.
      testZ.appendInstruction(Binary.create(INT_USHR, ZX, tempZX.copyRO(),
					    new OPT_IntConstantOperand(26)));
      // Extract ZE bit.
      testZ.appendInstruction(Binary.create(INT_AND, tempZE, ppc2ir.getFPSCRRegister(),
					    new OPT_IntConstantOperand(0x10)));
      // Shift to the last bit position.
      testZ.appendInstruction(Binary.create(INT_USHR, ZE, tempZE.copyRO(),
					    new OPT_IntConstantOperand(4)));
      // AND them together.
      testZ.appendInstruction(Binary.create(INT_AND, andZ, ZX.copyRO(),
					    ZE.copyRO()));
      // If the result is 1, jump to set the FEX bit.
      guardResult = null;
      testZ.appendInstruction(IfCmp.create(INT_IFCMP, guardResult, andZ,
					   new OPT_IntConstantOperand(1), OPT_ConditionOperand.EQUAL(),
					   setFEX.makeJumpTarget(),
					   OPT_BranchProfileOperand.unlikely()));
      // Add out edge to the jump block. 
      currentBlock.insertOut(setFEX);
    }

    /* Now test (XX & XE), i.e. (bit 6 & bit 28). */
    {
      OPT_RegisterOperand tempXX = ppc2ir.getTempInt(1);
      OPT_RegisterOperand tempXE = ppc2ir.getTempInt(2);
      OPT_RegisterOperand XX = ppc2ir.getTempInt(3);
      OPT_RegisterOperand XE = ppc2ir.getTempInt(4);
      OPT_RegisterOperand andX = ppc2ir.getTempInt(5);
      // Extract XX bit.
      testX.appendInstruction(Binary.create(INT_AND,  tempXX, ppc2ir.getFPSCRRegister(),
					    new OPT_IntConstantOperand(0x2000000)));
      // Shift to the last bit position.
      testX.appendInstruction(Binary.create(INT_USHR, XX, tempXX.copyRO(),
					    new OPT_IntConstantOperand(25)));
      // Extract XE bit.
      testX.appendInstruction(Binary.create(INT_AND, tempXE, ppc2ir.getFPSCRRegister(),
					    new OPT_IntConstantOperand(0x8)));
      // Shift to the last bit position.
      testX.appendInstruction(Binary.create(INT_USHR, XE, tempXE.copyRO(),
					    new OPT_IntConstantOperand(3)));
      // AND them together.
      testX.appendInstruction(Binary.create(INT_AND, andX, XX.copyRO(),
					    XE.copyRO()));
      // If the result is not 1, jump to avoid setting the FEX bit.
      guardResult = null;
      testX.appendInstruction(IfCmp.create(INT_IFCMP, guardResult, andX,
					   new OPT_IntConstantOperand(1), OPT_ConditionOperand.NOT_EQUAL(),
					   end.makeJumpTarget(),
					   OPT_BranchProfileOperand.likely()));
      // Add out edge to the end block
      testX.insertOut(end);
    }
    // Block entered from all paths through setting the FEX bit
    ppc2ir.setCurrentBlock(end);
  }

  /**
   * Translate mtfsf (move to fspscr fields)
   */
  protected int translateXFL_FORM(PPC2IR ppc2ir, PPC2IR.Laziness lazy, int pc, int inst, int opcode, int zero, int FM, int zero2, int frB, int secondaryOpcode, int Rc) {

    // FM has one bit per field of the FPSCR (set to 1 if that field
    // is to be written to). Make FMask have 4 identical bits per
    // field, to correspond to the full 32 bits of the FPSCR.
    int FMask = 0;
    for(int f = 0 ; f < 8 ; f++)	{
      // Extract bit number f from FM.
      int FMf = ( FM & (0x80 >>> f) ) >>> (7 -f);
      if(FMf == 1) {
		  int partMask = (0xf0000000 >>> (4 * f));
		  FMask |= partMask;
      }
    }

    // First, get the bit pattern from frB
    OPT_RegisterOperand tempLong = ppc2ir.getTempLong(0);
    ppc2ir.appendInstructionToCurrentBlock(Unary.create(DOUBLE_AS_LONG_BITS, tempLong,
																		  ppc2ir.getFPRegister(frB)));
	 
    // Extract the low order word
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(LONG_AND, tempLong.copyRO(),
																			tempLong.copyRO(),
																			new OPT_LongConstantOperand(0xffffffffL)));

    // Copy to an int
    OPT_RegisterOperand tempInt = ppc2ir.getTempInt(0);
    ppc2ir.appendInstructionToCurrentBlock(Unary.create(LONG_2INT, tempInt,
																		  tempLong.copyRO()));

	 // This is a MTFS
	 if (FMask == 0xffffffff){
		ppc2ir.appendInstructionToCurrentBlock(Move.create(INT_MOVE, ppc2ir.getFPSCRRegister(),
																			tempInt.copyRO()));
	 }
	 else {
		// Calculate tempInt = tempInt & mask
		ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_AND, tempInt.copyRO(),
																			  tempInt.copyRO(),
																			  new OPT_IntConstantOperand(FMask)));
		// Calculate FPSCR & NOT mask
		OPT_RegisterOperand tempInt2 = ppc2ir.getTempInt(1);
		ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_AND, tempInt2,
																			  ppc2ir.getFPSCRRegister(),
																			  new OPT_IntConstantOperand(~FMask)));
		
		// Calculate ((tempInt) & mask) | (FPSCR & NOT mask), and put into FPSCR
		ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_OR, ppc2ir.getFPSCRRegister(),
																			  tempInt.copyRO(),
																			  tempInt2.copyRO()));
	 }

    // If we are setting field 0, bits 0 and 3 are copied from frB, but
    // bits 1 and 2 are set according to the usual rules (page 2-8 of
    // the programming environments manual)
    if( (FM & 0x80) == 1) {
      throw new Error("Todo mtfsf 1!!!");
      //setVX();
      //setFEX();
    }

    // Check the condition register bit. */
    if(Rc != 0){
      throw new Error("Todo mtfsf 2!!!");
      //setCR1(ppc2ir.getFPRegister(frB));
    }
    return pc+4;
  }
}

// -oO Start of XFX form instructions Oo-

/**
 * The translator for the mtcrf instruction
 */
class mtcrf_translator extends PPC_InstructionTranslator {
  /**
   * Translate mtcrf (move to condition register fields)
   * <listing>
   * mask <- (4)(CRM[0]) || (4)(CRM[1]) ||... (4)(CRM[7])
   * CR <- (rS & mask) | (CR & mask)
   * </listing>
   */
  protected int translateXFX_FORM(PPC2IR ppc2ir, PPC2IR.Laziness lazy, int pc, int inst, int opcode, int rS, int CRM0, int secondaryOpcode, int zero) {
    int CRM = CRM0 >> 1;

    if(CRM == 0xff) { // mtcr rS
      ppc2ir.appendInstructionToCurrentBlock(Move.create(INT_MOVE,
							 ppc2ir.getCRRegister(),
							 ppc2ir.getGPRegister(rS)));
      for(int f=0; f < 8; f++) {
		  lazy.setConditionFieldLaziness(f, PPC2IR.Laziness.NOT_LAZY);
      }
    }
    else {
      int CRMask = 0; // CR mask (1 bit per bit of the CR)

      // CRM has one bit per field of the CR (set to 1 if that field
      // is to be written to). Make CRMask have 4 identical bits per
      // field, to correspond to the full 32 bits of the CR.
      for(int f=0 ; f < 8 ; f++) {
		  if(((CRM >>> (7 - f)) & 0x1) != 0) {
			 CRMask |= (0xf0000000 >>> (4 * f));
			 lazy.setConditionFieldLaziness(f, PPC2IR.Laziness.NOT_LAZY);
		  }
      }
      // Set bits in CR register
      OPT_RegisterOperand tempInt = ppc2ir.getTempInt(0);
      OPT_RegisterOperand cr = ppc2ir.getCRRegister();

      ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_AND, cr,
							   cr.copyRO(),
							   new OPT_IntConstantOperand(~CRMask)));
      ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_AND, tempInt,
							   ppc2ir.getGPRegister(rS),
							   new OPT_IntConstantOperand(CRMask)));
      ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_OR, cr.copyRO(),
							   tempInt.copyRO(),
							   cr.copyRO()));
    }
    return pc + 4;
  }
}
/**
 * The translator for the mfspr instruction
 */
class mfspr_translator extends PPC_InstructionTranslator {
  /**
   * Translate mfspr (move from special purpose register)
   * <listing>
   * n <- spr[5-9] || spr[0-4]
   * rS <- SPR(n)
   * </listing>
   */
  protected int translateXFX_FORM(PPC2IR ppc2ir, PPC2IR.Laziness lazy, int pc, int inst, int opcode, int rS, int spr, int secondaryOpcode, int zero) {
    spr = ((spr & 0x1f) << 5) | (spr >> 5);

    switch(spr) {
    case 1: // XER
      {
	OPT_RegisterOperand XER = ppc2ir.getTempInt(0);
	OPT_RegisterOperand XER_SO_Set = ppc2ir.getTempInt(1);
	ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_OR, XER_SO_Set,
							     ppc2ir.getXER_ByteCountRegister(),
							     new OPT_IntConstantOperand(1 << 31)));
	ppc2ir.appendInstructionToCurrentBlock(CondMove.create(INT_COND_MOVE, XER,
							       ppc2ir.getXER_SO_Register(),
							       new OPT_IntConstantOperand(0),
							       OPT_ConditionOperand.EQUAL(),
							       ppc2ir.getXER_ByteCountRegister(),
							       XER_SO_Set.copyRO()));
	OPT_RegisterOperand XER_OV_Set = ppc2ir.getTempInt(1);
	ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_OR, XER_OV_Set,
							     XER.copyRO(),
							     new OPT_IntConstantOperand(1 << 30)));
	ppc2ir.appendInstructionToCurrentBlock(CondMove.create(INT_COND_MOVE, XER.copyRO(),
							       ppc2ir.getXER_OV_Register(),
							       new OPT_IntConstantOperand(0),
							       OPT_ConditionOperand.EQUAL(),
							       XER.copyRO(),
							       XER_OV_Set.copyRO()));
	OPT_RegisterOperand XER_CA_Long = ppc2ir.getTempLong(1);
	OPT_RegisterOperand XER_CA_Int = ppc2ir.getTempInt(1);
	ppc2ir.appendInstructionToCurrentBlock(Binary.create(LONG_SHR, XER_CA_Long,
							     ppc2ir.getXER_CA_Register(),
							     new OPT_LongConstantOperand(3)));	
	ppc2ir.appendInstructionToCurrentBlock(Unary.create(LONG_2INT, XER_CA_Int,
							    XER_CA_Long.copyRO()));
	ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_AND, XER_CA_Int.copyRO(),
							     XER_CA_Int.copyRO(),
							     new OPT_IntConstantOperand(1 << 29)));	
	ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_OR, XER.copyRO(),
							     XER.copyRO(),
							     XER_CA_Int.copyRO()));
      }
      break;
    case 8: // LR
      ppc2ir.appendInstructionToCurrentBlock(Move.create(INT_MOVE,
							      ppc2ir.getGPRegister(rS),
							      ppc2ir.getLRRegister()));
      break;
    case 9: // CTR
      ppc2ir.appendInstructionToCurrentBlock(Move.create(INT_MOVE,
							      ppc2ir.getGPRegister(rS),
							      ppc2ir.getCTRRegister()));
      break;
    default:
      throw new Error("Unknown spr "+spr);
    }
    return pc+4;
  }
}
/**
 * The translator for the mtspr instruction
 */
class mtspr_translator extends PPC_InstructionTranslator {
  /**
   * Translate mtspr (move to special purpose register)
   * <listing>
   * n <- spr[5-9] || spr[0-4]
   * SPR(n) <- rS
   * </listing>
   */
  protected int translateXFX_FORM(PPC2IR ppc2ir, PPC2IR.Laziness lazy, int pc, int inst, int opcode, int rS, int spr, int secondaryOpcode, int zero) {
    spr = ((spr & 0x1f) << 5) | (spr >> 5);

    switch(spr) {
    case 1: // XER
      {
		  OPT_RegisterOperand rS_RegOp = ppc2ir.getGPRegister(rS);
		  // Copy SO bit into XER[SO] register
		  ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_AND, ppc2ir.getXER_SO_Register(),
																				 rS_RegOp,
																				 new OPT_IntConstantOperand(1 << 31)));
		  // Copy OV bit into XER[OV] register
		  ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_AND, ppc2ir.getXER_OV_Register(),
																				 rS_RegOp.copyRO(),
																				 new OPT_IntConstantOperand(1 << 30)));
		  // Copy byte count bits out
		  ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_AND, ppc2ir.getXER_ByteCountRegister(),
																				 rS_RegOp.copyRO(),
																				 new OPT_IntConstantOperand(0x3f)));
		  // Copy CA bit into XER[CA] register
		  OPT_RegisterOperand XER_CA = ppc2ir.getXER_CA_Register();
		  ppc2ir.appendInstructionToCurrentBlock(Unary.create(LONG_2INT, XER_CA, rS_RegOp.copyRO()));
		  ppc2ir.appendInstructionToCurrentBlock(Binary.create(LONG_SHL, XER_CA.copyRO(),
																				 XER_CA.copyRO(),
																				 new OPT_IntConstantOperand(3)));
      }
      break;
    case 8: // LR
      ppc2ir.appendInstructionToCurrentBlock(Move.create(INT_MOVE,
							      ppc2ir.getLRRegister(),
							      ppc2ir.getGPRegister(rS)));
      break;
    case 9: // CTR
      ppc2ir.appendInstructionToCurrentBlock(Move.create(INT_MOVE,
																				  ppc2ir.getCTRRegister(),
																				  ppc2ir.getGPRegister(rS)));
      break;
    default:
      throw new Error("Unknown spr "+spr);
    }
    return pc+4;
  }
}

// -oO Start of X form instructions Oo-

/**
 * The translator for the cmp instruction
 */
class cmp_translator extends PPC_InstructionTranslator {
  /**
   * Translate cmp (compare):
   * <listing>
   * if L=0 then a <- EXTS(rA)
   *             b <- EXTS(rB)
   * else        a <- rA
   *             b <- rB
   * if a < b
   *    then c <- 0b100
   *    else if a > EXTS(SIMM)
   *            then c <- 0b010
   *            else c <- 0b001
   * CR[(4 * crfD)-(4 * crfD + 3)] <- c || XER[SO]
   * </listing>
   * The L bit has no effect on 32-it operations.
   * Other simplified mnemonics: cmpw
   */
  protected int translateX_FORM(PPC2IR ppc2ir, PPC2IR.Laziness lazy, int pc, int inst, int opcode, int crfD0L, int rA, int rB, int zero, int zero2) {
    if (VM.VerifyAssertions) VM._assert((zero == zero2) && (zero2 == 0));
    int crfD = crfD0L >>> 2;

    setCRfield(ppc2ir, lazy, crfD, ppc2ir.getGPRegister(rA), ppc2ir.getGPRegister(rB), PPC2IR.Laziness.SIGNED_INT_CMP);

    return pc+4;
  }
}
/**
 * The translator for the tw instruction
 */
class tw_translator extends PPC_InstructionTranslator {
  /**
   * Translate tw (trap word)
   */
  protected int translateX_FORM(PPC2IR ppc2ir, PPC2IR.Laziness lazy, int pc, int inst, int opcode, int crfD0L, int rA, int rB, int zero, int zero2) {
	 // @TODO
	 ppc2ir.plantThrowBadInstruction((PPC2IR.Laziness)lazy.clone(), pc);
	 return -1;
  }
}
/**
 * The translator for the lvsl instruction
 */
class lvsl_translator extends PPC_InstructionTranslator {}
/**
 * The translator for the lvebx instruction
 */
class lvebx_translator extends PPC_InstructionTranslator {}
/**
 * The translator for the mulhdu instruction
 */
class mulhdu_translator extends PPC_InstructionTranslator {}
/**
 * The translator for the mfcr instruction
 */
class mfcr_translator extends PPC_InstructionTranslator {
  /**
   * Translate mfcr (move from condition register):
   * <listing>
   * rD <- CR
   * </listing>
   */
  protected int translateX_FORM(PPC2IR ppc2ir, PPC2IR.Laziness lazy, int pc, int inst, int opcode, int rD, int zero1, int zero2, int secodaryOpcode, int zero3) {
    if (VM.VerifyAssertions) VM._assert((zero1 == zero2) && (zero2 == zero3) && (zero3 == 0));
    ppc2ir.resolveLaziness(lazy);

    ppc2ir.appendInstructionToCurrentBlock(Move.create(INT_MOVE, ppc2ir.getGPRegister(rD),
						       ppc2ir.getCRRegister()));
	 return pc+4;
  }
}
/**
 * The translator for the lwarx instruction
 */
class lwarx_translator extends PPC_InstructionTranslator {
  /**
   * Translate lwarx (load word and reserve indexed)
   * <listing>
   * if rA = 0 then b <- 0
   * else b <- (rA)
   * EA <- b + (rB)
	* RESERVE <- 1
	* RESERVE_ADDR <- physical_addr(EA)
   * rD <- MEM(EA, 4)
   * </listing>
	* See also: stwcx
   */
  protected int translateX_FORM(PPC2IR ppc2ir, PPC2IR.Laziness lazy, int pc, int inst, int opcode, int rD, int rA, int rB, int secondaryOpcode, int zero) {
    OPT_RegisterOperand EA = ppc2ir.getTempInt(PPC2IR.TEMP_INT_SAFE_FROM_LOAD_CLOBBER);

    if (rA == 0) {
      // EA = rB
      ppc2ir.appendInstructionToCurrentBlock(Move.create(INT_MOVE,
																			EA,
																			ppc2ir.getGPRegister(rB)));
    }
    else {
      // EA = rA + rB
      ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_ADD, EA,
																			  ppc2ir.getGPRegister(rA),
																			  ppc2ir.getGPRegister(rB)));
    }
    // Load value from memory at EA into rD.
    ppc2ir.memoryLoad(ppc2ir.getGPRegister(rD), EA.copyRO());
	 // Ignore the reserve part of the instruction for the time being
    return pc+4;
  }
}
/**
 * The translator for the ldx instruction
 */
class ldx_translator extends PPC_InstructionTranslator {}
/**
 * The translator for the lwzx instruction
 */
class lwzx_translator extends PPC_InstructionTranslator {
  /**
   * Translate lwzx (load word and zero indexed)
   * <listing>
   * if rA = 0 then b <- 0
   * else b <- (rA)
   * EA <- b + (rB)
   * rD <- MEM(EA, 4)
   * </listing>
   */
  protected int translateX_FORM(PPC2IR ppc2ir, PPC2IR.Laziness lazy, int pc, int inst, int opcode, int rD, int rA, int rB, int secondaryOpcode, int zero) {
    OPT_RegisterOperand EA = ppc2ir.getTempInt(PPC2IR.TEMP_INT_SAFE_FROM_LOAD_CLOBBER);

    if (rA == 0) {
      // EA = rB
      ppc2ir.appendInstructionToCurrentBlock(Move.create(INT_MOVE,
							 EA,
							 ppc2ir.getGPRegister(rB)));
    }
    else {
      // EA = rA + rB
      ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_ADD, EA,
							   ppc2ir.getGPRegister(rA),
							   ppc2ir.getGPRegister(rB)));
    }
    // Load value from memory at EA into rD.
    ppc2ir.memoryLoad(ppc2ir.getGPRegister(rD), EA.copyRO());
    return pc+4;
  }
}

/**
 * The translator for the slw instruction
 */
class slw_translator extends PPC_InstructionTranslator {
  /**
   * Translate slw (shift left word)
   * <listing>
   * n <- rB[27-31]
   * r <- ROTL(rS,n)
   * </listing>
   * Equivalent to:
   * <listing>
   * rD <- rA << rB
   * </listing>
   */
  protected int translateX_FORM(PPC2IR ppc2ir, PPC2IR.Laziness lazy, int pc, int inst, int opcode, int rS, int rA, int rB, int secondaryOpcode, int Rc) {
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_SHL, ppc2ir.getGPRegister(rA),
							 ppc2ir.getGPRegister(rS),
							 ppc2ir.getGPRegister(rB)));
    if (Rc != 0) {
      // Set condition codes
      setCRfield(ppc2ir, lazy, 0, ppc2ir.getGPRegister(rA), new OPT_IntConstantOperand(0), PPC2IR.Laziness.SIGNED_INT_CMP);
    }
	 return pc+4;
  }
}

/**
 * The translator for the cntlzw instruction
 */
class cntlzw_translator extends PPC_InstructionTranslator {
  /**
   * Translate cntlzw (count leading zeros word)
   * <listing>
   * n <- 0
   * do while n < 32
   *   if rS[n] = then leave
   *   n <- n + 1
   * rA <- n
   * </listing>
   *
   * There are a number of count leading zero algorithms (a good list
   * is at {@url http://www.hackersdelight.org/HDcode/nlz.cc}). This
   * routine delegates to the routine it conceives to be more optimal
   */
  protected int translateX_FORM(PPC2IR ppc2ir, PPC2IR.Laziness lazy, int pc, int inst, int opcode, int rS, int rA, int zero, int secondaryOpcode, int Rc) {
    nlz_count_no_branches(ppc2ir, rS, rA);
    if (Rc != 0) {
      // Set condition codes
      setCRfield(ppc2ir, lazy, 0, ppc2ir.getGPRegister(rA), new OPT_IntConstantOperand(0), PPC2IR.Laziness.SIGNED_INT_CMP);
    }
	 return pc + 4;
  }
  /**
   * Perform a count of the number of leading zeros with no branch operations:
   * <listing>
   * int nlz4(unsigned x) {
   *   int y, m, n;
   *
   *   y = -(x >> 16);      // If left half of x is 0,
   *   m = (y >> 16) & 16;  // set n = 16.  If left half
   *   n = 16 - m;          // is nonzero, set n = 0 and
   *   x = x >> m;          // shift x right 16.
   *                        // Now x is of the form 0000xxxx.
   *   y = x - 0x100;       // If positions 8-15 are 0,
   *   m = (y >> 16) & 8;   // add 8 to n and shift x left 8.
   *   n = n + m;
   *   x = x << m;
   *
   *   y = x - 0x1000;      // If positions 12-15 are 0,
   *   m = (y >> 16) & 4;   // add 4 to n and shift x left 4.
   *   n = n + m;
   *   x = x << m;

   *   y = x - 0x4000;      // If positions 14-15 are 0,
   *   m = (y >> 16) & 2;   // add 2 to n and shift x left 2.
   *   n = n + m;
   *   x = x << m;
   *
   *   y = x >> 14;         // Set y = 0, 1, 2, or 3.
   *   m = y & ~(y >> 1);   // Set m = 0, 1, 2, or 2 resp.
   *   return n + 2 - m;
   * }
   * </listing>
   */
  private static void nlz_count_no_branches(PPC2IR ppc2ir, int rS, int rA) {
    OPT_RegisterOperand x = ppc2ir.getTempInt(0);
    OPT_RegisterOperand y = ppc2ir.getTempInt(1);
    OPT_RegisterOperand m = ppc2ir.getTempInt(2);
    OPT_RegisterOperand n = ppc2ir.getTempInt(3);

    ppc2ir.appendInstructionToCurrentBlock(Move.create(INT_MOVE, x,
																		 ppc2ir.getGPRegister(rS)));

    // y = -(x >> 16);      // If left half of x is 0,
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_USHR, y,
																			x.copyRO(),
																			new OPT_IntConstantOperand(16)));
    ppc2ir.appendInstructionToCurrentBlock(Unary.create(INT_NEG, y.copyRO(),
																		  y.copyRO()));
    // m = (y >> 16) & 16;  // set n = 16.  If left half
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_SHR, m,
							 y.copyRO(),
							 new OPT_IntConstantOperand(16)));
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_AND, m.copyRO(),
							 m.copyRO(),
							 new OPT_IntConstantOperand(16)));
    // n = 16 - m;          // is nonzero, set n = 0 and
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_SUB, n,
							 new OPT_IntConstantOperand(16),
							 m.copyRO()));
    // x = x >> m;          // shift x right 16.
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_USHR, x.copyRO(),
							 x.copyRO(),
							 m.copyRO()));
    //                      // Now x is of the form 0000xxxx.
    // y = x - 0x100;       // If positions 8-15 are 0,
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_SUB, y.copyRO(),
							 x.copyRO(),
							 new OPT_IntConstantOperand(0x100)));
    //  m = (y >> 16) & 8;   // add 8 to n and shift x left 8.
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_SHR, m.copyRO(),
							 y.copyRO(),
							 new OPT_IntConstantOperand(16)));
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_AND, m.copyRO(),
							 m.copyRO(),
							 new OPT_IntConstantOperand(8)));

    // n = n + m;
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_ADD, n.copyRO(),
							 n.copyRO(),
							 m.copyRO()));
    // x = x << m;
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_SHL, x.copyRO(),
							 x.copyRO(),
							 m.copyRO()));
    // y = x - 0x1000;      // If positions 12-15 are 0,
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_SUB, y.copyRO(),
							 x.copyRO(),
							 new OPT_IntConstantOperand(0x1000)));
    // m = (y >> 16) & 4;   // add 4 to n and shift x left 4.
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_SHR, m.copyRO(),
							 y.copyRO(),
							 new OPT_IntConstantOperand(16)));
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_AND, m.copyRO(),
							 m.copyRO(),
							 new OPT_IntConstantOperand(4)));
    // n = n + m;
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_ADD, n.copyRO(),
							 n.copyRO(),
							 m.copyRO()));
    // x = x << m;
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_SHL, x.copyRO(),
							 x.copyRO(),
							 m.copyRO()));
    // y = x - 0x4000;      // If positions 14-15 are 0,
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_SUB, y.copyRO(),
							 x.copyRO(),
							 new OPT_IntConstantOperand(0x4000)));
    // m = (y >> 16) & 2;   // add 2 to n and shift x left 2.
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_SHR, m.copyRO(),
							 y.copyRO(),
							 new OPT_IntConstantOperand(16)));
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_AND, m.copyRO(),
							 m.copyRO(),
							 new OPT_IntConstantOperand(2)));
    // n = n + m;
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_ADD, n.copyRO(),
							 n.copyRO(),
							 m.copyRO()));
    // x = x << m;
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_SHL, x.copyRO(),
							 x.copyRO(),
							 m.copyRO()));
    // y = x >> 14;         // Set y = 0, 1, 2, or 3.
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_USHR, y.copyRO(),
							 x.copyRO(),
							 new OPT_IntConstantOperand(14)));
    // m = y & ~(y >> 1);   // Set m = 0, 1, 2, or 2 resp.
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_SHR, m.copyRO(),
							 y.copyRO(),
							 new OPT_IntConstantOperand(1)));
    ppc2ir.appendInstructionToCurrentBlock(Unary.create(INT_NOT, m.copyRO(),
							m.copyRO()));
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_AND, m.copyRO(),
							 m.copyRO(),
							 y.copyRO()));
    // return n + 2 - m;
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_ADD, n.copyRO(),
							 n.copyRO(),
							 new OPT_IntConstantOperand(2)));    
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_SUB, ppc2ir.getGPRegister(rA),
							 n.copyRO(),
							 m.copyRO()));    
  }
  /**
   * Perform leading zero count using a loop
   */
  private static void nlz_count_loop(PPC2IR ppc2ir, int rS, int rA) {
    throw new Error("Todo");
  }
}
/**
 * The translator for the sld instruction
 */
class sld_translator extends PPC_InstructionTranslator {}
/**
 * The translator for the and instruction
 */
class and_translator extends PPC_InstructionTranslator {
  /**
   * Translate and
   * <listing>
   * rA <- (rS) & (rB)
   * </listing>
   */
  protected int translateX_FORM(PPC2IR ppc2ir, PPC2IR.Laziness lazy, int pc, int inst, int opcode, int rS, int rA, int rB, int secondaryOpcode, int Rc) {
    // Perform and
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_AND, ppc2ir.getGPRegister(rA),
							 ppc2ir.getGPRegister(rS),
							 ppc2ir.getGPRegister(rB)));
    if (Rc != 0) {
      // Set condition codes
      setCRfield(ppc2ir, lazy, 0, ppc2ir.getGPRegister(rA), new OPT_IntConstantOperand(0), PPC2IR.Laziness.SIGNED_INT_CMP);
    }

    return pc+4;
  }
}
/**
 * The translator for the cmpl instruction
 */
class cmpl_translator extends PPC_InstructionTranslator {
  /**
   * Translate cmpli (compare logical immediate):
   * <listing>
   * a <- (rA)
   * if a <U ((16)0 || UIMM)
   *    then c <- 0b100
   *    else if a >U ((16)0 || UIMM)
   *            then c <- 0b010
   *            else c <- 0b001
   * CR[(4 * crfD)-(4 * crfD + 3)] <- c || XER[SO]
   * </listing>
   */
  protected int translateX_FORM(PPC2IR ppc2ir, PPC2IR.Laziness lazy, int pc, int inst, int opcode, int crfD0L, int rA, int rB, int secondaryOpcode, int Rc) {
    int crfD = crfD0L >>> 2;

    setCRfield(ppc2ir, lazy, crfD, ppc2ir.getGPRegister(rA), ppc2ir.getGPRegister(rB), PPC2IR.Laziness.UNSIGNED_INT_CMP);

    return pc+4;
  }
}
/**
 * The translator for the lvsr instruction
 */
class lvsr_translator extends PPC_InstructionTranslator {}
/**
 * The translator for the lvehx instruction
 */
class lvehx_translator extends PPC_InstructionTranslator {}
/**
 * The translator for the ldux instruction
 */
class ldux_translator extends PPC_InstructionTranslator {}
/**
 * The translator for the dcbst instruction
 */
class dcbst_translator extends PPC_InstructionTranslator {
  /**
   * Translate dcbst (data cache block store)
   */
  protected int translateX_FORM(PPC2IR ppc2ir, PPC2IR.Laziness lazy, int pc, int inst, int opcode, int zero, int rA, int rB, int secondaryOpcode, int zero2) {
	 // translated into a NOP
	 return pc+4;
  }
}
/**
 * The translator for the lwzux instruction
 */
class lwzux_translator extends PPC_InstructionTranslator {}
/**
 * The translator for the andc instruction
 */
class andc_translator extends PPC_InstructionTranslator {
  /**
   * Translate andc (and with complement)
   * <listing>
   * rA <- (rS) & (rB)
   * </listing>
   */
  protected int translateX_FORM(PPC2IR ppc2ir, PPC2IR.Laziness lazy, int pc, int inst, int opcode, int rS, int rA, int rB, int secondaryOpcode, int Rc) {
    OPT_RegisterOperand tempInt = ppc2ir.getTempInt(0);
    // Perform complement
    ppc2ir.appendInstructionToCurrentBlock(Unary.create(INT_NOT, tempInt,
																		  ppc2ir.getGPRegister(rB)));

    // Perform and
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_AND, ppc2ir.getGPRegister(rA),
																			tempInt.copyRO(),
																			ppc2ir.getGPRegister(rS)));
    if (Rc != 0) {
      // Set condition codes
      setCRfield(ppc2ir, lazy, 0, ppc2ir.getGPRegister(rA), new OPT_IntConstantOperand(0), PPC2IR.Laziness.SIGNED_INT_CMP);
    }
    return pc+4;
  }
}
/**
 * The translator for the td instruction
 */
class td_translator extends PPC_InstructionTranslator {}
/**
 * The translator for the mulhd instruction
 */
class mulhd_translator extends PPC_InstructionTranslator {}
/**
 * The translator for the mfmsr instruction
 */
class mfmsr_translator extends PPC_InstructionTranslator {}
/**
 * The translator for the ldarx instruction
 */
class ldarx_translator extends PPC_InstructionTranslator {}
/**
 * The translator for the dcbf instruction
 */
class dcbf_translator extends PPC_InstructionTranslator {}
/**
 * The translator for the lbzx instruction
 */
class lbzx_translator extends PPC_InstructionTranslator {
  /**
   * Translate lbz (load byte and zero indexed)
   * <listing>
   * if rA = 0 then b <- 0
   * else b <- (rA)
   * EA <- b + (rB)
   * rD <- (24)0 || MEM(EA, 1)
   * </listing>
   */
  protected int translateX_FORM(PPC2IR ppc2ir, PPC2IR.Laziness lazy, int pc, int inst, int opcode, int rD, int rA, int rB, int secondaryOpcode, int zero) {
    OPT_RegisterOperand EA = ppc2ir.getTempInt(PPC2IR.TEMP_INT_SAFE_FROM_LOAD_CLOBBER);

    if (rA == 0) {
      // EA = rB
      ppc2ir.appendInstructionToCurrentBlock(Move.create(INT_MOVE,
							 EA,
							 ppc2ir.getGPRegister(rB)));
    }
    else {
      // EA = rA + rB
      ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_ADD, EA,
							   ppc2ir.getGPRegister(rA),
							   ppc2ir.getGPRegister(rB)));
    }
    // Load value from memory at EA into rD.
    OPT_RegisterOperand result = ppc2ir.getGPRegister(rD);
    ppc2ir.memoryLoad(result, EA.copyRO());
    // Turn 32bit address into appropriate shift using last 2 bits
    // of the address
    //-#if RVM_FOR_POWERPC
    // EA = (EA & 0x3) * 8
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_AND, EA.copyRO(),
																			EA.copyRO(),
																			new OPT_IntConstantOperand(3)));
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_SHL, EA.copyRO(),
																			EA.copyRO(),
																			new OPT_IntConstantOperand(3)));
	 //-#else
    // EA = (3 - (EA & 0x3)) * 8
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_AND, EA.copyRO(),
																			EA.copyRO(),
																			new OPT_IntConstantOperand(3)));
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_SUB, EA.copyRO(),
																			new OPT_IntConstantOperand(3),
																			EA.copyRO()));
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_SHL, EA.copyRO(),
																			EA.copyRO(),
																			new OPT_IntConstantOperand(3)));
	 //-#endif
    // rD >>>= EA
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_USHR, result.copyRO(),
																			result.copyRO(),
																			EA.copyRO()));
    // rD &= 0xff
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_AND, result.copyRO(),
																			result.copyRO(),
																			new OPT_IntConstantOperand(0xff)));
    return pc+4;
  }
}
/**
 * The translator for the lbzux instruction
 */
class lbzux_translator extends PPC_InstructionTranslator {
  /**
   * Translate lbzux (load byte and zero with update indexed)
   * <listing>
   * if rA = 0 then b <- 0
   * else b <- (rA)
   * EA <- b + (rB)
   * rD <- (24)0 || MEM(EA, 1)
   * </listing>
   */
  protected int translateX_FORM(PPC2IR ppc2ir, PPC2IR.Laziness lazy, int pc, int inst, int opcode, int rD, int rA, int rB, int secondaryOpcode, int zero) {
    OPT_RegisterOperand EA = ppc2ir.getTempInt(PPC2IR.TEMP_INT_SAFE_FROM_LOAD_CLOBBER);

    if ((rA == 0)||(rA == rD)) {
      throw new Error("Invalid form of lbzux at 0x" + Integer.toHexString(pc));
    }
    else {
      // EA = rA + rB
      ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_ADD, EA,
							   ppc2ir.getGPRegister(rA),
							   ppc2ir.getGPRegister(rB)));
    }
    // Load value from memory at EA into rD.
    OPT_RegisterOperand result = ppc2ir.getGPRegister(rD);
    ppc2ir.memoryLoad(result, EA.copyRO());
    // Place EA into rA
    ppc2ir.appendInstructionToCurrentBlock(Move.create(INT_MOVE, ppc2ir.getGPRegister(rA), 
																		 EA.copyRO()));
    // Turn 32bit address into appropriate shift using last 2 bits
    // of the address
    //-#if RVM_FOR_POWERPC
    // EA = (EA & 0x3) * 8
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_AND, EA.copyRO(),
																			EA.copyRO(),
																			new OPT_IntConstantOperand(3)));
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_SHL, EA.copyRO(),
																			EA.copyRO(),
																			new OPT_IntConstantOperand(3)));
	 //-#else
    // EA = (3 - (EA & 0x3)) * 8
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_AND, EA.copyRO(),
																			EA.copyRO(),
																			new OPT_IntConstantOperand(3)));
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_SUB, EA.copyRO(),
																			new OPT_IntConstantOperand(3),
																			EA.copyRO()));
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_SHL, EA.copyRO(),
																			EA.copyRO(),
																			new OPT_IntConstantOperand(3)));
	 //-#endif
    // rD >>>= EA
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_USHR, result.copyRO(),
																			result.copyRO(),
																			EA.copyRO()));
    // rD &= 0xff
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_AND, result.copyRO(),
																			result.copyRO(),
																			new OPT_IntConstantOperand(0xff)));
    return pc+4;
  }
}
/**
 * The translator for the nor instruction
 */
class nor_translator extends PPC_InstructionTranslator {
  /**
   * Translate nor
   * <listing>
   * rA <- ((rS) | (rB))
   * </listing>
   */
  protected int translateX_FORM(PPC2IR ppc2ir, PPC2IR.Laziness lazy, int pc, int inst, int opcode, int rS, int rA, int rB, int secondaryOpcode, int Rc) {
    OPT_RegisterOperand tempInt = ppc2ir.getTempInt(0);
    // Perform or
    if (rS == rB) {
      ppc2ir.appendInstructionToCurrentBlock(Move.create(INT_MOVE,
							 tempInt,
							 ppc2ir.getGPRegister(rB)));
    }
    else {
      ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_OR, tempInt,
							   ppc2ir.getGPRegister(rS),
							   ppc2ir.getGPRegister(rB)));
    }
    // Perform complement
    ppc2ir.appendInstructionToCurrentBlock(Unary.create(INT_NOT, ppc2ir.getGPRegister(rA),
																		  tempInt.copyRO()));
    if (Rc != 0) {
      // Set condition codes
      setCRfield(ppc2ir, lazy, 0, ppc2ir.getGPRegister(rA), new OPT_IntConstantOperand(0), PPC2IR.Laziness.SIGNED_INT_CMP);
    }
    return pc+4;
  }
}
/**
 * The translator for the mtmsr instruction
 */
class mtmsr_translator extends PPC_InstructionTranslator {}
/**
 * The translator for the stdx instruction
 */
class stdx_translator extends PPC_InstructionTranslator {}
/**
 * The translator for the stwcx instruction
 */
class stwcx_translator extends PPC_InstructionTranslator {
  /**
   * Translate stwx (store word conditional indexed)
   * <listing>
   * if rA = 0 then b <- 0
   * else b <- (rA)
   * EA <- b + rB
   * MEM(EA, 4) <- rS
	* if RESERVE then
	*  if RESERVE_ADDR = physical_addr(EA)
	*   MEM(EA,4) <- rS
	*   CR0 <- 0b00 || 0b1 || XER[SO]
	*  else
	*   u <- undefined 1-bit value
	*   if u then MEM(EA,4) <- rS
	*   CR0 <- 0b00 || u || XER[SO]
	*  RESERVE <- 0
	* else
	*   CR0 <- 0b00 || 0b0 || XER[SO]
   * </listing>
	* See also: lwarx
   */
  protected int translateX_FORM(PPC2IR ppc2ir, PPC2IR.Laziness lazy, int pc, int inst, int opcode, int rS, int rA, int rB, int secondaryOpcode, int zero) {
    OPT_RegisterOperand EA = ppc2ir.getTempInt(PPC2IR.TEMP_INT_SAFE_FROM_STORE_CLOBBER);

    if (rA == 0) {
      // EA = rB
      ppc2ir.appendInstructionToCurrentBlock(Move.create(INT_MOVE,
							 EA,
							 ppc2ir.getGPRegister(rB)));
    }
    else {
      // EA = rA + rB
      ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_ADD, EA,
							   ppc2ir.getGPRegister(rA),
							   ppc2ir.getGPRegister(rB)));
    }
    // Store value from rS into memory at address EA
    ppc2ir.memoryStore(ppc2ir.getGPRegister(rS), EA.copyRO());
	 // Set CR field to say store succeeded
	 setCRfield(ppc2ir, lazy, 0, new OPT_IntConstantOperand(0), new OPT_IntConstantOperand(0), PPC2IR.Laziness.SIGNED_INT_CMP);
    return pc+4;
  }
}
/**
 * The translator for the stwx instruction
 */
class stwx_translator extends PPC_InstructionTranslator {
  /**
   * Translate stwx (store word indexed)
   * <listing>
   * if rA = 0 then b <- 0
   * else b <- (rA)
   * EA <- b + rB
   * MEM(EA, 4) <- rS
   * </listing>
   */
  protected int translateX_FORM(PPC2IR ppc2ir, PPC2IR.Laziness lazy, int pc, int inst, int opcode, int rS, int rA, int rB, int secondaryOpcode, int zero) {
    OPT_RegisterOperand EA = ppc2ir.getTempInt(PPC2IR.TEMP_INT_SAFE_FROM_STORE_CLOBBER);

    if (rA == 0) {
      // EA = rB
      ppc2ir.appendInstructionToCurrentBlock(Move.create(INT_MOVE,
							 EA,
							 ppc2ir.getGPRegister(rB)));
    }
    else {
      // EA = rA + rB
      ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_ADD, EA,
							   ppc2ir.getGPRegister(rA),
							   ppc2ir.getGPRegister(rB)));
    }
    // Store value from rS into memory at address EA
    ppc2ir.memoryStore(ppc2ir.getGPRegister(rS), EA.copyRO());
    return pc+4;
  }
}
/**
 * The translator for the stdux instruction
 */
class stdux_translator extends PPC_InstructionTranslator {}
/**
 * The translator for the stwux instruction
 */
class stwux_translator extends PPC_InstructionTranslator {
  /**
   * Translate stwux (store word with update indexed)
   * <listing>
   * if rA = 0 then b <- 0
   * else b <- (rA)
   * EA <- b + rB
   * MEM(EA, 4) <- rS
   * </listing>
   */
  protected int translateX_FORM(PPC2IR ppc2ir, PPC2IR.Laziness lazy, int pc, int inst, int opcode, int rS, int rA, int rB, int secondaryOpcode, int zero) {
    OPT_RegisterOperand EA = ppc2ir.getTempInt(PPC2IR.TEMP_INT_SAFE_FROM_STORE_CLOBBER);

    if (rA == 0) {
      // EA = rB
      ppc2ir.appendInstructionToCurrentBlock(Move.create(INT_MOVE,
							 EA,
							 ppc2ir.getGPRegister(rB)));
    }
    else {
      // EA = rA + rB
      ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_ADD, EA,
							   ppc2ir.getGPRegister(rA),
							   ppc2ir.getGPRegister(rB)));
    }
    // Store value from rS into memory at address EA
    ppc2ir.memoryStore(ppc2ir.getGPRegister(rS), EA.copyRO());
    // Place EA into rA
    ppc2ir.appendInstructionToCurrentBlock(Move.create(INT_MOVE, ppc2ir.getGPRegister(rA), 
						       EA.copyRO()));
    return pc+4;
  }
}
/**
 * The translator for the mtsr instruction
 */
class mtsr_translator extends PPC_InstructionTranslator {}
/**
 * The translator for the stdcx instruction
 */
class stdcx_translator extends PPC_InstructionTranslator {}
/**
 * The translator for the stbx instruction
 */
class stbx_translator extends PPC_InstructionTranslator {
  /**
   * Translate stbx (store byte indexed)
   * <listing>
   * if rA = 0 then b <- 0
   * else b <- (rA)
   * EA <- b + rB
   * MEM(EA, 1) <- rS[24-31]
   * </listing>
   */
  protected int translateX_FORM(PPC2IR ppc2ir, PPC2IR.Laziness lazy, int pc, int inst, int opcode, int rS, int rA, int rB, int secondaryOpcode, int zero) {
    OPT_RegisterOperand EA = ppc2ir.getTempInt(PPC2IR.TEMP_INT_SAFE_FROM_STORE_BYTE_CLOBBER);

    if (rA == 0) {
      // EA = rB
      ppc2ir.appendInstructionToCurrentBlock(Move.create(INT_MOVE,
							 EA,
							 ppc2ir.getGPRegister(rB)));
    }
    else {
      // EA = rA + rB
      ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_ADD, EA.copyRO(),
							   ppc2ir.getGPRegister(rA),
							   ppc2ir.getGPRegister(rB)));
    }
    // Store value from rS into memory at address EA
    ppc2ir.memoryStoreByte(ppc2ir.getGPRegister(rS), EA.copyRO());
    return pc+4;
  }
}
/**
 * The translator for the mulld instruction
 */
class mulld_translator extends PPC_InstructionTranslator {}
/**
 * The translator for the mtsrin instruction
 */
class mtsrin_translator extends PPC_InstructionTranslator {}
/**
 * The translator for the dcbtst instruction
 */
class dcbtst_translator extends PPC_InstructionTranslator {
  /**
   * Translate dcbtst (data cache block touch for store)
   */
  protected int translateX_FORM(PPC2IR ppc2ir, PPC2IR.Laziness lazy, int pc, int inst, int opcode, int zero, int rA, int rB, int secondaryOpcode, int zero2) {
	 // translated into a NOP
	 return pc+4;
  }
}
/**
 * The translator for the stbux instruction
 */
class stbux_translator extends PPC_InstructionTranslator {}
/**
 * The translator for the dcbt instruction
 */
class dcbt_translator extends PPC_InstructionTranslator {}
/**
 * The translator for the lhzx instruction
 */
class lhzx_translator extends PPC_InstructionTranslator {
  /**
   * Translate lhzx (load half word and zero indexed)
   * <listing>
   * if rA = 0 then b <- 0
   * else b <- (rA)
   * EA <- b + (rB)
   * rD <- (16)0 || MEM(EA, 2)
   * </listing>
   */
  protected int translateX_FORM(PPC2IR ppc2ir, PPC2IR.Laziness lazy, int pc, int inst, int opcode, int rD, int rA, int rB, int secondaryOpcode, int zero) {
    OPT_RegisterOperand EA = ppc2ir.getTempInt(PPC2IR.TEMP_INT_SAFE_FROM_LOAD_CLOBBER);

    if (rA == 0) {
      // EA = rB
      ppc2ir.appendInstructionToCurrentBlock(Move.create(INT_MOVE,
																			EA,
																			ppc2ir.getGPRegister(rB)));
    }
    else {
      // EA = rA + rB
      ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_ADD, EA,
																			  ppc2ir.getGPRegister(rA),
																			  ppc2ir.getGPRegister(rB)));
    }
    // Load value from memory at EA into rD.
    OPT_RegisterOperand result = ppc2ir.getGPRegister(rD);
    ppc2ir.memoryLoad(result, EA.copyRO());
    // Turn 32bit address into appropriate shift using last 2 bits
    // of the address
    //-#if RVM_FOR_POWERPC
    // EA = (EA & 0x2) * 8
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_AND, EA.copyRO(),
							 EA.copyRO(),
							 new OPT_IntConstantOperand(0x2)));
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_SHL, EA.copyRO(),
							 EA.copyRO(),
							 new OPT_IntConstantOperand(3)));
	 //-#else
    // EA = (~EA & 0x2) * 8
    ppc2ir.appendInstructionToCurrentBlock(Unary.create(INT_NOT, EA.copyRO(),
																		  EA.copyRO()));
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_AND, EA.copyRO(),
							 EA.copyRO(),
							 new OPT_IntConstantOperand(0x2)));
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_SHL, EA.copyRO(),
							 EA.copyRO(),
							 new OPT_IntConstantOperand(3)));
	 //-#endif
    // rD >>>= EA
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_USHR, result.copyRO(),
							 result.copyRO(),
							 EA.copyRO()));
    // rD &= 0xffff
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_AND, result.copyRO(),
							 result.copyRO(),
							 new OPT_IntConstantOperand(0xffff)));
    return pc+4;
  }
}
/**
 * The translator for the eqv instruction
 */
class eqv_translator extends PPC_InstructionTranslator {}
/**
 * The translator for the tlbie instruction
 */
class tlbie_translator extends PPC_InstructionTranslator {}
/**
 * The translator for the eciwx instruction
 */
class eciwx_translator extends PPC_InstructionTranslator {}
/**
 * The translator for the lhzux instruction
 */
class lhzux_translator extends PPC_InstructionTranslator {}
/**
 * The translator for the xor instruction
 */
class xor_translator extends PPC_InstructionTranslator {
  /**
   * Translate xor
   * <listing>
   * rA <- (rS) (+) (rB)
   * </listing>
   */
  protected int translateX_FORM(PPC2IR ppc2ir, PPC2IR.Laziness lazy, int pc, int inst, int opcode, int rS, int rA, int rB, int secondaryOpcode, int Rc) {
    // Perform xor
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_XOR, ppc2ir.getGPRegister(rA),
							 ppc2ir.getGPRegister(rS),
							 ppc2ir.getGPRegister(rB)));
    if (Rc != 0) {
      // Set condition codes
      setCRfield(ppc2ir, lazy, 0, ppc2ir.getGPRegister(rA), new OPT_IntConstantOperand(0), PPC2IR.Laziness.SIGNED_INT_CMP);
    }
    return pc+4;
  }
}
/**
 * The translator for the lwax instruction
 */
class lwax_translator extends PPC_InstructionTranslator {}
/**
 * The translator for the dst instruction
 */
class dst_translator extends PPC_InstructionTranslator {}
/**
 * The translator for the lhax instruction
 */
class lhax_translator extends PPC_InstructionTranslator {
  /**
   * Translate lhax (load half word algabraeic indexed)
   * <listing>
   * if rA = 0 then b <- 0
   * else b <- (rA)
   * EA <- b+ (rB)
   * rD <- EXTS(MEM(EA, 2))
   * </listing>
   */
  protected int translateX_FORM(PPC2IR ppc2ir, PPC2IR.Laziness lazy, int pc, int inst, int opcode, int rD, int rA, int rB, int secondaryOpcode, int zero) {
    OPT_RegisterOperand EA = ppc2ir.getTempInt(PPC2IR.TEMP_INT_SAFE_FROM_LOAD_CLOBBER);
    if (rA == 0) {
      // EA = rB
      ppc2ir.appendInstructionToCurrentBlock(Move.create(INT_MOVE,
																			EA,
																			ppc2ir.getGPRegister(rB)));
    }
    else {
      // EA = rA + rB
      ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_ADD, EA,
																			  ppc2ir.getGPRegister(rA),
																			  ppc2ir.getGPRegister(rB)));
    }
    // Load value from memory at EA into rD.
    OPT_RegisterOperand result = ppc2ir.getGPRegister(rD);
    ppc2ir.memoryLoad(result, EA.copyRO());
    // Turn 32bit address into appropriate shift using last 2 bits
    // of the address
    //-#if RVM_FOR_POWERPC
    // EA = (EA & 0x2) * 8
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_AND, EA.copyRO(),
							 EA.copyRO(),
							 new OPT_IntConstantOperand(0x2)));
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_SHL, EA.copyRO(),
							 EA.copyRO(),
							 new OPT_IntConstantOperand(3)));
	 //-#else
    // EA = (~EA & 0x2) * 8
    ppc2ir.appendInstructionToCurrentBlock(Unary.create(INT_NOT, EA.copyRO(),
																		  EA.copyRO()));
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_AND, EA.copyRO(),
																			EA.copyRO(),
																			new OPT_IntConstantOperand(0x2)));
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_SHL, EA.copyRO(),
																			EA.copyRO(),
																			new OPT_IntConstantOperand(3)));
	 //-#endif
    // rD >>>= EA
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_SHR, result.copyRO(),
																			result.copyRO(),
																			EA.copyRO()));
    // rD = (rD << 16) >> 16
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_SHL, result.copyRO(),
																			result.copyRO(),
																			new OPT_IntConstantOperand(16)));
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_SHR, result.copyRO(),
																			result.copyRO(),
																			new OPT_IntConstantOperand(16)));
    return pc+4;
  }
}
/**
 * The translator for the lvxl instruction
 */
class lvxl_translator extends PPC_InstructionTranslator {}
/**
 * The translator for the tlbia instruction
 */
class tlbia_translator extends PPC_InstructionTranslator {}
/**
 * The translator for the mftb instruction
 */
class mftb_translator extends PPC_InstructionTranslator {}
/**
 * The translator for the lwaux instruction
 */
class lwaux_translator extends PPC_InstructionTranslator {}
/**
 * The translator for the dstst instruction
 */
class dstst_translator extends PPC_InstructionTranslator {}
/**
 * The translator for the lhaux instruction
 */
class lhaux_translator extends PPC_InstructionTranslator {}
/**
 * The translator for the sthx instruction
 */
class sthx_translator extends PPC_InstructionTranslator {
  /**
   * Translate sth (store half word)
   * <listing>
   * if rA = 0 then b <- 0
   * else b <- (rA)
   * EA <- b+ EXTS(d)
   * MEM(EA, 2) <- rS[16-31]
   * </listing>
   */
  protected int translateX_FORM(PPC2IR ppc2ir, PPC2IR.Laziness lazy, int pc, int inst, int opcode, int rS, int rA, int rB, int secondaryOpcode, int zero) {
    OPT_RegisterOperand EA = ppc2ir.getTempInt(PPC2IR.TEMP_INT_SAFE_FROM_STORE_HALF_WORD_CLOBBER);

    if (rA == 0) {
      // EA = rB
      ppc2ir.appendInstructionToCurrentBlock(Move.create(INT_MOVE,
																			EA,
																			ppc2ir.getGPRegister(rB)));
    }
    else {
      // EA = rA + rB
      ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_ADD, EA,
																			  ppc2ir.getGPRegister(rA),
																			  ppc2ir.getGPRegister(rB)));
    }
    // Store value from rS into memory at address EA
    ppc2ir.memoryStoreHalfWord(ppc2ir.getGPRegister(rS), EA.copyRO());
    return pc+4;
  }
}
/**
 * The translator for the orc instruction
 */
class orc_translator extends PPC_InstructionTranslator {
  /**
   * Translate orc (or with complement)
   * <listing>
   * rA <- (rS) | (rB)
   * </listing>
   */
  protected int translateX_FORM(PPC2IR ppc2ir, PPC2IR.Laziness lazy, int pc, int inst, int opcode, int rS, int rA, int rB, int secondaryOpcode, int Rc) {
    OPT_RegisterOperand tempInt = ppc2ir.getTempInt(0);
    // Perform complement
    ppc2ir.appendInstructionToCurrentBlock(Unary.create(INT_NOT, tempInt,
							ppc2ir.getGPRegister(rB)));
    // Perform or
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_OR, ppc2ir.getGPRegister(rA),
							 ppc2ir.getGPRegister(rS),
							 tempInt.copyRO()));
    if (Rc != 0) {
      // Set condition codes
      setCRfield(ppc2ir, lazy, 0, ppc2ir.getGPRegister(rA), new OPT_IntConstantOperand(0), PPC2IR.Laziness.SIGNED_INT_CMP);
    }
    return pc+4;
  }
}
/**
 * The translator for the slbie instruction
 */
class slbie_translator extends PPC_InstructionTranslator {}
/**
 * The translator for the ecowx instruction
 */
class ecowx_translator extends PPC_InstructionTranslator {}
/**
 * The translator for the sthux instruction
 */
class sthux_translator extends PPC_InstructionTranslator {}
/**
 * The translator for the or instruction
 */
class or_translator extends PPC_InstructionTranslator {
  /**
   * Translate or
   * <listing>
   * rA <- (rS) | (rB)
   * </listing>
   */
  protected int translateX_FORM(PPC2IR ppc2ir, PPC2IR.Laziness lazy, int pc, int inst, int opcode, int rS, int rA, int rB, int secondaryOpcode, int Rc) {
    // Perform or
	 if (rS == rB) {
		ppc2ir.appendInstructionToCurrentBlock(Move.create(INT_MOVE, ppc2ir.getGPRegister(rA),
																			ppc2ir.getGPRegister(rS)));
	 }
	 else {
		ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_OR, ppc2ir.getGPRegister(rA),
																			  ppc2ir.getGPRegister(rS),
																			  ppc2ir.getGPRegister(rB)));
	 }
    if (Rc != 0) {
      // Set condition codes
      setCRfield(ppc2ir, lazy, 0, ppc2ir.getGPRegister(rA), new OPT_IntConstantOperand(0), PPC2IR.Laziness.SIGNED_INT_CMP);
    }
    return pc+4;
  }
}

/**
 * The translator for the divdu instruction
 */
class divdu_translator extends PPC_InstructionTranslator {}
/**
 * The translator for the dcbi instruction
 */
class dcbi_translator extends PPC_InstructionTranslator {}
/**
 * The translator for the nand instruction
 */
class nand_translator extends PPC_InstructionTranslator {
  /**
   * Translate and
   * <listing>
   * rA <- ((rS) & (rB))
   * </listing>
   */
  protected int translateX_FORM(PPC2IR ppc2ir, PPC2IR.Laziness lazy, int pc, int inst, int opcode, int rS, int rA, int rB, int secondaryOpcode, int Rc) {
    OPT_RegisterOperand tempInt = ppc2ir.getTempInt(0);
    // Perform and
    if (rS == rB) {
      ppc2ir.appendInstructionToCurrentBlock(Move.create(INT_MOVE,
							 tempInt,
							 ppc2ir.getGPRegister(rB)));
    }
    else {
      ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_AND, tempInt,
							   ppc2ir.getGPRegister(rS),
							   ppc2ir.getGPRegister(rB)));
    }
    // Perform complement
    ppc2ir.appendInstructionToCurrentBlock(Unary.create(INT_NOT, ppc2ir.getGPRegister(rA),
							tempInt.copyRO()));
    if (Rc != 0) {
      // Set condition codes
      setCRfield(ppc2ir, lazy, 0, ppc2ir.getGPRegister(rA), new OPT_IntConstantOperand(0), PPC2IR.Laziness.SIGNED_INT_CMP);
    }
    return pc+4;
  }
}
/**
 * The translator for the duvd instruction
 */
class divd_translator extends PPC_InstructionTranslator {}
/**
 * The translator for the slbia instruction
 */
class slbia_translator extends PPC_InstructionTranslator {}
/**
 * The translator for the mcrxr instruction
 */
class mcrxr_translator extends PPC_InstructionTranslator {}
/**
 * The translator for the lswx instruction
 */
class lswx_translator extends PPC_InstructionTranslator {}
/**
 * The translator for the lwbrx instruction
 */
class lwbrx_translator extends PPC_InstructionTranslator {}
/**
 * The translator for the lfsx instruction
 */
class lfsx_translator extends PPC_InstructionTranslator {
  /**
   * Translate lfsx (load floating-point single indexed)
   * <listing>
   * if rA = 0 then b <- 0
   * else b <- (rA)
   * EA <- b+ (rB)
   * frD <- DOUBLE(MEM(EA, 4))
   * </listing>
   */
  protected int translateX_FORM(PPC2IR ppc2ir, PPC2IR.Laziness lazy, int pc, int inst, int opcode, int frD, int rA, int rB, int secondaryOpcode, int Rc) {
    OPT_RegisterOperand EA = ppc2ir.getTempInt(PPC2IR.TEMP_INT_SAFE_FROM_LOAD_CLOBBER);
    if (rA == 0) {
      // EA = rB
      ppc2ir.appendInstructionToCurrentBlock(Move.create(INT_MOVE,
																			EA,
																			ppc2ir.getGPRegister(rB)));
    }
    else {
      // EA = rA + rB
      ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_ADD, EA,
																			  ppc2ir.getGPRegister(rA),
																			  ppc2ir.getGPRegister(rB)));
    }
    // Load value from memory at EA into tempInt
    OPT_RegisterOperand tempInt = ppc2ir.getTempInt(PPC2IR.TEMP_INT_SAFE_FROM_LOAD_CLOBBER+1);
    ppc2ir.memoryLoad(tempInt, EA.copyRO());
    // Convert bit pattern to a float.
    OPT_RegisterOperand tempFloat = ppc2ir.getTempFloat(0);
    ppc2ir.appendInstructionToCurrentBlock(Unary.create(INT_BITS_AS_FLOAT, tempFloat,
																		  tempInt.copyRO()));

    // Now to a double, and put into register.
    ppc2ir.appendInstructionToCurrentBlock(Unary.create(ppc2ir.FLOAT_2DOUBLE, ppc2ir.getFPRegister(frD),
																		  tempFloat.copyRO()));
    return pc+4;
  }
}
/**
 * The translator for the srw instruction
 */
class srw_translator extends PPC_InstructionTranslator {
  /**
   * Translate srw (shift right word)
   * <listing>
   * n <- rB[27-31]
   * r <- ROTL(rS,32-n)
   * </listing>
   * Equivalent to:
   * <listing>
   * rD <- rA >>> rB
   * </listing>
   */
  protected int translateX_FORM(PPC2IR ppc2ir, PPC2IR.Laziness lazy, int pc, int inst, int opcode, int rS, int rA, int rB, int secondaryOpcode, int Rc) {
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_USHR, ppc2ir.getGPRegister(rA),
							 ppc2ir.getGPRegister(rS),
							 ppc2ir.getGPRegister(rB)));
    if (Rc != 0) {
      // Set condition codes
      setCRfield(ppc2ir, lazy, 0, ppc2ir.getGPRegister(rA), new OPT_IntConstantOperand(0), PPC2IR.Laziness.SIGNED_INT_CMP);
    }
	 return pc+4;
  }
}
/**
 * The translator for the srd instruction
 */
class srd_translator extends PPC_InstructionTranslator {}
/**
 * The translator for the cntlzd instruction
 */
class cntlzd_translator extends PPC_InstructionTranslator {}
/**
 * The translator for the lfsux instruction
 */
class lfsux_translator extends PPC_InstructionTranslator {}
/**
 * The translator for the mfsr instruction
 */
class mfsr_translator extends PPC_InstructionTranslator {}
/**
 * The translator for the lswi instruction
 */
class lswi_translator extends PPC_InstructionTranslator {}
/**
 * The translator for the sync instruction
 */
class sync_translator extends PPC_InstructionTranslator {}
/**
 * The translator for the lfdx instruction
 */
class lfdx_translator extends PPC_InstructionTranslator {
  /**
   * Translate lfdx (load floating-point double indexed)
   * <listing>
   * if rA = 0 then b <- 0
   * else b <- (rA)
   * EA <- b+ (rB)
   * frD <- MEM(EA, 8)
   * </listing>
   */
  protected int translateX_FORM(PPC2IR ppc2ir, PPC2IR.Laziness lazy, int pc, int inst, int opcode, int frD, int rA, int rB, int secondaryOpcode, int zero) {
    OPT_RegisterOperand EA = ppc2ir.getTempInt(PPC2IR.TEMP_INT_SAFE_FROM_LOAD_CLOBBER);
    OPT_RegisterOperand EA2 = ppc2ir.getTempInt(PPC2IR.TEMP_INT_SAFE_FROM_LOAD_CLOBBER+1);
    if (rA == 0) {
      // EA = rB
      ppc2ir.appendInstructionToCurrentBlock(Move.create(INT_MOVE,
																			EA,
																			ppc2ir.getGPRegister(rB)));
    }
    else {
      // EA = rA + rB
      ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_ADD, EA,
																			  ppc2ir.getGPRegister(rA),
																			  ppc2ir.getGPRegister(rB)));
    }
	 ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_ADD, EA2,
																			EA.copyRO(),
																			new OPT_IntConstantOperand(4)));

    // Load value from memory at EA & EA2 into tempInt & tempInt2 (msb at low address)
    OPT_RegisterOperand tempInt  = ppc2ir.getTempInt(PPC2IR.TEMP_INT_SAFE_FROM_LOAD_CLOBBER+2);
    OPT_RegisterOperand tempInt2 = ppc2ir.getTempInt(PPC2IR.TEMP_INT_SAFE_FROM_LOAD_CLOBBER+3);
    ppc2ir.memoryLoad(tempInt, EA.copyRO());
    ppc2ir.memoryLoad(tempInt2,EA2.copyRO());

    // Merge ints into a long
    // tempLong = ((long)tempInt << 32) | (long)tempInt2
    OPT_RegisterOperand tempLong = ppc2ir.getTempLong(0);
    OPT_RegisterOperand tempLong2 = ppc2ir.getTempLong(1);
    ppc2ir.appendInstructionToCurrentBlock(Unary.create(INT_2LONG, tempLong, tempInt.copyRO()));
    ppc2ir.appendInstructionToCurrentBlock(Unary.create(INT_2LONG, tempLong2, tempInt2.copyRO()));
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(LONG_SHL, tempLong.copyRO(), 
																			tempLong.copyRO(),
																			new OPT_IntConstantOperand(32)));
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(LONG_AND, tempLong2.copyRO(), 
																			tempLong2.copyRO(),
																			new OPT_LongConstantOperand(0xffffffffl)));
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(LONG_OR, tempLong.copyRO(), 
																			tempLong.copyRO(),
																			tempLong2.copyRO()));
    // Convert long to a double, and put into register.
    ppc2ir.appendInstructionToCurrentBlock(Unary.create(LONG_BITS_AS_DOUBLE, ppc2ir.getFPRegister(frD),
																		  tempLong.copyRO()));
    return pc+4;
  }
}
/**
 * The translator for the lfdux instruction
 */
class lfdux_translator extends PPC_InstructionTranslator {}
/**
 * The translator for the mfsrin instruction
 */
class mfsrin_translator extends PPC_InstructionTranslator {}
/**
 * The translator for the stswx instruction
 */
class stswx_translator extends PPC_InstructionTranslator {}
/**
 * The translator for the stwbrx instruction
 */
class stwbrx_translator extends PPC_InstructionTranslator {}
/**
 * The translator for the stfsx instruction
 */
class stfsx_translator extends PPC_InstructionTranslator {}
/**
 * The translator for the stfsux instruction
 */
class stfsux_translator extends PPC_InstructionTranslator {}
/**
 * The translator for the stswi instruction
 */
class stswi_translator extends PPC_InstructionTranslator {}
/**
 * The translator for the stfdx instruction
 */
class stfdx_translator extends PPC_InstructionTranslator {
  /**
   * Translate stfdx (store floating-point double indexed)
   * <listing>
   * if rA = 0 then b <- 0
   * else b <- (rA)
   * EA <- b+ (rB)
   * MEM(EA, 8) <- (frS)
   * </listing>
   */
  protected int translateX_FORM(PPC2IR ppc2ir, PPC2IR.Laziness lazy, int pc, int inst, int opcode, int frS, int rA, int rB, int secondaryOpcode, int zero) {
    OPT_RegisterOperand EA = ppc2ir.getTempInt(PPC2IR.TEMP_INT_SAFE_FROM_STORE_CLOBBER);
    OPT_RegisterOperand EA2 = ppc2ir.getTempInt(PPC2IR.TEMP_INT_SAFE_FROM_STORE_CLOBBER+1);
    if (rA == 0) {
      // EA = rB
      ppc2ir.appendInstructionToCurrentBlock(Move.create(INT_MOVE,
																			EA,
																			ppc2ir.getGPRegister(rB)));
    }
    else {
      // EA = rA + rB
      ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_ADD, EA,
																			  ppc2ir.getGPRegister(rA),
																			  ppc2ir.getGPRegister(rB)));
    }
	 ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_ADD, EA2,
																			EA.copyRO(),
																			new OPT_IntConstantOperand(4)));

    // Split double into ints
	 // tempLong = DOUBLE_AS_LONG_BITS(frS)
	 // tempLong2 = tempLong >>> 32
	 // tempLong = tempLong & 0xffffffff
	 // tempInt = (int)tempLong
	 // tempInt2 = (int)tempLong2 <- msb
	 // mem[EA] = tempInt2
	 // mem[EA2] = tempInt
    OPT_RegisterOperand tempLong = ppc2ir.getTempLong(0);
    OPT_RegisterOperand tempLong2 = ppc2ir.getTempLong(1);
    OPT_RegisterOperand tempInt = ppc2ir.getTempInt(PPC2IR.TEMP_INT_SAFE_FROM_STORE_CLOBBER+2);
    OPT_RegisterOperand tempInt2 = ppc2ir.getTempInt(PPC2IR.TEMP_INT_SAFE_FROM_STORE_CLOBBER+3);
    ppc2ir.appendInstructionToCurrentBlock(Unary.create(DOUBLE_AS_LONG_BITS, tempLong,
																		  ppc2ir.getFPRegister(frS)));
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(LONG_USHR, tempLong2, tempLong.copyRO(),
																			new OPT_IntConstantOperand(32)));
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(LONG_AND, tempLong.copyRO(), tempLong.copyRO(),
																			new OPT_LongConstantOperand(0xffffffffl)));
    ppc2ir.appendInstructionToCurrentBlock(Unary.create(LONG_2INT, tempInt, tempLong.copyRO()));
    ppc2ir.appendInstructionToCurrentBlock(Unary.create(LONG_2INT, tempInt2, tempLong2.copyRO()));

    // Store value into memory at address EA (msb at low address)
    ppc2ir.memoryStore(tempInt2.copyRO(), EA.copyRO());
    ppc2ir.memoryStore(tempInt.copyRO(), EA2.copyRO());
    return pc+4;
  }
}
/**
 * The translator for the dcba instruction
 */
class dcba_translator extends PPC_InstructionTranslator {}
/**
 * The translator for the stfdux instruction
 */
class stfdux_translator extends PPC_InstructionTranslator {}
/**
 * The translator for the lhbrx instruction
 */
class lhbrx_translator extends PPC_InstructionTranslator {}
/**
 * The translator for the sraw instruction
 */
class sraw_translator extends PPC_InstructionTranslator {
  /**
   * Translate sraw (shift right algebraic word)
   * <listing>
   * n <- rB[27-31]
   * r <- ROTL(rS,32-n)
	* if rB[26] = 0 then m <- MASK(n)
	* else m <- (32)0
   * S <- rS
   * rA <- r & m | S & m
   * XER[CA] <- S & ((r & m) != 0)
   * </listing>
   */
  protected int translateX_FORM(PPC2IR ppc2ir, PPC2IR.Laziness lazy, int pc, int inst, int opcode, int rS, int rA, int rB, int secondaryOpcode, int Rc) {
    OPT_BasicBlock rB26_set = ppc2ir.createBlockAfterCurrent();
    OPT_BasicBlock rB26_clear = ppc2ir.createBlockAfter(rB26_set);
    OPT_BasicBlock performSraw = ppc2ir.createBlockAfter(rB26_clear);

	 // test rB[26]	 
	 OPT_RegisterOperand mask = ppc2ir.getTempInt(0);
	 ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_AND, mask,
																			ppc2ir.getGPRegister(rB),
																			new OPT_IntConstantOperand(0x10)));
	 ppc2ir.appendInstructionToCurrentBlock(IfCmp.create(INT_IFCMP, null,
																		  mask.copyRO(),
																		  new OPT_IntConstantOperand(0),
																		  OPT_ConditionOperand.EQUAL(),
																		  rB26_clear.makeJumpTarget(),
																		  OPT_BranchProfileOperand.likely()));
	 ppc2ir.getCurrentBlock().insertOut(rB26_clear);

	 // if rB[26] == 1 then mask = 0
	 rB26_set.appendInstruction(Move.create(INT_MOVE, mask.copyRO(),
														 new OPT_IntConstantOperand(0)));
	 rB26_set.appendInstruction(Goto.create(GOTO, performSraw.makeJumpTarget()));
	 rB26_set.deleteNormalOut();
	 rB26_set.insertOut(performSraw);
	 // else mask=(0x1 << rB) - 1;
	 rB26_clear.appendInstruction(Binary.create(INT_SHL, mask.copyRO(),
															  new OPT_IntConstantOperand(1),
															  ppc2ir.getGPRegister(rB)));
	 rB26_clear.appendInstruction(Binary.create(INT_SUB, mask.copyRO(),
															  mask.copyRO(),
															  new OPT_IntConstantOperand(1)));
	 // perform sraw
	 ppc2ir.setCurrentBlock(performSraw);
	 // shifted bits out of rS
	 OPT_RegisterOperand shiftedBits = ppc2ir.getTempInt(1);
	 ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_AND, shiftedBits,
																			ppc2ir.getGPRegister(rS),
																			mask.copyRO()));
	 // do rA = rS >> rB
	 ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_SHR, ppc2ir.getGPRegister(rA),
																			ppc2ir.getGPRegister(rS),
																			ppc2ir.getGPRegister(rB)));
	 
	 // place either sign extended rA into XER CA register or 0
	 // dependent on whether bits were shifted out of rS
	 OPT_RegisterOperand XER_CA_Int = ppc2ir.getTempInt(2);
	 ppc2ir.appendInstructionToCurrentBlock(CondMove.create(INT_COND_MOVE, XER_CA_Int,
																			  shiftedBits.copyRO(),
																			  new OPT_IntConstantOperand(0),
																			  OPT_ConditionOperand.EQUAL(),
																			  new OPT_IntConstantOperand(0),
																			  ppc2ir.getGPRegister(rA)));
	 ppc2ir.appendInstructionToCurrentBlock(Unary.create(INT_2LONG, ppc2ir.getXER_CA_Register(),
																		  XER_CA_Int.copyRO()));
	 if (Rc != 0) {
		// Set condition codes
		setCRfield(ppc2ir, lazy, 0, ppc2ir.getGPRegister(rA), new OPT_IntConstantOperand(0), PPC2IR.Laziness.SIGNED_INT_CMP);
	 }
	 return pc+4;
  }
}
/**
 * The translator for the srad instruction
 */
class srad_translator extends PPC_InstructionTranslator {}
/**
 * The translator for the dss instruction
 */
class dss_translator extends PPC_InstructionTranslator {}
/**
 * The translator for the srawi instruction
 */
class srawi_translator extends PPC_InstructionTranslator {
  /**
   * Translate srawi (shift right algebraic word immediate)
   * <listing>
   * n <- SH
   * r <- ROTL(rS,32-n)
   * m <- MASK(n)
   * S <- rS
   * rA <- r & m | S & m
   * XER[CA] <- S & ((r & m) != 0)
   * </listing>
   */
  protected int translateX_FORM(PPC2IR ppc2ir, PPC2IR.Laziness lazy, int pc, int inst, int opcode, int rS, int rA, int SH, int secondaryOpcode, int Rc) {
    if(SH == 0) {
      ppc2ir.appendInstructionToCurrentBlock(Move.create(INT_MOVE, ppc2ir.getGPRegister(rA),
							 ppc2ir.getGPRegister(rS)));
      ppc2ir.appendInstructionToCurrentBlock(Move.create(LONG_MOVE, ppc2ir.getXER_CA_Register(),
							 new OPT_LongConstantOperand(0)));
    }
    else {
      // shifted bits out of rS
      OPT_RegisterOperand shiftedBits = ppc2ir.getTempInt(0);
		int mask=(0x1 << SH) - 1;
      ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_AND, shiftedBits,
																			  ppc2ir.getGPRegister(rS),
																			  new OPT_IntConstantOperand(mask)));
      // do rA = rS >> SH
      ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_SHR, ppc2ir.getGPRegister(rA),
																			  ppc2ir.getGPRegister(rS),
																			  new OPT_IntConstantOperand(SH)));

      // place either sign extended rA into XER CA register or 0
      // dependent on whether bits were shifted out of rS
      OPT_RegisterOperand XER_CA_Int = ppc2ir.getTempInt(1);
      ppc2ir.appendInstructionToCurrentBlock(CondMove.create(INT_COND_MOVE, XER_CA_Int,
																				 shiftedBits.copyRO(),
																				 new OPT_IntConstantOperand(0),
																				 OPT_ConditionOperand.EQUAL(),
																				 new OPT_IntConstantOperand(0),
																				 ppc2ir.getGPRegister(rA)));
      ppc2ir.appendInstructionToCurrentBlock(Unary.create(INT_2LONG, ppc2ir.getXER_CA_Register(),
																			 XER_CA_Int.copyRO()));
    }
    if (Rc != 0) {
      // Set condition codes
      setCRfield(ppc2ir, lazy, 0, ppc2ir.getGPRegister(rA), new OPT_IntConstantOperand(0), PPC2IR.Laziness.SIGNED_INT_CMP);
    }
	 return pc+4;
  }
}
/**
 * The translator for the sradi instruction
 */
class sradi_translator extends PPC_InstructionTranslator {}
/**
 * The translator for the eieio instruction
 */
class eieio_translator extends PPC_InstructionTranslator {}
/**
 * The translator for the sthbrx instruction
 */
class sthbrx_translator extends PPC_InstructionTranslator {}
/**
 * The translator for the extsh instruction
 */
class extsh_translator extends PPC_InstructionTranslator {
  /**
   * Translate extsh (extend sign half word)
   * <listing>
   * S <- rS[16]
   * rA[16-31] <- rS[16-31]
	* ra[0-15] <- (16)S
   * </listing>
   */
  protected int translateX_FORM(PPC2IR ppc2ir, PPC2IR.Laziness lazy, int pc, int inst, int opcode, int rS, int rA, int zero, int secondaryOpcode, int Rc) {
	 if (rA == rS) {
      ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_SHL, ppc2ir.getGPRegister(rA),
																			  ppc2ir.getGPRegister(rA),
																			  new OPT_IntConstantOperand(16)));		
      ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_SHR, ppc2ir.getGPRegister(rA),
																			  ppc2ir.getGPRegister(rA),
																			  new OPT_IntConstantOperand(16)));		
	 }
	 else {
		OPT_RegisterOperand tempInt = ppc2ir.getTempInt(0);
      ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_SHL, tempInt,
																			  ppc2ir.getGPRegister(rS),
																			  new OPT_IntConstantOperand(16)));		
      ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_SHR, ppc2ir.getGPRegister(rA),
																			  tempInt.copyRO(),
																			  new OPT_IntConstantOperand(16)));		
	 }
    if (Rc != 0) {
      // Set condition codes
      setCRfield(ppc2ir, lazy, 0, ppc2ir.getGPRegister(rA), new OPT_IntConstantOperand(0), PPC2IR.Laziness.SIGNED_INT_CMP);
    }
	 return pc+4;
  }
}
/**
 * The translator for the extsb instruction
 */
class extsb_translator extends PPC_InstructionTranslator {}
/**
 * The translator for the tlbld instruction
 */
class tlbld_translator extends PPC_InstructionTranslator {}
/**
 * The translator for the icbi instruction
 */
class icbi_translator extends PPC_InstructionTranslator {}
/**
 * The translator for the stfiwx instruction
 */
class stfiwx_translator extends PPC_InstructionTranslator {}
/**
 * The translator for the extsw instruction
 */
class extsw_translator extends PPC_InstructionTranslator {}
/**
 * The translator for the tlbli instruction
 */
class tlbli_translator extends PPC_InstructionTranslator {}
/**
 * The translator for the dcbz instruction
 */
class dcbz_translator extends PPC_InstructionTranslator {
  /**
   * Translate dcbz (data cache block clear to zero)
   */
  protected int translateX_FORM(PPC2IR ppc2ir, PPC2IR.Laziness lazy, int pc, int inst, int opcode, int zero, int rA, int rB, int secondaryOpcode, int zero2) {
	 // translated into a NOP
	 return pc+4;
  }
}
/**
 * The translator for the fcmpu instruction
 */
class fcmpu_translator extends PPC_InstructionTranslator {
  /**
   * Translate fcmpu (floating point compare unordered)
   * <listing>
	* if (frA) is a NaN or (frB) is a NaN then c <- 0b0001
	* else if (frA) < (frB) then c <- 0b1000
	* else if (frA) > (frB) then c <- 0b0100
	* else c <- 0b0010
	* fpscr[FPCC] <- c
	* CR[4 * crfD-4 * crfD + 3] <- c
	* if (frA) is an SNaN or (frB) is an SNaN then fpscr[VXSNAN] <- 1
   * </listing>
   */
  protected int translateX_FORM(PPC2IR ppc2ir, PPC2IR.Laziness lazy, int pc, int inst, int opcode, int crfD00, int frA, int frB, int secondaryOpcode, int zero) {
    int crfD = crfD00 >>> 2;

	 // Lazily set the condition codes in the CR register
    setCRfield(ppc2ir, lazy, crfD, ppc2ir.getFPRegister(frA), ppc2ir.getFPRegister(frB), PPC2IR.Laziness.FP_CMPU);
	 // Set mapping of them to FPCC bits of the FPRF in the FPSCR
	 lazy.setCrfToFprfMap(crfD);

    return pc+4;
  }
}
/**
 * The translator for the frsp instruction
 */
class frsp_translator extends PPC_InstructionTranslator {
  /**
   * Translate frsp (floating round to single)
   */
  protected int translateX_FORM(PPC2IR ppc2ir, PPC2IR.Laziness lazy, int pc, int inst, int opcode, int frD, int zero, int frB, int secondaryOpcode, int Rc) {
	 // This is truly hideous!!!! It depends on the FPSCR flags to
	 // define how to round to single precision.  So I'll fudge this
	 // and do something simpler, just cast to single precision using
	 // whatever Jikes does.
	 
	 OPT_RegisterOperand tempFloat = ppc2ir.getTempFloat(0);
	 ppc2ir.appendInstructionToCurrentBlock(Unary.create(DOUBLE_2FLOAT, tempFloat,
																		  ppc2ir.getFPRegister(frB)));
	 ppc2ir.appendInstructionToCurrentBlock(Unary.create(FLOAT_2DOUBLE, ppc2ir.getFPRegister(frD),
																		  tempFloat.copyRO()));


	 // Set the FPSCR
	 //?????????????? NOT YET IMPLEMENTED

	 // Check the condition register bit. */
    if (Rc != 0) {
      // Set condition codes
		throw new Error("Todo: record on frsp");
    }
	 return pc + 4;
  }
}
/**
 * The translator for the fctiw instruction
 */
class fctiw_translator extends PPC_InstructionTranslator {
  /**
   * Translate fctiw (floating convert to integer word)
   */
  protected int translateX_FORM(PPC2IR ppc2ir, PPC2IR.Laziness lazy, int pc, int inst, int opcode, int frD, int zero, int frB, int secondaryOpcode, int Rc) {
	 // @todo: rounding modes
    OPT_RegisterOperand tempLong = ppc2ir.getTempLong(0);
    ppc2ir.appendInstructionToCurrentBlock(Unary.create(DOUBLE_AS_LONG_BITS, tempLong,
																		  ppc2ir.getFPRegister(frB)));
    ppc2ir.appendInstructionToCurrentBlock(Unary.create(LONG_BITS_AS_DOUBLE, ppc2ir.getFPRegister(frD),
																		  tempLong.copyRO()));
	 // Check the condition register record
	 if(Rc != 0){
		throw new Error("fctiw: setting CR1 not yet implemented.");
	 }
	 return pc+4;
  }
}  
/**
 * The translator for the fctiwz instruction
 */
class fctiwz_translator extends PPC_InstructionTranslator {
  /**
   * Translate fctiwz (floating convert to integer word with round toward zero)
   */
  protected int translateX_FORM(PPC2IR ppc2ir, PPC2IR.Laziness lazy, int pc, int inst, int opcode, int frD, int zero, int frB, int secondaryOpcode, int Rc) {
	 // @todo: rounding modes
    OPT_RegisterOperand tempLong = ppc2ir.getTempLong(0);
    ppc2ir.appendInstructionToCurrentBlock(Unary.create(DOUBLE_AS_LONG_BITS, tempLong,
																		  ppc2ir.getFPRegister(frB)));
    ppc2ir.appendInstructionToCurrentBlock(Unary.create(LONG_BITS_AS_DOUBLE, ppc2ir.getFPRegister(frD),
																		  tempLong.copyRO()));
	 // Check the condition register record
	 if(Rc != 0){
		throw new Error("fctiwz: setting CR1 not yet implemented.");
	 }
	 return pc+4;
  }
}  
/**
 * The translator for the fcmpo instruction
 */
class fcmpo_translator extends PPC_InstructionTranslator {}
/**
 * The translator for the mtfsb1 instruction
 */
class mtfsb1_translator extends PPC_InstructionTranslator {
  /**
   * Translate mtfsb1 (move to fpscr bit 1)
   */
  protected int translateX_FORM(PPC2IR ppc2ir, PPC2IR.Laziness lazy, int pc, int inst, int opcode, int crbD, int zero, int zero2, int secondaryOpcode, int Rc) {
	 if((crbD == 1) || (crbD == 2)){
		throw new Error("mtfsb1: attempting to set bit " + crbD);
	 }
	 int setBit = 0x80000000 >>> crbD;
	 ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_OR, ppc2ir.getFPSCRRegister(),
																			ppc2ir.getFPSCRRegister(),
																			new OPT_IntConstantOperand(setBit)));

	 // Check the condition register record
	 if(Rc != 0){
		throw new Error("mtfsb1: setting CR1 not yet implemented.");
	 }
	 return pc+4;
  }
}
/**
 * The translator for the fneg instruction
 */
class fneg_translator extends PPC_InstructionTranslator {
  /**
   * Translate fneg (floating negate)
   */
  protected int translateX_FORM(PPC2IR ppc2ir, PPC2IR.Laziness lazy, int pc, int inst, int opcode, int frD, int zero, int frB, int secondaryOpcode, int Rc) {
	 // @todo setting of fpscr
	 ppc2ir.appendInstructionToCurrentBlock(Unary.create(DOUBLE_NEG, ppc2ir.getFPRegister(frD),
																		  ppc2ir.getFPRegister(frB)));
	 if (Rc != 0) {
		throw new Error("todo: record within a fneg instruction");
	 }
	 return pc + 4;
  }
}
/**
 * The translator for the mcrfs instruction
 */
class mcrfs_translator extends PPC_InstructionTranslator {}
/**
 * The translator for the mtfsb0 instruction
 */
class mtfsb0_translator extends PPC_InstructionTranslator {
  /**
   * Translate mtfsb0 (move to fpscr bit 0)
   */
  protected int translateX_FORM(PPC2IR ppc2ir, PPC2IR.Laziness lazy, int pc, int inst, int opcode, int crbD, int zero, int zero2, int secondaryOpcode, int Rc) {
	 if((crbD == 1) || (crbD == 2)){
		throw new Error("mtfsb0: attempting to clear bit " + crbD);
	 }
	 int clearBit = ~(0x80000000 >>> crbD);
	 ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_AND, ppc2ir.getFPSCRRegister(),
																			ppc2ir.getFPSCRRegister(),
																			new OPT_IntConstantOperand(clearBit)));

	 // Check the condition register record
	 if(Rc != 0){
		throw new Error("mtfsb0: setting CR1 not yet implemented.");
	 }
	 return pc+4;
  }
}
/**
 * The translator for the fmr instruction
 */
class fmr_translator extends PPC_InstructionTranslator {
  /**
   * Translate fmr (floating move register)
   */
  protected int translateX_FORM(PPC2IR ppc2ir, PPC2IR.Laziness lazy, int pc, int inst, int opcode, int frD, int zero, int frB, int secondaryOpcode, int Rc) {
	 ppc2ir.appendInstructionToCurrentBlock(Move.create(DOUBLE_MOVE, ppc2ir.getFPRegister(frD),
																		 ppc2ir.getFPRegister(frB)));
	 if(Rc != 0) {
		throw new Error("fmr: setting CR1 not yet implemented.");
	 }
	 return pc+4;
  }
}
/**
 * The translator for the mtfsfi instruction
 */
class mtfsfi_translator extends PPC_InstructionTranslator {
  /**
   * Translate mtfsfi (move to FPSCR field immediate)
   */
  protected int translateX_FORM(PPC2IR ppc2ir, PPC2IR.Laziness lazy, int pc, int inst, int opcode, int crfD00, int zero, int imm, int secondaryOpcode, int Rc) {
	 int crfD = crfD00 >>> 2;
	 if(crfD == 0) {
		throw new Error("mtfsfi: attempting to set field " + crfD);
	 }
	 int clearBitsMask = ~(0xf0000000 >>> (crfD * 4));
	 int bitsToSet = (imm << 28) >>> (crfD * 4);
	 ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_AND, ppc2ir.getFPSCRRegister(),
																			ppc2ir.getFPSCRRegister(),
																			new OPT_IntConstantOperand(clearBitsMask)));
	 ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_OR, ppc2ir.getFPSCRRegister(),
																			ppc2ir.getFPSCRRegister(),
																			new OPT_IntConstantOperand(bitsToSet)));
	 if(Rc != 0) {
		throw new Error("mtfsfi: setting CR1 not yet implemented.");
	 }
	 return pc+4;
  }
}
/**
 * The translator for the fnabs instruction
 */
class fnabs_translator extends PPC_InstructionTranslator {
  /**
   * Translate fnabs (floating negative absolute value)
   */
  protected int translateX_FORM(PPC2IR ppc2ir, PPC2IR.Laziness lazy, int pc, int inst, int opcode, int frD, int zero, int frB, int secondaryOpcode, int Rc) {
	 OPT_BasicBlock fallThrough = ppc2ir.getNextBlock();
	 OPT_BasicBlock negate = ppc2ir.createBlockAfterCurrent();
	 OPT_BasicBlock lteZeroTest = ppc2ir.getCurrentBlock();
	 ppc2ir.appendInstructionToCurrentBlock(Move.create(INT_MOVE, ppc2ir.getFPRegister(frD),
																		 ppc2ir.getFPRegister(frB)));
	 lteZeroTest.appendInstruction(IfCmp.create(INT_IFCMP, null,
															  ppc2ir.getFPRegister(frB),
															  new OPT_DoubleConstantOperand(0),
															  OPT_ConditionOperand.LESS_EQUAL(),
															  fallThrough.makeJumpTarget(),
															  OPT_BranchProfileOperand.likely()));
	 lteZeroTest.insertOut(fallThrough);
	 negate.appendInstruction(Unary.create(DOUBLE_NEG, ppc2ir.getFPRegister(frD),
														ppc2ir.getFPRegister(frD)));
	 if(Rc != 0) {
		throw new Error("fabs: setting CR1 not yet implemented.");
	 }
	 return pc+4;
  }
}
/**
 * The translator for the fabs instruction
 */
class fabs_translator extends PPC_InstructionTranslator {
  /**
   * Translate fabs (floating absolute value)
   */
  protected int translateX_FORM(PPC2IR ppc2ir, PPC2IR.Laziness lazy, int pc, int inst, int opcode, int frD, int zero, int frB, int secondaryOpcode, int Rc) {
	 OPT_BasicBlock fallThrough = ppc2ir.getNextBlock();
	 OPT_BasicBlock negate = ppc2ir.createBlockAfterCurrent();
	 OPT_BasicBlock gteZeroTest = ppc2ir.getCurrentBlock();
	 ppc2ir.appendInstructionToCurrentBlock(Move.create(INT_MOVE, ppc2ir.getFPRegister(frD),
																		 ppc2ir.getFPRegister(frB)));
	 gteZeroTest.appendInstruction(IfCmp.create(INT_IFCMP, null,
															  ppc2ir.getFPRegister(frB),
															  new OPT_DoubleConstantOperand(0),
															  OPT_ConditionOperand.GREATER_EQUAL(),
															  fallThrough.makeJumpTarget(),
															  OPT_BranchProfileOperand.likely()));
	 gteZeroTest.insertOut(fallThrough);
	 negate.appendInstruction(Unary.create(DOUBLE_NEG, ppc2ir.getFPRegister(frD),
														ppc2ir.getFPRegister(frD)));
	 if(Rc != 0) {
		throw new Error("fabs: setting CR1 not yet implemented.");
	 }
	 return pc+4;
  }
}
/**
 * The translator for the mffs instruction
 */
class mffs_translator extends PPC_InstructionTranslator {
  /**
   * Translate mffs (move from fpscr)
	* <listing>
	* frD[32-63] <- FPSCR
	* </listing>
   */
  protected int translateX_FORM(PPC2IR ppc2ir, PPC2IR.Laziness lazy, int pc, int inst, int opcode, int frD, int zero, int frB, int secondaryOpcode, int Rc) {
	 OPT_RegisterOperand tempLong = ppc2ir.getTempLong(0);

	 ppc2ir.appendInstructionToCurrentBlock(Unary.create(INT_2LONG, tempLong, ppc2ir.getFPSCRRegister()));
    ppc2ir.appendInstructionToCurrentBlock(Binary.create(LONG_AND, tempLong.copyRO(),
																			tempLong.copyRO(),
																			new OPT_LongConstantOperand(0xffffffffL)));

	 ppc2ir.appendInstructionToCurrentBlock(Unary.create(LONG_BITS_AS_DOUBLE, 
																		  ppc2ir.getFPRegister(frD), 
																		  tempLong.copyRO()));
	 if(Rc != 0) {
		throw new Error("mffs: setting CR1 not yet implemented.");
	 }
	 return pc+4;
  }
}
/**
 * The translator for the fctid instruction
 */
class fctid_translator extends PPC_InstructionTranslator {}
/**
 * The translator for the fctidz instruction
 */
class fctidz_translator extends PPC_InstructionTranslator {}
/**
 * The translator for the fcfid instruction
 */
class fcfid_translator extends PPC_InstructionTranslator {}

// -oO Start of SC form instructions Oo-

/**
 * The translator for the sc instruction
 */
class sc_translator extends PPC_InstructionTranslator {
  /**
   * Translate sc (system call)
   */
  protected int translateSC_FORM(PPC2IR ppc2ir, PPC2IR.Laziness lazy, int pc, int inst, int opcode, int zero1, int zero2, int zero3, int one, int zero4){
    if (VM.VerifyAssertions) VM._assert((zero1 == zero2) && (zero2 == zero3) && (zero3 == zero4) && (zero4 == 0) && (one == 1));
    ppc2ir.plantSystemCall(lazy);
	 return pc + 4;
  }
}

// -oO Start of M form instructions Oo-

/**
 * The translator for the rlwimi instruction
 */
class rlwimi_translator extends PPC_InstructionTranslator {
  /**
   * Translate rlwimi (rotate left word immediate then mask insert)
	* <listing>
	* n <- SH
	* r <- ROTL(rS,n)
	* m <- MASK(MB,ME)
	* rA <- (r & m) | (rA & m)
	* </listing>
   * Other simplified mnemonics: inslwi, inslrwi
   */
  protected int translateM_FORM(PPC2IR ppc2ir, PPC2IR.Laziness lazy, int pc, int inst, int opcode, int rS, int rA, int SH, int MB, int ME, int Rc){
	 // Mask has 1 bits from MB to ME, other bits are 0.
	 int m = 0;
	 // Simple if MB <= ME.
	 if(MB <= ME) {
      for(int b = MB; b <= ME; b++) {
		  m |= (0x80000000 >>> b);
		}
    }
	 else { // Otherwise, need to wrap.
      for(int b = MB; b <= 31; b++) {		
		  m |= (0x80000000 >>> b);
		}
      for(int b = 0; b <= ME; b++) {
		  m |= (0x80000000 >>> b);
		}
    }

	 if (m == 0) {
		ppc2ir.appendInstructionToCurrentBlock(Move.create(INT_MOVE, ppc2ir.getGPRegister(rA),
																			ppc2ir.getGPRegister(rA)));
	 }
	 else {
		OPT_RegisterOperand tempr = ppc2ir.getTempInt(0);
		if (SH != 0){
		  OPT_RegisterOperand tempInt1 = ppc2ir.getTempInt(1);
		  OPT_RegisterOperand tempInt2 = ppc2ir.getTempInt(2);
		  // 1. calculate rS << SH and put into tempInt1
		  // 2. calculate rs >>> (32-SH) and put into tempInt2
		  // 3. put the bitwise OR of these into tempr (the variable r in
		  // the prog. env. manual entry for this instruction).
		  ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_SHL, tempInt1, ppc2ir.getGPRegister(rS), 
																				 new OPT_IntConstantOperand(SH)));
		  ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_USHR, tempInt2, ppc2ir.getGPRegister(rS),
																				 new OPT_IntConstantOperand(32-SH)));
		  ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_OR, tempr, tempInt1.copyRO(), 
																				 tempInt2.copyRO()));
		}
		else {
		  ppc2ir.appendInstructionToCurrentBlock(Move.create(INT_MOVE, tempr,
																			  ppc2ir.getGPRegister(rS)));
		}
		if (m == 0xffffffff) {
		  ppc2ir.appendInstructionToCurrentBlock(Move.create(INT_MOVE, ppc2ir.getGPRegister(rA),
																			  tempr.copyRO()));
		}
		else {
		  OPT_RegisterOperand tempInt3 = ppc2ir.getTempInt(1);
		  OPT_RegisterOperand tempInt4 = ppc2ir.getTempInt(2);
		  // AND the value in tempr with the value m, and put into tempInt3.
		  ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_AND, tempInt3, 
																				 tempr.copyRO(), 
																				 new OPT_IntConstantOperand(m)));
		  // AND the current value of rA with NOT mask, and put into tempInt4.
		  ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_AND, tempInt4,
																				 ppc2ir.getGPRegister(rA),
																				 new OPT_IntConstantOperand(~m)));
		  // OR these two and put back into rA.
		  ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_OR, ppc2ir.getGPRegister(rA),
																				 tempInt3.copyRO(), tempInt4.copyRO()));
		}
	 }
    if (Rc != 0) {
      // Set condition codes
      setCRfield(ppc2ir, lazy, 0, ppc2ir.getGPRegister(rA), new OPT_IntConstantOperand(0), PPC2IR.Laziness.SIGNED_INT_CMP);
    }
	 return pc + 4;
  }
}
/**
 * The translator for the rlwinm instruction
 */
class rlwinm_translator extends PPC_InstructionTranslator {
  /**
   * Translate rlwinm (rotate left word immediate then and with mask)
	* <listing>
	* n <- SH
	* r <- ROTL(rS,n)
	* m <- MASK(MB,ME)
	* rA <- r & m
	* </listing>
   * Other simplified mnemonics: extlwi, extrwi, rotlwi, rotrwi, slwi, srwi, clrlwi, clrrwi, clrlslwi
   */
  protected int translateM_FORM(PPC2IR ppc2ir, PPC2IR.Laziness lazy, int pc, int inst, int opcode, int rS, int rA, int SH, int MB, int ME, int Rc){
	 // Mask has 1 bits from MB to ME, other bits are 0.
	 int m = 0;
	 // Simple if MB <= ME.
	 if(MB <= ME) {
      for(int b = MB; b <= ME; b++) {
		  m |= (0x80000000 >>> b);
		}
    }
	 else { // Otherwise, need to wrap.
      for(int b = MB; b <= 31; b++) {		
		  m |= (0x80000000 >>> b);
		}
      for(int b = 0; b <= ME; b++) {
		  m |= (0x80000000 >>> b);
		}
    }

	 if (m == 0) {
		ppc2ir.appendInstructionToCurrentBlock(Move.create(INT_MOVE, ppc2ir.getGPRegister(rA),
																			new OPT_IntConstantOperand(0)));
	 }
	 else {
		OPT_RegisterOperand tempr = ppc2ir.getTempInt(0);
		if (SH != 0){
		  OPT_RegisterOperand tempInt1 = ppc2ir.getTempInt(1);
		  OPT_RegisterOperand tempInt2 = ppc2ir.getTempInt(2);
		  // 1. calculate rS << SH and put into tempInt1
		  // 2. calculate rS >>> (32-SH) and put into tempInt2
		  // 3. put the bitwise OR of these into tempr (the variable r in
		  // the prog. env. manual entry for this instruction).
		  ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_SHL, tempInt1, ppc2ir.getGPRegister(rS), 
																				 new OPT_IntConstantOperand(SH)));
		  ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_USHR, tempInt2, ppc2ir.getGPRegister(rS),
																				 new OPT_IntConstantOperand(32-SH)));
		  ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_OR, tempr, tempInt1.copyRO(), 
																				 tempInt2.copyRO()));
		}
		else {
		  ppc2ir.appendInstructionToCurrentBlock(Move.create(INT_MOVE, tempr,
																			  ppc2ir.getGPRegister(rS)));
		}
		if (m == 0xffffffff) {
		  ppc2ir.appendInstructionToCurrentBlock(Move.create(INT_MOVE, ppc2ir.getGPRegister(rA),
																			  tempr.copyRO()));
		}
		else {
		  // AND the value in tempr with the value m, and put into rA
		  ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_AND, ppc2ir.getGPRegister(rA),
																				 tempr.copyRO(), 
																				 new OPT_IntConstantOperand(m)));
		}
	 }
    if (Rc != 0) {
      // Set condition codes
      setCRfield(ppc2ir, lazy, 0, ppc2ir.getGPRegister(rA), new OPT_IntConstantOperand(0), PPC2IR.Laziness.SIGNED_INT_CMP);
    }
	 return pc + 4;
  }
}
/**
 * The translator for the rlwnm instruction
 */
class rlwnm_translator extends PPC_InstructionTranslator {
  /**
   * Translate rlwnm (rotate left word then and with mask)
	* <listing>
	* n <- rB[27-31]
	* r <- ROTL(rS,n)
	* m <- MASK(MB,ME)
	* rA <- r & m
	* </listing>
   * Other simplified mnemonics: rtolw
   */
  protected int translateM_FORM(PPC2IR ppc2ir, PPC2IR.Laziness lazy, int pc, int inst, int opcode, int rS, int rA, int rB, int MB, int ME, int Rc){
	 // Mask has 1 bits from MB to ME, other bits are 0.
	 int m = 0;
	 // Simple if MB <= ME.
	 if(MB <= ME) {
      for(int b = MB; b <= ME; b++) {
		  m |= (0x80000000 >>> b);
		}
    }
	 else { // Otherwise, need to wrap.
      for(int b = MB; b <= 31; b++) {		
		  m |= (0x80000000 >>> b);
		}
      for(int b = 0; b <= ME; b++) {
		  m |= (0x80000000 >>> b);
		}
    }

	 if (m == 0) {
		ppc2ir.appendInstructionToCurrentBlock(Move.create(INT_MOVE, ppc2ir.getGPRegister(rA),
																			new OPT_IntConstantOperand(0)));
	 }
	 else {
		OPT_RegisterOperand tempr = ppc2ir.getTempInt(0);

		OPT_RegisterOperand tempInt1 = ppc2ir.getTempInt(1);
		OPT_RegisterOperand tempInt2 = ppc2ir.getTempInt(2);

		OPT_RegisterOperand tempIntSH = ppc2ir.getTempInt(3);

		ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_AND, tempIntSH,
																			  ppc2ir.getGPRegister(rB),
																			  new OPT_IntConstantOperand(0x1f)));

		// 1. calculate rS << rB and put into tempInt1
		// 2. calculate rS >>> (32-rB) and put into tempInt2
		// 3. put the bitwise OR of these into tempr (the variable r in
		// the prog. env. manual entry for this instruction).
		ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_SHL, tempInt1, ppc2ir.getGPRegister(rS), 
																			  tempIntSH.copyRO()));
		ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_SUB, tempIntSH.copyRO(),
																			  new OPT_IntConstantOperand(32),
																			  tempIntSH.copyRO()));
		ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_USHR, tempInt2, ppc2ir.getGPRegister(rS),
																			  tempIntSH.copyRO()));
		ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_OR, tempr, tempInt1.copyRO(), 
																			  tempInt2.copyRO()));
		if (m == 0xffffffff) {
		  ppc2ir.appendInstructionToCurrentBlock(Move.create(INT_MOVE, ppc2ir.getGPRegister(rA),
																			  tempr.copyRO()));
		}
		else {
		  // AND the value in tempr with the value m, and put into rA
		  ppc2ir.appendInstructionToCurrentBlock(Binary.create(INT_AND, ppc2ir.getGPRegister(rA),
																				 tempr.copyRO(), 
																				 new OPT_IntConstantOperand(m)));
		}
	 }
    if (Rc != 0) {
      // Set condition codes
      setCRfield(ppc2ir, lazy, 0, ppc2ir.getGPRegister(rA), new OPT_IntConstantOperand(0), PPC2IR.Laziness.SIGNED_INT_CMP);
    }
	 return pc + 4;
  }
}

// -oO Start of B form instructions Oo-

/**
 * The translator for the bc instruction
 */
class bc_translator extends PPC_InstructionTranslator {
  /**
   * Translate bc (branch conditional)
   * <listing>
   * if LK then LR <-iea CIA + 4
   * if  BO[2] then CTR <- CTR - 1
   * ctr_ok <- BO[2] | ((CTR != 0) (+) BO[3])
   * cond_ok <- BO[0] | (CR[BI] == BO[1])
   * if ctr_ok & cond_ok then
   *   if AA then NIA <-iea EXTS(BD || 0b00)
   *   else NIA <-iea CIA + EXTS(BD || 0b00)
   * </listing>
   * Other simplified mnemonics: bltctr, bnectr
   */
  protected int translateB_FORM(PPC2IR ppc2ir, PPC2IR.Laziness lazy, int pc, int inst, int opcode, int BO, int BI, int BD, int AA, int LK){
    // Translation process:
    // if lk alter lr value
    // decode BO and optionally plant any of:
    //  1: decrement CTR
    //  2: branch to instructionEndBlock if CTR == 0 (or CTR != 0)
    //  3: condition field test and branch to instructionEndBlock if true (or false)
    // plant branch block

    // The block gone to if the branch isn't taken
    OPT_BasicBlock instructionEndBlock = ppc2ir.getNextBlock();
	 // Did we find the block we're going to already translated?
	 boolean found_mapping = false;
    // Calculate branch target address
    int target_address;
    if (AA != 0) {
      target_address = EXTS(BD, 14) << 2;
    }
    else {
      target_address = pc + (EXTS(BD, 14) << 2);
    }

    if (LK != 0){
		ppc2ir.registerBranchAndLink(pc, target_address);
      OPT_RegisterOperand lr = ppc2ir.getLRRegister();
      ppc2ir.appendInstructionToCurrentBlock(Move.create(INT_MOVE,
																			lr,
																			new OPT_IntConstantOperand(pc+4)));
    }

    // Decode BO
    boolean likely_to_fallthrough = ((BO & 1) == 0);
    boolean branch_if_ctr_zero = ((BO & 2) == 0);
    if ((BO & 4) == 0) {
      plantDecrementCTR(ppc2ir);
      plantBranchToBlockDependentOnCTR(ppc2ir, instructionEndBlock, branch_if_ctr_zero, likely_to_fallthrough);
    }
    if ((BO & 16) == 0) {
      boolean branch_if_cond_true = ((BO & 8) != 0);
      plantBranchToBlockDependentOnCondition(ppc2ir, instructionEndBlock, lazy, BI, !branch_if_cond_true, likely_to_fallthrough);
    }

	 if ((LK == 0) || (ppc2ir.traceContinuesAfterBranchAndLink(pc))) {
		// Plant branch block
		OPT_BasicBlock targetBlock = ppc2ir.findMapping(target_address, lazy);
		if (targetBlock != null) {
		  found_mapping = true;
		  ppc2ir.appendInstructionToCurrentBlock(Goto.create(GOTO, targetBlock.makeJumpTarget()));
		  ppc2ir.getCurrentBlock().deleteNormalOut();
		  ppc2ir.getCurrentBlock().insertOut(targetBlock);
		  if (VM.VerifyAssertions) VM._assert(ppc2ir.getCurrentBlock().getNumberOfNormalOut() == 1);
		}
		else {
		  OPT_Instruction gotoInstr = Goto.create(GOTO, null);
		  ppc2ir.appendInstructionToCurrentBlock(gotoInstr);
		  ppc2ir.registerGotoTargetUnresolved(gotoInstr, target_address, (PPC2IR.Laziness)lazy.clone());
		}
		// stop translation on branch always
		if (BO == 0x14) {
		  return -1;
		}
		else {
		  // continue translation of next instruction if branch forward
		  // or the backward branch is already translated or if we're
		  // not trying to optimize here
		  if ((target_address > pc)||(found_mapping)||(!PPC_Options.optimizeBackwardBranches)) {
			 return pc+4;
		  }
		  else {
			 // force translation of backward branch first:
			 // place a goto pc+4 in the next block, then return the
			 // target_address to translate into an empty block
			 ppc2ir.setCurrentBlock(instructionEndBlock);
			 ppc2ir.setNextBlock(ppc2ir.createBlockAfterCurrent());
			 OPT_Instruction gotoInstr = Goto.create(GOTO, null);
			 ppc2ir.appendInstructionToCurrentBlock(gotoInstr);
			 ppc2ir.registerGotoTargetUnresolved(gotoInstr, pc+4, (PPC2IR.Laziness)lazy.clone());			 
			 return target_address;
		  }
		}
	 }
	 else {
		// This was a branch and link and the trace should stop, so end
		// gracefully
		ppc2ir.setReturnValueResolveLazinessAndBranchToFinish((PPC2IR.Laziness)lazy.clone(), new OPT_IntConstantOperand(target_address));
		return -1;		
	 }
  }
}

// -oO Start of I form instructions Oo-

/**
 * The translator for the b instruction
 */
class b_translator extends PPC_InstructionTranslator {
  /**
   * Translate b (branch)
	* <listing>
	* if AA then NIA <-iea EXTS(LI || 0b00)
	* else NIA <- CIA + EXTS(LI || 0b00)
	* if LK then LR <-iea CIA + 4
	* </listing>
   * Other simplified mnemonics: ba, bl, bla
   */
  protected int translateI_FORM(PPC2IR ppc2ir, PPC2IR.Laziness lazy, int pc, int instr, int opcode, int li, int aa, int lk) {
	 li = EXTS(li, 24) << 2; // This is what the programming environment manual denotes as EXTS(LI || 0b00)

	 int target_addr;
	 if (aa != 0) {
		 target_addr = li;
	 }
	 else {
		 target_addr = pc + li;
	 }
	 if (lk != 0) {
		// @todo: record this as a likely return address...
      ppc2ir.appendInstructionToCurrentBlock(Move.create(INT_MOVE, ppc2ir.getLRRegister(),
																			new OPT_IntConstantOperand(pc + 4)));
		ppc2ir.registerBranchAndLink(pc, target_addr);
		if (ppc2ir.traceContinuesAfterBranchAndLink(pc)) {
		  return target_addr;
		}
		else {
		  ppc2ir.setReturnValueResolveLazinessAndBranchToFinish((PPC2IR.Laziness)lazy.clone(), new OPT_IntConstantOperand(target_addr));
		  return -1;
		}
	 }
	 else {
		return target_addr;
	 }
  }
}


// -oO Start of A form instructions Oo-

/**
 * The translator for the fdivs instruction
 */
class fdivs_translator extends PPC_InstructionTranslator {
  /**
   * Translate fdivs (floating divide single)
   */
  protected int translateA_FORM(PPC2IR ppc2ir, PPC2IR.Laziness lazy, int pc, int instr, int opcode, int frD, int frA, int frB, int zero, int secondaryOpcode, int Rc){
	 // @todo setting of fpscr
	 OPT_RegisterOperand tempFloat1 = ppc2ir.getTempFloat(0);
	 OPT_RegisterOperand tempFloat2 = ppc2ir.getTempFloat(1);
	 ppc2ir.appendInstructionToCurrentBlock(Unary.create(DOUBLE_2FLOAT, tempFloat1,
																		  ppc2ir.getFPRegister(frA)));
	 ppc2ir.appendInstructionToCurrentBlock(Unary.create(DOUBLE_2FLOAT, tempFloat2,
																		  ppc2ir.getFPRegister(frB)));
	 ppc2ir.appendInstructionToCurrentBlock(Binary.create(FLOAT_DIV, tempFloat1.copyRO(),
																			tempFloat1.copyRO(),
																			tempFloat2.copyRO()));
	 ppc2ir.appendInstructionToCurrentBlock(Unary.create(FLOAT_2DOUBLE, ppc2ir.getFPRegister(frD),
																		  tempFloat1.copyRO()));
	 if (Rc != 0) {
		throw new Error("todo: record within a fdivs instruction");
	 }
	 return pc + 4;
  }
}
/**
 * The translator for the fsubs instruction
 */
class fsubs_translator extends PPC_InstructionTranslator {
  /**
   * Translate fsubs (floating subtract single)
   */
  protected int translateA_FORM(PPC2IR ppc2ir, PPC2IR.Laziness lazy, int pc, int instr, int opcode, int frD, int frA, int frB, int zero, int secondaryOpcode, int Rc){
	 // @todo setting of fpscr
	 OPT_RegisterOperand tempFloat1 = ppc2ir.getTempFloat(0);
	 OPT_RegisterOperand tempFloat2 = ppc2ir.getTempFloat(1);
	 ppc2ir.appendInstructionToCurrentBlock(Unary.create(DOUBLE_2FLOAT, tempFloat1,
																		  ppc2ir.getFPRegister(frA)));
	 ppc2ir.appendInstructionToCurrentBlock(Unary.create(DOUBLE_2FLOAT, tempFloat2,
																		  ppc2ir.getFPRegister(frB)));
	 ppc2ir.appendInstructionToCurrentBlock(Binary.create(FLOAT_SUB, tempFloat1.copyRO(),
																			tempFloat1.copyRO(),
																			tempFloat2.copyRO()));
	 ppc2ir.appendInstructionToCurrentBlock(Unary.create(FLOAT_2DOUBLE, ppc2ir.getFPRegister(frD),
																		  tempFloat1.copyRO()));
	 if (Rc != 0) {
		throw new Error("todo: record within a fsubs instruction");
	 }
	 return pc + 4;
  }
}
/**
 * The translator for the fadds instruction
 */
class fadds_translator extends PPC_InstructionTranslator {
  /**
   * Translate fadds (floating add single)
   */
  protected int translateA_FORM(PPC2IR ppc2ir, PPC2IR.Laziness lazy, int pc, int instr, int opcode, int frD, int frA, int frB, int zero, int secondaryOpcode, int Rc){
	 // @todo setting of fpscr
	 OPT_RegisterOperand tempFloat1 = ppc2ir.getTempFloat(0);
	 OPT_RegisterOperand tempFloat2 = ppc2ir.getTempFloat(1);
	 ppc2ir.appendInstructionToCurrentBlock(Unary.create(DOUBLE_2FLOAT, tempFloat1,
																		  ppc2ir.getFPRegister(frA)));
	 ppc2ir.appendInstructionToCurrentBlock(Unary.create(DOUBLE_2FLOAT, tempFloat2,
																		  ppc2ir.getFPRegister(frB)));
	 ppc2ir.appendInstructionToCurrentBlock(Binary.create(FLOAT_ADD, tempFloat1.copyRO(),
																			tempFloat1.copyRO(),
																			tempFloat2.copyRO()));
	 ppc2ir.appendInstructionToCurrentBlock(Unary.create(FLOAT_2DOUBLE, ppc2ir.getFPRegister(frD),
																		  tempFloat1.copyRO()));
	 if (Rc != 0) {
		throw new Error("todo: record within a fsubs instruction");
	 }
	 return pc + 4;
  }
}
/**
 * The translator for the fsqrts instruction
 */
class fsqrts_translator extends PPC_InstructionTranslator {}
/**
 * The translator for the fres instruction
 */
class fres_translator extends PPC_InstructionTranslator {}
/**
 * The translator for the fmuls instruction
 */
class fmuls_translator extends PPC_InstructionTranslator {
  /**
   * Translate fmuls (floating multiply single)
   */
  protected int translateA_FORM(PPC2IR ppc2ir, PPC2IR.Laziness lazy, int pc, int instr, int opcode, int frD, int frA, int zero, int frC, int secondaryOpcode, int Rc){
	 // @todo setting of fpscr
	 OPT_RegisterOperand tempFloat1 = ppc2ir.getTempFloat(0);
	 OPT_RegisterOperand tempFloat2 = ppc2ir.getTempFloat(1);
	 ppc2ir.appendInstructionToCurrentBlock(Unary.create(DOUBLE_2FLOAT, tempFloat1,
																		  ppc2ir.getFPRegister(frA)));
	 ppc2ir.appendInstructionToCurrentBlock(Unary.create(DOUBLE_2FLOAT, tempFloat2,
																		  ppc2ir.getFPRegister(frC)));
	 ppc2ir.appendInstructionToCurrentBlock(Binary.create(FLOAT_MUL, tempFloat1.copyRO(),
																			tempFloat1.copyRO(),
																			tempFloat2.copyRO()));
	 ppc2ir.appendInstructionToCurrentBlock(Unary.create(FLOAT_2DOUBLE, ppc2ir.getFPRegister(frD),
																		  tempFloat1.copyRO()));
	 if (Rc != 0) {
		throw new Error("todo: record within a fmuls instruction");
	 }
	 return pc + 4;
  }
}
/**
 * The translator for the fmsubs instruction
 */
class fmsubs_translator extends PPC_InstructionTranslator {}  
/**
 * The translator for the fmadds instruction
 */
class fmadds_translator extends PPC_InstructionTranslator {}  
/**
 * The translator for the fnmadds instruction
 */
class fnmadds_translator extends PPC_InstructionTranslator {} 
/**
 * The translator for the fnmsubs instruction
 */
class fnmsubs_translator extends PPC_InstructionTranslator {}
/**
 * The translator for the fdiv instruction
 */
class fdiv_translator extends PPC_InstructionTranslator {
  /**
   * Translate fdiv (floating divide double-precision)
   */
  protected int translateA_FORM(PPC2IR ppc2ir, PPC2IR.Laziness lazy, int pc, int instr, int opcode, int frD, int frA, int frB, int zero, int secondaryOpcode, int Rc){
	 // @todo setting of fpscr
	 ppc2ir.appendInstructionToCurrentBlock(Binary.create(DOUBLE_DIV, ppc2ir.getFPRegister(frD),
																			ppc2ir.getFPRegister(frA),
																			ppc2ir.getFPRegister(frB)));
	 if (Rc != 0) {
		throw new Error("todo: record within a fdiv instruction");
	 }

	 return pc + 4;
  }
}
/**
 * The translator for the fsub instruction
 */
class fsub_translator extends PPC_InstructionTranslator {
  /**
   * Translate fsub (floating subtract double-precision)
   */
  protected int translateA_FORM(PPC2IR ppc2ir, PPC2IR.Laziness lazy, int pc, int instr, int opcode, int frD, int frA, int frB, int zero, int secondaryOpcode, int Rc){
	 // @todo setting of fpscr
	 ppc2ir.appendInstructionToCurrentBlock(Binary.create(DOUBLE_SUB, ppc2ir.getFPRegister(frD),
																			ppc2ir.getFPRegister(frA),
																			ppc2ir.getFPRegister(frB)));
	 if (Rc != 0) {
		throw new Error("todo: record within a fsub instruction");
	 }

	 return pc + 4;
  }
}
/**
 * The translator for the fadd instruction
 */
class fadd_translator extends PPC_InstructionTranslator {
  /**
   * Translate fadd (floating add double-precision)
   */
  protected int translateA_FORM(PPC2IR ppc2ir, PPC2IR.Laziness lazy, int pc, int instr, int opcode, int frD, int frA, int frB, int zero, int secondaryOpcode, int Rc){
	 // @todo setting of fpscr
	 ppc2ir.appendInstructionToCurrentBlock(Binary.create(DOUBLE_ADD, ppc2ir.getFPRegister(frD),
																			ppc2ir.getFPRegister(frA),
																			ppc2ir.getFPRegister(frB)));
	 if (Rc != 0) {
		throw new Error("todo: record within a fadd instruction");
	 }

	 return pc + 4;
  }
}
/**
 * The translator for the fmul instruction
 */
class fmul_translator extends PPC_InstructionTranslator {
  /**
   * Translate fmul (floating multiply double-precision)
   */
  protected int translateA_FORM(PPC2IR ppc2ir, PPC2IR.Laziness lazy, int pc, int instr, int opcode, int frD, int frA, int zero, int frC, int secondaryOpcode, int Rc){
	 // @todo setting of fpscr
	 ppc2ir.appendInstructionToCurrentBlock(Binary.create(DOUBLE_MUL, ppc2ir.getFPRegister(frD),
																			ppc2ir.getFPRegister(frA),
																			ppc2ir.getFPRegister(frC)));
	 if (Rc != 0) {
		throw new Error("todo: record within a fmul instruction");
	 }

	 return pc + 4;
  }
}
/**
 * The translator for the fmadd instruction
 */
class fmadd_translator extends PPC_InstructionTranslator {
  /**
   * Translate fmadd (floating multiply-add double-precision)
   */
  protected int translateA_FORM(PPC2IR ppc2ir, PPC2IR.Laziness lazy, int pc, int instr, int opcode, int frD, int frA, int frB, int frC, int secondaryOpcode, int Rc){
	 // @todo setting of fpscr
    OPT_RegisterOperand tempDouble = ppc2ir.getTempDouble(0);
	 ppc2ir.appendInstructionToCurrentBlock(Binary.create(DOUBLE_MUL, tempDouble,
																			ppc2ir.getFPRegister(frA),
																			ppc2ir.getFPRegister(frC)));
	 ppc2ir.appendInstructionToCurrentBlock(Binary.create(DOUBLE_ADD, ppc2ir.getFPRegister(frD),
																			tempDouble.copyRO(),
																			ppc2ir.getFPRegister(frB)));
	 if (Rc != 0) {
		throw new Error("todo: record within a fmadd instruction");
	 }

	 return pc + 4;
  }
}
/**
 * The translator for the fmsub instruction
 */
class fmsub_translator extends PPC_InstructionTranslator {
  /**
   * Translate fmsub (floating multiply-subtract double-precision)
   */
  protected int translateA_FORM(PPC2IR ppc2ir, PPC2IR.Laziness lazy, int pc, int instr, int opcode, int frD, int frA, int frB, int frC, int secondaryOpcode, int Rc){
	 // @todo setting of fpscr
    OPT_RegisterOperand tempDouble = ppc2ir.getTempDouble(0);
	 ppc2ir.appendInstructionToCurrentBlock(Binary.create(DOUBLE_MUL, tempDouble,
																			ppc2ir.getFPRegister(frA),
																			ppc2ir.getFPRegister(frC)));
	 ppc2ir.appendInstructionToCurrentBlock(Binary.create(DOUBLE_SUB, ppc2ir.getFPRegister(frD),
																			tempDouble.copyRO(),
																			ppc2ir.getFPRegister(frB)));
	 if (Rc != 0) {
		throw new Error("todo: record within a fmsub instruction");
	 }

	 return pc + 4;
  }
}
/**
 * The translator for the fnmsub instruction
 */
class fnmsub_translator extends PPC_InstructionTranslator {
  /**
   * Translate fnmsub (floating negative multiply-subtract)
   */
  protected int translateA_FORM(PPC2IR ppc2ir, PPC2IR.Laziness lazy, int pc, int instr, int opcode, int frD, int frA, int frB, int frC, int secondaryOpcode, int Rc){
	 // @todo setting of fpscr
    OPT_RegisterOperand tempDouble = ppc2ir.getTempDouble(0);
	 ppc2ir.appendInstructionToCurrentBlock(Binary.create(DOUBLE_MUL, tempDouble,
																			ppc2ir.getFPRegister(frA),
																			ppc2ir.getFPRegister(frC)));
	 ppc2ir.appendInstructionToCurrentBlock(Binary.create(DOUBLE_SUB, tempDouble.copyRO(),
																			tempDouble.copyRO(),
																			ppc2ir.getFPRegister(frB)));
	 ppc2ir.appendInstructionToCurrentBlock(Unary.create(DOUBLE_NEG, ppc2ir.getFPRegister(frD),
																		  tempDouble.copyRO()));
	 if (Rc != 0) {
		throw new Error("todo: record within a fnmsub instruction");
	 }

	 return pc + 4;
  }
}
/**
 * The translator for the fnmadd instruction
 */
class fnmadd_translator extends PPC_InstructionTranslator {
  /**
   * Translate fnmadd (floating negative multiply-add)
   */
  protected int translateA_FORM(PPC2IR ppc2ir, PPC2IR.Laziness lazy, int pc, int instr, int opcode, int frD, int frA, int frB, int frC, int secondaryOpcode, int Rc){
	 // @todo setting of fpscr
    OPT_RegisterOperand tempDouble = ppc2ir.getTempDouble(0);
	 ppc2ir.appendInstructionToCurrentBlock(Binary.create(DOUBLE_MUL, tempDouble,
																			ppc2ir.getFPRegister(frA),
																			ppc2ir.getFPRegister(frC)));
	 ppc2ir.appendInstructionToCurrentBlock(Binary.create(DOUBLE_ADD, tempDouble.copyRO(),
																			tempDouble.copyRO(),
																			ppc2ir.getFPRegister(frB)));
	 ppc2ir.appendInstructionToCurrentBlock(Unary.create(DOUBLE_NEG, ppc2ir.getFPRegister(frD),
																		  tempDouble.copyRO()));
	 if (Rc != 0) {
		throw new Error("todo: record within a fnmadd instruction");
	 }

	 return pc + 4;
  }
}
